c--------------------------------------------------------------------
c compToMoles.f
c
c Convert from computational to physical variables. 
c
c--------------------------------------------------------------------

      subroutine compToMoles(nvec,x,nSpecies,nComp,nR,nMinQ,
     $     formulaMatrix,mu0,nPhase,phaseIdentity,reactionType,
     $     prod,list)

      IMPLICIT NONE
      INTRINSIC exp

      integer nSpecies,nComp,nR,nMinQ,i,ind,nPhase(*),list(nR),
     $     phaseIdentity(nSpecies),reactionType(nR)

      real*8 nvec(nSpecies),prod(nR),x(nComp+nMinQ),
     $     formulaMatrix(nComp,nSpecies),mu0(nSpecies)

      logical mineral

c--------------------------------------------------------------------
c     EVALUATE THE SECONDARY UNKNOWNS
c--------------------------------------------------------------------

      do i = 1,nR
         list(i) = 0
      enddo

      call massAction(prod,x,formulaMatrix,mu0,list,nSpecies,nComp,nR)

c--------------------------------------------------------------------
c     THE FIRST NCOMP UNKNOWNS ARE THE NATURAL LOGARITHMS OF THE
c     COMPONENT-SPECIES CONCENTRATIONS
c--------------------------------------------------------------------

      do i = 1,nComp
         nvec(i) = exp(x(i))
      enddo

c--------------------------------------------------------------------
c     THE PRODUCT SPECIES ARE EITHER:
c
c     (1) Kinetic  --> Leave unchanged
c     (2) Regular equilibrium  --> Use mass-action expression
c     (3) Mineral equilibrium  --> Primary variable
c--------------------------------------------------------------------

      ind = 0
      do i = 1,nR
         if (reactionType(i).eq.0) then
            mineral = nPhase(phaseIdentity(nComp+i)).eq.1
            if (mineral) then
               ind = ind + 1
               nvec(nComp+i) = x(nComp+ind)
            else
               nvec(nComp+i) = prod(i)
            endif
         endif
      enddo


      return
      end
