c--------------------------------------------------------------------
c linearSolver.f
c
c Solve the neq*neq linear system. 
c--------------------------------------------------------------------
c     
c INPUTS:
c 
c job                    Indicates whether system to be solved 
c                          is symmetric (1) or not (0)
c                          Algorithm is chosen accordingly.
c debug                  Logical to compute condition number for debugging
c                          purposes
c outUnit                Output unit number for debugging info
c
c INPUT/OUTPUT:
c
c matrix(neq,neq)        On call: matrix for linear system
c                        On return: destroyed
c
c rhs(neq)               On call: right-hand side for the linear system
c                        On return: the computed solution
c     
c
c OUTPUTS:
c
c rcond                  Approximate condition-number of the problem
c     
c --------------------------------------------------------------------
c     

      subroutine linearSolver(matrix,rhs,rcond,job,debug,
     $     iwk,dimiwk,rwk,dimrwk,outUnit)

      IMPLICIT NONE
      
      integer n,me,mi,neq,niparam,nrparam
      integer dimiwk,dimrwk,iwk(dimiwk),job,info,outUnit
      logical debug
      real*8 matrix(neq,neq),rhs(neq),rwk(dimrwk),rcond

cta 17 July 2002
c      real*8 WLANGE,WLANSY,anorm
      real*8 dlange,dlansy,anorm

      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c
c     ERROR TRAP
c
      if (dimrwk .lt. 4*neq .or. dimiwk .lt. 2*neq)  then
         if(debug .or. dimrwk .lt. 4*neq .or. dimiwk .lt. neq) then
            write(*,*) 'Insufficient work-space in linearSolver'
            STOP
         endif
      endif

c     
c     SOLVE THE LINEAR SYSTEM MATRIX*V = RHS 
c     NOTE : RESULT IS STORED IN RHS. MATRIX IS OVERWRITTEN.
c 
clinpack      ('dgefa(matrix,neq,neq,iwk(1),info)' ,
clinpack      'dsifa(matrix,neq,neq,iwk(1),info)' can also be used.)
c

      if (job.eq.1) then
clinpack        call dsico(matrix,neq,neq,iwk(1),rcond,rwk(1))
clinpack        call dsisl(matrix,neq,neq,iwk(1),rhs)
c
c +++ SLB 10 Feb 98   cant find certain routines on Cray :-(.  leave 'em out!
c
        if(.not.debug) then
cta 17 July 2002
c           call WSYTRF('U',neq,matrix,neq,iwk(1),rwk(1),4*neq,info)
c           call WSYTRS('U',neq,1,matrix,neq,iwk(1),rhs,neq,info)
           call dsytrf('U',neq,matrix,neq,iwk(1),rwk(1),4*neq,info)
           call dsytrs('U',neq,1,matrix,neq,iwk(1),rhs,neq,info)
         else
cCRAY       anorm = WLANSY('1','U',neq,matrix,neq,rwk(1))
cta 17 July 2002
c            call WSYTRF('U',neq,matrix,neq,iwk(1),rwk(1),4*neq,info)
            call dsytrf('U',neq,matrix,neq,iwk(1),rwk(1),4*neq,info)
            if(info.ne.0) then
               write(outUnit,*) "ERROR IN LAPACK(SYTRF):",info
               if(info.gt.0) write(outUnit,*) "  (SINGULAR MATRIX)"
            endif
cta 17 July 2002
c            call WSYTRS('U',neq,1,matrix,neq,iwk(1),rhs,neq,info)
            call dsytrs('U',neq,1,matrix,neq,iwk(1),rhs,neq,info)
            if(info.ne.0)
     $          write(outUnit,*) "ERROR IN LAPACK(SYTRS):",info
cCRAY       call WSYCON('U',neq,matrix,neq,iwk(1),anorm,rcond,rwk(1),
cCRAY $          iwk(neq+1),info)
cCRAY       if(info.ne.0)
cCRAY $          write(outUnit,*) "ERROR IN LAPACK(SYCON):",info
        endif
      else
clinpack        call dgeco(matrix,neq,neq,iwk(1),rcond,rwk(1))
clinpack        call dgesl(matrix,neq,neq,iwk(1),rhs,0)

        if(.not.debug) then
cta 17 July 2002
c           call WGETRF(neq,neq,matrix,neq,iwk(1),info)
c           call WGETRS('N',neq,1,matrix,neq,iwk(1),rhs,neq,info)
           call dgetrf(neq,neq,matrix,neq,iwk(1),info)
           call dgetrs('N',neq,1,matrix,neq,iwk(1),rhs,neq,info)
         else
cCRAY       anorm = WLANGE('1',neq,neq,matrix,neq,rwk(1))
cta 17 July 2002
c            call WGETRF(neq,neq,matrix,neq,iwk(1),info)
            call dgetrf(neq,neq,matrix,neq,iwk(1),info)
            if(info.ne.0) then
               write(outUnit,*) "ERROR IN LAPACK(GETRF):",info
               if(info.gt.0) write(outUnit,*) "  (SINGULAR MATRIX)"
            endif
cta 17 July 2002
c            call WGETRS('N',neq,1,matrix,neq,iwk(1),rhs,neq,info)
            call dgetrs('N',neq,1,matrix,neq,iwk(1),rhs,neq,info)
            if(info.ne.0)
     $          write(outUnit,*) "ERROR IN LAPACK(GETRS):",info
cCRAY       call WGECON('1',neq,matrix,neq,anorm,rcond,
cCRAY $          rwk(1),iwk(1),info)
cCRAY       if(info.ne.0)
cCRAY $          write(outUnit,*) "ERROR IN LAPACK(GECON):",info
        endif
c
clinpack     (NOTE: If condition number is NOT needed, we can instead 
clinpack     factorize using: 'call dgefa(matrix,neq,neq,iwk(1),info)')
c

      endif

      return
      end	
