c--------------------------------------------------------------------
c xsiToMoles.f
c
c Compute molevector based on extents-of reaction
c--------------------------------------------------------------------
c INPUT:
c
c n0(nSpecies)      Some composition satisfying the mass-balance
c xsi(nRxn)         Extents of reaction
c V(nSpecies,nRxn)  Collection of nRxn linearly independent reaction vectors
c                   (Need not be the entire stoichiometric subspace)
c
c OUTPUT:
c
c n(nSpecies)       New composition
c
c ALGORITHM (DAXPY):
c
c n < --- n0 + V*xsi
c
c--------------------------------------------------------------------

      subroutine xsiToMoles(n,n0,xsi,V,nSpecies,nRxn)
      IMPLICIT NONE
      integer nSpecies,nRxn,i,j
      real*8 n(nSpecies),n0(nSpecies),V(nSpecies,nRxn),xsi(nRxn)
c
c     Column-oriented
c

      do i = 1,nSpecies
         n(i) = n0(i)
      enddo

      do j = 1,nRxn
         do i = 1,nSpecies
            n(i)  = n(i) + V(i,j)*xsi(j)
         enddo
      enddo


      return
      end
