c--------------------------------------------------------------------
c nleq.f
c
c  Main routine for the non-linear equations w/ constraints solver
c  for the chemistry problem (entry point for calling routines)
c
c --------------------------------------------------------------------
c
      subroutine nleq(solution,iparam,rparam,iwk,niwk,rwk,nrwk,
     $     relTol,alpha,tauMin,rho,maxIter,nViol,scaleFlag,
     $     interpFlag,verbosity,outUnit,iter,returnflag)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

      integer niwk,nrwk,iwk(niwk),iparam(niparam),iter,
     $     verbosity,outUnit,returnflag,rp_rhs,rp_matrix,
     $     rp_g,rp_NEXT,rwk_SIZE,ip_NEXT,iwk_SIZE,
     $     maxIter,nViol,scaleFlag,interpFlag

      real*8 relTol,alpha,tauMin,rho,solution(neq),rwk(nrwk),
     $     rparam(nrparam)

      logical interpolate,scaling

      scaling     = (scaleFlag.eq.1)
      interpolate = (interpFlag.eq.1)

c--------------------------------------------------------------------     
c     ALLOCATE STORAGE
c--------------------------------------------------------------------     

c     /INTEGER WORK-SPACE/

      ip_NEXT = 1
      iwk_SIZE = niwk - ip_NEXT + 1

c     /REAL WORK-SPACE/

      rp_g = 1
      rp_rhs = rp_g + max(mi,1)
      rp_matrix = rp_rhs + neq
      rp_NEXT =  rp_matrix + neq*neq
      rwk_SIZE = nrwk - rp_NEXT + 1

c--------------------------------------------------------------------     
c     ERROR TRAP
c--------------------------------------------------------------------   

      if ((iwk_SIZE.le.0).or.(rwk_SIZE.le.0)) goto 90
      if (n.le.0) goto 91

c--------------------------------------------------------------------     
c     CALL THE DRIVER
c-------------------------------------------------------------------- 

      call nleqDriver(solution(1),solution(1+n),rwk(rp_g),
     $     rwk(rp_matrix),rwk(rp_rhs),relTol,alpha,tauMin,rho,
     $     maxiter,nViol,interpolate,scaling,iparam,rparam,
     $     iwk(ip_NEXT),iwk_SIZE,rwk(rp_NEXT),rwk_SIZE,verbosity,
     $     outUnit,iter,returnflag)

      goto 100

 90   write(*,*)'OUT OF MEMORY IN NLEQ - STOP RUNNING'
      write(*,*)'Required real workspace:',rp_NEXT
      write(*,*)'Available real workspace:',nrwk
      write(*,*)'Required integer workspace:',ip_NEXT
      write(*,*)'Available integer workspace:',niwk
      stop
 91   write(*,*)'Must have n >= 1 - STOP RUNNING'
      stop
 92   write(*,*)'Must have 1 =< me =< n-1  - STOP RUNNING'
      stop
 93   write(*,*)'Must have mi >= 0 - STOP RUNNING'
      stop

 100  return
      end
