c--------------------------------------------------------------------
c stopTestEval.f
c
c "Standard interface for evaluation of stopping criteria"
c--------------------------------------------------------------------
c
c INPUT 
c
c x                      Independent variables
c iparam                 Integer parameters
c rparam                 Real parameters
c
c OUTPUT:
c
c DESCRIPTION:
c 
c This routine serves the following purposes:
c
c (1) UNPACK integer (iparam) and real (rparam) parameters
c     needed in evaluation of the stopping criteria for
c     a given application.
c 
c (2) Possibly ALLOCATE WORKSPACE (if neccesary)
c
c (3) Evaluate the stopping criterion.
c     Either a user-supplied routine is called for
c     this purpose, or else a standard test is performed
c
c ---------------------------------------------------------------

      subroutine stopTestEval(residual,x,z,g,iparam,rparam,rwk,dimrwk,
     $     iwk,dimiwk,job,tol,constraints,converged,error)

      IMPLICIT NONE
      INTRINSIC sqrt

c     Common areas for NIPS 

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c     Problem-independent declarations

      integer dimiwk,dimrwk,iparam(niparam),iwk(dimiwk),job,i

      real*8 residual(neq),x(n),z(*),g(*),rparam(nrparam),rwk(dimrwk),
     $     error,tol,epsConc,zero,dnrm2

      logical converged,constraints

c     Other declarations

      integer nComp,nSpecies,nRxn,nMin,ip_minList,ip_minIndex,
     $     rp_aa,rp_b,rp_mu0

      parameter (zero=0.0d0)

c --------------------------------------------------------------------
c     COMPUTE THE L-2 NORM OF THE RESIDUAL IN THE 
c     UNPERTURBED KKT-CONDITIONS.
c --------------------------------------------------------------------

      if (constraints) then
         error = zero
         do i = 1,neq-mi
            error = error + residual(i)**2
         enddo
         do i = 1,mi
            error = error + (g(i)*z(i))**2
         enddo
         error = sqrt(error)
      else
         error = dnrm2(neq,residual,1)
      endif

      if (job.eq.0) then

c --------------------------------------------------------------------
c     THIS IS THE DEFAULT BRANCH: STOPPING CRITERION 
c     IS BASED SOLELY ON THE L2-NORM OF THE ERROR.
c --------------------------------------------------------------------

         converged = error.le.tol
         RETURN

      else

c --------------------------------------------------------------------
c     THIS IS THE "PROBLEM-SPECIFIC" BRANCH.
c     GENERAL STOPPING CRITERIA THAT HAVE ACCESS TO 
c     ALL DATA PERTAINING TO THE PROBLEM ARE TESTED.
c     DATA IS FIRST UNPACKED AND THEN PASSED TO A SUITABLE
c     SUBROUTINE FOR THE PURPOSES OF CHECKING THE OPTIMALITY
c     OF THE CURRENT ITERATE.
c --------------------------------------------------------------------

c
c     INTEGER parameters
c
      nComp            = iparam(1) 
      nSpecies         = iparam(2)
      nRxn             = nSpecies - nComp
      nMin             = iparam(3)
      ip_minList       = 4
      ip_minIndex      = ip_minList + nRxn
c
c     REAL parameters
c
      epsConc = rparam(1)
      rp_aa   = 2
      rp_b    = rp_aa + nComp*nSpecies
      rp_mu0  = rp_b + nComp

         
c ---------------------------------------------------------------
c     CALL SPECIFIC STOP-TEST EVALUATION
c ---------------------------------------------------------------

      call stopTest(x,nSpecies,nComp,nRxn,nMin,rparam(rp_aa),
     $     rparam(rp_b),rparam(rp_mu0),iparam(ip_minList),
     $     iparam(ip_minIndex),rwk(1),epsConc,tol,
     $     converged,constraints)
      endif

      return
      end

