c--------------------------------------------------------------------
c chempotDeriv.f
c
c Compute the derivative dmu_(i1)/dn_(i2) where i1, i2 are two
c indicies in the global species list.
c --------------------------------------------------------------------
c     
c INPUTS:
c     
c nvec(nSpecies)           Composition vector
c nTot(pi)                 Vector of total moles in each phase.
c
c phaseIdentity(nSpecies)  Species phase-identity
c nPhase(pi)               Number of species in each phase
c i1                       index 1 into species list
c i2                       index 2 into species list
c nSpecies                 Total number of species
c pi                       Total number of phases
c
c OUTPUTS:
c     
c chempotDeriv             dmu_(i1)/dn_(i2)
c     
c --------------------------------------------------------------------

      real*8 function chempotDeriv(nvec,nTot,phaseIdentity,
     $     nSpecies,pi,nPhase,i1,i2)

      IMPLICIT NONE

      real*8 gasconst,abstemp,pressure
      logical ideal
      common /thermoparam/ gasconst,abstemp,pressure,ideal

      integer nSpecies,pi,phasei1,phasei2,phaseIdentity(nSpecies),
     $     i1,i2,nPhase(pi)

      real*8 nvec(nSpecies),nTot(pi),zero,one
      parameter(zero = 0.0d0, one = 1.0d0)

      phasei1 = phaseIdentity(i1)
      phasei2 = phaseIdentity(i2)

c --------------------------------------------------------------------     
c     TEST FOR SINGLE-SPECIES PHASE
c --------------------------------------------------------------------
     
      if (nphase(phasei1).eq.1.or.nphase(phasei2).eq.1) then
         chempotDeriv = zero
         RETURN
      endif
c --------------------------------------------------------------------     
c     TEST FOR DIFFERENT PHASE
c --------------------------------------------------------------------
     
      if (phasei1.ne.phasei2) then
         chempotDeriv = zero
         RETURN
      endif

c --------------------------------------------------------------------     
c     GENERAL CASE: i1,i2 ARE SPECIES IN THE SAME MULTI-SPECIES PHASE.
c     THE DERIVATIVE HAS THE FORM:
c     
c     dmu_ni1/dni2 = RT(delta_i1,i2/n(i1) - 1/nTot(phase(i1)))
c --------------------------------------------------------------------     

      if (i1.eq.i2) then
         chempotDeriv = one/nvec(i1) - one/nTot(phasei1)
      else
         chempotDeriv = -one/nTot(phasei1)
      endif

      return
      end







