# 1 "gradeval.F"
c--------------------------------------------------------------------
c  gradeval_SF.f
c
c "Standard interface for gradient evaluation"
c --------------------------------------------------------------------
c INPUT 
c
c x                      Independent variables
c iparam                 Integer parameters
c rparam                 Real parameters
c
c OUTPUT:
c 
c gradf, gradh, gradg    Gradients of objective function, constraints.
c
c
c DESCRIPTION:
c 
c This routine serves the following purposes:
c
c (1) UNPACK integer (iparam) and real (rparam) parameters
c     needed in gradient evaluation
c 
c (2) Possibly ALLOCATE WORKSPACE for gradient evaluations
c
c (3) CALL user-supplied routine with the appropriate arguments
c
c ---------------------------------------------------------------
      subroutine gradeval(gradf,gradh,gradg,x,iparam,rparam,
     $     rwk,dimrwk,iwk,dimiwk)

      IMPLICIT NONE

c     COMMON AREAS FOR NIPS 

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c     PROBLEM-INDEPENDENT DECLARATIONS

      integer dimiwk,dimrwk,iparam(niparam),iwk(dimiwk)
      real*8 gradf(n),gradh(n,*),gradg(n,mi),x(n),
     $     rparam(nrparam),rwk(dimrwk)
      
c     OTHER DECLARATIONS


      integer nRxn,nSpecies,nComp,pi,piMulti,piPure,ip_nphase,
     $     ip_phaseIdentity,ip_reactionType,rp_V,rp_mu0,rp_n0,
     $     rptr_nvec,rptr_ntot
      real*8 epsConc
# 60 "gradeval.F"

c ---------------------------------------------------------------
c     UNPACK PARAMETERS
c ---------------------------------------------------------------

c
c     INTEGER parameters
c
      nRxn             = iparam(1) 
      nSpecies         = iparam(2)
      nComp            = iparam(3)
      piMulti          = iparam(4)
      piPure           = iparam(5)
      pi               = piMulti + piPure
      ip_nphase        = 6
      ip_phaseIdentity = ip_nphase + pi
      ip_reactionType  = ip_phaseIdentity + nSpecies
c
c     REAL parameters
c
      epsConc = rparam(1)
      rp_V    = 2
      rp_mu0  = rp_V + nRxn*nSpecies
      rp_n0   = rp_mu0 + nSpecies

c
c     Real work-space
c
      rptr_nvec = 1
      rptr_ntot = rptr_nvec + nSpecies

# 110 "gradeval.F"
 
c ---------------------------------------------------------------
c     CALL SPECIFIC GRADIENT EVALUATION
c ---------------------------------------------------------------

      call grad(gradf,gradh,gradg,x,nSpecies,nRxn,nComp,pi,
     $     iparam(ip_nphase),iparam(ip_phaseIdentity),
     $     iparam(ip_reactionType),rparam(rp_V),rparam(rp_n0),
     $     rparam(rp_mu0),rwk(rptr_nvec),rwk(rptr_ntot))






      return
      end

