# 1 "gradeval.F"
c--------------------------------------------------------------------
c  gradeval_UNSF.f
c
c "Standard interface for gradient evaluation"
c --------------------------------------------------------------------
c INPUT 
c
c x                      Independent variables
c iparam                 Integer parameters
c rparam                 Real parameters
c
c OUTPUT:
c 
c gradf, gradh, gradg    Gradients of objective function, constraints.
c
c
c DESCRIPTION:
c 
c This routine serves the following purposes:
c
c (1) UNPACK integer (iparam) and real (rparam) parameters
c     needed in gradient evaluation
c 
c (2) Possibly ALLOCATE WORKSPACE for gradient evaluations
c
c (3) CALL user-supplied routine with the appropriate arguments
c
c ---------------------------------------------------------------
      subroutine gradeval(gradf,gradh,gradg,x,iparam,rparam,
     $     rwk,dimrwk,iwk,dimiwk)

      IMPLICIT NONE

c     COMMON AREAS FOR NIPS 

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c     PROBLEM-INDEPENDENT DECLARATIONS

      integer dimiwk,dimrwk,iparam(niparam),iwk(dimiwk)
      real*8 gradf(n),gradh(n,*),gradg(n,mi),x(n),
     $     rparam(nrparam),rwk(dimrwk)
      
c     OTHER DECLARATIONS








      integer  nComp,nSpecies,pi,piMulti,piPure,ip_nphase,
     $     ip_phaseIdentity,rp_aa,rp_b,rp_mu0

      real*8 epsConc



c ---------------------------------------------------------------
c     UNPACK PARAMETERS
c ---------------------------------------------------------------
# 92 "gradeval.F"
c
c     INTEGER parameters
c
      nComp            = iparam(1) 
      nSpecies         = iparam(2)
      piMulti          = iparam(3)
      piPure           = iparam(4)
      pi               = piMulti + piPure
      ip_nphase        = 5
      ip_phaseIdentity =  ip_nphase + pi
c     
c     REAL parameters
c     
      epsConc = rparam(1)
      rp_aa   = 2
      rp_b    = rp_aa + nComp*nSpecies
      rp_mu0  = rp_b + nComp

 
c ---------------------------------------------------------------
c     CALL SPECIFIC GRADIENT EVALUATION
c ---------------------------------------------------------------






      call grad(gradf,gradh,gradg,x,nComp,nSpecies,pi,
     $     iparam(ip_nphase),iparam(ip_phaseIdentity),
     $     rparam(rp_aa),rparam(rp_mu0),rwk(1))


      return
      end

