# 1 "hleval.F"
c--------------------------------------------------------------------
c hleval_SF.f
c
c Standard interface for analytic evaluation of the Hessian
c of the Lagrangian.
c -------------------------------------------------------------------- 
c
c INPUT 
c
c x                      Independent variables
c y,z                    Lagrange Multipliers
c iparam                 Integer parameters
c rparam                 Real parameters
c
c OUTPUT:
c 
c hessl
c
c DESCRIPTION:
c 
c This routine serves the following purposes:
c
c (1) UNPACK integer (iparam) and real (rparam) parameters
c     needed in Hessian evaluation
c 
c (2) Possibly ALLOCATE WORKSPACE for Hessian evaluation
c
c (3) CALL user-supplied routine with the appropriate arguments
c
c ---------------------------------------------------------------
      subroutine hleval(hessl,x,y,z,iparam,rparam,rwk,dimrwk,iwk,dimiwk)

      IMPLICIT NONE

c     COMMON AREAS FOR NIPS 

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c     PROBLEM-INDEPENDENT DECLARATIONS

      integer dimiwk,dimrwk,iparam(niparam),iwk(dimiwk)
      real*8 hessl(n,n),x(n),y(*),z(mi),rparam(nrparam),rwk(dimrwk)

c     OTHER DECLARATIONS

      integer nRxn,dimRxn,nSpecies,nComp,pi,piMulti,piPure,
     $     ip_nphase,ip_phaseIdentity,ip_reactionType,
     $     rp_V,rp_mu0,rp_n0,rptr_nvec,rptr_ntot,rptr_A,
     $     rptr_B,rptr_C,rptr_L,rptr_temp,iptr_index

      real*8 epsConc






c ---------------------------------------------------------------
c     UNPACK PARAMETERS 
c ---------------------------------------------------------------

c
c     INTEGER parameters
c
      nRxn             = iparam(1) 
      nSpecies         = iparam(2)
      nComp            = iparam(3)
      piMulti          = iparam(4)
      piPure           = iparam(5)
      pi               = piMulti + piPure
      dimRxn           = nSpecies - nComp
      ip_nphase        = 6
      ip_phaseIdentity = ip_nphase + pi
      ip_reactionType  = ip_phaseIdentity + nSpecies
c
c     INTEGER work-space
c

      iptr_index = 1
      
c
c     REAL parameters
c
      epsConc = rparam(1)
      rp_V    = 2
      rp_mu0  = rp_V + nRxn*nSpecies
      rp_n0   = rp_mu0 + nSpecies
c
c     REAL work-space
c
      rptr_nvec = 1
      rptr_ntot = rptr_nvec + nSpecies
      rptr_A    = rptr_ntot + pi
      rptr_B    = rptr_A + nComp*nComp
      rptr_C    = rptr_B + dimRxn*dimRxn
      rptr_L    = rptr_C + nComp*dimRxn
      rptr_temp = rptr_L + nRxn*dimRxn

# 121 "hleval.F"

c ---------------------------------------------------------------
c     CALL SPECIFIC HESSIAN EVALUATION ROUTINE
c ---------------------------------------------------------------


      call hessian(hessl,x,nSpecies,nComp,dimRxn,nRxn,pi,
     $     iparam(ip_nphase),iparam(ip_phaseIdentity),
     $     iparam(ip_reactionType),rparam(rp_V),rparam(rp_n0),
     $     rwk(rptr_nvec),rwk(rptr_ntot),rwk(rptr_A),rwk(rptr_B),
     $     rwk(rptr_C),rwk(rptr_L),rwk(rptr_temp),iwk(iptr_index))





      return
      end

