# 1 "hleval.F"
c--------------------------------------------------------------------
c hleval_UNSF.f
c
c Standard interface for analytic evaluation of the Hessian
c of the Lagrangian.
c -------------------------------------------------------------------- 
c
c INPUT 
c
c x                      Independent variables
c y,z                    Lagrange Multipliers
c iparam                 Integer parameters
c rparam                 Real parameters
c
c OUTPUT:
c 
c hessl
c
c DESCRIPTION:
c 
c This routine serves the following purposes:
c
c (1) UNPACK integer (iparam) and real (rparam) parameters
c     needed in Hessian evaluation
c 
c (2) Possibly ALLOCATE WORKSPACE for Hessian evaluation
c
c (3) CALL user-supplied routine with the appropriate arguments
c
c ---------------------------------------------------------------
      subroutine hleval(hessl,x,y,z,iparam,rparam,rwk,dimrwk,iwk,dimiwk)

      IMPLICIT NONE

c     COMMON AREAS FOR NIPS 

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c     PROBLEM-INDEPENDENT DECLARATIONS

      integer dimiwk,dimrwk,iparam(niparam),iwk(dimiwk)
      real*8 hessl(n,n),x(n),y(*),z(mi),rparam(nrparam),rwk(dimrwk)

c     OTHER DECLARATIONS
# 54 "hleval.F"
      integer nComp,nSpecies,pi,piMulti,piPure,ip_nphase,
     $     ip_phaseIdentity,rp_mu0,rp_aa,rp_b
      real*8 epsConc


c ---------------------------------------------------------------
c     UNPACK PARAMETERS 
c ---------------------------------------------------------------
# 101 "hleval.F"

c
c     INTEGER parameters
c
      nComp = iparam(1) 
      nSpecies = iparam(2)
      piMulti = iparam(3)
      piPure = iparam(4)
      pi = piMulti + piPure
      ip_nphase = 5
      ip_phaseIdentity =  ip_nphase + pi

c
c     REAL parameters
c
      epsConc = rparam(1)
      rp_aa   = 2
      rp_b    = rp_aa + nComp*nSpecies
      rp_mu0  = rp_b + nComp


c ---------------------------------------------------------------
c     CALL SPECIFIC HESSIAN EVALUATION ROUTINE
c ---------------------------------------------------------------








      call hessian(hessl,x,nSpecies,nComp,pi,iparam(ip_nphase),
     $     iparam(ip_phaseIdentity),rparam(rp_aa),rwk(1))


      return
      end

