c--------------------------------------------------------------------
c l2norm.f
c
c Double-precision function to compute the L2-norm of an often
c recurring vector.
c --------------------------------------------------------------------
c INPUTS:
c
c gradl(n)
c h(me)
c g(mi)      
c z(mi)
c     
c OUTPUTS:
c
c l2norm
c
c ALGORITHM:
c    
c Computes cost:=F_mu(x)'F_mu(x) with F_mu in R^neq given by:
c
c
c                  [  gradl           ]
c                  [  h               ]
c       F(x) =     [  g_i(x)*z_i - mu ]
c
c
c NOTE: With mu=0, return the l2-norm of the RESIDUALS, i.e.
c the ERROR at the current iterate. With mu > 0, it instead returns
c the value of (the square-root of) the COST FUNCTION.
c
c --------------------------------------------------------------------

      real*8 function l2norm(gradl,h,g,z,mu)

      IMPLICIT NONE
      INTRINSIC sqrt

      integer n, me, mi, neq, niparam, nrparam
      common /nips_sizes/n, me, mi, neq, niparam, nrparam

      integer i
      real*8 sum, h(*), g(mi), gradl(n), z(mi), mu, zero
      parameter (zero = 0.0d0)

      sum = zero
      do i = 1,n
         sum = sum + gradl(i)**2
      enddo
      do i = 1,me
         sum = sum + h(i)**2
      enddo
      do i = 1,mi
         sum = sum + (g(i)*z(i) - mu)**2
      enddo

      l2norm = sqrt(sum)

      return
      end	

