c--------------------------------------------------------------------
c nips.f
c
c Main routine for the NIPS constrained optimization package.
c (entry point for calling routines)
c
c --------------------------------------------------------------------

      subroutine nips(solution,iparam,rparam,iwk,niwk,rwk,nrwk,
     $     relTol,alpha,tauMin,rho,maxIter,nViol,
     $     scaleFlag,testSolFlag,interpFlag,hessFlag,
     $     verbosity,outUnit,iter,returnflag)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

      integer maxIter,nViol,scaleFlag,interpFlag,hessFlag,testSolFlag
      real*8  relTol,alpha,tauMin,rho
      logical hanalytic,interpolate,scaling,secSuffTest

      integer i,niwk,nrwk,iwk(niwk),iparam(niparam),
     $     iter,verbosity,outUnit,returnflag,rp_gradf,rp_gradl,
     $     rp_hessl,rp_rhs,rp_matrix,rp_NEXT,rp_eigval,
     $     rp_h,rp_g,rp_gradh,rp_gradg,rwk_SIZE,
     $     ip_bindlist,ip_NEXT,ncActive,ncRank,iwk_SIZE

      real*8 solution(neq),rwk(nrwk),rparam(nrparam)

      scaling     = (scaleFlag.eq.1)
      secSuffTest = (testSolFlag.eq.1)
      interpolate = (interpFlag.eq.1)
      hanalytic   = (hessFlag.eq.1)

c--------------------------------------------------------------------     
c     ALLOCATE STORAGE
c--------------------------------------------------------------------     

c     /INTEGER WORK-SPACE/

      ip_bindlist = 1
      ip_NEXT     = ip_bindlist + mi
      iwk_SIZE    = niwk - ip_NEXT + 1


c     /REAL WORK-SPACE/

      rp_h      = 1
      rp_g      = rp_h + max(1,me)
      rp_gradf  = rp_g + mi
      rp_gradh  = rp_gradf + n
      rp_gradg  = rp_gradh + n*max(1,me)
      rp_gradl  = rp_gradg + n*mi
      rp_hessl  = rp_gradl + n
      rp_rhs    = rp_hessl + n*n 
      rp_matrix = rp_rhs + neq
      rp_eigval = rp_matrix + neq*neq
      rp_NEXT   = rp_eigval + n
      rwk_SIZE  = nrwk - rp_NEXT + 1

c--------------------------------------------------------------------     
c     ERROR TRAP
c--------------------------------------------------------------------     

      if ((iwk_SIZE.le.0).or.(rwk_SIZE.le.0)) goto 90
      if (n.le.0) goto 91
      if (me.lt.0 .or. me.ge.n) goto 92
      if (mi.lt.0) goto 93

c--------------------------------------------------------------------     
c     CALL THE DRIVER
c--------------------------------------------------------------------     

      call nipsDriver(solution(1),solution(1+n),solution(1+n+me),
     $     rwk(rp_h),rwk(rp_g),rwk(rp_gradf),rwk(rp_gradh),
     $     rwk(rp_gradg),rwk(rp_gradl),rwk(rp_hessl),
     $     rwk(rp_matrix),rwk(rp_rhs),relTol,alpha,tauMin,rho,
     $     maxiter,nViol,hanalytic,interpolate,scaling,
     $     secSuffTest,iparam,rparam,iwk(ip_NEXT),iwk_SIZE,
     $     rwk(rp_NEXT),rwk_SIZE,verbosity,outUnit,iter,returnflag)

c--------------------------------------------------------------------     
c     TEST SOLUTION (IF SPECIFIED) AND GENERATE OUTPUT
c--------------------------------------------------------------------     

      if (secSuffTest.and.returnflag.eq.0) then
         if (.not.hanalytic) then
            write(*,1000)
            STOP
         endif
         call secsuff(solution(1),solution(1+n),solution(1+n+me),
     $        iwk(ip_bindlist),ncActive,ncRank,rwk(rp_eigval),iparam,
     $        rparam,rwk(rp_NEXT),rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)

         write(outUnit,2050) ncActive,ncRank
         if (mi.gt.0) write(outUnit,2052)
     $        (i, iwk(ip_bindlist-1+i), i=1,mi)
         write(outUnit,2053)
         write(outUnit,2054)(i, rwk(rp_eigval-1+i), i=1,n-ncRank)
         write(outUnit,*)
      endif

      goto 100

 90   write(*,*)'OUT OF MEMORY IN NIPS - STOP RUNNING'
      write(*,*)'Required real workspace:',rp_NEXT
      write(*,*)'Available real workspace:',nrwk
      write(*,*)'Required integer workspace:',ip_NEXT
      write(*,*)'Available integer workspace:',niwk
      stop
 91   write(*,*)'Must have n >= 1 - STOP RUNNING'
      stop
 92   write(*,*)'Must have 1 =< me =< n-1  - STOP RUNNING'
      stop
 93   write(*,*)'Must have mi >= 0 - STOP RUNNING'
      stop


 100  return

c--------------------------------------------------------------------
c     FORMAT STATEMENTS

 1000 format(/,'ERROR, MUST SPECIFY ANALYTIC HESSIAN',/,
     $     'IN ORDER TO CHECK SECOND-ORDER SUFFICIENCY')
 2050 format(/,'SECOND-ORDER SUFFICIENCY TEST:',/,
     $     '--------------------------------',//,
     $     'Number of active constraints = ........................',
     $     i2,//,
     $     'Approximate rank of gradient set (active) = ...........',
     $     i2,//,
     $     'List of binding/non-binding constraints',/,
     $     'Constraint nr',tr10,'Binding (0=NO, 1=YES)',/)

 2052 format(i2,tr23,i2)
 2053 format(/,'Eigenvalues of the projected hessian:',/,
     $     'nr',tr10,'eigenvalue',/)
 2054 format(i2,tr10,e10.4)


      end
