c--------------------------------------------------------------------
c setupmat.f
c
c Matrix set-up for the perturbed Newton system arising in the 
c Interior-point Method.
c --------------------------------------------------------------------     
c This code sets up the  n_eq - by- n_eq matrix, mat, 
c which has the structure:
c
c
c mat  = [ hessl(n,n)	   -gradh(n,me)     -gradg(n,mi)   ]
c        [ gradh'(me,n)     zeros(me,me)    zeros(me,mi)   ]
c        [ diag(z)          zeros(mi,me)       diag(g)     ]
c
c --------------------------------------------------------------------
c INPUTS:
c      
c hessl(n,n)
c gradh(n,me)
c gradg(n,mi)
c g(mi)
c z(mi)
c
c OUTPUTS:
c
c mat(neq,neq)
c                                                                       
c --------------------------------------------------------------------

      subroutine setupmat(mat,hessl,gradh,gradg,g,z)

      IMPLICIT NONE

      integer n, me, mi, neq, niparam, nrparam
      common /nips_sizes/n, me, mi, neq, niparam, nrparam
      
      integer i,j,iglob,jglob
      real*8 gradh(n,*), gradg(n,mi), g(mi), z(mi),
     $     mat(neq,neq), hessl(n,n), zero, one

      parameter (zero = 0.0d0, one = 1.0d0)

c --------------------------------------------------------------------
c     INITIALIZE
c --------------------------------------------------------------------

      do i = 1,neq
         do j = 1,neq
            mat(i,j) = zero
         enddo
      enddo

c --------------------------------------------------------------------
c      hessl
c --------------------------------------------------------------------

      do i = 1,n
         do j = 1,n
            mat(i,j) = hessl(i,j)
         enddo
      enddo

      if (me.gt.0) then

c --------------------------------------------------------------------
c      -gradh 
c --------------------------------------------------------------------

         do i = 1,n
            do j = 1,me
               iglob = i
               jglob = n + j
               mat(iglob,jglob) = -gradh(i,j)
            enddo
         enddo

c --------------------------------------------------------------------
c      gradh' 
c --------------------------------------------------------------------

         do i = 1,me
            do j = 1,n
               iglob = n + i 
               jglob = j
               mat(iglob,jglob) = gradh(j,i)
            enddo
         enddo

      endif

c --------------------------------------------------------------------
c      -gradg 
c --------------------------------------------------------------------

      do i = 1,n
         do j = 1,mi
            iglob = i
            jglob = n + me + j
            mat(iglob,jglob) = -gradg(i,j)
         enddo
      enddo

c --------------------------------------------------------------------
c      diag(z) 
c --------------------------------------------------------------------

      do i = 1,mi
         do j = 1,n
            iglob = n + me + i
            jglob = j
            mat(iglob,jglob) = z(i)*gradg(j,i)
         enddo
      enddo

c --------------------------------------------------------------------
c      diag(g) 
c --------------------------------------------------------------------

      do i = 1,mi
         iglob = n + me + i
         jglob = n + me + i
         mat(iglob,jglob) = g(i)
      enddo


      return
      end	

