c--------------------------------------------------------------------
c setuprhs.f
c
c Right-hand side set-up for the perturbed Newton system arising in the 
c Interior-point Method.
c --------------------------------------------------------------------
c     
c The following vector of length neq is constructed:
c
c                  [ -gradl   ]
c    residual =    [ -h       ]
c                  [ -g*z + mu]
c
c
c --------------------------------------------------------------------
c INPUTS:
c       
c z(mi)
c h(me)
c g(mi)
c gradl(n) 
c
c     
c OUTPUTS:
c residual(neq)
c 
c --------------------------------------------------------------------
c

      subroutine setuprhs(residual,h,g,gradl,z,mu)

      IMPLICIT NONE

      integer n, me, mi, neq, niparam, nrparam
      common /nips_sizes/n, me, mi, neq, niparam, nrparam

      integer i, ind

      real*8 residual(neq), gradl(n), h(*), g(mi), z(mi), mu

      do i = 1,n
         residual(i) = - gradl(i)
      enddo

      do i = 1,me
         ind  = i + n
         residual(ind) = - h(i)
      enddo

      do i = 1,mi
         ind = i + n + me
         residual(ind) = mu - g(i)*z(i)
      enddo


      return
      end	

