# 1 "stopTesteval.F"
c--------------------------------------------------------------------
c stopTestEval_SF.f
c
c "Standard interface for evaluation of stopping criteria"
c -------------------------------------------------------------------- 
c
c INPUT 
c
c x                      Independent variables
c iparam                 Integer parameters
c rparam                 Real parameters
c
c OUTPUT:
c
c DESCRIPTION:
c 
c This routine serves the following purposes:
c
c (1) UNPACK integer (iparam) and real (rparam) parameters
c     needed in evaluation of the stopping criteria for
c     a given application.
c 
c (2) Possibly ALLOCATE WORKSPACE (if neccesary)
c
c (3) Evaluate the stopping criterion.
c     Either a user-supplied routine is called for
c     this purpose, or else a standard test is performed
c
c ---------------------------------------------------------------


      subroutine stopTestEval(f,h,g,gradf,gradh,gradg,gradl,x,y,z,
     $     iparam,rparam,rwk,dimrwk,iwk,dimiwk,job,converged,tol,error)

      IMPLICIT NONE

c     COMMON AREAS FOR NIPS 

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c     PROBLEM-INDEPENDENT DECLARATIONS

      integer dimiwk,dimrwk,iparam(niparam),iwk(dimiwk),job
      real*8 f,h(*),g(mi),x(n),y(*),z(mi),gradf(n),gradh(n,*),
     $     gradg(n,mi),gradl(n),rparam(nrparam),rwk(dimrwk),error,tol,
     $     epsConc,zero,l2norm
      logical converged

c     OTHER DECLARATIONS

      integer nRxn,dimRxn,nSpecies,nComp,piMulti,piPure,pi,ip_nphase,
     $     ip_phaseIdentity,ip_reactionType,rp_V,rp_mu0,rp_n0,
     $     iptr_index,rptr_nvec,rptr_ntot





      parameter (zero=0.0d0)

c --------------------------------------------------------------------
c     COMPUTE THE L-2 NORM OF THE RESIDUAL IN THE 
c     UNPERTURBED KKT-CONDITIONS.
c --------------------------------------------------------------------


      error = l2norm(gradl,h,g,z,zero)

      if (job.eq.0) then

c --------------------------------------------------------------------
c     THIS IS THE DEFAULT BRANCH: STOPPING CRITERION 
c     IS BASED SOLELY ON THE L2-ERROR.
c --------------------------------------------------------------------

         converged = error.le.tol
         RETURN

      else

c --------------------------------------------------------------------
c     THIS IS THE "PROBLEM-SPECIFIC" BRANCH.
c     GENERAL STOPPING CRITERIA THAT HAVE ACCESS TO 
c     ALL DATA PERTAINING TO THE PROBLEM ARE TESTED.
c     DATA IS FIRST UNPACKED AND THEN PASSED TO A SUITABLE
c     SUBROUTINE FOR THE PURPOSES OF CHECKING THE OPTIMALITY
c     OF THE CURRENT ITERATE.
c --------------------------------------------------------------------


c     
c     INTEGER parameters
c     
         nRxn             = iparam(1) 
         nSpecies         = iparam(2)
         nComp            = iparam(3)
         piMulti          = iparam(4)
         piPure           = iparam(5)
         pi               = piMulti + piPure
         dimRxn           = nSpecies - nComp      

         ip_nphase        = 6
         ip_phaseIdentity = ip_nphase + pi
         ip_reactionType  = ip_phaseIdentity + nSpecies
         
c     
c     INTEGER work-space
c     
         iptr_index = 1
         
c     
c     REAL parameters
c     
         epsConc = rparam(1)
         rp_V   = 2
         rp_mu0 = rp_V + nRxn*nSpecies
         rp_n0  = rp_mu0 + nSpecies

c     
c     REAL work-space
c     

         rptr_nvec = 1
         rptr_ntot = rptr_nvec + nSpecies

# 147 "stopTesteval.F"


c ---------------------------------------------------------------
c     CALL SPECIFIC STOP-TEST EVALUATION
c ---------------------------------------------------------------


         call stopTest(x,gradf,nSpecies,nRxn,dimRxn,nComp,
     $        iparam(ip_reactionType),rparam(rp_V),rparam(rp_n0),
     $        rwk(rptr_nvec),iwk(iptr_index),epsConc,tol,converged)





         
      endif

      return
      end

