# 1 "stopTesteval.F"
c--------------------------------------------------------------------
c stopTestEval_UNSF.f
c
c "Standard interface for evaluation of stopping criteria"
c -------------------------------------------------------------------- 
c
c INPUT 
c
c x                      Independent variables
c iparam                 Integer parameters
c rparam                 Real parameters
c
c OUTPUT:
c
c DESCRIPTION:
c 
c This routine serves the following purposes:
c
c (1) UNPACK integer (iparam) and real (rparam) parameters
c     needed in evaluation of the stopping criteria for
c     a given application.
c 
c (2) Possibly ALLOCATE WORKSPACE (if neccesary)
c
c (3) Evaluate the stopping criterion.
c     Either a user-supplied routine is called for
c     this purpose, or else a standard test is performed
c
c ---------------------------------------------------------------


      subroutine stopTestEval(f,h,g,gradf,gradh,gradg,gradl,x,y,z,
     $     iparam,rparam,rwk,dimrwk,iwk,dimiwk,job,converged,tol,error)

      IMPLICIT NONE

c     COMMON AREAS FOR NIPS 

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

c     PROBLEM-INDEPENDENT DECLARATIONS

      integer dimiwk,dimrwk,iparam(niparam),iwk(dimiwk),job
      real*8 f,h(*),g(mi),x(n),y(*),z(mi),gradf(n),gradh(n,*),
     $     gradg(n,mi),gradl(n),rparam(nrparam),rwk(dimrwk),error,tol,
     $     epsConc,zero,l2norm
      logical converged

c     OTHER DECLARATIONS






      integer nComp,nSpecies,piMulti,piPure,nRxn,rp_aa,rptr_deltaG


      parameter (zero=0.0d0)

c --------------------------------------------------------------------
c     COMPUTE THE L-2 NORM OF THE RESIDUAL IN THE 
c     UNPERTURBED KKT-CONDITIONS.
c --------------------------------------------------------------------


      error = l2norm(gradl,h,g,z,zero)

      if (job.eq.0) then

c --------------------------------------------------------------------
c     THIS IS THE DEFAULT BRANCH: STOPPING CRITERION 
c     IS BASED SOLELY ON THE L2-ERROR.
c --------------------------------------------------------------------

         converged = error.le.tol
         RETURN

      else

c --------------------------------------------------------------------
c     THIS IS THE "PROBLEM-SPECIFIC" BRANCH.
c     GENERAL STOPPING CRITERIA THAT HAVE ACCESS TO 
c     ALL DATA PERTAINING TO THE PROBLEM ARE TESTED.
c     DATA IS FIRST UNPACKED AND THEN PASSED TO A SUITABLE
c     SUBROUTINE FOR THE PURPOSES OF CHECKING THE OPTIMALITY
c     OF THE CURRENT ITERATE.
c --------------------------------------------------------------------

# 128 "stopTesteval.F"
c     
c     INTEGER parameters
c     
         nComp            = iparam(1) 
         nSpecies         = iparam(2)
         piMulti          = iparam(3)
         piPure           = iparam(4)
         nRxn             = nSpecies-nComp
c     
c     REAL parameters
c     
         epsConc = rparam(1)
         rp_aa   = 2
c     
c     REAL work-space
c     

         rptr_deltaG = 1



c ---------------------------------------------------------------
c     CALL SPECIFIC STOP-TEST EVALUATION
c ---------------------------------------------------------------






         call stopTest(x,h,gradf,rparam(rp_aa),rwk(rptr_deltaG),
     $        nSpecies,nComp,nRxn,epsConc,tol,converged)


         
      endif

      return
      end

