c--------------------------------------------------------------------
c yscale.f
c 
c Perform scaling of the linear system
c   Ax = b 
c by applying to it the diagonal scaling matrix D
c The resulting system is A_hat*x = b_hat
c with
c A_hat = D^(-1)*A
c and
c b_hat = D^(-1)*b
c
c It is assumed that the entries of D are "representative" of the problem.
c
c --------------------------------------------------------------------
c INPUTS:
c       
c A(neq,neq)
c b(neq)     
c
c OUTPUTS:
c
c A_hat(neq,neq)
c b_hat(neq)     
c 
c --------------------------------------------------------------------
c

      subroutine yscale(a,b,hessl,h,g,z)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

      integer i,j

      real*8 a(neq,neq),b(neq),hessl(n,n),h(*),g(mi),z(mi),
     $     small,factor
      parameter (small = 1.0d-7)


      do i = 1,n
         factor = dmax1( dabs(hessl(i,i)),small )
         b(i) = b(i)/factor
         do j = 1,neq
            a(i,j) =  a(i,j)/factor 
         enddo
      enddo


      do i = 1,me
         factor = dmax1( dabs(h(i)),small )
         b(n+i) = b(n+i)/factor
         do j = 1,neq
            a(n+i,j) =  a(n+i,j)/factor 
         enddo
      enddo


      do i = 1,mi
         factor = dmax1( g(i)*z(i),small)
         b(n+me+i) = b(n+me+i)/factor
         do j = 1,neq
            a(n+me+i,j) =  a(n+me+i,j)/factor 
         enddo
      enddo


      return
      end	

