c--------------------------------------------------------------------
c stopTest.f
c 
c Evaluate stopping criteria for the interior-point method 
c applied to the stoichiometric formulation of the batch problem.
c --------------------------------------------------------------------
c
c INPUTS:
c
c x(n)                        Extents of reaction
c gradf(n)                    V^T*mu
c n0(nSpecies)                Reference mole-numbers
c V(nSpecies,nRxn)            Stoichiometric matrix
c nVec(nSpecies)              Work array
c
c
c OUTPUTS:
c 
c converged
c --------------------------------------------------------------------

      subroutine stopTest(x,gradf,nSpecies,nRxn,dimRxn,nComp,
     $     reactionType,V,n0,nVec,globalRxnInd,epsConc,tol,converged)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 
          
      integer i,nSpecies,nRxn,dimRxn,nComp,reactionType(dimRxn),
     $      globalRxnInd(nRxn),prodInd,ind
     
      real*8 gradf(n),x(n),V(nSpecies,nRxn),n0(nSpecies),
     $     nVec(nSpecies),epsConc,tol

      logical converged

c --------------------------------------------------------------------
c     SET-UP
c --------------------------------------------------------------------

c     Compute mole vector from extents-of-reaction

      call xsiToMoles(nVec,n0,x,V,nSpecies,nRxn)

c     Initialize array of mapping from list of equilibrium reactions
c     to global list of reactions

      ind = 0
      do i = 1,dimRxn
         if (reactionType(i).eq.0) then
            ind = ind + 1
            globalRxnInd(ind) = i
         endif
      enddo

c --------------------------------------------------------------------
c     CYCLE THROUGH REACTIONS AND CHECK EACH ONE FOR
c     SATISFACTION OF THE EQUILIBRIUM CONDITIONS
c
c     >> THE CONDITIONS HERE ARE REALLY TOO STRINGENT,
c     AND RESULT IN REDUNDANT WORK...>>
c --------------------------------------------------------------------

      converged = .true.

      do i = 1,nRxn
         prodInd = nComp + globalRxnInd(i)

         if (nVec(prodInd).gt.(1.5*epsConc)) then
c
c     Non-binding constraint for product-species 
c     This reaction must be required to be at equilibrium.
c
            converged = converged.and.(dabs(gradf(i)).lt.tol)

         else
c
c     Approximately binding constraint for product-species.
c     It must hold that any increase in this product-species
c     only results in a local increase in the free-energy, i.e.
c     must have gradf(i) > 0 for such a reaction.
c
            converged = converged.and.(gradf(i).ge.-tol)

         endif

      enddo
      
      return
      end	
