c--------------------------------------------------------------------
c hessian.f
c
c Analytic evaluation of the hessian of the lagrangian.
c Non-stoichiometric mole-number version.
c --------------------------------------------------------------------
c INPUTS:
c     
c x(n)                      molenumbers
c aa(nComp,nSpecies)        Formula matrix for the system
c nPhase(pi)                Number of species in each phase
c phaseIdentity(nSpecies)   Integer array of phase-identity for 
c                           all species
c nTot(pi)                  work vector
c
c
c OUTPUTS:
c
c hessl(n,n)                Analytic Hessian of the Lagrangian 
c
c EXTERNAL ROUTINES:
c
c phaseMoles
c chempotDeriv
c
c --------------------------------------------------------------------
      
      subroutine hessian(hessl,x,nSpecies,nComp,pi,nPhase,
     $     phaseIdentity,aa,nTot) 

      IMPLICIT NONE
 
      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 

      real*8 gasconst,abstemp,pressure
      logical ideal
      common /thermoparam/ gasconst,abstemp,pressure,ideal

      integer i,j,nSpecies,nComp,pi,nPhase(pi),
     $     phaseIdentity(nSpecies)

      real*8 hessl(n,n),x(n),aa(nComp,nSpecies),nTot(pi),chempotDeriv

c --------------------------------------------------------------------
c     ERROR-TRAP
c --------------------------------------------------------------------
      if (.not.ideal) then
         write(*,10)
 10      format(/,'ERROR in Hessian',/
     $        'Non-ideal case is not implemented ---',/,
     $        'STOPPING')
         STOP
      endif

c --------------------------------------------------------------------
c     Compute the nSpecies*nSpecies hessian
c --------------------------------------------------------------------

c     Compute total moles for multi-species phases

      call phaseMoles(nTot,x,pi,nSpecies,phaseIdentity)

c     Upper portion of Hessian

      do i = 1,nSpecies
         do j = i,nSpecies
            hessl(i,j) = chempotDeriv(x,nTot,phaseIdentity,
     $           nSpecies,pi,nPhase,i,j)
         enddo
      enddo

c
c     Fill in symmetric part
c

      do i = 2,nSpecies
         do j = 1,i-1
            hessl(i,j) = hessl(j,i)
         enddo
      enddo


      return
      end	


