c--------------------------------------------------------------------
c stopTest.f
c
c Evaluate stopping criteria for the interior-point method 
c applied to the non-stoichiometric formulation of the batch problem.
c --------------------------------------------------------------------

      subroutine stopTest(nVec,h,gradf,A,deltaG,nSpecies,nComp,
     $     nRxn,epsConc,tol,converged)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 
          
      integer i,j,nSpecies,nComp,nRxn
      real*8 gradf(nSpecies),nVec(nSpecies),A(nComp,nSpecies),
     $     h(nComp),deltaG(nRxn),epsConc,tol

      logical converged

c --------------------------------------------------------------------
c     SET-UP
c --------------------------------------------------------------------

      converged = .true.
c
c     Compute the free-energy change deltaG.
c

      do i = 1,nRxn
         deltaG(i) = gradf(nComp+i)
         do j = 1,nComp
            deltaG(i) = deltaG(i) - A(j,nComp+i)*gradf(j)
         enddo
      enddo

c --------------------------------------------------------------------
c     CHECK MASS-BALANCE
c --------------------------------------------------------------------

      do i = 1,nComp
         converged = converged.and.(dabs(h(i)).lt.tol)
      enddo

      if (.not.converged) RETURN

c --------------------------------------------------------------------
c     CYCLE THROUGH REACTIONS AND CHECK EACH ONE FOR
c     SATISFACTION OF THE EQUILIBRIUM CONDITIONS
c --------------------------------------------------------------------

      do i = 1,nRxn

         if (nVec(nComp+i).gt.(1.5*epsConc)) then
c
c     Non-binding constraint for product-species 
c     This reaction must be required to be at equilibrium.
c
            converged = converged.and.(dabs(deltaG(i)).lt.tol)

         else
c
c     Approximately binding constraint for product-species.
c     It must hold that any increase in this product-species
c     only results in a local increase in the free-energy, i.e.
c     must have deltaG(i) > 0 for such a reaction.
c
            converged = converged.and.(deltaG(i).ge.-tol)

         endif

      enddo
      

      return
      end	
