c-----------------------------------------------------------------------
c applyRates.f
c
c Predict new mole-numbers based on rate vector and
c stoichiometric matrix
c ----------------------------------------------------------------------    
c
c Revised 8 jan 97 by Steven Bryant
c
c Purpose of Revision: incorporate Monod style kinetics
c ---------------------------------------------------------------------- 
c
c INPUTS: 
c           
c nvec(nSpecies)     
c rate(nR)                    vector of reaction rates
c V(nSpecies,nR)              stoichiometric matrix 
c
c reactionType(nR)            Current classification of reactions 
c                             (0=equilibrium, 1=kinetic, 2=Monod kinetics)
c
c sigma                       scalar multiplying the rate
c     
c OUTPUT:
c
c newnvec(nSpecies)           predicted new molenumbers
c
c ALGORITHM:
c
c newvec   < --- nvec  + sigma*V*rate
c
c
c-----------------------------------------------------------------------
c     
      subroutine applyRates(nvec,newnvec,rate,V,reactionType,sigma)

      IMPLICIT NONE

      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

      integer i,j,reactionType(nR),prodInd
      real*8 rate(nR),nvec(nSpecies),newnvec(nSpecies),
     $     V(nSpecies,nR),sigma

      call dcopy(nSpecies, nvec, 1, newnvec, 1)

c$$$
c$$$      do j = 1,nR
c$$$         if (reactionType(j).eq.1) then
c$$$            do i = 1,nSpecies
c$$$               newnvec(i) = newnvec(i) + sigma*V(i,j)*rate(j)
c$$$            enddo
c$$$         endif
c$$$      enddo



      do j = 1,nR
cccc         if (reactionType(j).eq.1) then
c SLB 8 Jan 97
          if (reactionType(j).eq.1.or.reactionType(j).eq.2) then
c SLB end
c
c-----------------------------------------------------------------------
c     UPDATE THE KINETIC PRODUCT SPECIES 
c-----------------------------------------------------------------------

            prodInd = nComp + j
            newnvec(prodInd) = newnvec(prodInd) +  sigma*rate(j)

c-----------------------------------------------------------------------
c     UPDATE ALL THE COMPONENTS
c-----------------------------------------------------------------------

            do i = 1,nComp
               newnvec(i) = newnvec(i) + sigma*V(i,j)*rate(j)
            enddo

         endif
      enddo




      return
      end

