c-----------------------------------------------------------------------
c avgRates.f
c
c Compute averaged rates for the time-stepping scheme 
c specified by odeAlgType
c
c INPUTS: 
c           
c r(nR,4)              matrix of rates
c
c OUTPUT:
c
c avgr(nR)              averaged rate
c
c ALGORITHM:
c 
c Currently, Euler-Forward and Runge-Kutta 2 and 4 are implemented.
c 
c If odeAlgType = 0 (Euler-Forward):
c
c avgr   < --- r1
c
c If odeAlgType = 1 (Runge-Kutta 2):
c
c avgr   < --- 0.5*(r1+r2)
c
c If odeAlgType = 2 (Runge-Kutta 4):
c
c avgr   < --- (1/6)*(r1+2*(r2+r3) + r4)
c
c-----------------------------------------------------------------------
c     
      subroutine avgRates(r,avgr,odeAlgType)

      IMPLICIT NONE

      integer nSpecies,nComp,nR
      integer i,odeAlgType
      real*8 r(nR,4),avgr(nR),half,two,sixth

      common /chemSizes/nSpecies,nComp,nR

      parameter (half = 0.5d0,two = 2.0d0, sixth = 1.0d0/6.0d0)

      if (odeAlgType.eq.0) then
         do i = 1,nR
            avgr(i) = r(i,1)
         enddo
      elseif (odeAlgType.eq.1) then
         do i = 1,nR
            avgr(i) = half*(r(i,1) + r(i,2))
         enddo
      elseif (odeAlgType.eq.2) then
         do i = 1,nR
            avgr(i) =  sixth*(r(i,1) + two*(r(i,2)+r(i,3)) + r(i,4))
         enddo
      endif


      return
      end

