c-----------------------------------------------------------------------
c switchModule
c
c Effect a switch of species from KINETIC to EQUILIBRIUM description
c based on a set of heuristics
c-----------------------------------------------------------------------
c
c Revised 8 Jan 97 by Steven Bryant
c
c Purpose of revision: incorporate Monod style kinetics 
c
c Revisions made
c     allow for new reaction classification, reactionType=2
c --------------------------------------------------------------------
c
c INPUTS: 
c           
c n(nSpecies)        current mole-vector
c r(nR)              average rates used for last time-step
c r0(nR)             reference rates 
c idebug
c debugunit
c     
c INPUT/OUTPUT:
c
c reactionType      on call: species classification used for this step
c                   on exit: unchanged if switch=.FALSE.
c                            new classification if switch = .TRUE.
c
c nRQ               on call: number of equilibrium reactions for this step
c                   on exit: increased number (if switch = .TRUE.)
c                            or unchanged (switch=.FALSE.)
c
c nRK               on call: number of kinetic reactions for this step
c                   on exit: decreased number (if switch = .TRUE.)
c                            or unchanged (switch=.FALSE.
c 
c
c OUTPUT:
c
c switch            .TRUE.  if switching occurred
c                   .FALSE. if no switching occurred
c
c ALGORITHM:
c
c Currently, we switch a reaction if 
c
c |r| < frac*|r0|
c
c-----------------------------------------------------------------------
c     
      subroutine switchModule(nVec,r,r0,reactionType,
     $     nRQ,nRK,switch,idebug,debugunit)

      IMPLICIT NONE

      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

      integer reactionType(nR),nRQ,nRK,idebug,debugunit,i
      real*8 nVec(nSpecies),r(nR),r0(nR),zero,fraction,
     $     newRate,refRate,nProd

      logical switch

      parameter(zero = 0.0d0,fraction = 1.0d-7)

      switch = .false.

      if (idebug.ne.0) then 
         write(debugunit,150)
         do i = 1,nR
            write(debugunit,160)r0(i),r(i)
         enddo
      endif

      write(debugunit,*)

      do i = 1,nR
cccc         if (reactionType(i).eq.1) then
c SLB 8 Jan 97
         if (reactionType(i).eq.1.or.reactionType(i).eq.2) then
c SLB end
            newRate = dabs(r(i))
            refRate = dabs(r0(i))
            nProd = nVec(nComp+i)

            if (newRate.lt.fraction*refRate) then

               switch = .true.
               reactionType(i) = 0
               nRK = nRK - 1
               nRQ = nRQ + 1

               if (idebug.ne.0) write(debugunit,175)i,refRate,newRate

            endif
         endif
      enddo

c-----------------------------------------------------------------------
c     FORMAT STATEMENTS

 150  format('Rates in SwitchMod:',/
     $     'Reference rates',10x,'New rates',/
     $     '-----------------------------------------')
 160  format(2x,e10.4,10x,e10.4)
 175  format('Affinity violation',/,
     $     1x,'--->',1x,'Switching reaction',1x,i2,1x,
     $     'to equilibrium description',/,
     $     'Initial affinity =',1x,e10.4,/,
     $     'Final affinity =',1x,e10.4,/)

      return
      end

