c-----------------------------------------------------------------------
c testChemistry.f
c
c Batch test routine for general chemistry
c The mixture in gridblock(1,1,1) is reacted (without flow/transport)
c from tstart to tfinal
c-----------------------------------------------------------------------

      subroutine testChemistry(conc,deltat,tstart,tfinal,nSpecies,
     $     nIparm,nRparm,nCparm,Iparm,Rparm,Cparm,cbar,totCoefInt,
     $     traceType,rwk,dimrwk,iwk,dimiwk,nxc,nyc,nzc,
     $     padxc,padyc,padzc,unit,verbosity)

      IMPLICIT NONE

      integer  nSpecies,nIparm,nRparm,nCparm,nKinetic,dimiwk,dimrwk,
     $     Iparm(nIparm),iwk(dimiwk),error,speciesPtr,
     $     peakIter,cumIter,cumNtTaken,exceptionLoc(3),
     $     traceType(nSpecies),nxc,nyc,nzc,padxc,padyc,padzc,
     $     unit,verbosity,itc,i

      real*4  conc(-padxc:nxc+padxc+1,-padyc:nyc+padyc+1,
     $             -padzc:nzc+padzc+1,nSpecies)

      real*8  deltat,tc,tstart,tfinal,Rparm(nRparm),cbar(nSpecies),
     $     totCoefInt(nSpecies),rwk(dimrwk)

      character*15 Cparm(nCparm)

c-----------------------------------------------------------------------
c     SET-UP
c-----------------------------------------------------------------------

      nKinetic = Iparm(5)
      speciesPtr = Iparm(3)+Iparm(4)+1

      do i=1,nSpecies
         cbar(i) = conc(1,1,1,i)
         totCoefInt(i) = 1.
         traceType(i) = 1
      enddo

      itc = 1
      tc = tstart

c     Generate output-file header and write the initial 
c     concentration table

      write(unit,1)
      call debugOutput('Initial concentrations',cbar,
     $     Cparm(speciesPtr),nSpecies,unit,1)

c-----------------------------------------------------------------------
c       TOP OF TIME-STEPPING LOOP
c-----------------------------------------------------------------------

 100  if ((tfinal-tc).gt.deltat*.250d0) then

         if (nKinetic.eq.0) then
            write(*,*) 'Calling REACTION step'
         else
            write(*,5) 'Calling REACTION step: time step',itc,'t=',tc
         endif

         call reactionStep(cbar,totCoefInt,totCoefInt,
     $        traceType,traceType,deltat,tc,rwk(1),rwk(1),rwk(1),1,1,1,
     $        Iparm,Rparm,nIparm,nRparm,nCparm,rwk(2),dimrwk-1,
     $        iwk(1),dimiwk,verbosity,unit,0,cumNtTaken,cumIter,
     $        peakIter,exceptionLoc,error,Cparm)

         if(error.ne.0) then
            if (error.eq.1) then
               write(*,*)'Exceeded number of nonlinear iterations'
            elseif(error.eq.2) then
               write(*,*)'Exceeded number of back-tracks'
            elseif(error.eq.3) then
               write(*,*) 'Ill-conditioning of linear system'
            elseif(error.eq.4) then
               write(*,*) 'Infeasible initial point'
            elseif(error.eq.5) then
               write(*,*)'Time-integration error'
            endif
            RETURN
         endif

c-----------------------------------------------------------------------
c     OUTPUT NEW CONCENTRATION TABLE, GENERATE MONITOR-OUTPUT
c     AND UPDATE THE TIME-STEP. 
c     IF EQUILIBRIUM SIMULATION, WE ARE DONE.
c-----------------------------------------------------------------------

         itc = itc+1
         if (nKinetic.eq.0) then
            tc = tFinal
            write(unit,10) cumIter
            call debugOutput('Equilibrium concentrations',cbar(1),
     $           Cparm(speciesPtr),nSpecies,unit,1)
         else
            tc = tc + deltat
            write(unit,20) cumNtTaken,cumIter,peakIter,tc
            call debugOutput('New concentrations',cbar(1),
     $           Cparm(speciesPtr),nSpecies,unit,1)
         endif

         GOTO 100
      endif

c-----------------------------------------------------------------------
c     FORMAT STATEMENTS

 1    format(
     $     ' -------------------------------------------------',/,
     $     '|          Result of Batch Simulation             |',/,
     $     ' -------------------------------------------------',///)
     $     

 5    format(/,a,1x,i3,2x,a,1x,f6.1)
 10   format(1x,'---> Successful speciation',/,
     $     6x,'Number of nonlinear iterations=',1x,i3,/)

 20   format('---> Successful time step',/,
     $     5x,'Number of reaction-steps taken=,',1x, i3, /,
     $     5x,'Number of nonlinear iterations=,',1x, i3, /,
     $     5x,'Highest nonlinear iteration count,',1x, i3, //,
     $     'Time=',1x,e10.4,/)

      return
      end
