c ...
c ...   Title:  Matrix and RHS formation for BCFDNF solver
c ...
c     ...  Description: These routines for the Matrix and RHS of the
c ...                discretization of cell centered finite
c ...                differences for a diagonal tensor with a variety
c ...                of boundary conditions
c ...   
c ...   Created:        Thu May 28 10:06:01 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 10:06:01 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: bcfdMake.f,v 1.3 1997-08-20 11:46:25-05 arbogast Exp $
C ...  $RCSfile: bcfdMake.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Bcfdnf/bcfdMake.f,v $
C  
c
c     Routines for building the Cell Centered Finite Difference Matrix and 
c     Right Hand sides
c
c ... MMDD:  $Id: bcfdMake.f,v 1.3 1997-08-20 11:46:25-05 arbogast Exp $
c ... Nested Factorization: bcdfnf v0.1  12/12/90
c
c    Code by Alan Weiser
c    Permeability modifcations by Lawrence Cowsar
c
c
c Block Centered 3D Finite Difference Elliptic Solver using Nested Fatorization
c Provided by Alan Weiser 12/1/90
c
c Tue May  5 10:02:15 1992 Added low order term with coeff. alpha: cowsar
c Tue May 26 16:01:43 1992 Changed the boundary conditions to just input
c                          -D Grad p . n.
c-----------------------------------------------------------------------
c
caw   subroutine setupMAT123(n1,n2,n3,a,k1,k2,k3,kbc,x1,x2,x3)
      subroutine setupMAT123(n1,n2,n3,a,k1,k2,k3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,r,alpha)
caw
      implicit none
      include '../Include/BCtypes.hf'
      logical    lbadr
      integer    n1,n2,n3
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2),
     $        region
      real*8     a(n1,n2,n3,7),
     *           x1(0:n1),x2(0:n2),x3(0:n3),
caw  *           diag,t,h
     *           diag,t,h,r
      real*8 alpha(n1,n2,n3)
      real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
c
c     inputs :
c       rate(.,.,.)    : block source injection rate
c       k?(.,.,.)      : edge permeability in direction ?
c       bcType         : type of boundary conditions
clcc                     bcType = 1 : -D Grad p.n bc[id](.,.,il)
c                        bcType = 2 : pressure in bc[id](.,.,il)
caw
c                        bcType = 4 : Robin B.C. : u.n+rp = bc[id](.,.,il)
c                        bcType = anything else  : no-flow boundary
c       x[id](.)       : cell boundary locations in direction id
c       alpha          : builds matrix for  alpha * p - div K grad p 
c
c     outputs :
c       a, b 
c
      integer i1,i2,i3
c
c ... Inline Function
c
      real*8 xx,yy
      h(xx,yy) = 2.d0*xx*yy/(xx+yy)
caw
      lbadr = .false.
c
      do 10 i3=1,n3
      do 10 i2=1,n2
      do 10 i1=1,n1
        diag = 0.d0
c
        if (i3 .eq.  1) then
          a(i1,i2,i3,1) = 0.d0
          region = bcRegion3(i1,i2,1)
          if (bcType(region) .eq. BCDIRICHLET) then
            t = k3(i1,i2,0) *
     *          (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *          (x3(i3)-x3(i3-1)) * 2.d0
            diag = diag + t
caw
          else if (bcType(region) .eq. BCROBIN 
     $           .and. k3(i1,i2,0) .gt. 0.d0) then
            t = r * (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *         (r * (x3(i3)-x3(i3-1)) * .5d0 / k3(i1,i2,0) + 1.d0)
            if (t .lt. 0.d0) lbadr = .true.
            diag = diag + t
          endif
        else
          a(i1,i2,i3,1) = - k3(i1,i2,i3-1)*
     *                    (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *                    (x3(i3)-x3(i3-2)) * 2.d0
          diag = diag - a(i1,i2,i3,1)
        endif
c
        if (i2 .eq.  1) then
          a(i1,i2,i3,2) = 0.d0
          region = bcRegion2(i1,i3,1)
          if (bcType(region) .eq. BCDIRICHLET) then
            t = k2(i1,0,i3) *
     *          (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *          (x2(i2)-x2(i2-1)) * 2.d0
            diag = diag + t
caw
          else if (bcType(region) .eq. BCROBIN 
     $           .and. k2(i1,0,i3) .gt. 0.d0) then
            t = r * (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *         (r * (x2(i2)-x2(i2-1)) * .5d0 / k2(i1,0,i3) + 1.d0)
            if (t .lt. 0.d0) lbadr = .true.
            diag = diag + t
          endif
        else
          a(i1,i2,i3,2) = - k2(i1,i2-1,i3)  *
     *                    (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *                    (x2(i2)-x2(i2-2)) * 2.d0
          diag = diag - a(i1,i2,i3,2)
        endif
c
        if (i1 .eq.  1) then
          a(i1,i2,i3,3) = 0.d0
          region = bcRegion1(i2,i3,1)
          if (bcType(region) .eq. BCDIRICHLET) then
            t = k1(0,i2,i3) *
     *          (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *          (x1(i1)-x1(i1-1)) * 2.d0
            diag = diag + t
caw
          else if (bcType(region) .eq. BCROBIN 
     $           .and. k1(0,i2,i3) .gt. 0.d0) then
            t = r * (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *         (r * (x1(i1)-x1(i1-1)) * .5d0 / k1(0,i2,i3) + 1.d0)
            if (t .lt. 0.d0) lbadr = .true.
            diag = diag + t
          endif
        else
          a(i1,i2,i3,3) = - k1(i1-1,i2,i3)*
     *                    (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *                    (x1(i1)-x1(i1-2)) * 2.d0
          diag = diag - a(i1,i2,i3,3)
        endif
c
        if (i1 .eq. n1) then
          a(i1,i2,i3,5) = 0.d0
          region = bcRegion1(i2,i3,2)
          if (bcType(region) .eq. BCDIRICHLET) then
            t = k1(n1,i2,i3) *
     *          (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *          (x1(i1)-x1(i1-1)) * 2.d0
            diag = diag + t
caw
          else if (bcType(region) .eq. BCROBIN 
     $           .and. k1(n1,i2,i3) .gt. 0.d0) then
            t = -r * (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *         (-r * (x1(i1)-x1(i1-1)) * .5d0 / k1(n1,i2,i3) + 1.d0)
            if (t .lt. 0.d0) lbadr = .true.
            diag = diag + t
          endif
        else
          a(i1,i2,i3,5) = - k1(i1,i2,i3)*
     *                    (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *                    (x1(i1+1)-x1(i1-1)) * 2.d0
          diag = diag - a(i1,i2,i3,5)
        endif
c
        if (i2 .eq. n2) then
          a(i1,i2,i3,6) = 0.d0
          region = bcRegion2(i1,i3,2)
          if (bcType(region) .eq. BCDIRICHLET) then
            t = k2(i1,n2,i3) *
     *          (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *          (x2(i2)-x2(i2-1)) * 2.d0
            diag = diag + t
caw
          else if (bcType(region) .eq. BCROBIN 
     $           .and. k2(i1,n2,i3) .gt. 0.d0) then
            t = -r * (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *         (-r * (x2(i2)-x2(i2-1)) * .5d0 / k2(i1,n2,i3) + 1.d0)
            if (t .lt. 0.d0) lbadr = .true.
            diag = diag + t
          endif
       else
          a(i1,i2,i3,6) = - k2(i1,i2,i3)*
     *                    (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *                    (x2(i2+1)-x2(i2-1)) * 2.d0
          diag = diag - a(i1,i2,i3,6)
        endif
c
        if (i3 .eq. n3) then
          a(i1,i2,i3,7) = 0.d0
          region = bcRegion3(i1,i2,2)
          if (bcType(region) .eq. BCDIRICHLET) then
            t = k3(i1,i2,n3) *
     *          (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *          (x3(i3)-x3(i3-1)) * 2.d0
            diag = diag + t
caw
          else if (bcType(region) .eq. BCROBIN 
     $           .and. k3(i1,i2,n3) .gt. 0.d0) then
            t = -r * (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *         (-r * (x3(i3)-x3(i3-1)) * .5d0 / k3(i1,i2,n3) + 1.d0)
            if (t .lt. 0.d0) lbadr = .true.
            diag = diag + t
          endif
        else
          a(i1,i2,i3,7) = - k3(i1,i2,i3)*
     *                    (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *                    (x3(i3+1)-x3(i3-1)) * 2.d0
          diag = diag - a(i1,i2,i3,7)
        endif
c
        a(i1,i2,i3,4) = diag + alpha(i1,i2,i3)*(x1(i1)-x1(i1-1))*
     $       (x2(i2)-x2(i2-1))*(x3(i3)-x3(i3-1))
c
   10 continue
caw
      if (lbadr) write(*,*)
     *   ' setupMAT123 warning : r too small.  r = ',r
      return
      end
c
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c Tue May 26 16:01:43 1992 Changed the boundary conditions to just input
c                          -D Grad p . n.  Changes are the (-1)* terms.
c
      subroutine setupRHS123(n1,n2,n3,b,rate,k1,k2,k3,
caw  $     kbc,bc1,bc2,bc3,x1,x2,x3)
     *     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     bc1,bc2,bc3,x1,x2,x3,r,aflag)
      implicit none 
      include '../Include/BCtypes.hf'
      integer    n1,n2,n3,aflag
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8     b(n1,n2,n3),rate(n1,n2,n3),
     *           bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2),
     *           x1(0:n1),x2(0:n2),x3(0:n3),
caw  *           rhs,t
     *           rhs,t,r,a
	real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
c
c     inputs :
c       rate(.,.,.)    : block source injection rate
c       k?(.,.,.)      : edge permeability in direction ?
c       bcType         : boundary condition type
c Tue May 26 16:01:43 1992 Changed the boundary conditions to just input
c                          -D Grad p . n.
c                        bcType = 1 : flux 
c                        bcType = 2 : pressure 
caw
c                        bcType = 4 : Robin B.C. : u.n+rp 
c                        bcType = anything else : no-flow boundary
c       bc[id](.,.,il) : as above
c       x[id](.)       : cell boundary locations in direction id
c
c     outputs :
c       b 
c
      integer i1,i2,i3,region
c
      do 10 i3=1,n3
      do 10 i2=1,n2
      do 10 i1=1,n1
         rhs  = 0.d0
c
         if (i3 .eq.  1) then
            region = bcRegion3(i1,i2,1)
            if (bcType(region) .eq. BCNEUMANN) then
               a = (x1(i1)-x1(i1-1))*(x2(i2)-x2(i2-1))
               if (aflag.eq.VOLUMETRIC) a = 1
               rhs = rhs + (-1)*bc3(i1,i2,1)*a
            else if (bcType(region) .eq. BCDIRICHLET) then
               t = k3(i1,i2,0) *
     *              (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *              (x3(i3)-x3(i3-1)) * 2.d0
               rhs = rhs + t * bc3(i1,i2,1)
caw
          else if (bcType(region) .eq. BCROBIN) then
            if (r .eq. 0.d0) then
              rhs = rhs + bc3(i1,i2,1) *
     *              (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1))
            else if (k3(i1,i2,0) .gt. 0.d0) then
              rhs = rhs + bc3(i1,i2,1) * 
     *              (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *              (r*(x3(i3)-x3(i3-1))*.5d0/k3(i1,i2,0) + 1.d0)
            endif
          endif
        endif
c
        if (i2 .eq.  1) then
           region = bcRegion2(i1,i3,1)
           if (bcType(region) .eq. BCNEUMANN) then
              a = (x1(i1)-x1(i1-1))*(x3(i3)-x3(i3-1))
              if (aflag.eq.VOLUMETRIC) a = 1
              rhs = rhs + (-1)*bc2(i1,i3,1)*a
           else if (bcType(region) .eq. BCDIRICHLET) then
              t = k2(i1,0,i3) *
     *             (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *             (x2(i2)-x2(i2-1)) * 2.d0
              rhs = rhs + t * bc2(i1,i3,1)
caw
           else if (bcType(region) .eq. BCROBIN) then
              if (r .eq. 0.d0) then
                 rhs = rhs + bc2(i1,i3,1) *
     *                (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1))
              else if (k2(i1,0,i3) .gt. 0.d0) then
                 rhs = rhs + bc2(i1,i3,1) * 
     *                (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *                (r*(x2(i2)-x2(i2-1))*.5d0/k2(i1,0,i3) + 1.d0)
              endif
           endif
        endif
c
        if (i1 .eq.  1) then
           region = bcRegion1(i2,i3,1)
           if (bcType(region) .eq. BCNEUMANN) then
              a = (x2(i2)-x2(i2-1))*(x3(i3)-x3(i3-1))
              if (aflag.eq.VOLUMETRIC) a = 1
              rhs = rhs + (-1)*bc1(i2,i3,1)*a
           else if (bcType(region) .eq. BCDIRICHLET) then
              t = k1(0,i2,i3) *
     *             (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *             (x1(i1)-x1(i1-1)) * 2.d0
              rhs = rhs + t * bc1(i2,i3,1)
caw
           else if (bcType(region) .eq. BCROBIN) then
              if (r .eq. 0.d0) then
                 rhs = rhs + bc1(i2,i3,1) *
     *                (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1))
              else if (k1(0,i2,i3) .gt. 0.d0) then
                 rhs = rhs + bc1(i2,i3,1) * 
     *                (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *                (r*(x1(i1)-x1(i1-1))*.5d0/k1(0,i2,i3) + 1.d0)
              endif
           endif
        endif
c
        if (i1 .eq. n1) then
           region = bcRegion1(i2,i3,2)
           if (bcType(region) .eq. BCNEUMANN) then
              a = (x2(i2)-x2(i2-1))*(x3(i3)-x3(i3-1))
              if (aflag.eq.VOLUMETRIC) a = 1
              rhs = rhs - bc1(i2,i3,2)*a
           else if (bcType(region) .eq. BCDIRICHLET) then
              t = k1(n1,i2,i3) *
     *             (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *             (x1(i1)-x1(i1-1)) * 2.d0
              rhs = rhs + t * bc1(i2,i3,2)
caw   
           else if (bcType(region) .eq. BCROBIN) then
              if (r .eq. 0.d0) then
                 rhs = rhs - bc1(i2,i3,2) *
     *                (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1))
              else if (k1(n1,i2,i3) .gt. 0.d0) then
                 rhs = rhs - bc1(i2,i3,2) * 
     *                (x3(i3)-x3(i3-1)) * (x2(i2)-x2(i2-1)) /
     *                (-r*(x1(i1)-x1(i1-1))*.5d0/k1(n1,i2,i3) + 1.d0)
              endif
           endif
        endif
c
        if (i2 .eq. n2) then
           region = bcRegion2(i1,i3,2)
           if (bcType(region) .eq. BCNEUMANN) then
              a = (x1(i1)-x1(i1-1))*(x3(i3)-x3(i3-1))
              if (aflag.eq.VOLUMETRIC) a = 1
              rhs = rhs - bc2(i1,i3,2)*a
          else if (bcType(region) .eq. BCDIRICHLET) then
            t = k2(i1,n2,i3) *
     *          (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *          (x2(i2)-x2(i2-1)) * 2.d0
            rhs = rhs + t * bc2(i1,i3,2)
caw
          else if (bcType(region) .eq. BCROBIN) then
            if (r .eq. 0.d0) then
              rhs = rhs - bc2(i1,i3,2) *
     *              (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1))
            else if (k2(i1,n2,i3) .gt. 0.d0) then
              rhs = rhs - bc2(i1,i3,2) * 
     *              (x3(i3)-x3(i3-1)) * (x1(i1)-x1(i1-1)) /
     *              (-r*(x2(i2)-x2(i2-1))*.5d0/k2(i1,n2,i3) + 1.d0)
            endif
          endif
        endif
c
        if (i3 .eq. n3) then
           region = bcRegion3(i1,i2,2)
           if (bcType(region) .eq. BCNEUMANN) then
              a = (x1(i1)-x1(i1-1))*(x2(i2)-x2(i2-1))
              if (aflag.eq.VOLUMETRIC) a = 1
              rhs = rhs - bc3(i1,i2,2)*a
           else if (bcType(region) .eq. BCDIRICHLET) then
              t = k3(i1,i2,n3) *
     *             (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *             (x3(i3)-x3(i3-1)) * 2.d0
              rhs = rhs + t * bc3(i1,i2,2)
caw   
           else if (bcType(region) .eq. BCROBIN) then
              if (r .eq. 0.d0) then
                 rhs = rhs - bc3(i1,i2,2) *
     *                (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1))
              else if (k3(i1,i2,n3) .gt. 0.d0) then
                 rhs = rhs - bc3(i1,i2,2) * 
     *                (x2(i2)-x2(i2-1)) * (x1(i1)-x1(i1-1)) /
     *                (-r*(x3(i3)-x3(i3-1))*.5d0/k3(i1,i2,n3) + 1.d0)
              endif
           endif
        endif
c
        b(i1,i2,i3) = rhs + rate(i1,i2,i3) * (x1(i1)-x1(i1-1)) *
     *                   (x2(i2)-x2(i2-1)) * (x3(i3)-x3(i3-1))
c
   10 continue
      return
      end
