      subroutine InitEXACT(n1,n2,n3,x1,x2,x3,alpha,D,bcType,
     $     nBCRegions,bcRegion1,bcRegion2,bcRegion3,SDparam,
     $     rwk,rlen,SDptrs)
      implicit none
c
      integer n1,n2,n3,rlen
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),alpha(*)
      real*8 SDparam(*),rwk(rlen)
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 D
      external D
c
      include '../Include/sdsolver.hf'
      include '../Include/BCtypes.hf'
      include '../Include/defines.hf'
c
      integer iSDparam,ik1,ik2,ik3,iamat,im0,info
c
      integer SDptrs(MAXSDPTRS)
c
c ... Internal Common
c
      integer TensorType
      common /parcelcom/ TensorType
      integer Lsing
c
c ... Functions
c
      integer memalr
c
      integer n,m,region
      integer iBmat,m1,m2,m3
c
c ... Copy the parameters
c
      iSDparam = memalr(numSDParam,rwk)
cta 17 July 2002
c      call WCOPY(numSDParam,SDparam,1,rwk(iSDparam),1)
      call dcopy(numSDParam,SDparam,1,rwk(iSDparam),1)
c
c ... Check that the tensor is diagonal
c
      if (TensorType.eq.UserFuncType) then
         call verifyDiagonalTensorFcn(n1,n2,n3,D)
      endif
c
c ... Allocate for the banded matrix
c
      n = n1*n2*n3
      m = min(n1*n2+1,min(n1*n3+1,n2*n3+1))
      im0 = memalr(n*m,rwk)
c
      Lsing = 1
      do region = 1,nBCRegions
         if ((bcType(region).eq.BCDIRICHLET).or.
     $        (bcType(region).eq.BCROBIN)) Lsing = 0
      enddo
c
c ... Make edge transmissibilities
c
      ik1   = memalr((n1+1)*n2*n3,rwk)
      ik2   = memalr(n1*(n2+1)*n3,rwk)
      ik3   = memalr(n1*n2*(n3+1),rwk)
      call Cell2EdgePerm2(n1,n2,n3,D,rwk(ik1),rwk(ik2),rwk(ik3))
c
c ... Setup Finite Difference Matrix
c
      iAmat = memalr(7*n1*n2*n3,rwk)
      call setupMAT123(n1,n2,n3,rwk(iAmat),rwk(ik1),rwk(ik2),rwk(ik3),
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $     0.0d0,alpha)
c
c ... Reorder the matrix to decrease band width and copy to rwk(im0)
c
      iBmat = memalr(7*n1*n2*n3,rwk)
      call ordermat(n1,n2,n3,rwk(iAmat),rwk(iBmat),m1,m2,m3)
      call copyBands(m1,m2,m3,n,m,rwk(iBmat),rwk(im0),Lsing)
c
      call memfreer(iBmat,rwk)
      call memfreer(iAmat,rwk)
c
c ... Factor
c
cta 17 July 2002
c      call WPBTRF('U',n,m-1,rwk(im0),m,info)
      call dpbtrf('U',n,m-1,rwk(im0),m,info)
      if (info.ne.0) then
         write(*,*) 'Matrix is not positive def:',n,info
         call Parcelcrash('WPBTRF() in exact.f fails',-30)
      endif
c
c ... Save the pointers in the pointer array      
c
      SDptrs(1) = iSDparam
      SDptrs(2) = iAmat
      SDptrs(3) = ik1
      SDptrs(4) = ik2
      SDptrs(5) = ik3
      SDptrs(6) = im0
      SDptrs(7) = Lsing
c
      end


      subroutine copyBands(n1,n2,n3,n,m,A,ABD,sing)
      implicit none
      integer n1,n2,n3,n,m
      real*8 A(n,7)
      real*8 ABD(m,n)
      integer sing
c
      integer i
c
cta 17 July 2002
c      call WCOPY(n*m,0.0d0,0,ABD,1)
      call dcopy(n*m,0.0d0,0,ABD,1)
      do i = 1,n
         ABD(m,i)    = A(i,4)
         ABD(m-1,i)  = A(i,3)
         ABD(m-n1,i) = A(i,2)
      enddo
      if (n3.gt.1) then
         do i = 1,n
            ABD(1,i) = A(i,1)
         enddo
      endif
      if (sing .eq. 1) then
         do i = 1,m-1
            ABD(i,n) = 0
         enddo
         ABD(m,n) = 1
      endif
      end
c ...
c ...   Title:  Solve the subdomain problem
c ...
c ...  Description: These routines solve the subdomain problem
c ...   
c ...   Created:        Tue May 26 13:45:11 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Mon Jul 20 11:53:47 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: exact.f,v 1.5 1997-08-20 11:46:27-05 arbogast Exp $
C ...  $RCSfile: exact.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Bcfdnf/exact.f,v $
C  
c ... Version $Id: exact.f,v 1.5 1997-08-20 11:46:27-05 arbogast Exp $
c
c ... Routine: Exact Solve
c
c ... Main function:
c     Solves Local Elliptic Problem using Direct Solver
c
c ... In:
c     outUnit:   Output unit number
c     mynod :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     mynbrSize: The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     bcType:     BCs for 1 - Neumann, 2 - Dirichlet, 3 - Noflow
c     x?    :     point centered grids
c     hx?   :     mesh widths hx?(i) = x?(i)-x?(i-1)
c     bc?   :     Boundary condition values
c     pres  :     Guess for Pressure
c     neuman:     logical denoting if local problem is Neumann problem
c     rate  :     Right hand side of -div A grad (pres) = (rate)
c     k?    :     Edge Permeability matrix
c
c ... Out:
c     pres  :  Pressure values in subdomain
c     bf?   :  Boundary fluxes for this subdomain
c     df?   :  Jump in Boundary fluxes across sumdomain
c     b     :   Right hand side for solving A pres = b
c
      subroutine exactSolve(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,bc1,bc2,bc3,pres,vel1,vel2,vel3,
     $     bf1,bf2,bf3,df1,df2,df3,rate,rwk,SDptrs)
c
      implicit none
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2)
      integer n1,n2,n3
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2),pres(n1,n2,n3)
      real*8  vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3)
      real*8  bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8  df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      real*8  rwk(*)
      real*8  rate(n1,n2,n3)
c
      integer n
      integer memalr
      external memalr
      integer itmp,ib,id,m,ibp,m1,m2,m3
c
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
      real*8 RobinR
      common/Robin/RobinR
c
      include '../Include/BCtypes.hf'
      include '../Include/sdsolver.hf'
c
      integer ik1,ik2,ik3,im0,info
c
      integer SDptrs(MAXSDPTRS)
      integer Lsing
c
c ... Unpack the pointers
c
      ik1      = SDptrs(3) 
      ik2      = SDptrs(4)
      ik3      = SDptrs(5)
      im0      = SDptrs(6)
      Lsing    = SDptrs(7)
c
c ... Setup RightHandSide
c
      ib = memalr(n1*n2*n3,rwk)
      ibp = memalr(n1*n2*n3,rwk)
      call setupRHS123(n1,n2,n3,rwk(ib),
     $     rate,rwk(ik1),rwk(ik2),rwk(ik3),
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     bc1,bc2,bc3,x1,x2,x3,RobinR,VOLUMETRIC)
c
c ... Solve system using the cholesky factored matrix
c
      n = n1*n2*n3
      m = min(n1*n2+1,min(n1*n3+1,n2*n3+1))
      call ordervec(n1,n2,n3,rwk(ib),rwk(ibp),m1,m2,m3)
      if (Lsing .eq. 1) rwk(ibp+n-1) = 0
cta 17 July 2002
c      call WPBTRS('U',n,m-1,1,rwk(im0),m,rwk(ibp),n,info)
      call dpbtrs('U',n,m-1,1,rwk(im0),m,rwk(ibp),n,info)
      call reordervec(n1,n2,n3,pres,rwk(ibp))
c
      call memfreer(ibp,rwk)
      call memfreer(ib,rwk)
c
c ... Compute velocities from pressure and boundary conditions
c
      itmp = 2*max(n1,n2)*max(n2,n3)
      id = memalr(3*itmp,rwk)
      call pres2vel123(n1,n2,n3,pres,bc1,bc2,bc3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     rwk(ik1),rwk(ik2),rwk(ik3),hx1,hx2,hx3,rwk(id),
     $     rwk(id+itmp),rwk(id+2*itmp),vel1,vel2,vel3,VOLUMETRIC)
      call memfreer(id,rwk)
c
c ... Calculate Flux across domain boundaries
c
clccVOLUMETRIC
      call bdryflux123(n1,n2,n3,pres,bc1,bc2,bc3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     rwk(ik1),rwk(ik2),rwk(ik3),hx1,hx2,hx3,
     $     bf1,bf2,bf3,VOLUMETRIC)
c
      call bdrypres123(n1,n2,n3,pres,bc1,bc2,bc3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     rwk(ik1),rwk(ik2),rwk(ik3),hx1,hx2,hx3,
     $     df1,df2,df3,VOLUMETRIC)
c     
      return
      end


      subroutine swapx2x3mat(n1,n2,n3,a,m1,m2,m3,b)
      implicit none
      integer n1,n2,n3,m1,m2,m3
      real*8 a(n1,n2,n3,7),b(m1,m2,m3,7)
      integer i1,i2,i3
c
c .... assert(m1=n1,m2=n3,m3=n2)
c
      do i3 = 1,n3
         do i2 = 1,n2
            do i1 = 1,n1
               b(i1,i3,i2,4) = a(i1,i2,i3,4)
               b(i1,i3,i2,3) = a(i1,i2,i3,3)
               b(i1,i3,i2,2) = a(i1,i2,i3,1)
               b(i1,i3,i2,1) = a(i1,i2,i3,2)
               b(i1,i3,i2,5) = a(i1,i2,i3,5)
               b(i1,i3,i2,6) = a(i1,i2,i3,7)
               b(i1,i3,i2,7) = a(i1,i2,i3,6)
            enddo
         enddo
      enddo
      end
      subroutine swapx1x3mat(n1,n2,n3,a,m1,m2,m3,b)
      implicit none
      integer n1,n2,n3,m1,m2,m3
      real*8 a(n1,n2,n3,7),b(m1,m2,m3,7)
      integer i1,i2,i3
c
c .... assert(m1=n3,m2=n2,m3=n1)
c
      do i3 = 1,n3
         do i2 = 1,n2
            do i1 = 1,n1
               b(i3,i2,i1,4) = a(i1,i2,i3,4)
               b(i3,i2,i1,3) = a(i1,i2,i3,1)
               b(i3,i2,i1,2) = a(i1,i2,i3,2)
               b(i3,i2,i1,1) = a(i1,i2,i3,3)
               b(i3,i2,i1,5) = a(i1,i2,i3,7)
               b(i3,i2,i1,6) = a(i1,i2,i3,6)
               b(i3,i2,i1,7) = a(i1,i2,i3,5)
            enddo
         enddo
      enddo
      end

      subroutine swapx2x3vec(n1,n2,n3,a,m1,m2,m3,b)
      implicit none
      integer n1,n2,n3,m1,m2,m3
      real*8 a(n1,n2,n3),b(m1,m2,m3)
      integer i1,i2,i3
c
c .... assert(m1=n1,m2=n3,m3=n2)
c
      do i3 = 1,n3
         do i2 = 1,n2
            do i1 = 1,n1
               b(i1,i3,i2) = a(i1,i2,i3)
            enddo
         enddo
      enddo
      end

      subroutine swapx1x3vec(n1,n2,n3,a,m1,m2,m3,b)
      implicit none
      integer n1,n2,n3,m1,m2,m3
      real*8 a(n1,n2,n3),b(m1,m2,m3)
      integer i1,i2,i3
c
c .... assert(m1=n3,m2=n2,m3=n1)
c
      do i3 = 1,n3
         do i2 = 1,n2
            do i1 = 1,n1
               b(i3,i2,i1) = a(i1,i2,i3)
            enddo
         enddo
      enddo
      end

      subroutine ordervec(n1,n2,n3,a,b,m1,m2,m3)
      implicit none
      integer n1,n2,n3,m1,m2,m3
      real*8 a(n1,n2,n3),b(*)
c
      if ((n1*n2.le.n1*n3).and.(n1*n2.le.n2*n3)) then
         m1 = n1
         m2 = n2
         m3 = n3
cta 17 July 2002
c         call WCOPY(n1*n2*n3,a,1,b,1)
         call dcopy(n1*n2*n3,a,1,b,1)
      else if (n1*n3.le.n2*n3) then
         m1 = n1
         m2 = n3
         m3 = n2
         call swapx2x3vec(n1,n2,n3,a,n1,n3,n2,b)
      else
         m1 = n3
         m2 = n2
         m3 = n1
         call swapx1x3vec(n1,n2,n3,a,n3,n2,n1,b)
      endif
      return
      end

      subroutine reordervec(n1,n2,n3,a,b)
      implicit none
      integer n1,n2,n3
      real*8 a(n1,n2,n3),b(*)
c
      if ((n1*n2.le.n1*n3).and.(n1*n2.le.n2*n3)) then
cta 17 July 2002
c         call WCOPY(n1*n2*n3,b,1,a,1)
         call dcopy(n1*n2*n3,b,1,a,1)
      else if (n1*n3.le.n2*n3) then
         call swapx2x3vec(n1,n3,n2,b,n1,n2,n3,a)
      else
         call swapx1x3vec(n3,n2,n1,b,n1,n2,n3,a)
      endif
      return
      end

      subroutine ordermat(n1,n2,n3,a,b,m1,m2,m3)
      implicit none
      integer n1,n2,n3,m1,m2,m3
      real*8 a(n1,n2,n3,7),b(*)
c
      if ((n1*n2.le.n1*n3).and.(n1*n2.le.n2*n3)) then
         m1 = n1
         m2 = n2
         m3 = n3
cta 17 July 2002
c         call WCOPY(n1*n2*n3*7,a,1,b,1)
         call dcopy(n1*n2*n3*7,a,1,b,1)
      else if (n1*n3.le.n2*n3) then
         m1 = n1
         m2 = n3
         m3 = n2
         call swapx2x3mat(n1,n2,n3,a,n1,n3,n2,b)
      else
         m1 = n3
         m2 = n2
         m3 = n1
         call swapx1x3mat(n1,n2,n3,a,n3,n2,n1,b)
      endif
      return
      end

      subroutine verifyDiagonalTensorFcn(n1,n2,n3,Dfcn)
      implicit none
      integer n1,n2,n3,i1,i2,i3,i,j
      real*8 Dfcn
      external Dfcn
c
      do i3 = 1,n3
         do i2 = 1,n2
            do i1 = 1,n1
               do j = 1,3
                  do i = 1,3
                     if ((i.ne.j).and.(Dfcn(i,j,i1,i2,i3).ne.0)) then
                        write(*,*) i,j,i1,i2,i3
                        call Parcelcrash(
     $                   'EXACT: Unable to handle a nondiag tensor',
     $                   -30)
                     endif
                  enddo
               enddo
            enddo
         enddo
      enddo
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: exact.f,v 1.5 1997-08-20 11:46:27-05 arbogast Exp $
c
c ... Routine: Cell2EdgePerm2
c
c ... File:  util.f
c
c ... Main function:
c     Computes permeabilities on the edges between cells given the
c     permeability in each cell, i.e. just does harmonic averaging
c
c ... Calls: none
c
c ... In: 
c     n? :    dimensions of grid
c     k  :    cell centered permeablilities
c     k? :    permeabilities in the direction ?
c
c ... Out: none
c     
c ... Interal Variables of note: none
c
c ... Comments:
c
      subroutine Cell2EdgePerm2(n1,n2,n3,D,k1,k2,k3)
      implicit none
      integer n1,n2,n3
      real*8 D,k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
c
      include '../Include/defines.hf'
c
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
c ... Internal Common
c
      integer TensorType
      common /parcelcom/ TensorType     
c
c$$$c ... Kludge for Jim Jennings, ARCO data
c$$$c ... Read in the edge transmissibilities from a file
c$$$c      
c$$$      if ((itest.eq.22).or.(itest.eq.23).or.(itest.eq.24)) then
c$$$         call ReadJennings(n1,n2,n3,k1,k2,k3)
c$$$         return
c$$$      endif
c
      if (TensorType.eq.UserFuncType) then
         call GetEdgeTransFcn(n1,n2,n3,D,k1,k2,k3)
      else if (TensorType.eq.EdgeTransType) then
         call GetEdgeTransEdgeTrans(n1,n2,n3,D,k1,k2,k3)
      else 
         call ParcelCrash('Cell2EdgePerm2: Unknown Tensor Type rep',-30)
      endif
      return
      end
c
      subroutine GetEdgeTransFcn(n1,n2,n3,userDfcn,k1,k2,k3)
      implicit none
      integer n1,n2,n3
      real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
      real*8 userDfcn
      external userDfcn

      integer i1,i2,i3
c
c ... Inline Function
c
      real*8 x,y,h
      h(x,y) = 2.d0*x*y/(x+y)
      
      do 130 i3 = 1,n3
         do 120 i2 = 1,n2
            do 110 i1 = 1,n1-1
               k1(i1,i2,i3) = h(userDfcn(1,1,i1,i2,i3),
     $              userDfcn(1,1,i1+1,i2,i3))
 110        continue
            k1(0,i2,i3) = userDfcn(1,1,1,i2,i3)
            k1(n1,i2,i3) = userDfcn(1,1,n1,i2,i3)
 120     continue
 130  continue
c     
      do 230 i3 = 1,n3
         do 220 i2 = 1,n2-1
            do 210 i1 = 1,n1
               k2(i1,i2,i3) = h(userDfcn(2,2,i1,i2,i3),
     $              userDfcn(2,2,i1,i2+1,i3))
 210        continue
 220     continue
 230  continue
      do 250 i3 = 1,n3
         do 240 i1 = 1,n1
            k2(i1,0,i3) = userDfcn(2,2,i1,1,i3)
            k2(i1,n2,i3) = userDfcn(2,2,i1,n2,i3)
 240     continue
 250  continue
c
      do 330 i3 = 1,n3-1
         do 320 i2 = 1,n2
            do 310 i1 = 1,n1
               k3(i1,i2,i3) = h(userDfcn(3,3,i1,i2,i3),
     $              userDfcn(3,3,i1,i2,i3+1))
 310        continue
 320     continue
 330  continue
      do 350 i2 = 1,n2
         do 340 i1 = 1,n1
            k3(i1,i2,0) = userDfcn(3,3,i1,i2,1)
            k3(i1,i2,n3) = userDfcn(3,3,i1,i2,n3)
 340     continue
 350  continue
c
      return
      end

      subroutine GetEdgeTransEdgeTrans(n1,n2,n3,D,k1,k2,k3)
      implicit none
      integer n1,n2,n3
      real*8 D(*),k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
      integer iet1,iet2,iet3
      iet1 = 1
      iet2 = iet1 + (n1+1)*n2*n3
      iet3 = iet2 + n1*(n2+1)*n3
cta 17 July 2002
c      call WCOPY((n1+1)*n2*n3,D(iet1),1,k1,1)
c      call WCOPY(n1*(n2+1)*n3,D(iet2),1,k2,1)
c      call WCOPY(n1*n2*(n3+1),D(iet3),1,k3,1)
      call dcopy((n1+1)*n2*n3,D(iet1),1,k1,1)
      call dcopy(n1*(n2+1)*n3,D(iet2),1,k2,1)
      call dcopy(n1*n2*(n3+1),D(iet3),1,k3,1)
      return
      end
