c Set up 19 band stencil
c
c     inputs :
c       rate(.,.,.)    : block source injection rate
c       k?(.,.,.)      : edge permeability in direction ?
c       x[id](.)       : cell boundary locations in direction id
c       bcType         : type of boundary conditions
clcc                     bcType = 1 : -D Grad p.n bc[id](.,.,il)
c                        bcType = 2 : pressure in bc[id](.,.,il)
caw
c                        bcType = 4 : Robin B.C. : u.n+rp = bc[id](.,.,il)
c                        bcType = anything else  : no-flow boundary
c       alpha          : builds matrix for  alpha * p - div K grad p
c
c     outputs :
c       a, b
c
      subroutine setup19bd(n1,n2,n3,a,k11,k12,k13,k22,k23,k33,
     $     bc31,bc32,bc21,bc22,bc11,bc12,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $     r,alpha,h1,h2,h3,byd1,byd2,byd3)

      implicit none
      include '../Include/BCtypes.hf'
      integer    n1,n2,n3
      integer    nBCRegions,bcType(nBCRegions),bcRegion1(n2,n3,2),
     $           bcRegion2(n1,n3,2),bcRegion3(n1,n2,2),region
      real*8     a(0:n1+1,0:n2+1,0:n3+1,-9:9),
     *           x1(0:n1),x2(0:n2),x3(0:n3),r
      real*8     alpha(n1,n2,n3)
      real*8     k11(n1,n2,n3),k12(n1,n2,n3),k13(n1,n2,n3),
     $           k22(n1,n2,n3),k23(n1,n2,n3),k33(n1,n2,n3)
      real*8     bc31(n1,n2,5),bc32(n1,n2,5),bc21(n1,n3,5),
     $           bc22(n1,n3,5),bc11(n2,n3,5),bc12(n2,n3,5)
      real*8     h1(0:n1+1),h2(0:n2+1),h3(0:n3+1),
     $           byd1(0:n1),byd2(0:n2),byd3(0:n3)
      integer i,j,k,index,km1,kp1,jm1,jp1,im1,ip1
      real*8 diag,fac
c
c Initialize
c
      call makehs(n1,n2,n3,x1,x2,x3,h1,h2,h3)
      call makebyds(n1,n2,n3,x1,x2,x3,byd1,byd2,byd3)
c
c Initialize matrix bands
c
      do 10 k=0,n3+1
      do 10 j=0,n2+1
      do 10 i=0,n1+1
      do 10 index=-9,9
         a(i,j,k,index) = 0.0d0
 10   continue
c
      do 159 k=1,n3 
      km1 = max(k-1,1)
      do 159 j=1,n2
      jm1 = max(j-1,1)
      do 159 i=1,n1
         a(i,j,k,-9) =
     $        byd2(j-1)*byd3(k-1)*k23(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
     $      + byd3(k-1)*byd2(j-1)*k23(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
 159  continue

      do 158 k=1,n3 
      km1 = max(k-1,1)
      do 158 j=1,n2
      do 158 i=1,n1
      im1 = max(i-1,1)
         a(i,j,k,-8) = 
     $        byd1(i-1)*byd3(k-1)*k13(im1,j,k)*h1(i-1)*(-h2(j))*h3(k)
     $      + byd3(k-1)*byd1(i-1)*k13(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
 158  continue

      do 156 k=1,n3 
      km1 = max(k-1,1)
      do 156 j=1,n2
      do 156 i=1,n1
      ip1 = min(i+1,n1)
         a(i,j,k,-6) =
     $        byd1(i)*byd3(k-1)*k13(ip1,j,k)*h1(i+1)*h2(j)*h3(k)
     $      - byd3(k-1)*byd1(i)*k13(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
 156  continue

      do 155 k=1,n3 
      km1 = max(k-1,1)
      do 155 j=1,n2
      jp1 = min(j+1,n2)
      do 155 i=1,n1
         a(i,j,k,-5) =
     $        byd2(j)*byd3(k-1)*k23(i,jp1,k)*h2(j+1)*h1(i)*h3(k)
     $      - byd3(k-1)*byd2(j)*k23(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
 155  continue

      do 154 k=1,n3 
      do 154 j=1,n2
      jm1 = max(j-1,1)
      do 154 i=1,n1
      im1 = max(i-1,1)
         a(i,j,k,-4) =
     $        byd1(i-1)*byd2(j-1)*k12(im1,j,k)*h1(i-1)*(-h2(j))*h3(k)
     $      + byd2(j-1)*byd1(i-1)*k12(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
 154  continue

      do 152 k=1,n3 
      do 152 j=1,n2
      jm1 = max(j-1,1)
      do 152 i=1,n1
      ip1 = min(i+1,n1)
         a(i,j,k,-2) =
     $        byd1(i)*byd2(j-1)*k12(ip1,j,k)*h1(i+1)*h2(j)*h3(k)
     $      - byd2(j-1)*byd1(i)*k12(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
 152  continue

      do 157 k=1,n3
      km1 = max(k-1,1)
      do 157 j=1,n2
      do 157 i=1,n1
         fac =
     $        2.*byd3(k-1)*byd3(k-1)*k33(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
     $      + 2.*byd3(k-1)*byd3(k-1)*k33(i,j,k)*h3(k)*(-h1(i))*h2(j)
     $      + byd1(i)*byd3(k-1)*k13(i,j,k)*h1(i)*h2(j)*h3(k)
     $      + byd1(i-1)*byd3(k-1)*k13(i,j,k)*h1(i)*(-h2(j))*h3(k)
     $      + byd2(j-1)*byd3(k-1)*k23(i,j,k)*h2(j)*(-h1(i))*h3(k)
     $      + byd2(j)*byd3(k-1)*k23(i,j,k)*h2(j)*h1(i)*h3(k)
         a(i,j,k,0) = a(i,j,k,0) + fac
c
         a(i,j,k,-7) = fac 
     $      + byd3(k-1)*byd1(i)*k13(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
     $      - byd3(k-1)*byd2(j-1)*k23(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
     $      + byd3(k-1)*byd2(j)*k23(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
     $      - byd3(k-1)*byd1(i-1)*k13(i,j,km1)*h3(k-1)*(-h1(i))*h2(j)
 157  continue

      do 153 k=1,n3
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 153 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 153 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
        fac = 
     $       2.*byd2(j-1)*byd2(j-1)*k22(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
     $      + 2.*byd2(j-1)*byd2(j-1)*k22(i,j,k)*h2(j)*(-h1(i))*h3(k)
     $      + byd1(i)*byd2(j-1)*k12(i,j,k)*h1(i)*h2(j)*h3(k)
     $      + byd1(i-1)*byd2(j-1)*k12(i,j,k)*h1(i)*(-h2(j))*h3(k)
     $      + byd3(k)*byd2(j-1)*k23(i,j,k)*h3(k)*h1(i)*h2(j)
     $      + byd3(k-1)*byd2(j-1)*k23(i,j,k)*h3(k)*(-h1(i))*h2(j)
        a(i,j,k,0) = a(i,j,k,0) + fac
c
        a(i,j,k,-3) = fac 
     $      - byd2(j-1)*byd1(i-1)*k12(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
     $      + byd2(j-1)*byd1(i)*k12(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
     $      - byd2(j-1)*byd3(k-1)*k23(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
     $      + byd2(j-1)*byd3(k)*k23(i,jm1,k)*h2(j-1)*(-h1(i))*h3(k)
 153  continue

      do 151 k=1,n3
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 151 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 151 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
        fac = 
     $        2.*byd1(i-1)*byd1(i-1)*k11(im1,j,k)*h1(i-1)*(-h2(j))*h3(k)
     $      + 2.*byd1(i-1)*byd1(i-1)*k11(i,j,k)*h1(i)*(-h2(j))*h3(k)
     $      + byd2(j)*byd1(i-1)*k12(i,j,k)*h2(j)*h1(i)*h3(k)
     $      + byd2(j-1)*byd1(i-1)*k12(i,j,k)*h2(j)*(-h1(i))*h3(k)
     $      + byd3(k)*byd1(i-1)*k13(i,j,k)*h3(k)*h1(i)*h2(j)
     $      + byd3(k-1)*byd1(i-1)*k13(i,j,k)*h3(k)*(-h1(i))*h2(j)
        a(i,j,k,0) = a(i,j,k,0) + fac
c
        a(i,j,k,-1) = fac 
     $      - byd1(i-1)*byd2(j-1)*k12(im1,j,k)*h1(i-1)*(-h2(j))*h3(k)
     $      + byd1(i-1)*byd2(j)*k12(im1,j,k)*h1(i-1)*(-h2(j))*h3(k)
     $      - byd1(i-1)*byd3(k-1)*k13(im1,j,k)*h1(i-1)*(-h2(j))*h3(k)
     $      + byd1(i-1)*byd3(k)*k13(im1,j,k)*h1(i-1)*(-h2(j))*h3(k)
 151   continue

      do 51 k=1,n3
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 51 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 51 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
        fac = 
     $        2.*byd1(i)*byd1(i)*k11(i,j,k)*h1(i)*(-h2(j))*h3(k)
     $      + 2.*byd1(i)*byd1(i)*k11(ip1,j,k)*h1(i+1)*(-h2(j))*h3(k)
     $      - byd1(i)*byd2(j-1)*k12(i,j,k)*h1(i)*(-h2(j))*h3(k)
     $      + byd1(i)*byd2(j)*k12(i,j,k)*h1(i)*(-h2(j))*h3(k)
     $      - byd1(i)*byd3(k-1)*k13(i,j,k)*h1(i)*(-h2(j))*h3(k)
     $      + byd1(i)*byd3(k)*k13(i,j,k)*h1(i)*(-h2(j))*h3(k)
        a(i,j,k,0) = a(i,j,k,0) + fac
c
        a(i,j,k,1) = fac 
     $      + byd2(j)*byd1(i)*k12(ip1,j,k)*h2(j)*h1(i+1)*h3(k)
     $      + byd2(j-1)*byd1(i)*k12(ip1,j,k)*h2(j)*(-h1(i+1))*h3(k)
     $      + byd3(k)*byd1(i)*k13(ip1,j,k)*h3(k)*h1(i+1)*h2(j)
     $      + byd3(k-1)*byd1(i)*k13(ip1,j,k)*h3(k)*(-h1(i+1))*h2(j)
 51   continue

      do 53 k=1,n3
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 53 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 53 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
        fac = 
     $        2.*byd2(j)*byd2(j)*k22(i,j,k)*h2(j)*(-h1(i))*h3(k)
     $      + 2.*byd2(j)*byd2(j)*k22(i,jp1,k)*h2(j+1)*(-h1(i))*h3(k)
     $      - byd2(j)*byd1(i-1)*k12(i,j,k)*h2(j)*(-h1(i))*h3(k)
     $      + byd2(j)*byd1(i)*k12(i,j,k)*h2(j)*(-h1(i))*h3(k)
     $      - byd2(j)*byd3(k-1)*k23(i,j,k)*h2(j)*(-h1(i))*h3(k)
     $      + byd2(j)*byd3(k)*k23(i,j,k)*h2(j)*(-h1(i))*h3(k)
        a(i,j,k,0) = a(i,j,k,0) + fac
c
        a(i,j,k,3) = fac 
     $      + byd1(i)*byd2(j)*k12(i,jp1,k)*h1(i)*h2(j+1)*h3(k)
     $      + byd1(i-1)*byd2(j)*k12(i,jp1,k)*h1(i)*(-h2(j+1))*h3(k)
     $      + byd3(k)*byd2(j)*k23(i,jp1,k)*h3(k)*h1(i)*h2(j+1)
     $      + byd3(k-1)*byd2(j)*k23(i,jp1,k)*h3(k)*(-h1(i))*h2(j+1)
 53   continue

      do 57 k=1,n3
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 57 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 57 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
         fac =    
     $        2.*byd3(k)*byd3(k)*k33(i,j,k)*h3(k)*(-h1(i))*h2(j)
     $      + 2.*byd3(k)*byd3(k)*k33(i,j,kp1)*h3(k+1)*(-h1(i))*h2(j)
     $      + byd3(k)*byd1(i)*k13(i,j,k)*h3(k)*(-h1(i))*h2(j)
     $      - byd3(k)*byd2(j-1)*k23(i,j,k)*h3(k)*(-h1(i))*h2(j)
     $      + byd3(k)*byd2(j)*k23(i,j,k)*h3(k)*(-h1(i))*h2(j)
     $      - byd3(k)*byd1(i-1)*k13(i,j,k)*h3(k)*(-h1(i))*h2(j)
         a(i,j,k,0) = a(i,j,k,0) + fac
c
         a(i,j,k,7) = fac
     $      + byd1(i)*byd3(k)*k13(i,j,kp1)*h1(i)*h2(j)*h3(k+1)
     $      + byd1(i-1)*byd3(k)*k13(i,j,kp1)*h1(i)*(-h2(j))*h3(k+1)
     $      + byd2(j-1)*byd3(k)*k23(i,j,kp1)*h2(j)*(-h1(i))*h3(k+1)
     $      + byd2(j)*byd3(k)*k23(i,j,kp1)*h2(j)*h1(i)*h3(k+1)
 57   continue

      do 50 k=1,n3
      do 50 j=1,n2
      do 50 i=1,n1
         a(i,j,k,0) = -a(i,j,k,0) + alpha(i,j,k)*h1(i)*h2(j)*h3(k)
 50   continue

      do 52 k=1,n3 
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 52 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 52 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
         a(i,j,k,2) =
     $        byd1(i-1)*byd2(j)*k12(i,jp1,k)*h1(i)*h2(j+1)*h3(k)
     $      - byd2(j)*byd1(i-1)*k12(im1,j,k)*h2(j)*(-h1(i-1))*h3(k)
 52   continue

      do 54 k=1,n3 
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 54 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 54 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
         a(i,j,k,4) =
     $        byd1(i)*byd2(j)*k12(i,jp1,k)*h1(i)*(-h2(j+1))*h3(k)
     $      + byd2(j)*byd1(i)*k12(ip1,j,k)*h2(j)*(-h1(i+1))*h3(k)
 54   continue

      do 55 k=1,n3 
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 55 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 55 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
         a(i,j,k,5) =
     $        byd2(j-1)*byd3(k)*k23(i,j,kp1)*h2(j)*h1(i)*h3(k+1)
     $      - byd3(k)*byd2(j-1)*k23(i,jm1,k)*h3(k)*(-h1(i))*h2(j-1)
 55   continue

      do 56 k=1,n3 
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 56 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 56 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
         a(i,j,k,6) =
     $        byd1(i-1)*byd3(k)*k13(i,j,kp1)*h1(i)*h2(j)*h3(k+1)
     $      - byd3(k)*byd1(i-1)*k13(im1,j,k)*h3(k)*(-h1(i-1))*h2(j)
 56   continue

      do 58 k=1,n3 
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 58 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 58 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
         a(i,j,k,8) =
     $        byd1(i)*byd3(k)*k13(i,j,kp1)*h1(i)*(-h2(j))*h3(k+1)
     $      + byd3(k)*byd1(i)*k13(ip1,j,k)*h3(k)*(-h1(i+1))*h2(j)
 58   continue

      do 59 k=1,n3 
      km1 = max(k-1,1)
      kp1 = min(k+1,n3) 
      do 59 j=1,n2
      jm1 = max(j-1,1)
      jp1 = min(j+1,n2)
      do 59 i=1,n1
      im1 = max(i-1,1)
      ip1 = min(i+1,n1)
         a(i,j,k,9) =
     $        byd2(j)*byd3(k)*k23(i,j,kp1)*h2(j)*(-h1(i))*h3(k+1)
     $      + byd3(k)*byd2(j)*k23(i,jp1,k)*h3(k)*(-h1(i))*h2(j+1)
 59   continue
c
c Now for boundary conditions...
c
c bottom face
c
      do j = 1,n2
         do i = 1,n1
            region = bcRegion3(i,j,1)
            if ((bcType(region).eq.BCROBIN)
     $              .OR.(bcType(region).eq.BCNEUMANN)
     $              .OR.(bcType(region).eq.BCNOFLOW)) then
               a(i,j,0,0) = 2.*byd3(0)*byd3(0)*
     $              k33(i,j,1)*h3(1)*h1(i)*h2(j)
               a(i,j,0,5) = byd3(0)*byd2(j-1)*
     $              k23(i,j,1)*h3(1)*h1(i)*h2(j)
               a(i,j,0,6) = byd3(0)*byd1(i-1)*
     $              k13(i,j,1)*h3(1)*h1(i)*h2(j)
               a(i,j,0,8) = -byd3(0)*byd1(i)*
     $              k13(i,j,1)*h3(1)*h1(i)*h2(j)
               a(i,j,0,9) = -byd3(0)*byd2(j)*
     $              k23(i,j,1)*h3(1)*h1(i)*h2(j)
               a(i,j,0,7) = -a(i,j,0,0) -
     $              a(i,j,0,5) - a(i,j,0,6) -
     $              a(i,j,0,8) - a(i,j,0,9)
               if (bcType(region).eq.BCROBIN) then
                  a(i,j,0,0) = a(i,j,0,0) - r*h1(i)*h2(j)
               endif
            endif
         enddo
      enddo
c
c top face
c            
      do j = 1,n2
         do i = 1,n1
            region = bcRegion3(i,j,2)
            if ((bcType(region).eq.BCROBIN)
     $              .or.(bcType(region).eq.BCNEUMANN)
     $              .OR. (bcType(region).eq.BCNOFLOW)) then
               a(i,j,n3+1,0) = 2.*byd3(n3)*byd3(n3)*
     $              k33(i,j,n3)*h3(n3)*h1(i)*h2(j)
               a(i,j,n3+1,-5) = byd3(n3)*byd2(j)*
     $              k23(i,j,n3)*h3(n3)*h1(i)*h2(j)
               a(i,j,n3+1,-6) = byd3(n3)*byd1(i)*
     $              k13(i,j,n3)*h3(n3)*h1(i)*h2(j)
               a(i,j,n3+1,-8) = -byd3(n3)*byd1(i-1)*
     $              k13(i,j,n3)*h3(n3)*h1(i)*h2(j)
               a(i,j,n3+1,-9) = -byd3(n3)*byd2(j-1)*
     $              k23(i,j,n3)*h3(n3)*h1(i)*h2(j)
               a(i,j,n3+1,-7) = -a(i,j,n3+1,0) -
     $              a(i,j,n3+1,-5) - a(i,j,n3+1,-6) -
     $              a(i,j,n3+1,-8) - a(i,j,n3+1,-9)
               if (bcType(region).eq.BCROBIN) then
                  a(i,j,n3+1,0) = a(i,j,n3+1,0)
     $                 - r*h1(i)*h2(j)
               endif
            endif
         enddo
      enddo
c
c front face
c
      do k = 1,n3
         do i = 1,n1
            region = bcRegion2(i,k,1)
            if ((bcType(region).eq.BCROBIN)
     $              .or.(bcType(region).eq.BCNEUMANN)
     $              .OR. (bcType(region).eq.BCNOFLOW)) then
               a(i,0,k,0) = 2.*byd2(0)*byd2(0)*
     $              k22(i,1,k)*h2(1)*h1(i)*h3(k)
               a(i,0,k,-5) = byd2(0)*byd3(k-1)*
     $              k23(i,1,k)*h2(1)*h1(i)*h3(k)
               a(i,0,k,2) = byd2(0)*byd1(i-1)*
     $              k12(i,1,k)*h2(1)*h1(i)*h3(k)
               a(i,0,k,9) = -byd2(0)*byd3(k)*
     $              k23(i,1,k)*h2(1)*h1(i)*h3(k)
               a(i,0,k,4) = -byd2(0)*byd1(i)*
     $              k12(i,1,k)*h2(1)*h1(i)*h3(k)
               a(i,0,k,3) = -a(i,0,k,0) -
     $              a(i,0,k,-5) - a(i,0,k,2) -
     $              a(i,0,k,4) - a(i,0,k,9)
               if (bcType(region).eq.BCROBIN) then
                  a(i,0,k,0) = a(i,0,k,0) - r*h1(i)*h3(k)
               endif
            endif
         enddo
      enddo
c
c back face
c
      do k = 1,n3
         do i = 1,n1
            region = bcRegion2(i,k,2)
            if ((bcType(region).eq.BCROBIN)
     $              .or.(bcType(region).eq.BCNEUMANN)
     $              .OR. (bcType(region).eq.BCNOFLOW)) then
               a(i,n2+1,k,0) = 2.*byd2(n2)*byd2(n2)*
     $              k22(i,n2,k)*h2(n2)*h1(i)*h3(k)
               a(i,n2+1,k,5) = byd2(n2)*byd3(k)*
     $              k23(i,n2,k)*h2(n2)*h1(i)*h3(k)
               a(i,n2+1,k,-2) = byd2(n2)*byd1(i)*
     $              k12(i,n2,k)*h2(n2)*h1(i)*h3(k)
               a(i,n2+1,k,-9) = -byd2(n2)*byd3(k-1)*
     $              k23(i,n2,k)*h2(n2)*h1(i)*h3(k)
               a(i,n2+1,k,-4) = -byd2(n2)*byd1(i-1)*
     $              k12(i,n2,k)*h2(n2)*h1(i)*h3(k)
               a(i,n2+1,k,-3) = -a(i,n2+1,k,0) -
     $              a(i,n2+1,k,5) - a(i,n2+1,k,-2) -
     $              a(i,n2+1,k,-4) - a(i,n2+1,k,-9)
               if (bcType(region).eq.BCROBIN) then
                  a(i,n2+1,k,0) = a(i,n2+1,k,0)
     $                 - r*h1(i)*h3(k)
               endif
            endif
         enddo
      enddo
c
c left face
c
      do k = 1,n3
         do j = 1,n2
            region = bcRegion1(j,k,1)
            if ((bcType(region).eq.BCROBIN)
     $              .or.(bcType(region).eq.BCNEUMANN)
     $              .OR. (bcType(region).eq.BCNOFLOW)) then
               a(0,j,k,0) = 2.*byd1(0)*byd1(0)*
     $              k11(1,j,k)*h1(1)*h2(j)*h3(k)
               a(0,j,k,-6) = byd1(0)*byd3(k-1)*
     $              k13(1,j,k)*h1(1)*h2(j)*h3(k)
               a(0,j,k,-2) = byd1(0)*byd2(j-1)*
     $              k12(1,j,k)*h1(1)*h2(j)*h3(k)
               a(0,j,k,4) = -byd1(0)*byd2(j)*
     $              k12(1,j,k)*h1(1)*h2(j)*h3(k)
               a(0,j,k,8) = -byd1(0)*byd3(k)*
     $              k13(1,j,k)*h1(1)*h2(j)*h3(k)
               a(0,j,k,1) = -a(0,j,k,0) -
     $              a(0,j,k,-6) - a(0,j,k,-2) -
     $              a(0,j,k,4) - a(0,j,k,8)
               if (bcType(region).eq.BCROBIN) then
                  a(0,j,k,0) = a(0,j,k,0) - r*h2(j)*h3(k)
               endif
            endif
         enddo
      enddo
c
c right face
c
      do k = 1,n3
         do j = 1,n2
            region = bcRegion1(j,k,2)
            if ((bcType(region).eq.BCROBIN)
     $              .or.(bcType(region).eq.BCNEUMANN)
     $              .OR. (bcType(region).eq.BCNOFLOW)) then
               a(n1+1,j,k,0) = 2.*byd1(n1)*byd1(n1)*
     $              k11(n1,j,k)*h1(n1)*h2(j)*h3(k)
               a(n1+1,j,k,6) = byd1(n1)*byd3(k)*
     $              k13(n1,j,k)*h1(n1)*h2(j)*h3(k)
               a(n1+1,j,k,2) = byd1(n1)*byd2(j)*
     $              k12(n1,j,k)*h1(n1)*h2(j)*h3(k)
               a(n1+1,j,k,-4) = -byd1(n1)*byd2(j-1)*
     $              k12(n1,j,k)*h1(n1)*h2(j)*h3(k)
               a(n1+1,j,k,-8) = -byd1(n1)*byd3(k-1)*
     $              k13(n1,j,k)*h1(n1)*h2(j)*h3(k)
               a(n1+1,j,k,-1) = -a(n1+1,j,k,0) -
     $              a(n1+1,j,k,6) - a(n1+1,j,k,2) -
     $              a(n1+1,j,k,-4) - a(n1+1,j,k,-8)
               if (bcType(region).eq.BCROBIN) then
                  a(n1+1,j,k,0) = a(n1+1,j,k,0)
     $                 - r*h2(j)*h3(k)
               endif
            endif
         enddo
      enddo
c
c Dirichlet BC
c
c ... bottom face
c
      do j = 1,n2
         do i = 1,n1
            region = bcRegion3(i,j,1)
            if (bcType(region) .eq. BCDIRICHLET) then
               a(i,j,0,0) = 1.0d0*h1(i)*h2(j)
c
               bc31(i,j,1) = a(i,j-1,1,-5)
               bc31(i,j,2) = a(i-1,j,1,-6)
               bc31(i,j,3) = a(i,j,1,-7)
               bc31(i,j,4) = a(i+1,j,1,-8)
               bc31(i,j,5) = a(i,j+1,1,-9)
c
               a(i,j-1,1,-5) = 0.0
               a(i-1,j,1,-6) = 0.0
               a(i,j,1,-7) = 0.0
               a(i+1,j,1,-8) = 0.0
               a(i,j+1,1,-9) = 0.0
            endif
         enddo
      enddo
c
c top face
c            
      do j = 1,n2
         do i = 1,n1
            region = bcRegion3(i,j,2)
            if (bcType(region).eq.BCDIRICHLET) then
               a(i,j,n3+1,0) = 1.0d0*h1(i)*h2(j)
c
               bc32(i,j,1) = a(i,j+1,n3,5)
               bc32(i,j,2) = a(i+1,j,n3,6)
               bc32(i,j,3) = a(i,j,n3,7)
               bc32(i,j,4) = a(i-1,j,n3,8)
               bc32(i,j,5) = a(i,j-1,n3,9)
c
               a(i,j+1,n3,5) = 0.0
               a(i+1,j,n3,6) = 0.0
               a(i,j,n3,7) = 0.0
               a(i-1,j,n3,8) = 0.0
               a(i,j-1,n3,9) = 0.0
            endif
         enddo
      enddo
c
c front face
c
      do k = 1,n3
         do i = 1,n1
            region = bcRegion2(i,k,1)
            if (bcType(region).eq.BCDIRICHLET) then
               a(i,0,k,0) = 1.0d0*h1(i)*h3(k)
c
               bc21(i,k,1) = a(i,1,k+1,-9)
               bc21(i,k,2) = a(i+1,1,k,-4)
               bc21(i,k,3) = a(i,1,k,-3)
               bc21(i,k,4) = a(i-1,1,k,-2)
               bc21(i,k,5) = a(i,1,k-1, 5)
c
               a(i,1,k+1,-9) = 0.0
               a(i+1,1,k,-4) = 0.0
               a(i,1,k,-3) = 0.0
               a(i-1,1,k,-2) = 0.0
               a(i,1,k-1, 5) = 0.0
            endif
         enddo
      enddo
c
c back face
c
      do k = 1,n3
         do i = 1,n1
            region = bcRegion2(i,k,2)
            if (bcType(region).eq.BCDIRICHLET) then
               a(i,n2+1,k,0) = 1.0d0*h1(i)*h3(k)
c
               bc22(i,k,1) = a(i,n2,k-1,9)
               bc22(i,k,2) = a(i-1,n2,k,4)
               bc22(i,k,3) = a(i,n2,k,3)
               bc22(i,k,4) = a(i+1,n2,k,2)
               bc22(i,k,5) = a(i,n2,k+1,-5)
c
               a(i,n2,k-1,9) = 0.0
               a(i-1,n2,k,4) = 0.0
               a(i,n2,k,3) = 0.0
               a(i+1,n2,k,2) = 0.0
               a(i,n2,k+1,-5) = 0.0
            endif
         enddo
      enddo
c
c left face
c
      do k = 1,n3
         do j = 1,n2
            region = bcRegion1(j,k,1)
            if (bcType(region).eq.BCDIRICHLET) then
               a(0,j,k,0) = 1.0d0*h2(j)*h3(k)
c
               bc11(j,k,1) = a(1,j,k+1,-8)
               bc11(j,k,2) = a(1,j+1,k,-4)
               bc11(j,k,3) = a(1,j,k,-1)
               bc11(j,k,4) = a(1,j-1,k, 2)
               bc11(j,k,5) = a(1,j,k-1, 6)
c
               a(1,j,k+1,-8) = 0.0
               a(1,j+1,k,-4) = 0.0
               a(1,j,k,-1) = 0.0
               a(1,j-1,k,2) = 0.0
               a(1,j,k-1,6) = 0.0
            endif
         enddo
      enddo
c
c right face
c
      do k = 1,n3
         do j = 1,n2
            region = bcRegion1(j,k,2)
            if (bcType(region).eq.BCDIRICHLET) then
               a(n1+1,j,k,0) = 1.0d0*h2(j)*h3(k)
c
               bc12(j,k,1) = a(n1,j,k-1,8)
               bc12(j,k,2) = a(n1,j-1,k,4)
               bc12(j,k,3) = a(n1,j,k,1)
               bc12(j,k,4) = a(n1,j+1,k,-2)
               bc12(j,k,5) = a(n1,j,k+1,-6)
c
               a(n1,j,k-1,8) = 0.0
               a(n1,j-1,k,4) = 0.0
               a(n1,j,k,1) = 0.0
               a(n1,j+1,k,-2) = 0.0
               a(n1,j,k+1,-6) = 0.0
            endif
         enddo
      enddo
c
c Account for edge unknowns
c
         do i = 0,n1+1
            a(i,0,0,0) = 1.0
            a(i,0,n3+1,0) = 1.0
            a(i,n2+1,0,0) = 1.0
            a(i,n2+1,n3+1,0) = 1.0
         enddo
         do j = 0, n2+1
            a(0,j,0,0) = 1.0
            a(n1+1,j,0,0) = 1.0
            a(0,j,n3+1,0) = 1.0
            a(n1+1,j,n3+1,0) = 1.0
         enddo
         do k = 0, n3+1
            a(0,0,k,0) = 1.0
            a(n1+1,0,k,0) = 1.0
            a(0,n2+1,k,0) = 1.0
            a(n1+1,n2+1,k,0) = 1.0
         enddo 
         return
         end
c-----------------------------------------------------------------------
c Tue May 26 16:01:43 1992 Changed the boundary conditions to just input
c                          -D Grad p . n.  Changes are the (-1)* terms.
c
      subroutine setupTensRHS(n1,n2,n3,b,rate,
     $        h1,h2,h3,bc31,bc32,bc21,bc22,bc11,bc12,
     $        bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        bc1,bc2,bc3,x1,x2,x3,r,aflag)
      implicit none 
      include '../Include/BCtypes.hf'
      integer    n1,n2,n3,aflag
      integer    nBCRegions,bcType(nBCRegions),bcRegion1(n2,n3,2),
     $           bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8     b(0:n1+1,0:n2+1,0:n3+1),rate(n1,n2,n3),
     *           bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2),
     *           x1(0:n1),x2(0:n2),x3(0:n3),r,a
      real*8     bc31(n1,n2,5),bc32(n1,n2,5),bc21(n1,n3,5),
     $           bc22(n1,n3,5),bc11(n2,n3,5),bc12(n2,n3,5)
      real*8     h1(0:n1+1),h2(0:n2+1),h3(0:n3+1)
c
c     inputs :
c       rate(.,.,.)    : block source injection rate
c       bcType         : boundary condition type
c Tue May 26 16:01:43 1992 Changed the boundary conditions to just input
c                          -D Grad p . n.
c                        bcType = 1 : flux in bc
c                        bcType = 2 : pressure in bc
c                        bcType = 4 : Robin B.C. : u.n+rp = bc
c                        bcType = anything else : no-flow boundary
c       bc[id](.,.,il) : as above
c       x[id](.)       : cell boundary locations in direction id
c
c     outputs :
c       b 
c
      integer i,j,k,region
c
      call makehs(n1,n2,n3,x1,x2,x3,h1,h2,h3)
c
      do k = 0,n3+1
         do j = 0,n2+1
            do i = 0,n1+1
               b(i,j,k) = 0.0
            enddo
         enddo
      enddo
c
c bottom face
c
      do 100 j = 1,n2
      do 100 i = 1,n1
         region = bcRegion3(i,j,1)
         a = h1(i)*h2(j)
         if (bcType(region) .eq. BCDIRICHLET) then
            b(i,j,0) = bc3(i,j,1)*a
c
            b(i,j-1,1) = b(i,j-1,1) - bc31(i,j,1)*bc3(i,j,1)
            b(i-1,j,1) = b(i-1,j,1) - bc31(i,j,2)*bc3(i,j,1)
            b(i,j,  1) = b(i,j,  1) - bc31(i,j,3)*bc3(i,j,1)
            b(i+1,j,1) = b(i+1,j,1) - bc31(i,j,4)*bc3(i,j,1)
            b(i,j+1,1) = b(i,j+1,1) - bc31(i,j,5)*bc3(i,j,1)
         else if ((bcType(region).eq.BCNEUMANN)
     $           .or.(bcType(region).eq.BCROBIN).or.
     $           (bcType(region).eq.BCNOFLOW)) then
            if (aflag.EQ.VOLUMETRIC) a = 1.
            b(i,j,0) = b(i,j,0) - bc3(i,j,1)*a
         endif
 100  continue
c
c top face
c
      do 200 j = 1,n2
      do 200 i = 1,n1
         region = bcRegion3(i,j,2)
         a = h1(i)*h2(j)
         if (bcType(region) .eq. BCDIRICHLET) then
            b(i,j,n3+1) = bc3(i,j,2)*a
c
            b(i,j+1,n3) = b(i,j+1,n3) - bc32(i,j,1)*bc3(i,j,2)
            b(i+1,j,n3) = b(i+1,j,n3) - bc32(i,j,2)*bc3(i,j,2)
            b(i,j,n3)   = b(i,j,n3)   - bc32(i,j,3)*bc3(i,j,2)
            b(i-1,j,n3) = b(i-1,j,n3) - bc32(i,j,4)*bc3(i,j,2)
            b(i,j-1,n3) = b(i,j-1,n3) - bc32(i,j,5)*bc3(i,j,2)
         else if ((bcType(region).eq.BCNEUMANN)
     $           .or.(bcType(region).eq.BCROBIN).or.
     $           (bcType(region).eq.BCNOFLOW)) then
            if (aflag.EQ.VOLUMETRIC) a = 1.
            b(i,j,n3+1) = b(i,j,n3+1) - bc3(i,j,2)*a
         endif
 200  continue
c
c front face
c
      do 300 k = 1,n3
      do 300 i = 1,n1
         region = bcRegion2(i,k,1)
         a = h1(i)*h3(k)
         if (bcType(region) .eq. BCDIRICHLET) then
            b(i,0,k) = bc2(i,k,1)*a
c
            b(i,1,k+1) = b(i,1,k+1) - bc21(i,k,1)*bc2(i,k,1)
            b(i+1,1,k) = b(i+1,1,k) - bc21(i,k,2)*bc2(i,k,1)
            b(i,1,k)   = b(i,1,k)   - bc21(i,k,3)*bc2(i,k,1)
            b(i-1,1,k) = b(i-1,1,k) - bc21(i,k,4)*bc2(i,k,1)
            b(i,1,k-1) = b(i,1,k-1) - bc21(i,k,5)*bc2(i,k,1)
         else if ((bcType(region).eq.BCNEUMANN)
     $           .or.(bcType(region).eq.BCROBIN).or.
     $           (bcType(region).eq.BCNOFLOW)) then
            if (aflag.EQ.VOLUMETRIC) a = 1.
            b(i,0,k) = b(i,0,k) - bc2(i,k,1)*a
         endif
 300  continue
c
c back face
c
      do 400 k = 1,n3
      do 400 i = 1,n1
         region = bcRegion2(i,k,2)
         a = h1(i)*h3(k)
         if (bcType(region) .eq. BCDIRICHLET) then
            b(i,n2+1,k) = bc2(i,k,2)*a
c
            b(i,n2,k-1) = b(i,n2,k-1) - bc22(i,k,1)*bc2(i,k,2)
            b(i-1,n2,k) = b(i-1,n2,k) - bc22(i,k,2)*bc2(i,k,2)
            b(i,n2,k)   = b(i,n2,k)   - bc22(i,k,3)*bc2(i,k,2)
            b(i+1,n2,k) = b(i+1,n2,k) - bc22(i,k,4)*bc2(i,k,2)
            b(i,n2,k+1) = b(i,n2,k+1) - bc22(i,k,5)*bc2(i,k,2)
         else if ((bcType(region).eq.BCNEUMANN)
     $           .or.(bcType(region).eq.BCROBIN).or.
     $           (bcType(region).eq.BCNOFLOW)) then
            if (aflag.EQ.VOLUMETRIC) a = 1.
            b(i,n2+1,k) = b(i,n2+1,k) - bc2(i,k,2)*a
         endif
 400  continue
c
c left face
c
      do 500 k = 1,n3
      do 500 j = 1,n2
         region = bcRegion1(j,k,1)
         a = h2(j)*h3(k)
         if (bcType(region) .eq. BCDIRICHLET) then
            b(0,j,k) = bc1(j,k,1)*a
c
            b(1,j,k+1) = b(1,j,k+1) - bc11(j,k,1)*bc1(j,k,1)
            b(1,j+1,k) = b(1,j+1,k) - bc11(j,k,2)*bc1(j,k,1)
            b(1,j,k)   = b(1,j,k)   - bc11(j,k,3)*bc1(j,k,1)
            b(1,j-1,k) = b(1,j-1,k) - bc11(j,k,4)*bc1(j,k,1)
            b(1,j,k-1) = b(1,j,k-1) - bc11(j,k,5)*bc1(j,k,1)
         else if ((bcType(region).eq.BCNEUMANN)
     $           .or.(bcType(region).eq.BCROBIN).or.
     $           (bcType(region).eq.BCNOFLOW)) then
            if (aflag.EQ.VOLUMETRIC) a = 1.
            b(0,j,k) = b(0,j,k) - bc1(j,k,1)*a
         endif
 500  continue
c
c right face
c
      do 600 k = 1,n3
      do 600 j = 1,n2
         region = bcRegion1(j,k,2)
         a = h2(j)*h3(k)
         if (bcType(region) .eq. BCDIRICHLET) then
            b(n1+1,j,k) = bc1(j,k,2)*a
c
            b(n1,j,k-1) = b(n1,j,k-1) - bc12(j,k,1)*bc1(j,k,2)
            b(n1,j-1,k) = b(n1,j-1,k) - bc12(j,k,2)*bc1(j,k,2)
            b(n1,j,k)   = b(n1,j,k)   - bc12(j,k,3)*bc1(j,k,2)
            b(n1,j+1,k) = b(n1,j+1,k) - bc12(j,k,4)*bc1(j,k,2)
            b(n1,j,k+1) = b(n1,j,k+1) - bc12(j,k,5)*bc1(j,k,2)
         else if ((bcType(region).eq.BCNEUMANN)
     $           .or.(bcType(region).eq.BCROBIN).or.
     $           (bcType(region).eq.BCNOFLOW)) then
            if (aflag.EQ.VOLUMETRIC) a = 1.
            b(n1+1,j,k) = b(n1+1,j,k) - bc1(j,k,2)*a
         endif
 600  continue
c
      do k = 1,n3
         do j = 1,n2
            do i = 1,n1
               b(i,j,k) = b(i,j,k) + rate(i,j,k)*h1(i)*h2(j)*h3(k)
            enddo
         enddo
      enddo
c
      return
      end
c
c
      subroutine makehs(n1,n2,n3,x1,x2,x3,hx1,hx2,hx3)
      implicit none
      integer n1,n2,n3,i1,i2,i3
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),hx1(0:n1+1),
     $     hx2(0:n2+1),hx3(0:n3+1)
c
      do 10 i1=1,n1
        hx1(i1) = x1(i1)-x1(i1-1)
 10   continue
      do 20 i2=1,n2
        hx2(i2) = x2(i2)-x2(i2-1)
 20   continue
      do 30 i3=1,n3
        hx3(i3) = x3(i3)-x3(i3-1)
 30   continue
c
      hx1(0) = 0.0d0
      hx2(0) = 0.0d0
      hx3(0) = 0.0d0
      hx1(n1+1) = 0.0d0
      hx2(n2+1) = 0.0d0
      hx3(n3+1) = 0.0d0
c
      return
      end
c
      subroutine makebyds(n1,n2,n3,x1,x2,x3,byd1,byd2,byd3)
      implicit none
      integer n1,n2,n3,i1,i2,i3
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),
     $       byd1(0:n1),byd2(0:n2),byd3(0:n3)
c
      byd1(0) = 1./(x1(1) - x1(0))
      do i1 = 1,n1-1
         byd1(i1) = 1./(x1(i1+1) - x1(i1-1))
      enddo
      byd1(n1) = 1./(x1(n1) - x1(n1-1))
c
      byd2(0) = 1./(x2(1) - x2(0))
      do i2 = 1,n2-1
         byd2(i2) = 1./(x2(i2+1) - x2(i2-1))
      enddo
      byd2(n2) = 1./(x2(n2) - x2(n2-1))
c
      byd3(0) = 1./(x3(1) - x3(0))
      do i3 = 1,n3-1
         byd3(i3) = 1./(x3(i3+1) - x3(i3-1))
      enddo
      byd3(n3) = 1./(x3(n3) - x3(n3-1))
c
      return
      end

