      subroutine InitTensor(n1,n2,n3,x1,x2,x3,alpha,D,bcType,
     $     nBCRegions,bcRegion1,bcRegion2,bcRegion3,SDparam,
     $     rwk,rlen,SDptrs)
c     
      implicit none
c
c     Handles the initialization of workspace and factorization of the 
c     system matrix for the full symmetric tensor case
c
      integer n1,n2,n3,rlen
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),alpha(*)
      real*8 SDparam(*),rwk(rlen)
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 D
      external D
c
      include '../Include/sdsolver.hf'
      include '../Include/BCtypes.hf'
      include '../Include/defines.hf'
c
      integer iSDparam,ik11,ik12,ik13,ik22,ik23,ik33,iamat,im0,info
      integer ibc31,ibc32,ibc21,ibc22,ibc11,ibc12
c
      integer SDptrs(MAXSDPTRS)
      integer Lsing
c
c ... Internal Common
c
      integer TensorType
      common /parcelcom/ TensorType
c
c ... Functions
c
      integer memalr
c
      integer n,m
      integer iBmat,m1,m2,m3
      integer ih1,ih2,ih3,ibyd1,ibyd2,ibyd3
c
c ... Copy the parameters
c
      iSDparam = memalr(numSDParam,rwk)
cta 17 July 2002
c      call WCOPY(numSDParam,SDparam,1,rwk(iSDparam),1)
      call dcopy(numSDParam,SDparam,1,rwk(iSDparam),1)
c
c ... Allocate for the banded matrix
c
      n = (n1+2)*(n2+2)*(n3+2)
      m = min((n1+2)*(n2+3)+1,min((n1+2)*(n3+3)+1,min((n2+2)*(n3+3)+1,
     $    min((n2+2)*(n1+3)+1,min((n3+2)*(n1+3)+1,(n3+2)*(n2+3)+1)))))
      im0 = memalr(n*m,rwk)
c
      call checkSing(n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,
     $     bcRegion3,Lsing)
c
c ... Make edge transmissibilities
c
      ik11   = memalr(n1*n2*n3,rwk)
      ik12   = memalr(n1*n2*n3,rwk)
      ik13   = memalr(n1*n2*n3,rwk)
      ik22   = memalr(n1*n2*n3,rwk)
      ik23   = memalr(n1*n2*n3,rwk)
      ik33   = memalr(n1*n2*n3,rwk)
      call PermCond(n1,n2,n3,D,rwk(ik11),rwk(ik12),rwk(ik13),
     $              rwk(ik22),rwk(ik23),rwk(ik33))
c
c ... Setup Finite Difference Matrix
c
      ibc31 = memalr(5*n1*n2,rwk)
      ibc32 = memalr(5*n1*n2,rwk)
      ibc21 = memalr(5*n1*n3,rwk)
      ibc22 = memalr(5*n1*n3,rwk)
      ibc11 = memalr(5*n2*n3,rwk)
      ibc12 = memalr(5*n2*n3,rwk)
c
      iAmat = memalr(19*(n1+2)*(n2+2)*(n3+2),rwk)
c
c     Set up the 19 band matrix
c
      ih1 = memalr(n1+2,rwk)
      ih2 = memalr(n2+2,rwk)
      ih3 = memalr(n3+2,rwk)
      ibyd1 = memalr(n1+1,rwk)
      ibyd2 = memalr(n2+1,rwk)
      ibyd3 = memalr(n3+1,rwk)
c
      call setup19bd(n1,n2,n3,rwk(iAmat),rwk(ik11),rwk(ik12),
     $     rwk(ik13),rwk(ik22),rwk(ik23),rwk(ik33),
     $     rwk(ibc31),rwk(ibc32),rwk(ibc21),
     $     rwk(ibc22),rwk(ibc11),rwk(ibc12),
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $     0.0d0,alpha,rwk(ih1),rwk(ih2),rwk(ih3),rwk(ibyd1),
     $     rwk(ibyd2),rwk(ibyd3))
c
      call memfreer(ih1,rwk)
c
c ... Reorder the matrix to decrease band width and copy to rwk(im0)
c
      iBmat = memalr(19*(n1+2)*(n2+2)*(n3+2),rwk)
      call orderTensmat(n1+2,n2+2,n3+2,rwk(iAmat),rwk(iBmat),m1,m2,m3,m)
      call copyTensBands(m1,m2,m3,n,m,rwk(iBmat),rwk(im0),Lsing)
c
      call memfreer(iBmat,rwk)
      call memfreer(iAmat,rwk)
c
c ... Factor
c     
cta 17 July 2002
c      call WPBTRF('U',n,m-1,rwk(im0),m,info)
      call dpbtrf('U',n,m-1,rwk(im0),m,info)
      if (info.ne.0) then
         write(*,*) 'Matrix is not positive def:',n,info
         call Parcelcrash('WPBTRF() in tensor.f fails',-30)
      endif
c
c ... Save the pointers in the pointer array      
c
      SDptrs(1)  = iSDparam
      SDptrs(2)  = iAmat
      SDptrs(3)  = ik11
      SDptrs(4)  = ik12
      SDptrs(5)  = ik13
      SDptrs(6)  = ik22
      SDptrs(7)  = ik23
      SDptrs(8)  = ik33
      SDptrs(9)  = ibc31
      SDptrs(10) = ibc32
      SDptrs(11) = ibc21
      SDptrs(12) = ibc22
      SDptrs(13) = ibc11
      SDptrs(14) = ibc12
      SDptrs(15) = im0
      SDptrs(16) = Lsing
c
      return
      end


c ...   Title:  Solve the subdomain problem
c ...
c ...  Description: These routines solve the subdomain problem
c ...   
c ...   Created:        Tue May 26 13:45:11 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Mon Jul 20 11:53:47 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: tensor.f,v 1.6 1997-08-20 11:46:30-05 arbogast Exp $
C ...  $RCSfile: tensor.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Bcfdnf/tensor.f,v $
C  
c ... Version $Id: tensor.f,v 1.6 1997-08-20 11:46:30-05 arbogast Exp $
c
c ... Routine: exactTensSolve
c
c ... Main function:
c     Solves Local Elliptic Problem using Nested Factorization
c
c ... In:
c     outUnit: Output unit number
c     MYNOD :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     mynbrSize: The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     bcType:     BCs for 1 - Neumann, 2 - Dirichlet, 3 - Noflo
c     x?    :     point centered grids
c     hx?   :     mesh widths hx?(i) = x?(i)-x?(i-1)
c     bc?   :     Boundary condition values
c     pres  :     Guess for Pressure
c     niterSD :   Maximum iterations usuable to solve subdomain problems
c     epsSD :     Relative residual reduction in subdomains
c     tolSD :     Absolute residual required in subdomains
c     a     :     Finite difference matrix
c     rate  :     Right hand side of -div A grad (pres) = (rate)
c     k??   :     Edge Permeability matrix
c
c ... Out:
c     pres  :  Pressure values in subdomain
c     bf?   :  Boundary fluxes for this subdomain
c     df?   :  Boundary Pressures                          
c     b     :  Right hand side for solving A pres = b
c
      subroutine exactTensSolve(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,bc1,bc2,bc3,pres,vel1,vel2,vel3,
     $     bf1,bf2,bf3,df1,df2,df3,rate,rwk,SDptrs)
c
      implicit none
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2)
      integer n1,n2,n3
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3)
      real*8  bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2),pres(n1,n2,n3)
      real*8  vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3)
      real*8  bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8  df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      real*8  rwk(*),rate(n1,n2,n3)

      integer i,j,k
c
      integer n
      integer memalr
      external memalr
      integer ib,m,ibp,m1,m2,m3,ipres
c
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
      real*8 RobinR
      common/Robin/RobinR
c
      include '../Include/BCtypes.hf'
      include '../Include/sdsolver.hf'
c
      integer ik11,ik12,ik13,ik22,ik23,ik33,im0
      integer ibc31,ibc32,ibc21,ibc22,ibc11,ibc12
      integer ih1,ih2,ih3
c
      integer SDptrs(MAXSDPTRS)
      integer Lsing,info
c
c ... Unpack the pointers
c
c
      ik11     = SDptrs(3)
      ik12     = SDptrs(4)
      ik13     = SDptrs(5)
      ik22     = SDptrs(6)
      ik23     = SDptrs(7)
      ik33     = SDptrs(8)
      ibc31    = SDptrs(9)
      ibc32    = SDptrs(10)
      ibc21    = SDptrs(11)
      ibc22    = SDptrs(12)
      ibc11    = SDptrs(13)
      ibc12    = SDptrs(14)
      im0      = SDptrs(15)
      Lsing    = SDptrs(16)
c
c ... Setup RightHandSide
c
cta exchanged following 2 lines Aug 13, 1997
      ibp = memalr((n1+2)*(n2+2)*(n3+2),rwk)
      ib = memalr((n1+2)*(n2+2)*(n3+2),rwk)
c
      ih1 = memalr(n1+2,rwk)
      ih2 = memalr(n2+2,rwk)
      ih3 = memalr(n3+2,rwk)
c
      call setupTensRHS(n1,n2,n3,rwk(ib),
     $     rate,rwk(ih1),rwk(ih2),rwk(ih3),
     $     rwk(ibc31),rwk(ibc32),rwk(ibc21),
     $     rwk(ibc22),rwk(ibc11),rwk(ibc12),
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     bc1,bc2,bc3,x1,x2,x3,RobinR,VOLUMETRIC)
c
      call memfreer(ih1,rwk)
c
c ... Solve system using the cholesky factored matrix
c
      n = (n1+2)*(n2+2)*(n3+2)
      m = min((n1+2)*(n2+3)+1,min((n1+2)*(n3+3)+1,min((n2+2)*(n3+3)+1,
     $    min((n2+2)*(n1+3)+1,min((n3+2)*(n1+3)+1,(n3+2)*(n2+3)+1)))))
      call orderTensvec(n1+2,n2+2,n3+2,rwk(ib),rwk(ibp),m1,m2,m3,m)
      if (Lsing .eq. 1) rwk(ibp+n/2) = 0
cta 17 July 2002
c      call WPBTRS('U',n,m-1,1,rwk(im0),m,rwk(ibp),n,info)
      call dpbtrs('U',n,m-1,1,rwk(im0),m,rwk(ibp),n,info)
c
      call memfreer(ib,rwk)
      ipres = memalr((n1+2)*(n2+2)*(n3+2),rwk)
      call reorderTensVec(n1,n2,n3,pres,rwk(ipres),rwk(ibp),m,
     m     df1,df2,df3)
c
cta Aug 13, 1997:     call memfreer(ibp,rwk)
c
c     Calculate velocities from pressures
c
      ih1 = memalr(n1+2,rwk)
      ih2 = memalr(n2+2,rwk)
      ih3 = memalr(n3+2,rwk)
c
      call pres2velTens(n1,n2,n3,rwk(ipres),bc1,bc2,bc3,
     $     rwk(ih1),rwk(ih2),rwk(ih3),
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,
     $     rwk(ik11),rwk(ik12),rwk(ik13),rwk(ik22),rwk(ik23),rwk(ik33),
     $     bf1,bf2,bf3,vel1,vel2,vel3,VOLUMETRIC)
c
cta added line Aug 13, 1997:
      call memfreer(ibp,rwk)
cta Aug 13, 1997:      call memfreer(ih1,rwk)
cta Aug 13, 1997:      call memfreer(ipres,rwk)
c     
      return
      end
c-----------------------------------------------------------------------
      subroutine CellPerm(n1,n2,n3,D,k11,k12,k13,k22,k23,k33)
c
      implicit none
c
      integer n1,n2,n3
      real*8 D,k11(n1,n2,n3),k12(n1,n2,n3),k13(n1,n2,n3),
     $       k22(n1,n2,n3),k23(n1,n2,n3),k33(n1,n2,n3)
      integer i1,i2,i3
      external D
c
      do i3 = 1,n3
         do i2 = 1,n2
            do i1 = 1,n1
               k11(i1,i2,i3) = D(1,1,i1,i2,i3)
               k12(i1,i2,i3) = D(1,2,i1,i2,i3)
               k13(i1,i2,i3) = D(1,3,i1,i2,i3)
               k22(i1,i2,i3) = D(2,2,i1,i2,i3)
               k23(i1,i2,i3) = D(2,3,i1,i2,i3)
               k33(i1,i2,i3) = D(3,3,i1,i2,i3)
            enddo
         enddo
      enddo
c
      return
      end
c
      subroutine PermCond(n1,n2,n3,D,k11,k12,k13,k22,k23,k33)
      implicit none
c
      integer n1, n2, n3
      real*8 D, k11(n1,n2,n3),k12(n1,n2,n3),k13(n1,n2,n3),
     $       k22(n1,n2,n3),k23(n1,n2,n3),k33(n1,n2,n3)
      external D
c
      include '../Include/defines.hf'
      integer TensorType
      common /parcelcom/ TensorType
c
      if (TensorType.eq.UserFuncType) then
         call CellPerm(n1,n2,n3,D,k11,k12,k13,k22,k23,k33)
      else if (TensorType.eq.ArrayType) then
         call GetArrays(n1,n2,n3,D,k11,k12,k13,k22,k23,k33)
      else
         call ParcelCrash('PermCond: Unknown Tensor Type rep',-30)
      endif
c
      return
      end
c
c
      subroutine GetArrays(n1,n2,n3,D,k11,k12,k13,k22,k23,k33)
      implicit none
c
      integer n1, n2, n3, i, j, k
      real*8 D(6,n1,n2,n3), k11(n1,n2,n3),k12(n1,n2,n3),k13(n1,n2,n3),
     $       k22(n1,n2,n3),k23(n1,n2,n3),k33(n1,n2,n3)
c
      do k = 1, n3
         do j = 1, n2
            do i = 1, n1
               k11(i,j,k) = D(1,i,j,k)
               k12(i,j,k) = D(2,i,j,k)
               k13(i,j,k) = D(3,i,j,k)
               k22(i,j,k) = D(4,i,j,k)
               k23(i,j,k) = D(5,i,j,k)
               k33(i,j,k) = D(6,i,j,k)
            enddo
         enddo
      enddo
c
      return
      end
c-----------------------------------------------------------------------
      subroutine checkSing(n1,n2,n3,bcType,nBCRegions,bcRegion1,
     $     bcRegion2,bcRegion3,lsing)
      implicit none
c
      include '../Include/BCtypes.hf'
c
      integer n1,n2,n3
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      integer lsing
      integer i,j,k,region
c
      lsing = 1
c
c left face
c
      do k = 1,n3
         do j = 1,n2
            region = bcRegion1(j,k,1)
            if ((bcType(region).eq.BCDIRICHLET).or.
     $           (bcType(region).eq.BCROBIN)) then 
               lsing = 0
               return
            endif
         enddo
      enddo
c
c right face
c
      do k = 1,n3
         do j = 1,n2
            region = bcRegion1(j,k,2)
            if ((bcType(region).eq.BCDIRICHLET).or.
     $           (bcType(region).eq.BCROBIN)) then 
               lsing = 0
               return
            endif
         enddo
      enddo
c
c front face
c
      do k = 1,n3
         do i = 1,n1
            region = bcRegion2(i,k,1)
            if ((bcType(region).eq.BCDIRICHLET).or.
     $           (bcType(region).eq.BCROBIN)) then 
               lsing = 0
               return
            endif
         enddo
      enddo
c
c back face
c
      do k = 1,n3
         do i = 1,n1
            region = bcRegion2(i,k,2)
            if ((bcType(region).eq.BCDIRICHLET).or.
     $           (bcType(region).eq.BCROBIN)) then 
               lsing = 0
               return
            endif
         enddo
      enddo
c
c bottom face
c
      do j = 1,n2
         do i = 1,n1
            region = bcRegion3(i,j,1)
            if ((bcType(region).eq.BCDIRICHLET).or.
     $           (bcType(region).eq.BCROBIN)) then 
               lsing = 0
               return
            endif
         enddo
      enddo
c
c top face
c
      do j = 1,n2
         do i = 1,n1
            region = bcRegion3(i,j,2)
            if ((bcType(region).eq.BCDIRICHLET).or.
     $           (bcType(region).eq.BCROBIN)) then 
               lsing = 0
               return
            endif
         enddo
      enddo
c
      return
      end
