c ...
c ...   Title:  Memory allocation routines
c ...
c     ...  Description: These routines control the memory allocation of a
c ...                work space vector
c ...   
c ...   Created:        Thu May 28 11:05:46 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 11:05:46 1992"
C     
C ...  $Author: jeaton $
C ...  $Id: memory.F,v 1.1.1.1 1997-02-19 15:11:54-06 jeaton Exp $
C ...  $RCSfile: memory.F,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Memory/memory.F,v $
C  
c
c ... LRT3D:  $Id: memory.F,v 1.1.1.1 1997-02-19 15:11:54-06 jeaton Exp $
c
c
c ... Memory Management routines for Real and Integer work Vectors
c ... Based on an initial version by Jan Mandel
c
      integer function mrD(isize,rw,cmd,line,file)
      implicit none
      integer isize,line
      real*8 rw(*)
      character*(*) file,cmd
      integer last,next

      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)
      
      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4)
      
      if (int(rw(iMAGIC)).ne.MEMMAGIC) then
         call MEMcrash2('MEMALR: uninitialized work space',cmd,
     $        line,file)
      endif
      if (isize.lt.0) call MEMcrash2('MEMALR: Negative size',
     $     cmd,line,file)
      if (isize.eq.0) call MEMcrash2('MEMALR: Zero size',cmd,line,file)
      last = int(rw(iNAV))
      next = int(rw(iNAV)) + isize
      if (next.gt.int(rw(iMWORK))) goto 910
c ... Update Max Used
      rw(iNUSED) = max(int(rw(iNUSED)),next)
c ... If testing
      if (DEBUG) then
c ... Link back for checking
         rw(iNAV) = next + 1
         rw(next) = last
      else
         rw(iNAV) = next
      endif
c ... Return pointer
      mrD = last
      return
c ... Error Handling
 910  continue
      write(*,911) next,int(rw(iMWORK)),isize
 911  format('Out of real workspace.  Needed at least ',i8,
     $     ' have only ',i8,' words.',i8)
      call MEMcrash2('MEMALR: Out of real workspace',cmd,line,file)
      end
c
      integer function miD(isize,iw,cmd,line,file)
c***	allocate array size isize in integer workspace	
      implicit none
      integer isize,iw(*),line
      character*(*) file,cmd
      integer last,next
      
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4)
      
      if (iw(iMAGIC).ne.MEMMAGIC) then
         call MEMcrash2('MEMALI: uninitialized work space',
     $        cmd,line,file)
      endif
      if (isize.lt.0) call MEMcrash2('MEMALI: Negative size',
     $     cmd,line,file)
      if (isize.eq.0) call MEMcrash2('MEMALI: Zero size',cmd,line,file)
      last = iw(iNAV)
      next = iw(iNAV) + isize
      if (next.gt.iw(iMWORK)) goto 910
c ... Update Max Used
      iw(iNUSED) = max(iw(iNUSED),next)
c ... If testing
      if (DEBUG) then
c ... Link back for checking
         iw(iNAV) = next + 1
         iw(next) = last
      else
         iw(iNAV) = next
      endif
c ... Return pointer
      miD = last
      return
c ... Error Handling
 910  continue
      write(*,911) next,iw(iMWORK)
 911  format('Out of integer workspace.  Needed at least ',i8,
     $     ' have only ',i8,' words.')
      call MEMcrash2('MEMALI: Out of integer workspace',cmd,line,file)
      end

c
      integer function memalr(isize,rw)
c***	allocate array size isize in real workspace	
      implicit none
      integer isize
      real*8 rw(*)
      integer last,next

      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)
      
      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4)
      
      if (int(rw(iMAGIC)).ne.MEMMAGIC) 
     $     call MEMcrash('MEMALR: uninitialized work space')
      if (isize.lt.0) call MEMcrash('MEMALR: Negative size')
      if (isize.eq.0) call MEMcrash('MEMALR: Zero size')
      last = int(rw(iNAV))
      next = int(rw(iNAV)) + isize
      if (next.gt.int(rw(iMWORK))) goto 910
c ... Update Max Used
      rw(iNUSED) = max(int(rw(iNUSED)),next)
c ... If testing
      if (DEBUG) then
c ... Link back for checking
         rw(iNAV) = next + 1
         rw(next) = last
      else
         rw(iNAV) = next
      endif
c ... Return pointer
      memalr = last
      return
c ... Error Handling
 910  continue
      write(*,911) next,int(rw(iMWORK)),isize
 911  format('Out of real workspace.  Needed at least ',i8,
     $     ' have only ',i8,' words.',i8)
      call MEMcrash('MEMALR: Out of real workspace')
      end


      integer function memali(isize,iw)
c***	allocate array size isize in integer workspace	
      implicit none
      integer isize,iw(*)
      integer last,next

      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)
      
      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4)
      
      if (iw(iMAGIC).ne.MEMMAGIC) 
     $     call MEMcrash('MEMALI: uninitialized work space')
      if (isize.lt.0) call MEMcrash('MEMALI: Negative size')
      if (isize.eq.0) call MEMcrash('MEMALI: Zero size')
      last = iw(iNAV)
      next = iw(iNAV) + isize
      if (next.gt.iw(iMWORK)) goto 910
c ... Update Max Used
      iw(iNUSED) = max(iw(iNUSED),next)
c ... If testing
      if (DEBUG) then
c ... Link back for checking
         iw(iNAV) = next + 1
         iw(next) = last
      else
         iw(iNAV) = next
      endif
c ... Return pointer
      memali = last
      return
c ... Error Handling
 910  continue
      write(*,911) next,iw(iMWORK)
 911  format('Out of integer workspace.  Needed at least ',i8,
     $     ' have only ',i8,' words.')
      call MEMcrash('MEMALI: Out of integer workspace')
      end

      subroutine memfreei(ip,iw)
c ... call bdfree for integer array
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer ip,iw(*)
      real*8 dummy
      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4)
      if (iw(iMAGIC).ne.MEMMAGIC) 
     $     call MEMcrash('MEMFREEI: uninitialized work space')
      if (ip.le.iNUSED) 
     $     call MEMcrash('MEMFREEI: Small Free Integer work space')
      call memfree(ip,1,iw,dummy)
      end

      subroutine memfreer(ip,rw)
c     call bdfree for real array
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer ip,idummy
      real*8 rw(*)
      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4)

      if (int(rw(iMAGIC)).ne.MEMMAGIC) 
     $     call MEMcrash('MEMFREER: uninitialized work space')
      if (ip.le.iNUSED) 
     $     call MEMcrash('MEMFREER: Small Free Real work space')
      call memfree(ip,2,idummy,rw)
      end

      subroutine memfree(ip,iprec,iw,rw)
c***	description
c     free all arrays allocated AFTER ip
c     if DEBUG follow links back
c     and every iatest calls, check all links

c***	arguments
      integer ip,iprec,iw(*)
      real*8 rw(*)
c
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4,iHEADLEN=4)
      
      if (iprec.eq.1) then
         if (iw(iMAGIC).ne.MEMMAGIC) 
     $        call MEMcrash('MEMFREE: uninitialized work space')
      else if (iprec.eq.2) then
         if (int(rw(iMAGIC)).ne.MEMMAGIC) 
     $        call MEMcrash('MEMFREE: uninitialized work space')
      endif
c
c ... Check the memory
c
      if (DEBUG)  call memchkm(iHEADLEN+1,ip,iprec,iw,rw)
c ... release all allocations above ip
      if (iprec.eq.1) then
         iw(iNAV) = ip
      else if (iprec.eq.2) then
         rw(iNAV) = ip
      endif
      ip=0
      end

      subroutine memchkm(is,ip,iprec,iw,rw)
c     test integrity of memory allocations
c     link back and search for address istart
c***	arguments
      integer is,ip,iprec,iw(*)
      real*8 rw(*)
c***	local
      integer link,last
      logical found

      integer iMAGIC,iNAV,iMWORK,iNUSED
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4)

c	starting address
      if (iprec.eq.1) then
         link = iw(iNAV)-1
      else  if (iprec.eq.2) then
         link = int(rw(iNAV))-1
      endif
      found=.false.

c     one link back
 1    goto(10,20)iprec
      goto 85
 10   last=iw(link)
      goto 30
 20   last= int(rw(link))
 30   continue
c     ip found?
      if(last.eq.ip)found=.true.
c     end of chain?
      if(last.eq.is)then
c     found end of chain
         if(found)goto 90
c     but ip was not there
         goto 70
      endif
c     bad value of link?
      if(last.le.0.or.last.gt.link)goto 80
      link=last-1
      goto 1

 70   print *,'chain of back links does not contain ',ip
 80   print *,'bad memory chain ! last=',last,' link=',link,
     $     ' iprec=',iprec
      call MEMcrash(
     $     'INTERNAL ERR: bad free memory request or corrupted stack')
 85   call MEMcrash('bdchkm: bad iprec')

 90   return
      end


      subroutine meminiti(iw,nlen)
c     initialize allocations
      integer nlen,iw(nlen)
c
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer iMAGIC,iNAV,iMWORK,iNUSED,iHEADLEN
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4,iHEADLEN=4)
c
      if (nlen.le.iHEADLEN) 
     $     call MEMcrash(
     $     'MEMINITI:Cannot have workspace of size .le. 5')
      iw(iMAGIC) = MEMMAGIC
      iw(iNAV) = iHEADLEN+1
      iw(iMWORK) = nlen
      iw(iNUSED) = iHEADLEN
      return
      end

      subroutine meminitr(rw,nlen)
c     initialize allocations
      integer nlen
      real*8 rw(nlen)
c
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer iMAGIC,iNAV,iMWORK,iNUSED,iHEADLEN
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4,iHEADLEN=4)
c
      if (nlen.le.iHEADLEN) 
     $     call MEMcrash(
     $     'MEMINITR:Cannot have workspace of size .le. 5')
      rw(iMAGIC) = MEMMAGIC
      rw(iNAV) = iHEADLEN+1
      rw(iMWORK) = nlen
      rw(iNUSED) = iHEADLEN
      return
      end
c
      subroutine memused(niwork,nrwork,iwk,rwk)
      implicit none
      integer niwork,nrwork,iwk(*)
      real*8 rwk(*)
c
      integer iMAGIC,iNAV,iMWORK,iNUSED,iHEADLEN
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4,iHEADLEN=4)
c
      niwork=iwk(iNUSED)
      nrwork=rwk(iNUSED)
      end

      integer function memusedReal(rwk)
      implicit none
      real*8 rwk(*)
c
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer iMAGIC,iNAV,iMWORK,iNUSED,iHEADLEN
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4,iHEADLEN=4)
c
      if (int(rwk(iMAGIC)).ne.MEMMAGIC) 
     $        call MEMcrash('memusedReal: uninitialized work space')

      memusedReal=rwk(iNUSED)
      end

      integer function memusedInt(iwk)
      implicit none
      integer iwk(*)
c
      integer MEMMAGIC
      logical DEBUG
      parameter(MEMMAGIC = -22867, DEBUG = .false.)

      integer iMAGIC,iNAV,iMWORK,iNUSED,iHEADLEN
      parameter(iMAGIC=1,iNAV=2,iMWORK=3,iNUSED=4,iHEADLEN=4)
c
      if (iwk(iMAGIC).ne.MEMMAGIC) 
     $        call MEMcrash('memusedInt: uninitialized work space')

      memusedInt=iwk(iNUSED)
      end

      subroutine MEMcrash(str)
      character*(*) str
c
      write(*,*) str
      call Parcelcrash('Memory Allocation Error',-15)
      return
      end

      subroutine MEMcrash2(str,cmd,line,file)
      character*(*) str,cmd,file
      integer line
c
      write(*,*) str,' Line: ',line,' File: ',file
      write(*,*) 'Arguments: ',cmd
      call Parcelcrash('Memory Allocation Error',-15)
      return
      end


