c
c ...   Title:  Balancing Preconditioner
c ...
c     ...  Description: These files implement the balancing preconditioner
c ...                as described in Mandel-92, To appear in CANM
c ...                The N_i are the vectors that are one on the boundary
c ...                of a subdomain.
c ...
c ...   Created:        Mon Aug 24 11:05:57 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Mon Aug 24 11:05:57 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: bal.f,v 1.4 1997-08-20 11:46:38-05 arbogast Exp $
C ...  $RCSfile: bal.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/bal.f,v $
C  
c
c ... This routine returns z result of appling to preconditioner to r
c ...                 z <- BalPC(r)
c

      subroutine BalPC(r1,r2,r3,z1,z2,z3,
     $     outUnit,mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,DSDptrs,
     $     BalEval,BalEvec)
      implicit none
c
      integer mynod,mynbr(3,2),mynbrSize(3,2,2),numproc,n1,n2,n3
      integer outUnit
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8  r1(n2,n3,2),r2(n1,n3,2),r3(n1,n2,2)
      real*8  z1(n2,n3,2),z2(n1,n3,2),z3(n1,n2,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  alpha(*),D,rwk(*)
      integer iwk(*),DSDptrs(*)
      integer bcType_save
c
      include '../Include/sdsolver.hf'
      include 'mpif.h'
      include '../Include/msgtypes.hf'
      include '../Include/BCtypes.hf'
c
      real*8 BalEval(*),BalEvec(*)
      integer NSDptrs(MAXSDPTRS)
      common /BalCommon/ NSDptrs
      integer iwt1,iwt2,iwt3,iwt1z,iwt2z,iwt3z,ibal1,ibal2,ibal3,
     $     balSlvType
      common /BalCom/iwt1,iwt2,iwt3,iwt1z,iwt2z,iwt3z,ibal1,ibal2,ibal3,
     $     balSlvType
c
c ... Functions
c
      integer memalr
      external memalr,memfreer
      real*8 ddotBdry
      external ddotBdry
c
      integer dir,face,iMemHead,iDr1,iDr2,iDr3,stat,cnt
      logical internalBoundary
c
      internalBoundary = .false.
      do dir = 1,3
         do face = 1,2
            if (mynbr(dir,face).ne.-1) internalBoundary = .true.
         enddo
      enddo
c
c ... z <- 0      
      call dinitBdry(n1,n2,n3,z1,z2,z3,0.0d0)
c
c ... Correct the residual to be balanced (Eqn 12 and 13)
c        
c$$$      call balance(z1,z2,z3,r1,r2,r3,
c$$$c     $     .true.,.true.,.true.,.true.,BalEval,BalEvec,
c$$$     $     .true.,.true.,.true.,.false.,BalEval,BalEvec,
c$$$     $     mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
c$$$     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
c$$$     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,DSDptrs)
c
c ... Test Residual to see if it is still in balance
c
c$$$      call balance(z1,z2,z3,r1,r2,r3,
c$$$     $     .true.,.false.,.false.,.true.,BalEval,BalEvec,
c$$$     $     mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
c$$$     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
c$$$     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,DSDptrs)
c
c ... Allocate work space
      iMemHead = memalr(1,rwk)
      iDr1 = memalr(2*n2*n3,rwk)
      iDr2 = memalr(2*n1*n3,rwk)
      iDr3 = memalr(2*n1*n2,rwk)
c
c ... Dr <- D*r  where D is the weights (EQN 13) for s_i
c
      call dmultBdry(n1,n2,n3,r1,r2,r3,rwk(iwt1z),rwk(iwt2z),
     $     rwk(iwt3z),rwk(iDr1),rwk(iDr2),rwk(iDr3))
c ... z_i <- S_i^{-1} Dr  (EQN 14)
      if (.true.) then
         if (internalBoundary) then
            bcType_save = bcType(nBCRegions)
            bcType(nBCRegions) = BCNEUMANN
         endif
         call ApplySinv(rwk(iDr1),rwk(iDr2),rwk(iDr3),z1,z2,z3,
     $        outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,bcType,nBCRegions,
     $        bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $        hx1,hx2,hx3,stat,cnt,alpha,D,rwk,iwk,NSDptrs)
         if (internalBoundary) then
            bcType(nBCRegions) = bcType_save
         endif
      else
         call dscalBdry(n1,n2,n3,-1.0d0,rwk(iDr1),rwk(iDr2),rwk(iDr3),
     $        z1,z2,z3)
      endif
c
clcc.begin
c
c ... z <- \Sum D_i z_i
c
      call dmultBdry(n1,n2,n3,z1,z2,z3,rwk(iwt1z),rwk(iwt2z),
     $     rwk(iwt3z),rwk(iDr1),rwk(iDr2),rwk(iDr3))
      call addBdry(mynod,mynbr,mynbrSize,n1,n2,n3,
     $     rwk(iDr1),rwk(iDr2),rwk(iDr3),
     $     z1,z2,z3,rwk,BAL1EDGESUBASS,SCALAR)
c
clcc.end
c
c ... z gets balanced, (EQN 15 and 16)
c      
      call balance(z1,z2,z3,r1,r2,r3,
     $     .false.,.true.,.false.,.false.,BalEval,BalEvec,
     $     mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,DSDptrs)
c
      call dscalBdry(n1,n2,n3,-1.0d0,z1,z2,z3,z1,z2,z3)
c
c ... Check to see if the answer z is in the space
c
clcc      call inSpace(n1,n2,n3,z1,z2,z3,rwk)
c
c ... Free temp memory
c
      call memfreer(iMemHead,rwk)
      end
      
c ... This routine determines if r? is in balance and, depending on the 
c ... flag settings, replaces u? and/or r? to the correct balanced values
c ... This solves EQN 15 and EQN 16.
      subroutine balance(u1,u2,u3,r1,r2,r3,izero,iupdateU,iupdateR,
     $     itest,BalEval,BalEvec,
     $     mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs)
c
c ... Flags:
c     izero : is u? zero?
c     iupdateU : update u by solving EQN 15 and 16.
c     iupdateR : update r by r - S(updated u)
c     itest    : printout the amount that r is out of balance
c
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
      include '../Include/BCtypes.hf'
c
      logical izero,iupdateU,iupdateR,itest
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2),numproc
      integer n1,n2,n3
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 u1(n2,n3,2),u2(n1,n3,2),u3(n1,n2,2)
      real*8 r1(n2,n3,2),r2(n1,n3,2),r3(n1,n2,2)
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8 BalEval(numproc),BalEvec(numproc,numproc)
      real*8 alpha(*),D,rwk(*)
      integer iwk(*),SDptrs(*)
c
      integer iwt1,iwt2,iwt3,iwt1z,iwt2z,iwt3z,ibal1,ibal2,ibal3,
     $     balSlvType
      common /BalCom/iwt1,iwt2,iwt3,iwt1z,iwt2z,iwt3z,ibal1,ibal2,ibal3,
     $     balSlvType
c
c ... Functions
c
      integer memalr,error
      external memalr,memfreer
cta 17 July 2002
c      real*8 ddotBdry,WASUM
c      external ddotBdry,WASUM
      real*8 ddotBdry,dasum
      external ddotBdry,dasum
c
      integer iMemHead,iDu1,iDu2,iDu3,iNDu1,iNDu2,iNDu3
      integer stat,cnt,icgRHS,icgRHS2,icgANS,i
c
      if ((iupdateU.eqv..false.).and.(iupdateR.eqv..true.)) then
         if (mynod.eq.0) then
            write(*,*) 'Warning must update U to update R in balance'
         endif
      endif
c
c ... Allocate work space
c
      iMemHead = memalr(1,rwk)
      iDu1 = memalr(2*n2*n3,rwk)
      iDu2 = memalr(2*n1*n3,rwk)
      iDu3 = memalr(2*n1*n2,rwk)
      iNDu1 = memalr(2*n2*n3,rwk)
      iNDu2 = memalr(2*n1*n3,rwk)
      iNDu3 = memalr(2*n1*n2,rwk)
c
      icgRHS  = memalr(numproc,rwk)
      icgRHS2 = memalr(numproc,rwk)
      do i=0,numproc-1
         rwk(icgRHS+i) = 0
         rwk(icgRHS2+i) = 0
      enddo
c
c ... ZDNr = (NDZ)^{trans}(r - S \sum ( NDu ) )
c
      if (izero) then
c ... u <- 0
         if (iupdateU) call dinitBdry(n1,n2,n3,u1,u2,u3,0.0d0)
c ... RHS_i = (D_i Z_i)^{trans} r_i
         rwk(icgRHS+mynod) = ddotBdry(n1,n2,n3,hx1,hx2,hx3,r1,r2,r3,
     $        rwk(iwt1z),rwk(iwt2z),rwk(iwt3z))
      else
c
c ... RHS gets local portion of (N_i D_i Z_i)^{trans} S u
c
         call balSu(mynod,mynbr,numproc,n1,n2,n3,hx1,hx2,hx3,
     $        u1,u2,u3,rwk(ibal1),rwk(ibal2),rwk(ibal3),rwk(icgRHS))
c
c ... RHS2_i = (D_i Z_i)^{trans} r_i
c
         rwk(icgRHS2+mynod) = ddotBdry(n1,n2,n3,hx1,hx2,hx3,r1,r2,r3,
     $        rwk(iwt1z),rwk(iwt2z),rwk(iwt3z))
         do i = 0,numproc-1
            rwk(icgRHS+i) = rwk(icgRHS2+i) - rwk(icgRHS+i)
         enddo
      endif
c
c ... Collect the global RHS for EQN (15)
c      
      call concatvec(numproc,rwk(icgRHS),BALRHSMSG)
c
c      if ((itest).and.(mynod.eq.0)) then
c         write(outUnit,*) 'Out of balance by ',
c     $        dasum(numproc,rwk(icgRHS),1)
c      endif

      if ((.not.iupdateU).and.(.not.iupdateR)) goto 900
c
c ... Solve the Coarse grid matrix problem EQN (15)
c
      icgANS = memalr(numproc,rwk)
      if (mynod.eq.0) then
         call pseudoInv(numproc,rwk(icgRHS),rwk(icgANS),BalEval,BalEvec,
     $        1.0d-5,0,balSlvType)
      endif
c
c ... Broadcast the results
cmpi      call gdbcastx(BALANSMSG,rwk(icgANS),numproc,0)
      call MPI_BCAST(rwk(icgANS), numproc, MPI_REAL8, 0,
     $     MPI_COMM_WORLD, error)
c
c ... Update u <- D (u + Z ans) EQN(16)
c
      call dmultBdry(n1,n2,n3,u1,u2,u3,rwk(iwt1z),rwk(iwt2z),
     $     rwk(iwt3z),u1,u2,u3)
      call daxpbyBdry(n1,n2,n3,1.0d0,u1,u2,u3,rwk(icgANS+mynod),
     $     rwk(iwt1z),rwk(iwt2z),rwk(iwt3z),u1,u2,u3)
      call dcopyBdry(n1,n2,n3,u1,u2,u3,rwk(iDu1),rwk(iDu2),rwk(iDu3))
      call addBdry(mynod,mynbr,mynbrSize,n1,n2,n3,rwk(iDu1),rwk(iDu2),
     $     rwk(iDu3),u1,u2,u3,rwk,BAL2EDGESUBASS,SCALAR)
c
      if (iupdateR) then
c
c ... Update r <-  (r - Su)
c
         call ApplyS(u1,u2,u3,
     $        rwk(iDu1),rwk(iDu2),rwk(iDu3),
     $        outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,bcType,nBCRegions,
     $        bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $        hx1,hx2,hx3,stat,cnt,alpha,D,rwk,iwk,SDptrs)
         call daxpbyBdry(n1,n2,n3,1.0d0,r1,r2,r3,
     $        -1.0d0,rwk(iDu1),rwk(iDu2),rwk(iDu3),
     $        r1,r2,r3)
      endif
c
 900  continue
      call memfreer(iMemHead,rwk)
      end
c
c ... This routine initializes the balancing preconditioner by forming 
c ... the coarse grid Matrix (N_i D_i Z_i){trans} S \sum (N_j D_j Z_j)
c
      subroutine InitBalance(outUnit,mynod,mynbr,mynbrSize,numproc,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,rlen,iwk,ilen,
     $     BALMAT,BalEval,BalEvec,DSDptrs,NSDptrs,SDparam)
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
      include '../Include/defines.hf'
      include '../Include/BCtypes.hf'
c
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2),numproc
      integer n1,n2,n3,rlen,ilen
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8 BALMAT(numproc,numproc)
      real*8 BalEval(numproc),BalEvec(numproc,numproc)
      real*8 alpha(*),D,rwk(rlen),SDparam(*)
      integer iwk(ilen),DSDptrs(*),NSDptrs(*),info
c
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
      integer face,dir
      integer itol1,itol2,itol3
      integer stat,cnt,i,i1,i2,iproc,itmp,ierr
      integer ibf1,ibf2,ibf3,idf1,idf2,idf3,iv1,iv2,iv3,ipres,irate
      integer ipv1,ipv2,ipv3,iProbeMemHead
      integer irow(100),icol(100),icnt,ir,ic
      real*8  val(100),v,asymmetry
      integer LocalMat,iptr
      integer bcType_save
      integer ibalptr1,ibalptr2,ibalptr3
      logical internalBoundary
      integer status(MPI_STATUS_SIZE)
c
c ... Commons
c
      integer iwt1,iwt2,iwt3,iwt1z,iwt2z,iwt3z,ibal1,ibal2,ibal3,
     $     balSlvType
      common /BalCom/iwt1,iwt2,iwt3,iwt1z,iwt2z,iwt3z,
     $     ibal1,ibal2,ibal3,balSlvType
c
c ... Functions
c
      integer memalr
      external memalr,memfreer
      real*8 ddotBdry
      external ddotBdry
c
      internalBoundary = .false.
      do dir = 1,3
         do face = 1,2
            if (mynbr(dir,face).ne.-1) internalBoundary = .true.
         enddo
      enddo
c
c ... Initialize Subdomain Solvers for neumann problems
c
      NSDptrs(3) = DSDptrs(3)
      NSDptrs(4) = DSDptrs(4)
      NSDptrs(5) = DSDptrs(5)
c
      if (internalBoundary) then
         bcType_save = bcType(nBCRegions)
         bcType(nBCRegions) = BCNEUMANN
      endif
c
      call InitSDSolvers(n1,n2,n3,x1,x2,x3,alpha,D,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,SDparam,
     $     iwk,ilen,rwk,rlen,NSDptrs)
c
      if (internalBoundary) then
         bcType(nBCRegions) = bcType_save
      endif
c     
c  ... Determine the local weights
c
      iwt1   = memalr(2*n2*n3,rwk)
      iwt2   = memalr(2*n1*n3,rwk)
      iwt3   = memalr(2*n1*n2,rwk)
      iwt1z   = memalr(2*n2*n3,rwk)
      iwt2z   = memalr(2*n1*n3,rwk)
      iwt3z   = memalr(2*n1*n2,rwk)
      ibal1   = memalr(7*2*n2*n3,rwk)
      ibal2   = memalr(7*2*n1*n3,rwk)
      ibal3   = memalr(7*2*n1*n2,rwk)
      call WGHT2(rwk(iwt1),rwk(iwt2),rwk(iwt3),.false.,
     $     mynod,mynbr,mynbrSize,n1,n2,n3,D,rwk)
c
c ... Make them into a partition of unity  wght = (my wght) / (sum of wght)
c
      itol1   = memalr(2*n2*n3,rwk)
      itol2   = memalr(2*n1*n3,rwk)
      itol3   = memalr(2*n1*n2,rwk)
      call addBdry(mynod,mynbr,mynbrSize,n1,n2,n3,
     $     rwk(iwt1),rwk(iwt2),rwk(iwt3),
     $     rwk(itol1),rwk(itol2),rwk(itol3),rwk,BAL3EDGESUBASS,SCALAR)
      call ddivideBdry(n1,n2,n3,rwk(iwt1),rwk(iwt2),rwk(iwt3),
     $     rwk(itol1),rwk(itol2),rwk(itol3),
     $     rwk(iwt1),rwk(iwt2),rwk(iwt3))
      call memfreer(itol1,rwk)
c
c ... These are the D_i
      call copyIface(mynbr,n1,n2,n3,rwk(iwt1),rwk(iwt2),rwk(iwt3),
     $     rwk(iwt1z),rwk(iwt2z),rwk(iwt3z))
c
c ... We now compute the entries in the matrix 
c
c ... Allocate local memory for subdomain solves
c
      iProbeMemHead = memalr(1,rwk)
      irate = memalr(n1*n2*n3,rwk)
      ibf1  = memalr(2*n2*n3,rwk)
      ibf2  = memalr(2*n1*n3,rwk)
      ibf3  = memalr(2*n1*n2,rwk)
      idf1  = memalr(2*n2*n3,rwk)
      idf2  = memalr(2*n1*n3,rwk)
      idf3  = memalr(2*n1*n2,rwk)
      iv1   = memalr((n1+1)*n2*n3,rwk)
      iv2   = memalr(n1*(n2+1)*n3,rwk)
      iv3   = memalr(n1*n2*(n3+1),rwk)
      ipres = memalr(n1*n2*n3,rwk)
      ipv1  = memalr(2*n2*n3,rwk)
      ipv2  = memalr(2*n1*n3,rwk)
      ipv3  = memalr(2*n1*n2,rwk)
      call dinitPres(n1,n2,n3,rwk(irate),0.0d0)
c
c ... Solve with own local weights
c
      call dinitPres(n1,n2,n3,rwk(ipres),0.0d0)
      call copyIface(mynbr,n1,n2,n3,rwk(iwt1),rwk(iwt2),rwk(iwt3),
     $     rwk(ipv1),rwk(ipv2),rwk(ipv3))

      call sdSolver(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,hx1,hx2,hx3,
     $     rwk(ipv1),rwk(ipv2),rwk(ipv3),
     $     rwk(ipres),rwk(iv1),rwk(iv2),rwk(iv3),
     $     rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $     rwk(idf1),rwk(idf2),rwk(idf3),stat,cnt,
     $     alpha,D,rwk(irate),rwk,iwk,DSDptrs)
      call dscalBdry(n1,n2,n3,-1.0d0,rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $     rwk(ibf1),rwk(ibf2),rwk(ibf3))
c
c ... Compute the entries in the balancing matrix
c     
      icnt = 0
      call BalMatCol(mynod,mynbr,mynod,n1,n2,n3,hx1,hx2,hx3,
     $     rwk(iwt1z),rwk(iwt2z),rwk(iwt3z),
     $     rwk(ibf1),rwk(ibf2),rwk(ibf3),icnt,irow,icol,val)
c
c ... Extract and save the action of S N_j D_j Z_j 
c
      ibalptr1 = ibal1
      ibalptr2 = ibal2
      ibalptr3 = ibal3
      call copyIface(mynbr,n1,n2,n3,rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $     rwk(ibalptr1),rwk(ibalptr2),rwk(ibalptr3))
c
c ... Compute the Fluxes from your neighbors weights
c
      do dir = 1,3
      do face = 1,2
         ibalptr1 = ibalptr1 + 2*n2*n3
         ibalptr2 = ibalptr2 + 2*n1*n3
         ibalptr3 = ibalptr3 + 2*n1*n2
         if (mynbr(dir,face).ne.-1) then
            call dinitBdry(n1,n2,n3,rwk(ipv1),rwk(ipv2),rwk(ipv3),0.0d0)
            if (dir.eq.1) then
               call copyfaceComp(n2,n3,rwk(iwt1z+(face-1)*n2*n3),
     $              rwk(ipv1+(face-1)*n2*n3))
            else if (dir.eq.2) then
               call copyfaceComp(n1,n3,rwk(iwt2z+(face-1)*n1*n3),
     $              rwk(ipv2+(face-1)*n1*n3))
            else
               call copyfaceComp(n1,n2,rwk(iwt3z+(face-1)*n1*n2),
     $              rwk(ipv3+(face-1)*n1*n2))
            endif
            call dinitPres(n1,n2,n3,rwk(ipres),0.0d0)
            call sdSolver(outUnit,mynod,mynbr,mynbrSize,
     $           n1,n2,n3,bcType,nBCRegions,
     $           bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,hx1,hx2,hx3,
     $           rwk(ipv1),rwk(ipv2),rwk(ipv3),
     $           rwk(ipres),rwk(iv1),rwk(iv2),rwk(iv3),
     $           rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $           rwk(idf1),rwk(idf2),rwk(idf3),stat,cnt,
     $           alpha,D,rwk(irate),rwk,iwk,DSDptrs)
            call dscalBdry(n1,n2,n3,-1.0d0,
     $           rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $           rwk(ibf1),rwk(ibf2),rwk(ibf3))
            call BalMatCol(mynod,mynbr,mynbr(dir,face),n1,n2,n3,
     $           hx1,hx2,hx3,rwk(iwt1z),rwk(iwt2z),rwk(iwt3z),
     $           rwk(ibf1),rwk(ibf2),rwk(ibf3),icnt,irow,icol,val)
            call copyIface(mynbr,n1,n2,n3,rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $           rwk(ibalptr1),rwk(ibalptr2),rwk(ibalptr3))
         else
            call dinitBdry(n1,n2,n3,rwk(ibalptr1),rwk(ibalptr2),
     $           rwk(ibalptr3),0.0d0)
         endif
      enddo
      enddo
c
c ... Free the memory
c
      call memfreer(iProbeMemHead,rwk)
c
cmpi      call gsyncx()
      call MPI_BARRIER(MPI_COMM_WORLD,ierr)
c
c ... Assemble the Balancing Matrix 
c
      do i2 = 1,numproc
         do i1 = 1,numproc
            BALMAT(i1,i2) = 0
         enddo
      enddo
c
      LocalMat = memalr(49*3+1,rwk)
      iptr = 0
      rwk(LocalMat+iptr) = icnt
      iptr = iptr + 1
      do i = 1,icnt
         rwk(LocalMat+iptr)   = irow(i)
         rwk(LocalMat+iptr+1) = icol(i)
         rwk(LocalMat+iptr+2) = val(i)
         BALMAT(irow(i)+1,icol(i)+1) = BALMAT(irow(i)+1,icol(i)+1) 
     $        + val(i)
        iptr = iptr + 3
      enddo
c
cmpi
c      do iproc = 0,numproc-1
c         if (iproc.ne.mynod) then
c            call sendrealx(BALMATSUBASS,rwk(LocalMat),(icnt*3+1),iproc)
c         endif
c      enddo
cc
c      do iproc = 0,numproc-2
c         call recvrealx(BALMATSUBASS,rwk(LocalMat),(49*3+1))
c         iptr = 1
c         do i = 1,int(rwk(LocalMat))
c            ir = int(rwk(LocalMat+iptr))
c            ic = int(rwk(LocalMat+iptr+1))
c            v  = rwk(LocalMat+iptr+2)
c            iptr = iptr + 3
c            BALMAT(ir+1,ic+1) = BALMAT(ir+1,ic+1) + v
c         enddo
c      enddo
      itmp = memalr(49*3+1,rwk)
      do iproc = 0,numproc-1
         if(mynod.eq.iproc) then
            do i = 0,49*3
               rwk(itmp+i) = rwk(LocalMat+i)
            enddo
         endif
         call MPI_Bcast(rwk(itmp),(49*3+1),MPI_REAL8,iproc,
     $        MPI_COMM_WORLD,ierr)
         if(mynod.ne.iproc) then
            iptr = 1
            do i = 1,int(rwk(itmp))
               ir = int(rwk(itmp+iptr))
               ic = int(rwk(itmp+iptr+1))
               v  = rwk(itmp+iptr+2)
               iptr = iptr + 3
               BALMAT(ir+1,ic+1) = BALMAT(ir+1,ic+1) + v
            enddo
         endif
      enddo
      call memfreer(itmp,rwk)

c
c ... Write a copy of the matrix to unit 16.
c
      if ((mynod.eq.0).and.(iverb.gt.4)) then
         write(16,*) 'Balancing Matrix',numproc,' by ',numproc
         do i2 = 1,numproc
            write(16,*) i2,(BALMAT(i2,i1),i1=1,numproc)
         enddo
      endif
c
c ... Symmetrize the Matrix
c 
      asymmetry = 0
      do ic = 1,numproc
         do ir = ic+1,numproc
            asymmetry = max(asymmetry,abs(BALMAT(ir,ic)-BALMAT(ic,ir)))
            BALMAT(ir,ic) = 0.5*(BALMAT(ir,ic)+BALMAT(ic,ir))
            BALMAT(ic,ir) = BALMAT(ir,ic)
         enddo
      enddo
      if ((mynod.eq.0).and.(iverb.gt.1)) 
     $     write(outUnit,*) 'Asymmetry is:',asymmetry
c
      call memfreer(LocalMat,rwk)
c
c ... decompose (or factor) the matrix
c
cta 17 July 2002
c      call WCOPY(numproc*numproc,BALMAT,1,BalEvec,1)
c      call WPOTRF('U',numproc,BalEvec,numproc,info)
      call dcopy(numproc*numproc,BALMAT,1,BalEvec,1)
      call dpotrf('U',numproc,BalEvec,numproc,info)
      if (info.gt.0) then
         if ((mynod.eq.0).and.(iverb.gt.1)) then
            write(*,*) 'Balance matrix is structurally singular',
     $           numproc-info
            write(*,*) 'Continuing with eigen-decomposition'
         endif
         itmp = memalr(2*numproc,rwk)
         call rs(numproc,numproc,BALMAT,BalEval,1,BalEvec,rwk(itmp),
     $        rwk(itmp+numproc),ierr)
         call memfreer(itmp,rwk)
         balSlvType = EIGENTYPE
c$$$         call Parcelcrash('Balmat did not factor',-30)
      else
         balSlvType = CHOLESKYTYPE
      endif
c     
c ... Print the Eigenvalues if desired
c
      if ((balSlvType.eq.EIGENTYPE).and.
     $     (mynod.eq.0).and.(iverb.gt.4)) then
         write(16,'(A,1x,2000E11.2)') 
     $        'Evals: ',(real(BalEval(i2)),i2=1,numproc)
         write(16,*) 'Balancing Evecs',numproc,' by ',numproc
         do i2 = 1,numproc
            write(16,*) i2,(BalEvec(i2,i1),i1=1,numproc)
         enddo
      endif
      return
      end
c
c ... This routine calculates the partial  entries into the coarse grid
c ... matrix from the local solves.  The routines assume VOLUMETRIC FLUX.
c ... For NONVOLUMETRIC FLUX replace this routine by the one below that 
c ... is commented out.
c
      subroutine BalMatCol(mynod,mynbr,colnum,n1,n2,n3,hx1,hx2,hx3,
     $     wt1,wt2,wt3,bf1,bf2,bf3,icnt,irow,icol,val)
      implicit none
      integer mynod,mynbr(3,2),n1,n2,n3,icnt,irow(*),icol(*),colnum
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 wt1(n2,n3,2),wt2(n1,n3,2),wt3(n1,n2,2)
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 val(*)
c
      real*8 ddotBdry
      external ddotBdry
c
      integer i1,i2,i3
c
      icnt = icnt + 1
      irow(icnt) = mynod
      icol(icnt) = colnum
      val(icnt)  =  ddotBdry(n1,n2,n3,hx1,hx2,hx3,wt1,wt2,wt3,
     $     bf1,bf2,bf3)
      
      if (mynbr(1,1).ne.-1) then
         icnt = icnt + 1
         val(icnt) = 0
         irow(icnt) = mynbr(1,1)
         icol(icnt) = colnum
         do i3 = 1,n3
            do i2 = 1,n2
               val(icnt) = val(icnt) + (1-wt1(i2,i3,1))*
     $              bf1(i2,i3,1)
            enddo
         enddo
      endif
      if (mynbr(1,2).ne.-1) then
         icnt = icnt + 1
         val(icnt) = 0
         irow(icnt) = mynbr(1,2)
         icol(icnt) = colnum
         do i3 = 1,n3
            do i2 = 1,n2
               val(icnt) = val(icnt) + (1-wt1(i2,i3,2))*
     $              bf1(i2,i3,2)
            enddo
         enddo
      endif
      if (mynbr(2,1).ne.-1) then
         icnt = icnt + 1
         val(icnt) = 0
         irow(icnt) = mynbr(2,1)
         icol(icnt) = colnum
         do i3 = 1,n3
            do i1 = 1,n1
               val(icnt) = val(icnt) + (1-wt2(i1,i3,1))*
     $              bf2(i1,i3,1)
            enddo
         enddo
      endif
      if (mynbr(2,2).ne.-1) then
         icnt = icnt + 1
         val(icnt) = 0
         irow(icnt) = mynbr(2,2)
         icol(icnt) = colnum
         do i3 = 1,n3
            do i1 = 1,n1
               val(icnt) = val(icnt) + (1-wt2(i1,i3,2))*
     $              bf2(i1,i3,2)
            enddo
         enddo
      endif
      if (mynbr(3,1).ne.-1) then
         icnt = icnt + 1
         val(icnt) = 0
         irow(icnt) = mynbr(3,1)
         icol(icnt) = colnum
         do i2 = 1,n2
            do i1 = 1,n1
               val(icnt) = val(icnt) + (1-wt3(i1,i2,1))*
     $              bf3(i1,i2,1)
            enddo
         enddo
      endif
      if (mynbr(3,2).ne.-1) then
         icnt = icnt + 1
         val(icnt) = 0
         irow(icnt) = mynbr(3,2)
         icol(icnt) = colnum
         do i2 = 1,n2
            do i1 = 1,n1
               val(icnt) = val(icnt) + (1-wt3(i1,i2,2))*
     $              bf3(i1,i2,2)
            enddo
         enddo
      endif
      return
      end
c$$$
c$$$      subroutine BalMatCol(mynod,mynbr,colnum,n1,n2,n3,hx1,hx2,hx3,
c$$$     $     wt1,wt2,wt3,bf1,bf2,bf3,icnt,irow,icol,val)
c$$$      implicit none
c$$$      integer mynod,mynbr(3,2),n1,n2,n3,icnt,irow(*),icol(*),colnum
c$$$      real*8 hx1(n1),hx2(n2),hx3(n3)
c$$$      real*8 wt1(n2,n3,2),wt2(n1,n3,2),wt3(n1,n2,2)
c$$$      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
c$$$      real*8 val(*)
c$$$c
c$$$      real*8 ddotBdry
c$$$      external ddotBdry
c$$$c
c$$$      integer i1,i2,i3
c$$$c
c$$$      icnt = icnt + 1
c$$$      irow(icnt) = mynod
c$$$      icol(icnt) = colnum
c$$$      val(icnt)  =  ddotBdry(n1,n2,n3,hx1,hx2,hx3,wt1,wt2,wt3,
c$$$     $     bf1,bf2,bf3)
c$$$      
c$$$      if (mynbr(1,1).ne.-1) then
c$$$         icnt = icnt + 1
c$$$         val(icnt) = 0
c$$$         irow(icnt) = mynbr(1,1)
c$$$         icol(icnt) = colnum
c$$$         do i3 = 1,n3
c$$$            do i2 = 1,n2
c$$$               val(icnt) = val(icnt) + (1-wt1(i2,i3,1))*
c$$$     $              bf1(i2,i3,1)*hx2(i2)*hx3(i3)
c$$$            enddo
c$$$         enddo
c$$$      endif
c$$$      if (mynbr(1,2).ne.-1) then
c$$$         icnt = icnt + 1
c$$$         val(icnt) = 0
c$$$         irow(icnt) = mynbr(1,2)
c$$$         icol(icnt) = colnum
c$$$         do i3 = 1,n3
c$$$            do i2 = 1,n2
c$$$               val(icnt) = val(icnt) + (1-wt1(i2,i3,2))*
c$$$     $              bf1(i2,i3,2)*hx2(i2)*hx3(i3)
c$$$            enddo
c$$$         enddo
c$$$      endif
c$$$      if (mynbr(2,1).ne.-1) then
c$$$         icnt = icnt + 1
c$$$         val(icnt) = 0
c$$$         irow(icnt) = mynbr(2,1)
c$$$         icol(icnt) = colnum
c$$$         do i3 = 1,n3
c$$$            do i1 = 1,n1
c$$$               val(icnt) = val(icnt) + (1-wt2(i1,i3,1))*
c$$$     $              bf2(i1,i3,1)*hx1(i1)*hx3(i3)
c$$$            enddo
c$$$         enddo
c$$$      endif
c$$$      if (mynbr(2,2).ne.-1) then
c$$$         icnt = icnt + 1
c$$$         val(icnt) = 0
c$$$         irow(icnt) = mynbr(2,2)
c$$$         icol(icnt) = colnum
c$$$         do i3 = 1,n3
c$$$            do i1 = 1,n1
c$$$               val(icnt) = val(icnt) + (1-wt2(i1,i3,2))*
c$$$     $              bf2(i1,i3,2)*hx1(i1)*hx3(i3)
c$$$            enddo
c$$$         enddo
c$$$      endif
c$$$      if (mynbr(3,1).ne.-1) then
c$$$         icnt = icnt + 1
c$$$         val(icnt) = 0
c$$$         irow(icnt) = mynbr(3,1)
c$$$         icol(icnt) = colnum
c$$$         do i2 = 1,n2
c$$$            do i1 = 1,n1
c$$$               val(icnt) = val(icnt) + (1-wt3(i1,i2,1))*
c$$$     $              bf3(i1,i2,1)*hx1(i1)*hx2(i2)
c$$$            enddo
c$$$         enddo
c$$$      endif
c$$$      if (mynbr(3,2).ne.-1) then
c$$$         icnt = icnt + 1
c$$$         val(icnt) = 0
c$$$         irow(icnt) = mynbr(3,2)
c$$$         icol(icnt) = colnum
c$$$         do i2 = 1,n2
c$$$            do i1 = 1,n1
c$$$               val(icnt) = val(icnt) + (1-wt3(i1,i2,2))*
c$$$     $              bf3(i1,i2,2)*hx1(i1)*hx2(i2)
c$$$            enddo
c$$$         enddo
c$$$      endif
c$$$      return
c$$$      end
c
c ... df <- S*(bc)
c
      subroutine ApplyS(bc1,bc2,bc3,df1,df2,df3,
     $     outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $     hx1,hx2,hx3,stat,cnt,alpha,D,rwk,iwk,DSDptrs)
      implicit none
c
      integer n1,n2,n3
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      real*8 df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2)
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      integer cnt,stat
      real*8 alpha(*),D,rwk(*)
      integer iwk(*),DSDptrs(*)
c
c ... Functions
c
      integer memalr
      external memalr,memfreer
c
      integer iProbeMemHead,ibf1,ibf2,ibf3,iv1,iv2,iv3,ipres,irate
c
      iProbeMemHead = memalr(1,rwk)
      irate = memalr(n1*n2*n3,rwk)
      ibf1  = memalr(2*n2*n3,rwk)
      ibf2  = memalr(2*n1*n3,rwk)
      ibf3  = memalr(2*n1*n2,rwk)
      iv1   = memalr((n1+1)*n2*n3,rwk)
      iv2   = memalr(n1*(n2+1)*n3,rwk)
      iv3   = memalr(n1*n2*(n3+1),rwk)
      ipres = memalr(n1*n2*n3,rwk)
      
      call dinitPres(n1,n2,n3,rwk(irate),0.0d0)
      call dinitPres(n1,n2,n3,rwk(ipres),0.0d0)
      call compflux123(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,bc1,bc2,bc3,
     $     rwk(ipres),rwk(iv1),rwk(iv2),rwk(iv3),
     $     rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $     df1,df2,df3,stat,cnt,
     $     alpha,D,rwk(irate),rwk,iwk,DSDptrs)
c
      call memfreer(iProbeMemHead,rwk)
c
      end
c
c ... u <- S^{-1}*(s)
c
      subroutine ApplySinv(s1,s2,s3,u1,u2,u3,
     $     outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $     hx1,hx2,hx3,stat,cnt,alpha,D,rwk,iwk,SDptrs)
      implicit none
c
      integer mynod,mynbr(3,2),mynbrSize(3,2,2),n1,n2,n3
      integer outUnit
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8  s1(n2,n3,2),s2(n1,n3,2),s3(n1,n2,2)
      real*8  u1(n2,n3,2),u2(n1,n3,2),u3(n1,n2,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  alpha(*),D,rwk(*)
      integer iwk(*),SDptrs(*),stat,cnt
c
c ... Functions
c
      integer memalr
      external memalr,memfreer
c
      integer iMemHead,ibf1,ibf2,ibf3,iv1,iv2,iv3,ipres,irate
c
      iMemHead = memalr(1,rwk)
      irate = memalr(n1*n2*n3,rwk)
      ibf1  = memalr(2*n2*n3,rwk)
      ibf2  = memalr(2*n1*n3,rwk)
      ibf3  = memalr(2*n1*n2,rwk)
      iv1   = memalr((n1+1)*n2*n3,rwk)
      iv2   = memalr(n1*(n2+1)*n3,rwk)
      iv3   = memalr(n1*n2*(n3+1),rwk)
      ipres = memalr(n1*n2*n3,rwk)
c
      call dinitPres(n1,n2,n3,rwk(irate),0.0d0)
      call dinitPres(n1,n2,n3,rwk(ipres),0.0d0)
      call sdSolver(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,s1,s2,s3,
     $     rwk(ipres),rwk(iv1),rwk(iv2),rwk(iv3),
     $     rwk(ibf1),rwk(ibf2),rwk(ibf3),u1,u2,u3,
     $     stat,cnt,alpha,D,rwk(irate),rwk,iwk,SDptrs)
c
c ... Extract those faces that are interfaces
c
      call copyIface(mynbr,n1,n2,n3,u1,u2,u3,u1,u2,u3)
c
      call memfreer(iMemHead,rwk)
c
      end
c
c ... RHS_i <- (N_i D_i Z_i)^{trans} S N_j D_j u_j (Part of RHS of EQN(15)
c
      subroutine balSu(mynod,mynbr,nproc,n1,n2,n3,hx1,hx2,hx3,
     $     Du1,Du2,Du3,bal1,bal2,bal3,RHS)
      implicit none
c
      integer mynod,mynbr(3,2),nproc,n1,n2,n3
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 Du1(n2,n3,2),Du2(n1,n3,2),Du3(n1,n2,2)
      real*8 bal1(n2,n3,2,7),bal2(n1,n3,2,7),bal3(n1,n2,2,7)
      real*8 RHS(0:nproc-1)
c
      real*8 ddotBdry
      external ddotBdry
c
      integer dir,face,k
c
      RHS(mynod) =  ddotBdry(n1,n2,n3,hx1,hx2,hx3,
     $     bal1(1,1,1,1),bal2(1,1,1,1),bal3(1,1,1,1),Du1,Du2,Du3)
c
      k = 1
      do dir = 1,3
         do face = 1,2
            k = k + 1
            if (mynbr(dir,face).ne.-1) then
               RHS(mynbr(dir,face)) = 
     $              ddotBdry(n1,n2,n3,hx1,hx2,hx3,
     $              bal1(1,1,1,k),bal2(1,1,1,k),bal3(1,1,1,k),
     $              Du1,Du2,Du3)
            endif
         enddo
      enddo
      end
c
c ... y <- (Coarse Matrix)^{pseudo inverse} x.
c
      subroutine pseudoInv(n,x,y,eval,evec,cutoff,iounit,slvType)
c     find pseudoinverse solution, from eigenvalue/eignevecor decomposiiton
c     if singular or Cholesky factorization if not.
c     eigenvalues smaller that cutoff are taken to be zero
c     on first call, check for zero eigenvalues 
c     
      include '../Include/defines.hf'
c     
      integer n,iounit,slvType
      real*8 x(n),y(n),eval(n),evec(n,n),cutoff,ev
c***  calls
cta 17 July 2002
c      real*8 WDOT
      real*8 ddot
c***  local
      real*8 fourier
      integer i,j, info
      logical first
      data first/.true./
      save first
      
c***  executable
      if (slvType.eq.CHOLESKYTYPE) then
cta 17 July 2002
c         call WCOPY(n,x,1,y,1)
c         call WPOTRS('U',n,1,evec,n,y,n,info)
         call dcopy(n,x,1,y,1)
         call dpotrs('U',n,1,evec,n,y,n,info)
c
      else if (slvType.eq.EIGENTYPE) then
c
         if(first.and.iounit.ge.0)then
            first=.false.
            do 212 j = 1,n
               ev=eval(j)
               if (abs(ev).lt.cutoff.or.ev.le.-cutoff)then
                  if(iounit.gt.0) write(iounit,*) 
     $                 'Small e-val',eval(j),
     $                 'with e-Vec:',
     x                 (evec(i,j),i=1,n)
               endif
               if(ev.lt.-cutoff) then
                  write(*,*) 'Negative eigenvalue: ',ev
                  call  Parcelcrash
     1                 ('negative eigenvalue in balancing matrix',
     $                 -34)
                endif
 212        continue
         endif
c***  solve
         do i = 1,n
            y(i) = 0
         enddo
         do 500 i = 1,n
            if (abs(eval(i)).gt.cutoff) then
cta 17 July 2002
c               fourier = WDOT(n,x,1,evec(1,i),1)
c               call WAXPY(n,fourier/eval(i),evec(1,i),1,y,1)
               fourier = ddot(n,x,1,evec(1,i),1)
               call daxpy(n,fourier/eval(i),evec(1,i),1,y,1)
            endif
 500     continue
      else
         call Parcelcrash('Unknown solver type for coarse matrix',
     $              -30)
      endif
      end
c
c ... Just a couple of calls to check if balancing is working
c
      subroutine TestBalance(outUnit,mynod,mynbr,mynbrSize,numproc,
     $        n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,hx1,hx2,hx3,
     $        alpha,D,rwk,iwk,BALMAT,BalEval,BalEvec,SDptrs)
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2),numproc
      integer n1,n2,n3
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8 BALMAT(numproc,numproc)
      real*8 BalEval(numproc),BalEvec(numproc,numproc)
      real*8 alpha(*),D,rwk(*)
      integer iwk(*),SDptrs(*)
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
c
      integer iu1,iu2,iu3,ir1,ir2,ir3
c
c ... Functions
c
      integer memalr
      external memalr,memfreer
      real*8 ddotBdry
      external ddotBdry
c
      iu1 = memalr(2*n2*n3,rwk)
      iu2 = memalr(2*n1*n3,rwk)
      iu3 = memalr(2*n1*n2,rwk)
      ir1 = memalr(2*n2*n3,rwk)
      ir2 = memalr(2*n1*n3,rwk)
      ir3 = memalr(2*n1*n2,rwk)
c      
      call dinitBdry(n1,n2,n3,rwk(ir1),rwk(ir2),rwk(ir3),0.0d0)
      call balance(rwk(iu1),rwk(iu2),rwk(iu3),rwk(ir1),rwk(ir2),
     $     rwk(ir3),.true.,.true.,.true.,.true.,BalEval,BalEvec,
     $     mynod,mynbr,mynbrSize,numproc,n1,n2,n3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs)
c
      call dinitBdry(n1,n2,n3,rwk(ir1),rwk(ir2),rwk(ir3),4.0d0)
      call copyIface(mynbr,n1,n2,n3,rwk(ir1),rwk(ir2),rwk(ir3),
     $     rwk(ir1),rwk(ir2),rwk(ir3))
      call balance(rwk(iu1),rwk(iu2),rwk(iu3),rwk(ir1),rwk(ir2),
     $     rwk(ir3),.true.,.true.,.true.,.true.,BalEval,BalEvec,
     $     mynod,mynbr,mynbrSize,numproc,n1,n2,n3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs)
      call balance(rwk(iu1),rwk(iu2),rwk(iu3),rwk(ir1),rwk(ir2),
     $     rwk(ir3),.true.,.false.,.false.,.true.,BalEval,BalEvec,
     $     mynod,mynbr,mynbrSize,numproc,n1,n2,n3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs)
c
      call BalPC(rwk(ir1),rwk(ir2),rwk(ir3),rwk(iu1),rwk(iu2),rwk(iu3),
     $     outUnit,mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs,
     $     BalEval,BalEvec)
c
      call memfreer(iu1,rwk)
c
      return
      end


      subroutine WGHT2(diag1,diag2,diag3,SubAssemble,
     $     mynod,mynbr,mynbrSize,n1,n2,n3,D,rwk)
c
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
      include '../Include/defines.hf'
      include '../Include/BCtypes.hf'
c
      logical SubAssemble
      integer n1,n2,n3
      real*8 diag1(n2,n3,2),diag2(n1,n3,2),diag3(n1,n2,2)
c
      integer mynod,mynbr(3,2),mynbrSize(3,2,2)
      real*8  rwk(*),D(*)
c
c ... Internal Common
c
      integer TensorType
      common /parcelcom/ TensorType
c
      integer n,ip,idf1,idf2,idf3,iet1,iet2,iet3
c
      integer memalr
      external memalr
c
      if (TensorType.eq.UserFuncType)  then
         call GetWghtsFcn(n1,n2,n3,diag1,diag2,diag3,D)
      else if (TensorType.eq.ArrayType) then
         call GetWghtsArray(n1,n2,n3,diag1,diag2,diag3,D)
      else if (TensorType.eq.EdgeTransType)  then
         iet1 = 1
         iet2 = iet1 + (n1+1)*n2*n3
         iet3 = iet2 + n1*(n2+1)*n3
         call GetWghtsEdgeTrans(n1,n2,n3,diag1,diag2,diag3,
     $        D(iet1),D(iet2),D(iet3))
      else
         call ParcelCrash('WGHT2:Unknown Tensor Type rep',-30)
      endif
c
c
c ... Subassemby the global diag from the local diags
c
      if (SubAssemble) then
         n   = n1*n2*n3
         ip = memalr(2*n,rwk)
         idf1 = memalr(2*n2*n3,rwk)
         idf2 = memalr(2*n1*n3,rwk)
         idf3 = memalr(2*n1*n2,rwk)
         call dcopyBdry(n1,n2,n3,diag1,diag2,diag3,rwk(idf1),rwk(idf2),
     $        rwk(idf3))
         call addBdry(mynod,mynbr,mynbrSize,n1,n2,n3,
     $        rwk(idf1),rwk(idf2),rwk(idf3),
     $        diag1,diag2,diag3,rwk,BAL1EDGESUBASS,SCALAR)
         call memfreer(ip,rwk)
      endif
c
c ... Set outside faces to one.
c
      if (mynbr(1,1).eq.-1) then
         call dinitFace(n2,n3,diag1(1,1,1),1.0d0)
      endif
      if (mynbr(1,2).eq.-1) then
         call dinitFace(n2,n3,diag1(1,1,2),1.0d0)
      endif
      if (mynbr(2,1).eq.-1) then
         call dinitFace(n1,n3,diag2(1,1,1),1.0d0)
      endif
      if (mynbr(2,2).eq.-1) then
         call dinitFace(n1,n3,diag2(1,1,2),1.0d0)
      endif
      if (mynbr(3,1).eq.-1) then
         call dinitFace(n1,n2,diag3(1,1,1),1.0d0)
      endif
      if (mynbr(3,2).eq.-1) then
         call dinitFace(n1,n2,diag3(1,1,2),1.0d0)
      endif
      end

      subroutine GetWghtsFcn(n1,n2,n3,diag1,diag2,diag3,userDfcn)
      implicit none
      integer n1,n2,n3
      real*8 diag1(n2,n3,2),diag2(n1,n3,2),diag3(n1,n2,2)
      real*8 userDfcn
      external userDfcn
c
      integer i1,i2,i3

      do i3 = 1,n3
         do i2 = 1,n2
            diag1(i2,i3,1) = userDfcn(1,1,1,i2,i3)
            diag1(i2,i3,2) = userDfcn(1,1,n1,i2,i3)
         enddo
      enddo
      do i3 = 1,n3
         do i1 = 1,n1
            diag2(i1,i3,1) = userDfcn(2,2,i1,1,i3)
            diag2(i1,i3,2) = userDfcn(2,2,i1,n2,i3)
         enddo
      enddo
      do i2 = 1,n2
         do i1 = 1,n1
            diag3(i1,i2,1) = userDfcn(3,3,i1,i2,1)
            diag3(i1,i2,2) = userDfcn(3,3,i1,i2,n3)
         enddo
      enddo
      end

      subroutine GetWghtsEdgeTrans(n1,n2,n3,p1,p2,p3,k1,k2,k3)
      integer n1,n2,n3
      real*8 p1(n2,n3,2),p2(n1,n3,2),p3(n1,n2,2)
      real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
c
      integer i1,i2,i3
c
      do 10 i3=1,n3
         do 11 i2=1,n2
            p1(i2,i3,1) = k1(0,i2,i3)
            p1(i2,i3,2) = k1(n1,i2,i3)
 11      continue
 10   continue
c     
      do 20 i3=1,n3
         do 21 i1=1,n1
            p2(i1,i3,1) = k2(i1,0,i3)
            p2(i1,i3,2) = k2(i1,n2,i3)
 21      continue
 20   continue
c     
      do 30 i2=1,n2
         do 31 i1=1,n1
            p3(i1,i2,1) = k3(i1,i2,0)
            p3(i1,i2,2) = k3(i1,i2,n3)
 31      continue
 30   continue      
      end
c
c
      subroutine GetWghtsArray(n1,n2,n3,diag1,diag2,diag3,D)
      implicit none
c
      integer n1, n2, n3, i1, i2, i3
      real*8 diag1(n2,n3,2),diag2(n1,n3,2),diag3(n1,n2,2)
      real*8 D(6, n1, n2, n3)
c
      do i3 = 1,n3
         do i2 = 1,n2
            diag1(i2,i3,1) = D(1,1,i2,i3)
            diag1(i2,i3,2) = D(1,n1,i2,i3)
         enddo
      enddo
      do i3 = 1,n3
         do i1 = 1,n1
            diag2(i1,i3,1) = D(4,i1,1,i3)
            diag2(i1,i3,2) = D(4,i1,n2,i3)
         enddo
      enddo
      do i2 = 1,n2
         do i1 = 1,n1
            diag3(i1,i2,1) = D(6,i1,i2,1)
            diag3(i1,i2,2) = D(6,i1,i2,n3)
         enddo
      enddo
      end
