c ...
c ...   THIS FILE MUST GET CLEANED UP AT SOME POINT SOON
c ...
c ...   Title:  Conjugate gradients on the interface problem
c ...        
c ...
c     ...  Description: Solves the equation Jump in Flux of correction = -
c ...                Jump in flux of the guess.
c ...   
c ...   Created:        Thu May 28 11:12:20 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 11:12:20 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: cg.bal.f,v 1.4 1997-08-20 11:46:40-05 arbogast Exp $
C ...  $RCSfile: cg.bal.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/cg.bal.f,v $
C  
c
c ... Version $Id: cg.bal.f,v 1.4 1997-08-20 11:46:40-05 arbogast Exp $
c
c ... Routine: cg123
c
c ... File:  cg.f
c
c ... Main function:
c     Conjugate gradients to solve df(lc) = - df(lg)
c     Calls compflux to evaluate the bilinear form B 
c     See Rice Masc Tech Reports :
c     Cowsar-Wheeler TR90-37 and Glowinski-Wheeler TR87-11 for details.
c
c ... Include Files: 
c     msgtypes.h
c     
c ... Calls: 
c     checkOB,rmBdry,compflux,applyPC and
c     myBlas, myBlasGlobal
c
c ... In: 
c     iverb :     Verbosity of output
c     idebug:     Debug level
c     outUnit:    Output unit number
c     MYNOD :     processor id
c     mynbr :     processor ids of neighbor subdomains, -1 if bdry edge
c     mynbrSize:  The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     bcType:     BCs for 1 - Neumann, 2 - Dirichlet, 3 - Noflow
c     globalNeuman: 1 if global problem is Neuman, 0 o.w.
c     x?    :     point centered grids
c     hx?   :  mesh widths hx?(i) = x?(i)-x?(i-1)
c     lc?   :  Guess at interface Values
c     niterIF : Maximum number of CG iterations
c     epsIF :  Relative residual reduction for Interface
c     tolIF :  Absolute residual required for Interface
c     df?   :  Right hand side  to solve B(lc) = df
c     pc?   :  Preconditioner 
c
c     dfc?,bf?,g?,w?,gwt?,presw:  CG work variables
c
c     niterSD :  Maximum iterations usuable to solve subdomain problems
c     epsSD :  Relative residual reduction in subdomains
c     tolSD :  Absolute residual required in subdomains
c     neuman:  logical denoting if problem is Neumann problem
c     a     :  Finite difference matrix
c     rate  :  Right hand side of -div A grad (pres) = (rate)
c     k?    :  Edge Permeability matrix
c     kp1   :  Number of directions to orthogonalize against
c
c     m0,p,q,rhs3,rhs2,rhs1, : Nested Factorization Variables
c     del2,del1,gamma        : Nested Factorization Variables
c     
c ... Out:
c     presc  :  Pressure values in subdomain corresponding to lc? BV's
c     lc?    :  Interface boundary values
c     g?     :  Jump in Boundary fluxes across sumdomain
c     cgstat :  return status of conjugate gradients
c     NormIn :  Initial norm of residual (weighted by preconditioner) 
c     NormOut:  Final norm of residual (weighted by preconditioner) 
c
c ... Interal Variables of note:
c     CGRSTRT : number of iterations after which an increase in the residual
c               is assumed due to pollution in recurances (restarted)
c     cgiter : current conjugate gradient iteration
c
c ... Comments:
c     See user documentation for notation for this routine
c
      subroutine CG123(iverb,idebug,outUnit,mynod,mynbr,mynbrSize,
     $     numproc,n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,
     $     bcRegion3,globalNeuman,x1,x2,x3,hx1,hx2,hx3,alpha,D,
     $     lc1,lc2,lc3,presc,vel1c,vel2c,vel3c,
     $     dfc1,dfc2,dfc3,df1,df2,df3,bf1,bf2,bf3,pc1,pc2,pc3,
     $     niterIF,epsIF,tolIF,cgstat,cgiter,
     $     nfstat,nfcntT,NormIn,NormOut,rwk,iwk,SDptrs,PCstyle,cond,
     $     balEval,balEvec)
      implicit none
      INTRINSIC sqrt
c
      include '../Include/msgtypes.hf'
      include '../Include/defines.hf'
      include '../Include/sdsolver.hf'
      include '../Include/BCtypes.hf'
c
      integer CGRSTRT
      parameter (CGRSTRT = 50)
c
      integer iverb,idebug
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2),numproc
      integer n1,n2,n3,globalNeuman,PCstyle
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  alpha(*),D
      real*8  lc1(n2,n3,2),lc2(n1,n3,2),lc3(n1,n2,2),presc(n1,n2,n3)
      real*8  vel1c(0:n1,n2,n3),vel2c(n1,0:n2,n3),vel3c(n1,n2,0:n3)
      real*8  dfc1(n2,n3,2),dfc2(n1,n3,2),dfc3(n1,n2,2)
      real*8  df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      real*8  bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2),cond
      real*8  BalEval(*),BalEvec(*)
      integer niterIF,cgstat
      integer nfcntT,nfcnt,nfstat
      real*8 epsIF,tolIF,NormIn,NormOut
      integer iwk(*),SDptrs(*)
      real*8 rwk(*)
c      real*8 NormInReal
c
c ... Conjugate Gradient Variables
c
      real*8  pc1(*),pc2(*),pc3(*)
      real*8  rho,psi,Bww
      integer irate,iStartptr
      integer n12,n13,n23
      integer ig1,ig2,ig3,igwt1,igwt2,igwt3,iw1,iw2,iw3
      integer ipresw,ivelw1,ivelw2,ivelw3
c
c ... Functions
c
      real*8 dnorm2BdryG,ddotBdryG,dsumBdryGNoArea
      integer memalr
c
c ... Local Variables
c
      integer cgiter,cgrestart
      real*8 g0Norm2,gnNorm2,gnp1Norm2,gwt0Norm2
      integer n
c
      real*8 rho1,eigmin,eigmax
      logical eval,noest
      integer id,ie,icndw,ier
c
      integer NSDptrs(MAXSDPTRS)
      common /BalCommon/ NSDptrs
c
c ... No errors yet
c
      cgstat = 0
      cgiter = 0
      cgrestart = 0
      nfcntT = 0
c
      noest = .true.
      eval = .true.
c      eval = .false.
c
      iStartptr = memalr(1,rwk)
c
      n = n1*n2*n3
      irate = memalr(n,rwk)
      call dinitPres(n1,n2,n3,rwk(irate),0.0d0)
      n23 = 2*n2*n3
      n12 = 2*n1*n2
      n13 = 2*n1*n3
c
c ... Build work vectors
c
      ig1 = memalr(n23,rwk)
      ig2 = memalr(n13,rwk)
      ig3 = memalr(n12,rwk)
      igwt1 = memalr(n23,rwk)
      igwt2 = memalr(n13,rwk)
      igwt3 = memalr(n12,rwk)
      iw1 = memalr(n23,rwk)
      iw2 = memalr(n13,rwk)
      iw3 = memalr(n12,rwk)
      ipresw = memalr(n,rwk)
      ivelw1 = memalr((n1+1)*n2*n3,rwk)
      ivelw2 = memalr(n1*(n2+1)*n3,rwk)
      ivelw3 = memalr(n1*n2*(n3+1),rwk)
      if (eval) then
         id = memalr(niterIF,rwk)
         ie = memalr(niterIF,rwk)
         icndw = memalr(3*niterIF,rwk)
         rwk(id) = 0
         rwk(ie) = 0
      endif
c
c ... If global problem is neuman, remove mean of df
c
      if (globalNeuman.eq.1) then
         if (idebug.gt.1) then
            call checkOB(mynod,mynbr,n1,n2,n3,df1,df2,df3,'df-m')
         endif
         if (idebug.gt.1) write(outUnit,*) 
     &       'Global Neumann encountered',mynod
         call rmBdryFluxes(mynbr,n1,n2,n3,x1,x2,x3,hx1,hx2,hx3,
     $        df1,df2,df3,VOLUMETRIC)
         if (idebug.gt.1) write(outUnit,*) 'Avg: ',
     $        dsumBdryGNoArea(n1,n2,n3,df1,df2,df3,BDRYSUMTYPE)
         if (idebug.gt.1) then
            call checkOB(mynod,mynbr,n1,n2,n3,df1,df2,df3,'df-mr')
         endif
      endif
c
c ... CG needs a balanced initial guess 
c
      if (PCstyle.eq.BALANCESTYLE) then
         call balance(lc1,lc2,lc3,df1,df2,df3,
     $        .true.,.true.,.false.,.false.,BalEval,BalEvec,
     $        mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $        bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs)
      endif
c     
c ... Compute First Residual
c
      if (idebug.gt.1) then
         call checkOB(mynod,mynbr,n1,n2,n3,lc1,lc2,lc3,'lc: First')
      endif
      call compflux123(outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,lc1,lc2,lc3,
     $     presc,vel1c,vel2c,vel3c,bf1,bf2,bf3,dfc1,dfc2,dfc3,
     $     nfstat,nfcnt,alpha,D,rwk(irate),rwk,iwk,SDptrs)
c
c ... Handle Error State
c
      if (nfstat.lt.0) then
         cgstat = -3
         if (iverb.gt.0) then
            write(outUnit,*) 'Warning: Nested Factorization returns: ',
     $           nfstat,' on computing first residual. Node: ',mynod
         endif
      endif
c
      if (idebug.gt.1) then
         call checkOB(mynod,mynbr,n1,n2,n3,dfc1,dfc2,dfc3,'dfc')
      endif
      if (idebug.gt.1) then
         call checkOB(mynod,mynbr,n1,n2,n3,df1,df2,df3,'df')
      endif
c     
c ... g = df - B(lc) 
c
      call daxpbyBdry(n1,n2,n3,-1.0d0,dfc1,dfc2,dfc3,1.0d0,df1,df2,df3,
     $     rwk(ig1),rwk(ig2),rwk(ig3))
      if (idebug.gt.1) then
         call checkOB(mynod,mynbr,n1,n2,n3,rwk(ig1),rwk(ig2),rwk(ig3),
     $        'g')
      endif
c$$$c Could put this in, but probably need to balance again.
c$$$c ... If global problem is neuman, remove mean of g
c$$$c
c$$$      if (globalNeuman.eq.1) then
c$$$         call rmBdryFluxes(mynbr,n1,n2,n3,x1,x2,x3,hx1,hx2,hx3,
c$$$     $        rwk(ig1),rwk(ig2),rwk(ig3),VOLUMETRIC)
c$$$      endif
c
      g0Norm2 = dnorm2BdryG(n1,n2,n3,hx1,hx2,hx3,
     $     rwk(ig1),rwk(ig2),rwk(ig3),FNORMSUMTYPE)
      NormIn = sqrt(g0Norm2)
c$$$      NormInReal = NormIn
c
      if (g0Norm2.lt.tolIF*tolIF) then
         gnp1Norm2 = g0Norm2
         cgstat = 1
         goto 999
      endif
c
c ... gwt = PC(g) ; w = gwt
c
      if (PCstyle.eq.BALANCESTYLE) then
         call BalPC(rwk(ig1),rwk(ig2),rwk(ig3),
     $        rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $        outUnit,mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $        bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs,
     $        BalEval,BalEvec)
      else
         call applyPC(n1,n2,n3,pc1,pc2,pc3,rwk(ig1),rwk(ig2),rwk(ig3),
     $        rwk(igwt1),rwk(igwt2),rwk(igwt3),mynbr)
      endif
      call dcopyBdry(n1,n2,n3,rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $     rwk(iw1),rwk(iw2),rwk(iw3))
c
      gwt0Norm2 = ddotBdryG(n1,n2,n3,hx1,hx2,hx3,
     $     rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $     rwk(ig1),rwk(ig2),rwk(ig3),FNORMSUMTYPE)
      gnp1Norm2 = gwt0Norm2
      NormIn = sqrt(gwt0Norm2)
c
      if (mynod.eq.0) then
      if ((gwt0Norm2.lt.0).and.(abs(gwt0Norm2).gt.1.0e-25)) then
         write(*,*) g0Norm2,gwt0Norm2
         call Parcelcrash('Preconditioned operator is not pd',-30)
      endif
      endif

      if ((mynod.eq.0).and.(iverb.gt.4)) then 
         write(outUnit,7000) cgiter,sqrt(gwt0Norm2),sqrt(g0Norm2)
 7000    format('CGIter: ',i5,2x,'Residual Norm: ',E10.3,1x,
     $        '(',E10.3,')')
      endif
c
      do 10 cgiter = 1, niterIF
c
         gnNorm2 = gnp1Norm2
c
c ... Compute B(w)
c
         call dinitPres(n1,n2,n3,rwk(ipresw),0.0d0)
         call compflux123(outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,
     $        bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,hx1,hx2,hx3,rwk(iw1),rwk(iw2),rwk(iw3),
     $        rwk(ipresw),rwk(ivelw1),rwk(ivelw2),rwk(ivelw3),
     $        bf1,bf2,bf3,
     $        dfc1,dfc2,dfc3,nfstat,nfcnt,alpha,D,rwk(irate),rwk,iwk,
     $        SDptrs)
         nfcntT = nfcntT + nfcnt
c
c ... Handle Error State
c
         if (nfstat.lt.0) then
            cgstat = -3
            if (iverb.gt.0) then
               write(outUnit,*) 'Warning: Nested Factorization ',
     $              'returns: ',nfstat,' on iteration # ',cgiter,
     $              ' Node: ',mynod
            endif
         endif
c     
         Bww = ddotBdryG(n1,n2,n3,hx1,hx2,hx3,
     $        rwk(iw1),rwk(iw2),rwk(iw3),dfc1,dfc2,dfc3,FLUXDOTTYPE)
c
         rho = gnNorm2 / Bww
c
c ... lc = lc - (rho)w
c
         call daxpbyBdry(n1,n2,n3,1.0d0,lc1,lc2,lc3,rho,
     $        rwk(iw1),rwk(iw2),rwk(iw3),lc1,lc2,lc3)
         if (idebug.gt.1) then
            call checkOB(mynod,mynbr,n1,n2,n3,lc1,lc2,lc3,'lc: Second')
         endif
c
c ... presc = presc + (rho)presw
c 
         call daxpbyPres(n1,n2,n3,1.0d0,presc,rho,rwk(ipresw),presc)
         call daxpbyVel(n1,n2,n3,1.0d0,vel1c,vel2c,vel3c,rho,
     $        rwk(ivelw1),rwk(ivelw2),rwk(ivelw3),vel1c,vel2c,vel3c)
c     
c ... g = g - (rho) B(w)
c     
         call daxpbyBdry(n1,n2,n3,1.0d0,rwk(ig1),rwk(ig2),rwk(ig3),
     $        -rho,dfc1,dfc2,dfc3,rwk(ig1),rwk(ig2),rwk(ig3))
c
c ... gwt = PC(g)
c
         if (PCstyle.eq.BALANCESTYLE) then
            call BalPC(rwk(ig1),rwk(ig2),rwk(ig3),
     $           rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $           outUnit,mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $           bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $           x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs,
     $           BalEval,BalEvec)
         else
            call applyPC(n1,n2,n3,pc1,pc2,pc3,rwk(ig1),rwk(ig2),
     $           rwk(ig3),rwk(igwt1),rwk(igwt2),rwk(igwt3),mynbr)
         endif

         gnp1Norm2 = ddotBdryG(n1,n2,n3,hx1,hx2,hx3,
     $        rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $        rwk(ig1),rwk(ig2),rwk(ig3),FLUXDOTTYPE)

         if (gnp1Norm2.lt.0) then
            if (mynod.eq.0) then
               write(outunit,*) 'Negative norm encountered: ',gnp1Norm2
            endif
            if (abs(gnp1Norm2).lt.1.0e-25) then
               gnp1Norm2 = abs(gnp1Norm2)
            else
               goto 800
            endif
         endif
c
c ... Write progress
c
         if ((mynod.eq.0).and.(iverb.gt.4)) then 
            if (gnp1Norm2.le.gnNorm2) then
               write(outUnit,7001) cgiter,sqrt(gnp1Norm2),
     $              sqrt(gnp1Norm2/gwt0Norm2)
 7001          format('CGIter: ',i5,2x,'Residual Norm: ',E10.3,1x,
     $              '(',E10.3,')')
            else
               write(outUnit,7002) cgiter,sqrt(gnp1Norm2),
     $              sqrt(gnp1Norm2/gwt0Norm2)
 7002          format('CGIter: ',i5,2x,'Residual Norm: ',E10.3,1x,
     $              '(',E10.3,')*')
            endif
         endif
c
c ... Check Convergence
c
      if ((gnp1Norm2/gwt0Norm2).lt.epsIF*epsIF) then
         cgstat = 0
         goto 999
      endif
      if (gnp1Norm2.lt.tolIF*tolIF) then
         cgstat = 1
         goto 999
      endif
c
c ... Update Search Direction  w = g + (psi) w
c         
      psi = gnp1Norm2/gnNorm2
c
      call daxpbyBdry(n1,n2,n3,1.0d0,rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $     psi,rwk(iw1),rwk(iw2),rwk(iw3),rwk(iw1),rwk(iw2),rwk(iw3))
c
      if (eval) then
         rho1 = 1d0/rho
         rwk(id+cgiter-1) = rwk(id+cgiter-1) + rho1
         rwk(id+cgiter) = psi*rho1
         rwk(ie+cgiter) = -sqrt(psi)*rho1
      endif
c
c ... For Steepest Descent : w = g,  Replace above section
c
c      call dcopyBdry(n1,n2,n3,rwk(ig1),rwk(ig2),rwk(ig3),
c    $      rwk(iw1),rwk(iw2),rwk(iw3))
c
c ... Check for polution of residual
c
      if ( .not. ( (gnp1Norm2.gt.gnNorm2)
     $     .and.((cgiter-cgrestart).gt.CGRSTRT) ) ) goto 10
c
 800  cgrestart = cgiter
c
      if (PCstyle.eq.BALANCESTYLE) then
         if ((iverb.gt.3).and.(mynod.eq.0))
     $        write(outUnit,*) 'Rebalancing initial residual'
         call balance(lc1,lc2,lc3,df1,df2,df3,
     $        .false.,.true.,.false.,.true.,BalEval,BalEvec,
     $        mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $        bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs)
      endif
c
c ... Recompute residual
c
      call compflux123(outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,lc1,lc2,lc3,
     $     presc,rwk(ivelw1),rwk(ivelw2),rwk(ivelw3),
     $     bf1,bf2,bf3,dfc1,dfc2,dfc3,
     $     nfstat,nfcnt,alpha,D,rwk(irate),rwk,iwk,SDptrs)
      nfcntT = nfcntT + nfcnt
c
c ... Handle error return code
c
      if (nfstat.lt.0) then
         cgstat = -3
         if (iverb.gt.0) then
            write(outUnit,*) 'Warning: Nested Factorization ',
     $           'returns: ',nfstat,' on restart after iteration # ',
     $           cgiter,' Node: ',mynod
         endif
      endif
c     
c ... g = B(lc) - df
c
      call daxpbyBdry(n1,n2,n3,-1.0d0,dfc1,dfc2,dfc3,1.0d0,
     $     df1,df2,df3,rwk(ig1),rwk(ig2),rwk(ig3))
c
c ... If global problem is neuman, remove mean of g
c
      if (globalNeuman.eq.1) then
         call rmBdryFluxes(mynbr,n1,n2,n3,x1,x2,x3,hx1,hx2,hx3,
     $        rwk(ig1),rwk(ig2),rwk(ig3),VOLUMETRIC)
      endif
c
c ... Restart CG: gwt = PC(g) ; w = gwt
c
      if (PCstyle.eq.BALANCESTYLE) then
         call BalPC(rwk(ig1),rwk(ig2),rwk(ig3),
     $        rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $        outUnit,mynod,mynbr,mynbrSize,numproc,n1,n2,n3,
     $        bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,hx1,hx2,hx3,alpha,D,rwk,iwk,SDptrs,
     $        BalEval,BalEvec)
      else
         call applyPC(n1,n2,n3,pc1,pc2,pc3,rwk(ig1),rwk(ig2),
     $        rwk(ig3),rwk(igwt1),rwk(igwt2),rwk(igwt3),mynbr)
      endif
      call dcopyBdry(n1,n2,n3,rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $     rwk(iw1),rwk(iw2),rwk(iw3))
      gnp1Norm2 = ddotBdryG(n1,n2,n3,hx1,hx2,hx3,
     $     rwk(igwt1),rwk(igwt2),rwk(igwt3),
     $     rwk(ig1),rwk(ig2),rwk(ig3),FLUXDOTTYPE)
      if (mynod.eq.0) then
         write(outUnit,*) 'Possible polluted residual. Restart: ',
     $        gnp1Norm2
      endif
      
 10   continue
c
c  ... Did not converge in ncgiter iterations
c
      cgstat = -2
c ... since cgiter = niterIF + 1 if the loop runs to completion
      cgiter = niterIF
c
 999  continue
c
c ... Report the unweighted norms in determining a stoping criteria
c
c      if (mynod.eq.0) 
c     $     write(outUnit,*) 'Norm PC: ',NormIn,sqrt(gnp1Norm2)
c      gnp1Norm2 = ddotBdryG(n1,n2,n3,hx1,hx2,hx3,
c     $     rwk(ig1),rwk(ig2),rwk(ig3),
c     $    rwk(ig1),rwk(ig2),rwk(ig3),FLUXDOTTYPE)
  
c      NormOut = sqrt(gnp1Norm2)
c      NormIn = NormInReal
c      if (mynod.eq.0) 
c     $     write(outUnit,*) 'Norm woPC: ',NormInReal,sqrt(gnp1Norm2)
c
      NormOut = sqrt(gnp1Norm2)
      if (cgiter.gt.0) then
         nfcntT = nfcntT / cgiter
      else
         nfcntT = 0
      endif
      if (globalNeuman.eq.1) then
         call rmBdry(mynbr,n1,n2,n3,x1,x2,x3,hx1,hx2,hx3,
     $        lc1,lc2,lc3)
      endif
ccss fix for 1 iteration
      if ((eval) .and. (cgiter .le. 1)) then
         cond = 1.0d0
         eigmin = 1.0d0
         eigmax = 1.0d0
      else
         if (eval) then
            cgiter = min(niterIF,cgiter)
            call bdcone(noest,0,0,rwk(id),rwk(ie),
     $        rwk(icndw),rwk(icndw+niterIF),rwk(icndw+2*niterIF),
     $        cgiter,eigmin,eigmax,cond,ier)
         endif
      endif
c
      call memfreer(iStartptr,rwk)
c
      return
      end




      subroutine bdcone(call1,iounit,ising,d,e,w1,w2,ind,nt,
     2   eigmin,eigmax,cond,ier)
	implicit real*8(a-h,o-z)
	implicit integer (i-n)
c***begin prologue  bdcone
c***refer to  bdpcg
c***routines called  ratqr
c***revision date 880602   (yymmdd)
c***end prologue  bdcone
c
c     this subroutine computes the smallest and largest eigenvalues of
c     a symmetric positive definite tridiagonal matrix of order nt. the
c     diagonal is stored in the array d and the sub-diag is stored in
c     the last nt-1 positions of the array e.  eigmin and eigmax are
c     updated with each call to bdcone.  once these are determined, an
c     estimate for the condition number is computed.  the arrays d and
c     e are unaltered.  two real*8 work arrays (w1, w2) and one integer
c     work array (ind) must be passed.
c
c     the latest change to this routine is to improve error handling.
c     in the case of a zero eigenvalue or nonzero error code returned
c     by ratqr, this routine tries to take smaller sections of the
c     tridiagonal in order to get estimates.
c
c	changed jm to look for (ising)+1st smallest eigenvalue
c     *** declarations ***
      parameter (m=1, idef=1)
      dimension d(*),e(*),w1(*),w2(*),ind(*)
      logical type, call1
	integer ising,mising
	parameter(mising=6)
	
c
c***first executable statement  bdcone
 1    continue
c
c     *** initialize lasest, marker for size of last trid. matrix
c     *** estimation was performed on
      if (call1) then
        cond = 1.0
        eigmin = d(1)
        eigmax = d(1)
        lasest = 1
        call1 = .false.
      end if
c
c     *** find smallest eigenvalue ***
      nteff = nt
 21   epsubr1 = 0.0
      type = .true.
      do 10 i = 2, nteff
         w1(i) = e(i)**2
 10   continue
      ierr = 0
      call ratqr(nteff,epsubr1,d,e,w1,ising+1,w2,ind,w1,type,idef,ierr)
      if (ierr .ne. 0) then
         ierrsv = ierr
         nteff = nteff - 1
         if (nteff .gt. lasest) go to 21
      end if
      do 220 i=1,ising+1
      emin=w2(i)
      if(abs(emin).gt.10000*epsubr1)goto 221
220   continue
221   continue
      if (emin .eq. 0e0) then
        nteff = nteff - 1
        if (nteff .gt. lasest) go to 21
      end if
c
      if (nteff.ne.nt .and. iounit.gt.0) then
         if (emin .eq. 0e0) write (iounit,22)
 22      format (/,'bdcone: estimated zero eigval, disregarding.',/)
         if (ierrsv .ne. 0) write(iounit,23) ierrsv
 23      format(/,' bdcone: ratqr returned ierr = ', i5, /)
      end if
      if (nteff .le. lasest) emin = eigmin
c
c     *** find largest eigenvalue ***
      nteff = nt
 31   epsubr1 = 0.0
      type = .false.
      do 20 i = 2, nteff
         w1(i) = e(i)**2
 20   continue
      ierr = 0
      call ratqr(nteff,epsubr1,d,e,w1,m,w2,ind,w1,type,idef,ierr)
      emax = w2(1)
      if (ierr .ne. 0) then
         ierrsv = ierr
         nteff = nteff - 1
         if (nteff .gt. lasest) go to 31
      end if
c
      if (nteff.ne.nt .and. iounit.gt.0) then
         write(iounit,23) ierrsv
      end if
      if (nteff .le. lasest) emax = eigmax
c
c     *** update condition number estimate ***
      eigmin = min(eigmin,emin)
      eigmax = max(eigmax,emax)
      if (eigmax*eigmin .lt. 0e0) go to 993
      cond = max(1d0, eigmax/eigmin, cond)
      lasest = nt
c
      if (iounit .gt. 0) write(iounit,50) eigmin, eigmax, cond
 50   format(4x, ' min, max eigenvalues = ', e12.5, ' and ', e12.5, /,
     2       4x, ' condition number     = ', e12.5, /)
c
      return
c
c
c     *** error returns ***
 993  ier = -7
      if (iounit.ge.0) write (iounit,893)
 893  format (' error: need pos. definite iteration matrix.')
      return
      end
