c ...
c ...   Title:  Parcel : Parallel Cell Centered Elliptic
c ...
c ...   Description:  Intel Parallel Code
c ... User interface program for solving 3D elliptic problems with zero order
c ... term using mixed method domain decomposition with preconditioning.
c ... Solves Elliptic Problem:  alpha pres - div D grad pres = f  
c ... via substructuring.  Alpha is a vector >= 0, D is a symmetric 3x3 
c ... matrix.
c ...  Most of the work is done in pressure() and the routines called therein.
c ...   
c ...   Created:        Thu May 28 11:10:32 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 11:10:32 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: parcel3D.f,v 1.3 1997-08-20 11:46:42-05 arbogast Exp $
C ...  $RCSfile: parcel3D.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/parcel3D.f,v $
C  
c
c     parcel.f
c
c ... Include Files:
c
c ... Calls: WhereAmI(),pressure()
c
c ... See the Users Guide for a LIST OF INPUT AND OUTPUT arguments
c
      subroutine parcel3D(majv,minv,
     $     n1,n2,n3,x1,x2,x3,l1,l2,l3,pres,vel1,vel2,vel3,
     $     alpha,D,dtype,f,bcType_in,nBCRegions_in,bcRegion1,bcRegion2,
     $     bcRegion3,bc1,bc2,bc3,mynbr,icube,rparam,iparam,idecomp,
     $     SDparam,rwk,len,iwk,ilen)
c
      implicit none
c
      include '../Include/msgtypes.hf'
      include '../Include/defines.hf'
      include '../Include/sdsolver.hf'
c
      integer MajorVersion,MinorVersion
      parameter(MajorVersion=1,MinorVersion=4)
c
      integer majv,minv
      integer n1,n2,n3
      real*8  x1(0:n1),x2(0:n2),x3(0:n3)
      real*8  pres(n1,n2,n3),l1(n2,n3,2),l2(n1,n3,2),l3(n1,n2,2)
      real*8  vel1(n1+1,n2,n3),vel2(n1,n2+1,n3),vel3(n1,n2,n3+1)
c
CLcc  real*8 alpha,D(9,n1,n2,n3),f(n1,n2,n3)
      real*8 alpha(*),D,f(n1,n2,n3)
      real*8 bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      integer nBCRegions_in,nBCRegions,bcType_in(nBCRegions_in),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2),
     $        mynbr(3,2),Method,dtype
c
      integer icube(3),iparam(9),idecomp(3)
      real*8  rparam(20),SDparam(4)
      integer len,ilen,iwk(ilen)
      real*8  rwk(len)
c
c ... Internal variables
c
      integer MYNOD,CUBESIZE,OutUnit,ising,bcTypeptr
      integer PCstyle,niterIF,niterSD,istat,kp1,i
      integer ndiv1,ndiv2,ndiv3
      real*8  epsIF,tolIF,epsSD,tolSD,times(10)
      integer memali
c
c ... Internal Common
c
      integer TensorType
      common /parcelcom/ TensorType
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
      nBCRegions = nBCRegions_in
c
      if (majv.ne.MajorVersion) then
         call Parcelcrash('Parcel: Major Version number is incorrect',
     $        -31)
      endif
      if ( minv .ne. MinorVersion) then
         if (icube(1).eq.0) then
            write(*,*) 'Warning Parcel3D() version number is ',
     $           MajorVersion,'.',MinorVersion
         endif
      endif
c
      MYNOD = icube(1)
      CUBESIZE = icube(3)
c
      outUnit = iparam(1)
      itest   = iparam(2)
      iverb   = iparam(3)
      idebug  = iparam(4)
      ising   = iparam(5)
      PCstyle = iparam(6)
      niterIF = iparam(7)
      istat   = iparam(8)
c
      epsIF   = rparam(1)
      tolIF   = rparam(2)
c
      ndiv1   = idecomp(1)
      ndiv2   = idecomp(2)
      ndiv3   = idecomp(3)
c
      TensorType = dtype
c
c ... Initialize memory usage
c
      call meminiti(iwk,ilen)
      call meminitr(rwk,len)
c
c ... Figure out local boundary conditions
c
      bcTypeptr = memali(nBCRegions+1,iwk)
      Method = 2
      call LocalBCs(n1,n2,n3,mynbr,bcType_in,iwk(bcTypeptr),nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,Method)

      niterSD = int(SDparam(initerSD))
      epsSD   = SDparam(iepsSD)
      tolSD   = SDParam(itolSD)
      kp1     = int(SDParam(4))
c
      if ((CUBESIZE.gt.MAXCUBE).and.(PCstyle.eq.BALANCESTYLE)) then
       call  Parcelcrash('Cube larger than Include/msgtypes.h:MAXCUBE',
     $        -32)
      endif
c
c ... Set the outer boundaries of the lambda's to zero
c ... Set the inter boundaries of the boundary conditions to zero
c
      call dinitNIFBdry(mynbr,n1,n2,n3,l1,l2,l3,0.0d0)
      call dinitIFBdry(mynbr,n1,n2,n3,bc1,bc2,bc3,0.0d0)
c
      call pressure(MYNOD,CUBESIZE,outUnit,itest,iverb,idebug,
     $     n1,n2,n3,ndiv1,ndiv2,ndiv3,mynbr,ising,x1,x2,x3,
     $     niterIF,epsIF,tolIF,kp1,niterSD,epsSD,tolSD,
     $     pres,l1,l2,l3,istat,PCstyle,vel1,vel2,vel3,times,
     $     alpha,iwk(bcTypeptr),nBCRegions,bcRegion1,bcRegion2,
     $     bcRegion3,bc1,bc2,bc3,f,D,SDparam,rwk,len,iwk,ilen)
c
c ... Figure out how much memory was used
c
      call memused(ilen,len,iwk,rwk)
c
c ... Set the outer boundaries of the lambda's back to zero
c
      call dinitNIFBdry(mynbr,n1,n2,n3,l1,l2,l3,0.0d0)
c
      iparam(8) = istat
      do i = 1,10
         rparam(2+i) = times(i)
      enddo
c
      return
      end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: parcel3D.f,v 1.3 1997-08-20 11:46:42-05 arbogast Exp $
c
c ... Routine: LocalBCs
c
c ... File:  whereami.f
c
c ... Main function:
c     determines type of local boundary conditions form global boundary
c     conditions
c
c ... Include Files: types.h
c     
c ... Calls: none
c
c ... In: 
c     mynbr: ids of neighbor processors, -1 if on the boundary
c     bcType: BC type 1-Neumann,2-Dirichlet,3-NoFlow
c     bcRegion?: BC regions
c
c ... Out:
c     bcType: BC type 1-Neumann,2-Dirichlet,3-NoFlow
c     bcRegion?: BC regions
c
c ... Internal Variables of note:
c
c ... Comments:
c
      subroutine LocalBCs(n1,n2,n3,mynbr,bcType_in,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,Method)
      implicit none
c
      include 'mpif.h'
      include '../Include/BCtypes.hf'
      integer n1,n2,n3,mynbr(3,2),nBCRegions,bcType_in(*),bcType(*),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2),
     $        Method
c
      integer i1,i2,i3,region
      integer nnodesx,ierr
c
      do region = 1,nBCRegions
         bcType(region) = bcType_in(region)
      enddo
c
      call MPI_Comm_size(MPI_COMM_WORLD,nnodesx,ierr)
      if(nnodesx.eq.1) return
c
      nBCRegions = nBCRegions + 1
      region = nBCRegions
      if (mynbr(1,1).ne.-1) then
         do i3 = 1,n3
            do i2 = 1,n2
               bcRegion1(i2,i3,1) = region
            enddo
         enddo
      endif
      if (mynbr(1,2).ne.-1) then
         do i3 = 1,n3
            do i2 = 1,n2
               bcRegion1(i2,i3,2) = region
            enddo
         enddo
      endif
      if (mynbr(2,1).ne.-1) then
         do i3 = 1,n3
            do i1 = 1,n1
               bcRegion2(i1,i3,1) = region
            enddo
         enddo
      endif
      if (mynbr(2,2).ne.-1) then
         do i3 = 1,n3
            do i1 = 1,n1
               bcRegion2(i1,i3,2) = region
            enddo
         enddo
      endif
      if (mynbr(3,1).ne.-1) then
         do i2 = 1,n2
            do i1 = 1,n1
               bcRegion3(i1,i2,1) = region
            enddo
         enddo
      endif
      if (mynbr(3,2).ne.-1) then
         do i2 = 1,n2
            do i1 = 1,n1
               bcRegion3(i1,i2,2) = region
            enddo
         enddo
      endif
      if (Method.eq.2) then 
         bcType(nBCRegions) = BCDIRICHLET
      else if (Method.eq.3) then
         bcType(nBCRegions) = BCROBIN
      else
         call parcelCrash('Unknown interface matching method',-33)
      endif
c
c METHOD 4 IS NOT SUPPORTED
c
c$$$      do 10 i = 1,2
c$$$         do 20 j = 1,3
c$$$            if (Method.eq.4) then
c$$$               if (j.eq.2) then
c$$$                  mykbc(j,i) = BCNEUMANN
c$$$               else
c$$$                  mykbc(j,i) = BCDIRICHLET
c$$$               endif
c$$$            endif
c$$$ 20      continue
c$$$ 10   continue
c
      return
      end
c
c ... In: 
c     version :     Version number for ellip().      
c     n?    :       number of grid points in each direction
c     x?    :       point centered grid for this subdomain
c     l?    :       Guess for interfacial values
c     alpha :       Coeff. of zero order term
c     D     :       Tensor
c     dtype :       Type of information contained in D
c     f     :       RHS for elliptic problem
c     ibctype(3,2) : Types of boundary conditions for this subdomain
c                    1 - Neumann, 2 - Dirichlet, 3 - Noflo
c     bc?   :      Boundary values for this subdomain.  Values are used only
c                  for those edges that are real external boundary conditions
c     icube :      Cube information 
c        icube(1) =   MYNOD :       my processor id
c        icube(3) =   CUBESIZE:   size of cube
c     rparam :     Real parameters for interface problem
c        rparam(1) = epsIF :       Relative residual reduction for Interface
c        rparam(2) = tolIF :       Absolute residual required for Interface
c     iparam :     Integer parameters for interface problem
c        iparam(1) = outUnit:      Output unit number
c        iparam(2) = itest (it):   test problem number
c        iparam(3) = iverb (iv):   verbosity of output
c        iparam(4) = idebug (id):  debug level
c        iparam(5) = ising   1 if global problem is singular, 0 o.w.
c        iparam(6) = PCstyle Type of preconditioner used for interface problem
c        iparam(7) = niterIF : Maximum iterations to solve interface problem
c     idecomp :    Integer parameters used to describe the decomposition
c        idecomp(1:3) = ndiv[1-3] : number of subdomains in each direction
c     SDparam :    Real parameters used to describe the subdomain solver
c        SDparam(1) = niterSD : Maximum iterations to solve subdomain problems
c        SDparam(2) = epsSD :   Relative residual reduction in subdomains
c        SDParam(3) = tolSD :   Absolute residual required in subdomains
c        SDParam(4) = kp1   :   Number of directions to orthogonalize against
c     rwk(len) : Real vector of length of lenght len
c     
c ... Out:
c     pres  :        Pressure values in subdomain
c     vel?  :        Velocities at Mixed method places
c     l?    :        interfacial values (Lagrange Multiplier)
c     iparam(8) =    return status, >= 0 if no error
c     rparam(3:12) = Execution times for certain phases of the code
c
