c ...
c ...   Title:  Solves the interface problem
c ...
c ...   Created:        Thu May 28 11:17:58 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 11:17:58 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: pressure.f,v 1.3 1997-08-20 11:46:44-05 arbogast Exp $
C ...  $RCSfile: pressure.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/pressure.f,v $
C  
c
c     pressure.f
c
c ... MMDD:  $Id: pressure.f,v 1.3 1997-08-20 11:46:44-05 arbogast Exp $
c ... Pressure program for 3D MMDD (Mixed method domain decomposition)
c
c ... Main function:
c ... Solves Elliptic Problem:  - div K grad pres = f  via substructuring
c ... conjugate gradient main decomposition METHOD 2.  See Rice Masc Tech 
c ... Reports Cowsar-Wheeler TR90-37 and Glowinski-Wheeler TR87-11 for details
c
c ... Include Files:
c     msgtypes.h --- program defined message types and sizes
c
c ... Calls:
c
c ... Major Labels:
c      600  Continuation for one node only
c
c ... In: 
c     MYNOD :       my processor id
c     outUnit:      Output unit number
c     itest (it):   test problem number
c     iverb (iv):   verbosity of output
c     idebug (id):  debug level
c     n?    :       number of grid points in each direction
c     ndiv? :       number of subdomains in each direction
c     mynbr(3,2) :  processor ids of neighbor subdomains
c     bcType:       BC for subdomain: 1 - Neumann, 2 - Dirichlet, 3 - Noflo
c     globalNeuman: 1 if global problem is Neuman, 0 o.w.
c     x?    :       point centered grid for this subdomain
c     niterIF :     Maximum iterations usuable to solve interface problem
c     epsIF :       Relative residual reduction for Interface
c     tolIF :       Absolute residual required for Interface
c     kp1   :       Number of directions to orthogonalize against
c     niterSD :     Maximum iterations usuable to solve subdomain problems
c     epsSD :       Relative residual reduction in subdomains
c     tolSD :       Absolute residual required in subdomains
c     pres  :       Guess for Pressure
c     l?    :       Guess for interfacial values
c     times :         Execution times for certain phases of the code
c     vel?  :         Velocities at Mixed method places
c     
c ... Out:
c     pres  :       Pressure values in subdomain
c     l?    :       interfacial values (Lagrange Multiplier)
c     istat :       return status, >= 0 if no error.
c
c ... Internal Variables of note
c     lc?,presc:    Corrections for pressure and interfacial values
c     pc?   :       Preconditioner on the interface values
c     mynbrSize:    The dimension of neighbor faces 
c                   (used for local grid refinement)
c     nfcntT :      Average number of iterations to solve subdomain problem for
c                   a given smoothing
c     rate   :      Place for right hand side of - div K grad pres = f
c     K      :      Diagonal tensor of permeabilities
c     k?     :      Normal components permeability on iterfaces
c     a,b    :      Matrix representation and RHS of subdomain problem
c     df?    :      Jump in fluxes across interfaces
c
      subroutine pressure(MYNOD,CUBESIZE,outUnit,it,iv,id,
     $     n1,n2,n3,ndiv1,ndiv2,ndiv3,mynbr,globalNeuman,x1,x2,x3,
     $     niterIF,epsIF,tolIF,kp1,niterSD,epsSD,tolSD,
     $     pres,l1,l2,l3,istat,PCstyle,vel1,vel2,vel3,times,
     $     alpha,bcType,nBCRegions,bcRegion1,bcRegion2,
     $     bcRegion3,bc1,bc2,bc3,rate,D,SDparam,rwk,len,iwk,ilen)
c
      implicit none
c
      include '../Include/msgtypes.hf'
      include '../Include/BCtypes.hf'
      include '../Include/defines.hf'
      include '../Include/sdsolver.hf'
c
      integer MYNOD,CUBESIZE
      integer it,iv,id
      integer n1,n2,n3,ndiv1,ndiv2,ndiv3,mynbr(3,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),epsSD,tolSD,epsIF,tolIF
      integer niterSD,kp1,niterIF
      real*8  pres(n1,n2,n3),l1(n2,n3,2),l2(n1,n3,2),l3(n1,n2,2)
      real*8  bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      real*8  D,rate(n1,n2,n3),alpha(*)
      real*8  vel1(n1+1,n2,n3),vel2(n1,n2+1,n3),vel3(n1,n2,n3+1)
      real*8  times(10),cond
      integer istat,outUnit,nfstat,nfcntT,cgstat,cgiter,globalNeuman
      integer PCstyle
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      integer len,ilen,iwk(ilen)
      real*8  rwk(len),SDparam(*)
      integer DSDptrs(MAXSDPTRS)
c
      integer mynbrSize(3,2,2)
c
c ... Corrections 
c      
      integer ipresc,ivel1c,ivel2c,ivel3c
      integer ilc1,ilc2,ilc3
c
c ... Diagonal preconditioner
c
      integer ipc1,ipc2,ipc3
c
c ... Conjugate Gradient variables
c
      integer idfc1,idfc2,idfc3
c
c ... Counters
c
      integer region
c
c ... Related meshes
c
      integer ix1c,ix2c,ix3c,ihx1,ihx2,ihx3
      integer ip
c
c ... Memory Allocation
c
      integer memalr,memali
      external memalr,memali
      integer iMemStartI,iMemStartR
c
c ... Bilinear form rhs variables
c      
      integer ibf1,ibf2,ibf3,idf1,idf2,idf3
c
c ... Norms and times
c
      real*8 NormIn,NormOut,VNormStart,VNormIn,VNormOut
      real*8 stime,stime1,etime
      integer itmp
c
c ... Functions called
c
cmpi      integer clockx
      double precision MSG_f_Wtime
c
c ... Commons
c
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
      real*8 RobinR
      common/Robin/RobinR
c
      integer NSDptrs(MAXSDPTRS)
      common /BalCommon/ NSDptrs
      integer balMatPtr,BalEvalPtr,BalEvecPtr
c
c Initialize some pointers used only for balancing
c
      balMatPtr = 1
      BalEvalPtr = 1
      BalEvecPtr = 1
c
c ... No Return Error Yet
c
      istat = 0
c
c ... Initialize Common Variables
c
      itest = it
      iverb = iv
      idebug = id
c
c ... Start Timing
c
cmpi      stime = dble(clockx())*0.001d0
      stime = MSG_f_Wtime()

      if (iverb.gt.5) then
         write(outUnit,111) mynod,1,n1,n2,n3
 111     format('Node: ',i3,3x,'Level: ',i2,3x,'Grid: ',3i4)
      endif
c
c ... Mark start of memory
c
      iMemStartI = memali(1,iwk)
      iMemStartR = memalr(1,rwk)
c
c ... Share Dimensions with neighbors for transfer operators
c
      call GetNbrSize(mynod,mynbr,mynbrSize,n1,n2,n3)
c
c ... Make related Meshes
c
      ix1c = memalr(n1,rwk)
      ix2c = memalr(n2,rwk)
      ix3c = memalr(n3,rwk)
      ihx1 = memalr(n1,rwk)
      ihx2 = memalr(n2,rwk)
      ihx3 = memalr(n3,rwk)
      call makemesh(n1,n2,n3,x1,x2,x3,rwk(ix1c),rwk(ix2c),rwk(ix3c),
     $     rwk(ihx1),rwk(ihx2),rwk(ihx3))
c
c ... Put the boundary conditions in with lambda's
c
      call daxpbyBdry(n1,n2,n3,1.0d0,l1,l2,l3,1.0d0,bc1,bc2,bc3,
     $     l1,l2,l3)
c
c ... Initialize Subdomain Solvers
c
cmpi      times(4) = dble(clockx())*0.001d0
      times(4) = MSG_f_Wtime()
c
c ... Set this to zero so that k1,k2,k3 perms are made
      DSDptrs(3) = 0
      call InitSDSolvers(n1,n2,n3,x1,x2,x3,alpha,D,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,SDparam,iwk,ilen,rwk,len,
     $     DSDptrs)
cmpi      times(4) = dble(clockx())*0.001d0 - times(4)
      times(4) = MSG_f_Wtime() - times(4)
c
c ... Initialize Preconditioner for Interface Problem
c
      ipc1   = memalr(2*n2*n3,rwk)
      ipc2   = memalr(2*n1*n3,rwk)
      ipc3   = memalr(2*n1*n2,rwk)
c
      if (PCstyle.eq.NOPRECOND) then
c         call dinitBdry(n1,n2,n3,rwk(ipc1),rwk(ipc2),rwk(ipc3),1.0d0)  
         itmp = 16*max(max(n1*n2,n1*n3),n2*n3)
         ip   = memalr(5*itmp,rwk)
         call initPC(mynod,mynbr,mynbrSize,n1,n2,n3,D,
     $        rwk(ihx1),rwk(ihx2),rwk(ihx3),
     $        rwk(ipc1),rwk(ipc2),rwk(ipc3),
     $        rwk(ip+2*itmp),rwk(ip+3*itmp),rwk(ip+4*itmp),itmp,
     $        rwk(ip),rwk(ip+itmp),IDENTITY,VOLUMETRIC)
         call memfreer(ip,rwk)
      else if (PCstyle.eq.DIAGONAL) then
         itmp = 16*max(max(n1*n2,n1*n3),n2*n3)
         ip   = memalr(5*itmp,rwk)
         call initPC(mynod,mynbr,mynbrSize,n1,n2,n3,D,
     $        rwk(ihx1),rwk(ihx2),rwk(ihx3),
     $        rwk(ipc1),rwk(ipc2),rwk(ipc3),
     $        rwk(ip+2*itmp),rwk(ip+3*itmp),rwk(ip+4*itmp),itmp,
     $        rwk(ip),rwk(ip+itmp),ARITHMETIC,VOLUMETRIC)
         call memfreer(ip,rwk)
      else if (PCstyle.eq.BALANCESTYLE) then
         balMatPtr = memalr(CUBESIZE*CUBESIZE,rwk)
         BalEvalPtr = memalr(CUBESIZE,rwk)
         BalEvecPtr = memalr(CUBESIZE*CUBESIZE,rwk)
         call InitBalance(outUnit,mynod,mynbr,mynbrSize,CUBESIZE,
     $        n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,rwk(ihx1),rwk(ihx2),rwk(ihx3),
     $        alpha,D,rwk,len,iwk,ilen,
     $        rwk(balMatPtr),rwk(BalEvalPtr),rwk(BalEvecPtr),
     $        DSDptrs,NSDptrs,SDparam)
c$$$         call TestBalance(outUnit,mynod,mynbr,mynbrSize,CUBESIZE,
c$$$     $        n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,
c$$$     $        bcRegion3,x1,x2,x3,rwk(ihx1),rwk(ihx2),rwk(ihx3),
c$$$     $        alpha,D,rwk,iwk,rwk(balMatPtr),rwk(BalEvalPtr),
c$$$     $        rwk(BalEvecPtr),DSDptrs)
      else
         call Parcelcrash('Unknown pcstyle',-30)
      endif
c     
c ... End initialization
c
c ... Report Set-up Time
c
cmpi      etime = dble(clockx())*0.001d0
      etime = MSG_f_Wtime()
      times(1) = etime - stime
      if ((mynod.eq.0).and.(iverb.gt.2)) then
         write(outUnit,*) 'Setup Time: ',REAL(etime-stime),' (secs)'
      endif
c
c ... Begin Timing Residual formation time
c
cmpi      stime = dble(clockx())*0.001d0
      stime = MSG_f_Wtime()
c
c ... Solve Problem with initial guess. Pres(lg)
c
      ibf1   = memalr(2*n2*n3,rwk)
      ibf2   = memalr(2*n1*n3,rwk)
      ibf3   = memalr(2*n1*n2,rwk)
      idf1   = memalr(2*n2*n3,rwk)
      idf2   = memalr(2*n1*n3,rwk)
      idf3   = memalr(2*n1*n2,rwk)

      call compflux123(outUnit,mynod,mynbr,mynbrSize,n1,n2,n3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,rwk(ihx1),rwk(ihx2),rwk(ihx3),
     $     l1,l2,l3,pres,vel1,vel2,vel3,
     $     rwk(ibf1),rwk(ibf2),rwk(ibf3),rwk(idf1),rwk(idf2),rwk(idf3),
     $     nfstat,nfcntT,alpha,D,rate,rwk,iwk,DSDptrs)
c
c ... Handle Error State
c
      if (nfstat.lt.0) then
         istat = -2
         if (iverb.gt.0) then
            write(outUnit,*) 'Warning: Nested Factorization returns: ',
     $           nfstat,' on computing Residual Formulation. Node: ',
     $           mynod
         endif
      endif
c
c ... If no domain decomposition, then this is solution
c
      if (ndiv1*ndiv2*ndiv3.eq.1) then
cmpi         etime = dble(clockx())*0.001d0
         etime = MSG_f_Wtime()
         if ((mynod.eq.0).and.(iverb.gt.2)) then
            write(outUnit,*) 'Execution Time: ',
     &           REAL(etime-stime),' (secs)',
     $           ' with status: ',nfstat,istat
         endif
         goto 600
      endif
c
c ... (Change sign of delta flux) df  =  -(df)
c
      call dscalBdry(n1,n2,n3,-1.0d0,rwk(idf1),rwk(idf2),rwk(idf3),
     $     rwk(idf1),rwk(idf2),rwk(idf3))
c
c ... Let lg be lamda guess, lc be lamda residual
c ... Place in residual Form and seek:  
c ...         l=lg+lc; Pres(l)=Pres(lg)+Pres(lc)
c 
      do region = 1,nBCRegions-1
         if (bcType(region).eq.BCNEUMANN) bcType(region) = BCNOFLOW
      enddo
c
c ... Allocate work space
c
      ipresc = memalr(n1*n2*n3,rwk)
      ivel1c = memalr((n1+1)*n2*n3,rwk)
      ivel2c = memalr(n1*(n2+1)*n3,rwk)
      ivel3c = memalr(n1*n2*(n3+1),rwk)
      ilc1   = memalr(2*n2*n3,rwk)
      ilc2   = memalr(2*n1*n3,rwk)
      ilc3   = memalr(2*n1*n2,rwk)
      idfc1   = memalr(2*n2*n3,rwk)
      idfc2   = memalr(2*n1*n3,rwk)
      idfc3   = memalr(2*n1*n2,rwk)

      call dinitBdry(n1,n2,n3,rwk(ilc1),rwk(ilc2),rwk(ilc3),0.0d0)
      call dinitPres(n1,n2,n3,rwk(ipresc),0.0d0)
c
c ... Report Residual Formulation Time
c
cmpi      etime = dble(clockx())*0.001d0
      etime = MSG_f_Wtime()
      times(2) = etime - stime
      if ((mynod.eq.0).and.(iverb.gt.2)) then
         write(outUnit,*) 'Residual Time: ',
     &        REAL(etime-stime),' (secs)'
      endif
c
c ... Begin Timing Solution
c
cmpi      stime = dble(clockx())*0.001d0
      stime = MSG_f_Wtime()
      stime1 = stime
c
 1001 continue
c
c ... CG needs a balanced initial guess
c ... This is now done internal to CG123
c
c$$$      if (PCstyle.eq.BALANCESTYLE) then
c$$$         call balance(rwk(ilc1),rwk(ilc2),rwk(ilc3),rwk(idf1),rwk(idf2),
c$$$     $        rwk(idf3),.true.,.true.,.false.,.false.,
c$$$     $        rwk(BalEvalPtr),rwk(BalEvecPtr),
c$$$     $        mynod,mynbr,mynbrSize,CUBESIZE,n1,n2,n3,
c$$$     $        bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
c$$$     $        x1,x2,x3,rwk(ihx1),rwk(ihx2),rwk(ihx3),alpha,D,rwk,iwk,
c$$$     $        DSDptrs)
c$$$      endif
c
c ... Solve df(lc) = - df(lg) (Note: df(lg) now stores -df(lg))
c
      call CG123(iverb,idebug,outUnit,mynod,mynbr,
     $     mynbrSize,CUBESIZE,n1,n2,n3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,globalNeuman,x1,x2,x3,
     $     rwk(ihx1),rwk(ihx2),rwk(ihx3),
     $     alpha,D,rwk(ilc1),rwk(ilc2),rwk(ilc3),
     $     rwk(ipresc),rwk(ivel1c),rwk(ivel2c),rwk(ivel3c),
     $     rwk(idfc1),rwk(idfc2),rwk(idfc3),
     $     rwk(idf1),rwk(idf2),rwk(idf3),
     $     rwk(ibf1),rwk(ibf2),rwk(ibf3),
     $     rwk(ipc1),rwk(ipc2),rwk(ipc3),
     $     niterIF,epsIF,tolIF,cgstat,cgiter,nfstat,nfcntT,
     $     NormIn,NormOut,rwk,iwk,DSDptrs,PCstyle,cond,
     $     rwk(BalEvalPtr),rwk(BalEvecPtr))
c
      times(2) = cgiter
      times(10) = cond
c
c ... Handle error
c
      if (cgstat.lt.0) then
         istat = cgstat
      endif
c
c ... Save Values for Report if only One Level
c     
      VNormIn = NormIn
      VNormOut = NormOut
      VNormStart = NormIn
c
c ... Report Progress
c     
      if (iverb.gt.3) then
         call ReportLevel(mynod,CUBESIZE,outUnit,nfcntT,1,
     $        1,NormIn,NormOut,'Final:     ')
      endif
c
c ... Report Progress
c
      if (mynod.eq.0) then
         if (iverb.gt.1) then
            write(outUnit,6001) 'Norms:    ',1,VNormIn,VNormOut,
     $           VNormIn/VNormOut,VNormOut/VNormStart
 6001       format(1x,a,2x,i6,4d12.3)
         endif
      endif
c     
c ... Report Time
c
      if (mynod.eq.0) then
         if (iverb.gt.2) then
cmpi            etime = dble(clockx())*0.001d0
            etime = MSG_f_Wtime()
            if ((mynod.eq.0).and.(iverb.gt.2)) then
               write(outUnit,*) 'Soln Time: ',
     &              REAL(etime-stime1),' (secs)'
            endif
            stime1 = etime
         endif
      endif
c
c ... Test for Convergence (Absolute and Relative)
c
      if (VNormOut.lt.tolIF) then
         istat = 1
      endif
      if (((VNormOut/VNormStart).lt.epsIF)) then
         istat = 0
      endif
 999  continue
cmpi      etime = dble(clockx())*0.001d0
      etime = MSG_f_Wtime()
      times(3) = etime - stime
      stime = etime
c
c  ...  l = lg + lc; pres = presg + presc
c
      call daxpbyBdry(n1,n2,n3,1.0d0,l1,l2,l3,1.0d0,
     $     rwk(ilc1),rwk(ilc2),rwk(ilc3),l1,l2,l3)
      call daxpbyPres(n1,n2,n3,1.0d0,pres,1.0d0,rwk(ipresc),pres)
      call daxpbyVel(n1,n2,n3,1.0d0,vel1,vel2,vel3,1.0d0,
     $     rwk(ivel1c),rwk(ivel2c),rwk(ivel3c),vel1,vel2,vel3)
      call projVel(n1,n2,n3,vel1,vel2,vel3,mynod,mynbr,mynbrSize,rwk)
c
c ... Deallocate work space (using the fact that this frees everything 
c ... allocated after and including ipresc)
c
      call memfreer(ipresc,rwk)
c
c ... Continue if only one subdomain
c
 600  continue
c
c ... Remove Mean from Pressure Solution if global neuman
c
      if (globalNeuman.eq.1) then
         call rmPresG(n1,n2,n3,x1,x2,x3,
     $        rwk(ihx1),rwk(ihx2),rwk(ihx3),pres)
      endif
c
c ... Deallocate bilinear form rhs's
c
      call memfreer(idf1,rwk)
      call memfreer(ibf1,rwk)
c
c ... Deallocate Preconditioner work space
c
      call memfreer(ipc3,rwk)
      call memfreer(ipc2,rwk)
      call memfreer(ipc1,rwk)
c
c ... Deallocate Meshes
c
      call memfreer(ihx1,rwk)
c
c ... Deallocate everything: In particular this frees the stuff allocated 
c ... in InitSDSolvers
c
      call memfreei(iMemStartI,iwk)
      call memfreer(iMemStartR,rwk)
c
cmpi      etime = dble(clockx())*0.001d0
      etime = MSG_f_Wtime()
      times(4) = etime - stime
c
      return
      end
