c ...   
c ...   Title:  Subdomain Solver interface routines
c ...   
c     ...  Description: This file provides interfaces to the subdomain
c ...                solver called from pressure.f and compflux.f
c ...   
c ...   Created:        Mon May 25 15:37:51 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Mon Jul 20 10:48:01 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: sdsolver.f,v 1.3 1997-08-20 11:46:46-05 arbogast Exp $
C ...  $RCSfile: sdsolver.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/sdsolver.f,v $
C  
      subroutine InitSDSolvers(n1,n2,n3,x1,x2,x3,alpha,D,bcType,
     $     nBCRegions,bcRegion1,bcRegion2,bcRegion3,SDparam,
     $     iwk,ilen,rwk,rlen,SDptrs)
c
      implicit none
c
      include '../Include/sdsolver.hf'
c
      integer n1,n2,n3,ilen,rlen,iwk(ilen)
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),alpha(*),D
      real*8 SDparam(*),rwk(rlen)
      integer SDptrs(*)
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
c
      if (int(SDparam(isdType)).eq.EXACTtype) then
         call InitEXACT(n1,n2,n3,x1,x2,x3,alpha,D,bcType,
     $        nBCRegions,bcRegion1,bcRegion2,bcRegion3,SDparam,
     $        rwk,rlen,SDptrs)
      else if (int(SDparam(isdType)).eq.TENSOREXACTtype) then
         call InitTensor(n1,n2,n3,x1,x2,x3,alpha,D,bcType,
     $        nBCRegions,bcRegion1,bcRegion2,bcRegion3,SDparam,
     $        rwk,rlen,SDptrs)
      else
         call Parcelcrash('Unknown Solver Type',-30)
      endif
      return
      end
c
c ...
c ...   Title:  Compute Flux and Related Routines
c ...
c ...  Description: These routines compute the flux across the
c ...                subdomain boundaries by solving the subdomain
c ...                PDE's
c ...   
c ...   Created:        Tue May 26 13:45:11 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Tue May 26 13:45:11 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: sdsolver.f,v 1.3 1997-08-20 11:46:46-05 arbogast Exp $
C ...  $RCSfile: sdsolver.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/sdsolver.f,v $
C  
c ... Version $Id: sdsolver.f,v 1.3 1997-08-20 11:46:46-05 arbogast Exp $
c
c ... Routine: compflux123
c
c ... Main function:
c     Solves Local Elliptic Problem using Nested Factorization and 
c     Computes Jump in Fluxes across subdomains
c     Evaluates the bilinear form B(lamda,*) See Rice Masc Tech Reports 
c     Cowsar-Wheeler TR90-37 and Glowinski-Wheeler TR87-11 for details.
c
c ... Include Files: none
c     
c ... Calls:
c     setuprhs, nfrun, flux
c
c ... In:
c     outUnit: Output unit number
c     MYNOD :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     mynbrSize: The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     bcType:     BC's for 1 - Neumann, 2 - Dirichlet, 3 - Noflow
c     x?    :     point centered grids
c     hx?   :     mesh widths hx?(i) = x?(i)-x?(i-1)
c     bc?   :     Boundary condition values
c     pres  :     Guess for Pressure
c     niterSD :   Maximum iterations usuable to solve subdomain problems
c     epsSD :     Relative residual reduction in subdomains
c     tolSD :     Absolute residual required in subdomains
c     neuman:     logical denoting if local problem is Neumann problem
c     a     :     Finite difference matrix
c     rate  :     Right hand side of -div A grad (pres) = (rate)
c     k?    :     Edge Permeability matrix
c     kp1   :     Number of directions to orthogonalize against
c
c     m0,p,q,rhs3,rhs2,rhs1, : Nested Factorization Variables
c     del2,del1,gamma        : Nested Factorization Variables
c
c ... Out:
c     pres  :  Pressure values in subdomain
c     bf?   :  Boundary fluxes for this subdomain
c     df?   :  Jump in Boundary fluxes across sumdomain
c     nfstat :  return status of nested Factorization
c     nfcnt :   Number of nested factorization iterations
c     b     :   Right hand side for solving A pres = b
c
c
c ... Interal Variables of note:
c
c ... Comments:
c     p,q are also used as work space in call to flux
c
      subroutine compflux123(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,bc1,bc2,bc3,pres,vel1,vel2,vel3,
     $     bf1,bf2,bf3,df1,df2,df3,stat,cnt,
     $     alpha,D,rate,rwk,iwk,SDptrs)
c
      implicit none
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2),cnt
      integer n1,n2,n3,iwk(*)
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2),pres(n1,n2,n3)
      real*8  vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3)
      real*8  bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8  df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      real*8  rwk(*)
      real*8  alpha(*),D,rate(n1,n2,n3)
      integer SDptrs(*)

      integer stat

      integer iws
      integer memalr
      external memalr
      integer ip
c
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
      real*8 RobinR
      common/Robin/RobinR
c
      include '../Include/sdsolver.hf'
      include '../Include/defines.hf'
      include '../Include/BCtypes.hf'
c
c ... Switch for the solver type
c
      call sdSolver(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,bc1,bc2,bc3,pres,vel1,vel2,vel3,
     $     bf1,bf2,bf3,df1,df2,df3,stat,cnt,
     $     alpha,D,rate,rwk,iwk,SDptrs)
c
c ... Calculate Flux across domain boundaries
c
      iws = 16*max(max(n1*n2,n1*n3),n2*n3)
      ip = memalr(2*iws,rwk)
      call swapflux123(mynod,mynbr,mynbrSize,n1,n2,n3,bf1,bf2,bf3,
     $     df1,df2,df3,iws,rwk(ip),rwk(ip+iws),VOLUMETRIC)
      call memfreer(ip,rwk)
c
      return
      end

      subroutine sdSolver(outUnit,mynod,mynbr,mynbrSize,
     $     n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     x1,x2,x3,hx1,hx2,hx3,bc1,bc2,bc3,pres,vel1,vel2,vel3,
     $     bf1,bf2,bf3,df1,df2,df3,stat,cnt,
     $     alpha,D,rate,rwk,iwk,SDptrs)
c
      implicit none
      integer outUnit,mynod,mynbr(3,2),mynbrSize(3,2,2),cnt
      integer n1,n2,n3,iwk(*)
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2),pres(n1,n2,n3)
      real*8  vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3)
      real*8  bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8  df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      real*8  rwk(*)
      real*8  alpha(*),D,rate(n1,n2,n3)

      integer stat
c
      integer memalr
      external memalr
c
      integer itest,iverb,idebug
      common/ParcelTest1/itest,iverb,idebug
c
      real*8 RobinR
      common/Robin/RobinR
c
      include '../Include/sdsolver.hf'
      include '../Include/defines.hf'
c
      integer iSDparam,SDptrs(MAXSDPTRS)
c
c ... Switch for the solver type
c
      iSDparam = SDptrs(1)
c
      if (int(rwk(iSDparam+isdType-1)).eq.EXACTtype) then
c
         call exactSolve(outUnit,mynod,mynbr,mynbrSize,
     $        n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,hx1,hx2,hx3,bc1,bc2,bc3,pres,vel1,vel2,vel3,
     $        bf1,bf2,bf3,df1,df2,df3,rate,rwk,SDptrs)
c
      else if (int(rwk(iSDparam+isdType-1)).eq.TENSOREXACTtype) then
c
         call exactTensSolve(outUnit,mynod,mynbr,mynbrSize,
     $        n1,n2,n3,bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $        x1,x2,x3,bc1,bc2,bc3,pres,vel1,vel2,vel3,
     $        bf1,bf2,bf3,df1,df2,df3,rate,rwk,SDptrs)
c
      else
         call Parcelcrash('Unknown Solver Type',-30)
      endif
      return
      end

