/* a repository for global problem values that do not change after
   initialization, provided to help reduce the number of arguments
   passed around. */

/* naming convention:

l_ indicates a quantity numbered as though the problem on a node were
   the whole problem;
g_ indicates a quantity numbered within the entire problem, so that
   g_* = l_* if the problem is run on a single processor;
n_ indicates a quantity numbered over the nodes of the multiprocessor;
p_ indicates a quantity affected by padding.

*/

#include "ftn.h"

int g_concSize[3];    /* dimensions of global concentration grid */
int g_concStart[3];   /* first cSlice numbers for this node */

int n_dom[3];         /* position of this node */
int n_domSum;         /* sum of positions */
int n_domSize[3];     /* dimension of node array */

int concPad[3];       /* depth of replicated data at node boundary */
int nComps;           /* number of problem components */

int l_concSize[3];    /* dimensions of the local concentration grid */
int pl_concSize[3];   /* padded dimensions of the local concentration grid */

double* l_concPlane[3]; /* 3 lists of grid spacings */
double* pl_concPlane[3]; /* 3 lists of grid spacings, including padding */

int (*nbr)[3][3];	/* neighboring processors */

void CGlobals(int *ncomp, double *xc, double *yc, double *zc,
	      int *nxc, int *nyc, int *nzc,
	      int *padxc, int *padyc, int *padzc,
	      int *point_xc_myDom1, int *point_yc_myDom2, int *point_zc_myDom3,
	      int *nxc_g, int *nyc_g, int *nzc_g,
	      int *myDom, int *ndiv, int *mynbr)
{
  nComps = *ncomp;
  concPad[0] = *padxc;
  concPad[1] = *padyc;
  concPad[2] = *padzc;
  pl_concPlane[0] = xc;
  pl_concPlane[1] = yc;
  pl_concPlane[2] = zc;
  l_concSize[0] = *nxc;
  l_concSize[1] = *nyc;
  l_concSize[2] = *nzc;
  g_concStart[0] = *point_xc_myDom1;
  g_concStart[1] = *point_yc_myDom2;
  g_concStart[2] = *point_zc_myDom3;
  g_concSize[0] = *nxc_g;
  g_concSize[1] = *nyc_g;
  g_concSize[2] = *nzc_g;
  n_dom[0] = myDom[0];
  n_dom[1] = myDom[1];
  n_dom[2] = myDom[2];
  n_domSize[0] = ndiv[0];
  n_domSize[1] = ndiv[1];
  n_domSize[2] = ndiv[2];

  n_domSum = 0;
  {
    int j;
    for (j = 0; j < 3; ++j)
      {
        pl_concSize[j] = l_concSize[j] + 2*concPad[j];
        l_concPlane[j] = pl_concPlane[j] + concPad[j];
        n_domSum += n_dom[j];
      }
  }

  nbr = (int(*)[3][3])(mynbr + 9 + 3 + 1);
}
