/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%              */
/* %   This file was created by:                                            */
/* %   mclay@jedrik.ae.utexas.edu on Wed Mar  5 13:01:24 1997               */
/* %   $RCSfile: collectArr.c,v $                                          */
/* %   $Revision: 1.10 $                                                     */
/* %   $Date: 1997-08-15 17:32:23-05 $                                      */
/* %   $Name:  $                                                            */
/* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%             */
/* % l2h ignore miltime                                                     */
/* \documentclass{article}                                                  */
/* \usepackage{full}                                                        */
/* \usepackage{noweb}                                                       */
/* \usepackage{multicol}                                                    */
/* \usepackage{diss_inc}                                                    */
/* \usepackage{time}                                                        */
/* \title{Writing out data to a file via proc 0 with chain scheme}          */
/* \author{Robert McLay}                                                    */
/* \date{\today \\ \miltime}                                                */
/* \pagestyle{noweb}                                                        */
/* \noweboptions{}                                                          */
/* \begin{document}                                                         */
/*                                                                          */
/*                                                                          */
/* \pagenumbering{roman}                                                    */
/* \maketitle                                                               */
/* \tableofcontents                                                         */
/* \clearpage  % Can put a clear page at the end of the toc.                */
/* \pagenumbering{arabic}                                                   */
/*                                                                          */
/*                                                                          */
/* \section{Introduction}                                                   */
/*                                                                          */
/* This routine writes out parallel distributed data to a single output     */
/* file with a chaining scheme.  This chain starts at the last processor    */
/* and sends a block of memory from the n'th processor to the n-1'th        */
/* processor until it reaches processor 0 which writes out the data.        */
/*                                                                          */
/* The current block of memory has a start and end address.  Each           */
/* processor looks to see if it has any data to place in the buffer.        */
/*                                                                          */
/* The following is a table describing the arguments to the collectArr      */
/* functions.                                                               */
/*                                                                          */
/* \begin{center}                                                           */
/*   \begin{tabular}{|l|l|} \hline                                          */
/*      [[array[nDim]]]     & The global distributed array across processors \\ \hline */
/*      [[nDim]]            & The number of dimensions in the array $\geq$ 3 \\ \hline */
/*      [[pad[nDim]]]       & The padding in each dimension \\ \hline       */
/*      [[shared[nDim]]]    & 1 if node is shared in that dimension, 0 otherwise \\ \hline */
/*      [[lwrBnd[nDim]]]    & The lower index local to a processor\\ \hline */
/*      [[upBnd[nDim]]]     & The upper index local to a processor\\ \hline */
/*      [[arrSizeG[nDim]]]  & The global size across all processors\\ \hline */
/*      [[gblOrigin[nDim]]] & The global index of the lower left bottom     */
/*                              entry\\ \hline                              */
/*      [[nDom[3]]]         & The number of processors in each direction \\ \hline */
/*      [[myDom[3]]]        & My processor index in processor space. \\ \hline */
/*   \end{tabular}                                                          */
/* \end{center}                                                             */
/*                                                                          */
/* Note that the arrays [[lwrBnd]], [[upBnd]], [[arrSizeG]] are specified   */
/* assuming no padding.  The [[gblOrigin]] is the global index for the      */
/* local point specified by [[lwrBnd]].                                     */
/*                                                                          */
/* \bigskip                                                                 */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/*                                                                          */
/* <*>=                                                                     */
/* \section{Header files}                                                   */
/*                                                                          */
/*                                                                          */
/* <Header files>=                                                          */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "msg.h"
#define VN "$Id: collectArr.c,v 1.10 1997-08-15 17:32:23-05 arbogast Exp $\n"

/* \section{Global Variables and Defines}                                   */
/*                                                                          */
/*                                                                          */
/* <Global Variables and Defines>=                                          */
extern int Flush_Output;
/* #define MSG_RD_MINSIZE 22301 */
#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif
#define ALLOC(x,s,t) do { x = (t) malloc((s)); if (x == 0) \
                         exit(1);} while(0)
#define SIZE_BUF 32768
#ifdef __STDC__
#  ifndef va_arg
#    include <stdarg.h>
#  endif
#  define  VA_START(ap,last)  va_start(ap, last)
#  define  VA_ARG(ap,type)    va_arg(ap, type)
#  define  VA_END(ap)         va_end(ap)
#else
#  ifndef va_arg
#    include  <varargs.h>
#  endif
#  define  VA_START(ap,last)  va_start(ap)
#  define  VA_ARG(ap,type)    va_arg(ap, type)
#  define  VA_END(ap)         va_end(ap)
#endif

/* \section{Compute 1D Index}                                               */
/*                                                                          */
/*                                                                          */
/* <Compute 1D Index>=                                                      */
int index1D(int *index, int nDim, int* lwrBnd, int* arrFactor, int* pad)
{
  int ja;
  int result;

  result = index[0] - lwrBnd[0] + pad[0];

  for (ja = 1; ja < nDim; ja++)
    result += (index[ja] - lwrBnd[ja] + pad[ja])*arrFactor[ja];

  return result;
}

/* \section{Convert 1D Index to Array Index}                                */
/*                                                                          */
/*                                                                          */
/* <Convert 1D Index to Array Index>=                                       */
void convert1DtoIndex(int idx, int nDim, int* lwrBnd, int* pad, int* arrSizePad, 
                      int* index)
{
  int ja;

  for (ja = 0; ja < nDim - 1; ja++)
    {
      index[ja] = (idx%arrSizePad[ja]) + lwrBnd[ja] - pad[0];
      idx        = idx/arrSizePad[ja];
    }
  index[nDim - 1] = idx + lwrBnd[nDim-1] - pad[nDim-1];
}

/* \section{Convert Global Index to Local Index}                            */
/*                                                                          */
/*                                                                          */
/* <Convert Global Index to Local Index>=                                   */
int convertGblIndex2LocalIndex(int* arrIdxG, int nDim, int* lwrBnd,
                               int* gblOrigin, int* arrSizeL, int* indexL)
{
  int onproc = 1;
  int i;

  for (i = 0; onproc && (i < nDim); i++)
    {
      onproc    = ((gblOrigin[i] <= arrIdxG[i]                     )   &&
                   (arrIdxG[i]   <= gblOrigin[i] + arrSizeL[i] - 1));
      indexL[i] = arrIdxG[i] - gblOrigin[i] + lwrBnd[i];
    }

  return onproc;
}

#define TYPE double
#define RecvTypeX recvRealX
#define SendTypeX sendRealX
#define PROC_DIM 3              /* number of processor directions */
void collectDblArray(TYPE* array, int nDim, int* pad, int* shared,
                     int* lwrBnd, int* upBnd, int* arrSizeG,
                     int *gblOrigin, int* nDom, int* myDom, 
                     FILE* fp,
                     int (*writeBuf)(void *buf, int sizeper, int number,
                                     FILE* fp2))
{
  /* \subsection{Local Variables}                                             */
  /*                                                                          */
  /*                                                                          */
  /* <Local Variables>=                                                       */
  int     blksize;                /* block size */
  int     done;                   /* done flag */
  int     gblcnt;                 /* 1d Global counter */
  int     iblock;                 /* counter over blocks */
  int*    arrIdxG;                /* a global index for the array*/
  int*    arrIdxEndG;             /* global index at end of the current
                                     row of the array*/
  int*    arrIdxL;                /* a local index for the array*/
  int*    arrSizeL;               /* array size w/o padding Local*/
  int*    arrSizePadL;            /* array size w/ padding Local*/
  int*    padG;                   /* Global padding (it always zero)*/
  int*    arrFactorL;             /* array Factor w/ padding Local*/
  int*    arrFactorG;             /* array Factor w/ padding global*/
  int*    arrGblEnd;              /* array End index globally.  It is
                                     different from arrIdxEndG in that
                                     it computed once and has end index
                                     for all rows. */
  int     ip;                     /* counter over processor dimensions */
  int     i;                      /* local counter */
  int     ja;                     /* counter */
  int     i_am, nprocs;           /* current processor number, total procs. */
  int     nblocks;                /* number of blocks to send */
  int     arrTotSize;             /* Total array size */
  int     arrStartG;              /* 1D gbl index of current location in array */
  int     arrStartL;              /* local start of array */
  int     arrEndRowG;             /* 1D global location for end of row */
  int     bufStartG;              /* start 1D global index for current block  */ 
  int     bufEndG;                /* end 1D global index for current block  */ 
  int     bufStartL;              /* local start of buffer */
  int     rowsize;                /* the size of the row to copy */
  int     onproc;                 /* TRUE if index is on processor */
  int     resetInd;               /* The "index" location where arrIdxG */
                                  /* is distributed across processors.*/
  TYPE*   buffer;                 /* comm. buffer */


  /* i_am = mynodeX(); */
  /* nprocs = nnodesX(); */
  MPI_Comm_rank(comm,&i_am);
  MPI_Comm_size(comm,&nprocs);
  /* \subsection{if [[nprocs]]==1 and no padding, write out array directly, then [[return]]} */
  /*                                                                          */
  /*                                                                          */
  /* <if [[nprocs]]==1 and no padding, write out array directly, then [[return]]>= */
  if (nprocs == 1)
  {
    int size;                     /* total size of array */

    for (ja = 0; ja < nDim; ja++)
      if (pad[ja] != 0)
        goto havePad;
        
    size = 1;
    for (ja = 0; ja < nDim; size *= arrSizeG[ja++]);
    (*writeBuf)(array,sizeof(TYPE), size,fp);
    return;
  }
  havePad:


  /* \subsection{Allocate [[buffer]] of [[blksize]] on all processors}        */
  /*                                                                          */
  /*                                                                          */
  /* <Allocate [[buffer]] of [[blksize]] on all processors>=                  */
  blksize = 2*SIZE_BUF;
  do
    {
      blksize >>= 1;   /* divide by 2. */
      if (blksize <= 1)
        exit(1);
      buffer  = (TYPE *) malloc(blksize*sizeof(TYPE));
    }
  while (buffer == 0);

  /* giminX(MSG_RD_MINSIZE,&blksize,1); */
  MSG_Gimin1(&blksize,comm);

  /* \subsection{Allocation and Deallocation of work arrays}                  */
  /*                                                                          */
  /*                                                                          */
  /* <Allocate working arrays>=                                               */
  ALLOC(arrIdxG,    sizeof(int)*nDim, int *);
  ALLOC(arrIdxEndG, sizeof(int)*nDim, int *);
  ALLOC(arrIdxL,    sizeof(int)*nDim, int *);
  ALLOC(arrSizeL,   sizeof(int)*nDim, int *);
  ALLOC(arrSizePadL,sizeof(int)*nDim, int *);
  ALLOC(padG,       sizeof(int)*nDim, int *);
  ALLOC(arrFactorL, sizeof(int)*nDim, int *);
  ALLOC(arrFactorG, sizeof(int)*nDim, int *);
  ALLOC(arrGblEnd,  sizeof(int)*nDim, int *);

  /* \subsection{Initialize}                                                  */
  /* Initialize working arrays and find [[resetInd]] the ``index'' where      */
  /*  [[arrIdxG]] is first distributed across processors.                     */
  /* If [[nprocs]] is 1 and there is padding then we set [[resetInd]] to be   */
  /*  the number of dimension in the array minus the number of processor      */
  /*  directions, namely 3.                                                   */
  /*                                                                          */
  /*                                                                          */
  /* <Initialize>=                                                            */
  resetInd   = -1;
  arrTotSize = 1;
  for (ja = 0; ja < nDim; ja++)
    {
      arrIdxG[ja]     = gblOrigin[ja];
      arrSizeL[ja]    = upBnd[ja] - lwrBnd[ja] + 1;
      arrGblEnd[ja]   = gblOrigin[ja] + arrSizeL[ja] - 1;
      arrSizePadL[ja] = upBnd[ja] - lwrBnd[ja] + 1 + 2*pad[ja];
      padG[ja]        = 0;
      arrTotSize     *= arrSizeG[ja];
      if (arrSizeL[ja] != arrSizeG[ja] && resetInd == -1)
        resetInd = ja;
    }
  if (nprocs == 1)
    resetInd = nDim - PROC_DIM;
  assert(resetInd > -1);

  arrFactorL[0] = 1;
  arrFactorG[0] = 1;
  for (ja = 1; ja < nDim; ja++)
    {    
      arrFactorL[ja] = arrSizePadL[ja-1] * arrFactorL[ja-1];
      arrFactorG[ja] = arrSizeG[ja-1] * arrFactorG[ja-1];
    }

  gblcnt = 0;

  /* \subsection{Calculate number of blocks, [[nblocks]]}                     */
  /*                                                                          */
  /*                                                                          */
  /* <Calculate number of blocks, [[nblocks]]>=                               */
  nblocks = (arrTotSize + blksize - 1)/blksize;


  for (iblock = 0; iblock < nblocks; iblock++)
    {
      int size = MIN(blksize, arrTotSize - gblcnt);
      if (i_am != nprocs - 1)
        /* \section{Sending, Receiving and Writing Blocks}                          */
        /* Here we send and receive the blocks of memory between neighbor           */
        /* processors in the chain.                                                 */
        /*                                                                          */
        /*                                                                          */
        /*                                                                          */
        /* \subsection{Receive current block from the [[i_am + 1]] processor}       */
        /*                                                                          */
        /*                                                                          */
        /* <Receive current block from the [[i_am + 1]] processor>=                 */
        {
          /* RecvTypeX(1000+i_am,buffer,size); */
	  MPI_Status status;
	  MPI_Recv((void*)buffer,size,MPI_DOUBLE,MPI_ANY_SOURCE,
		   1000+i_am,comm,&status);
        }



      /* \section{Copy portions of my array into [[buffer]]}                      */
      /*                                                                          */
      /*                                                                          */
      /* <Copy portions of my array into [[buffer]]>=                             */
      done = 0;
      while (!done)
        {
          /* \subsection{Find Start location in [[buffer]] and in the Parallel Array} */
          /* Here we find the maximum of the global counter for buffer [[gblcnt]]     */
          /*  and [[arrStartG]] (the global 1D index of [[arrIdxG]]).                 */
          /* Later we will check to see if there is any overlap.                      */
          /* Note that we use the rule that processors don't copy out nodes at        */
          /*  ``left'' boundary unless there are at the ``left'' edge.                */
          /*                                                                          */
          /*                                                                          */
          /* <Find Start location in [[buffer]] and in the Parallel Array>=           */
          ip = -1;
          /* Move start of origin stored in [[arrIdxG]] if node is
             shared and note a left boundary */
          for (ja = resetInd; ja < nDim; ja++)
            {
              ip++;
              assert(ip < PROC_DIM);
              if (myDom[ip] > 0 && shared[ja] && arrIdxG[ja] == gblOrigin[ja])
                arrIdxG[ja] = gblOrigin[ja] + 1;
            }

          arrStartG = index1D(arrIdxG, nDim, lwrBnd, arrFactorG, padG);

          if (gblcnt <= arrStartG)
            bufStartG = arrStartG;
          else
            {
              bufStartG = gblcnt;
              convert1DtoIndex(bufStartG,nDim, lwrBnd, padG, arrSizeG, arrIdxG);
            }

          onproc = convertGblIndex2LocalIndex(arrIdxG, nDim, lwrBnd, gblOrigin,
                                              arrSizeL, arrIdxL);
          if (! onproc)
            {
              done = 1;
              break;
            }

          /* \subsection{Find End of row location in [[buffer]], [[break]] if no overlap} */
          /*                                                                          */
          /*                                                                          */
          /* <Find End of row location in [[buffer]], [[break]] if no overlap>=       */
          for (ja = 0; ja <= resetInd; ja++)
            arrIdxEndG[ja] = arrGblEnd[ja];

          for (ja = resetInd + 1; ja < nDim; ja++)
            arrIdxEndG[ja] = arrIdxG[ja];

          arrEndRowG = index1D(arrIdxEndG, nDim, lwrBnd, arrFactorG, padG);
          bufEndG    = MIN(arrEndRowG,gblcnt+size-1);

          /* break if no overlap. */
          if (bufEndG < bufStartG)
            break;

          /* \subsection{Store Row in [[buffer]]}                                     */
          /*                                                                          */
          /*                                                                          */
          /* <Store Row in [[buffer]]>=                                               */
          arrStartL = index1D(arrIdxL, nDim, lwrBnd, arrFactorL, pad);
          bufStartL = bufStartG - gblcnt;
          rowsize   = bufEndG - bufStartG + 1;
          for (ja = 0; ja < rowsize; ja++)
            buffer[ja + bufStartL] = array[ja + arrStartL];

          /* \subsection{bump counters past end of row, if necessary}                 */
          /* Here we need to move [[arrIdxG]] to the beginning of the next row.       */
          /* So we move the ``index'' up one and handle the carries.                  */
          /* If the buffer ended in the middle of a row then we set [[arrIdxG]] to    */
          /*  the next location and set the done flag as there can be no more to      */
          /*  add to the current buffer.                                              */
          /*                                                                          */
          /*                                                                          */
          /* <bump counters past end of row, if necessary>=                           */
          if (bufEndG < arrEndRowG)
            {
              convert1DtoIndex(bufEndG+1, nDim, lwrBnd, pad, arrSizeG,
                               arrIdxG);
              done = 1;
            }
          else
            {
              assert (bufEndG == arrEndRowG);
              for (ja = 0; ja <= resetInd; ja++)
                arrIdxG[ja] = gblOrigin[ja];
              for (ja = resetInd+1;  ; ja++)
                {
                  if (ja == nDim)
                    {
                      done = 1;
                      break;
                    }
                  i = arrIdxG[ja] + 1;
                  if ( i <= arrGblEnd[ja] )
                    {
                      arrIdxG[ja] = i;
                      break;
                    }
                  else
                    arrIdxG[ja] = gblOrigin[ja];
                }
            }

        }
      gblcnt += size;


      if (i_am > 0)
        /* \subsection{Send current block to the [[i_am - 1]] processor}            */
        /*                                                                          */
        /*                                                                          */
        /* <Send current block to the [[i_am - 1]] processor>=                      */
        {
          /* SendTypeX(1000+i_am-1, buffer, size, i_am-1); */
	  MPI_Send((void*)buffer,size,MPI_DOUBLE,i_am-1,1000+i_am-1,comm);
        }

      else
        /* \subsection{Write out current block to output stream}                    */
        /*                                                                          */
        /*                                                                          */
        /* <Write out current block to output stream>=                              */
        (*writeBuf)(buffer, sizeof(TYPE), size, fp);


    }
  /* <Deallocate working arrays>=                                             */
  free(arrGblEnd);
  free(arrFactorG);
  free(arrFactorL);
  free(padG);
  free(arrSizePadL);
  free(arrSizeL);
  free(arrIdxL);
  free(arrIdxEndG);
  free(arrIdxG);

  /* \subsection{Deallocate [[buffer]]}                                       */
  /*                                                                          */
  /*                                                                          */
  /* <Deallocate [[buffer]]>=                                                 */
  free(buffer);

}

#undef  TYPE
#undef  RecvTypeX
#undef  SendTypeX
#define TYPE float
#define RecvTypeX recvHalfX
#define SendTypeX sendHalfX

void collectFltArray(TYPE* array, int nDim, int* pad, int* shared,
                     int* lwrBnd, int* upBnd, int* arrSizeG,
                     int *gblOrigin, int* nDom, int* myDom, 
                     FILE* fp,
                     int (*writeBuf)(void *buf, int sizeper, int number,
                                     FILE* fp2))
{
  /* \subsection{Local Variables}                                             */
  /*                                                                          */
  /*                                                                          */
  /* <Local Variables>=                                                       */
  int     blksize;                /* block size */
  int     done;                   /* done flag */
  int     gblcnt;                 /* 1d Global counter */
  int     iblock;                 /* counter over blocks */
  int*    arrIdxG;                /* a global index for the array*/
  int*    arrIdxEndG;             /* global index at end of the current
                                     row of the array*/
  int*    arrIdxL;                /* a local index for the array*/
  int*    arrSizeL;               /* array size w/o padding Local*/
  int*    arrSizePadL;            /* array size w/ padding Local*/
  int*    padG;                   /* Global padding (it always zero)*/
  int*    arrFactorL;             /* array Factor w/ padding Local*/
  int*    arrFactorG;             /* array Factor w/ padding global*/
  int*    arrGblEnd;              /* array End index globally.  It is
                                     different from arrIdxEndG in that
                                     it computed once and has end index
                                     for all rows. */
  int     ip;                     /* counter over processor dimensions */
  int     i;                      /* local counter */
  int     ja;                     /* counter */
  int     i_am, nprocs;           /* current processor number, total procs. */
  int     nblocks;                /* number of blocks to send */
  int     arrTotSize;             /* Total array size */
  int     arrStartG;              /* 1D gbl index of current location in array */
  int     arrStartL;              /* local start of array */
  int     arrEndRowG;             /* 1D global location for end of row */
  int     bufStartG;              /* start 1D global index for current block  */ 
  int     bufEndG;                /* end 1D global index for current block  */ 
  int     bufStartL;              /* local start of buffer */
  int     rowsize;                /* the size of the row to copy */
  int     onproc;                 /* TRUE if index is on processor */
  int     resetInd;               /* The "index" location where arrIdxG */
                                  /* is distributed across processors.*/
  TYPE*   buffer;                 /* comm. buffer */


  /* i_am = mynodeX(); */
  /* nprocs = nnodesX();  */
  MPI_Comm_rank(comm,&i_am);
  MPI_Comm_size(comm,&nprocs);
  /* \subsection{if [[nprocs]]==1 and no padding, write out array directly, then [[return]]} */
  /*                                                                          */
  /*                                                                          */
  /* <if [[nprocs]]==1 and no padding, write out array directly, then [[return]]>= */
  if (nprocs == 1)
  {
    int size;                     /* total size of array */

    for (ja = 0; ja < nDim; ja++)
      if (pad[ja] != 0)
        goto havePad;
        
    size = 1;
    for (ja = 0; ja < nDim; size *= arrSizeG[ja++]);
    (*writeBuf)(array,sizeof(TYPE), size,fp);
    return;
  }
  havePad:


  /* \subsection{Allocate [[buffer]] of [[blksize]] on all processors}        */
  /*                                                                          */
  /*                                                                          */
  /* <Allocate [[buffer]] of [[blksize]] on all processors>=                  */
  blksize = 2*SIZE_BUF;
  do
    {
      blksize >>= 1;   /* divide by 2. */
      if (blksize <= 1)
        exit(1);
      buffer  = (TYPE *) malloc(blksize*sizeof(TYPE));
    }
  while (buffer == 0);

  /* giminX(MSG_RD_MINSIZE,&blksize,1); */
  MSG_Gimin1(&blksize,comm);

  /* \subsection{Allocation and Deallocation of work arrays}                  */
  /*                                                                          */
  /*                                                                          */
  /* <Allocate working arrays>=                                               */
  ALLOC(arrIdxG,    sizeof(int)*nDim, int *);
  ALLOC(arrIdxEndG, sizeof(int)*nDim, int *);
  ALLOC(arrIdxL,    sizeof(int)*nDim, int *);
  ALLOC(arrSizeL,   sizeof(int)*nDim, int *);
  ALLOC(arrSizePadL,sizeof(int)*nDim, int *);
  ALLOC(padG,       sizeof(int)*nDim, int *);
  ALLOC(arrFactorL, sizeof(int)*nDim, int *);
  ALLOC(arrFactorG, sizeof(int)*nDim, int *);
  ALLOC(arrGblEnd,  sizeof(int)*nDim, int *);

  /* \subsection{Initialize}                                                  */
  /* Initialize working arrays and find [[resetInd]] the ``index'' where      */
  /*  [[arrIdxG]] is first distributed across processors.                     */
  /* If [[nprocs]] is 1 and there is padding then we set [[resetInd]] to be   */
  /*  the number of dimension in the array minus the number of processor      */
  /*  directions, namely 3.                                                   */
  /*                                                                          */
  /*                                                                          */
  /* <Initialize>=                                                            */
  resetInd   = -1;
  arrTotSize = 1;
  for (ja = 0; ja < nDim; ja++)
    {
      arrIdxG[ja]     = gblOrigin[ja];
      arrSizeL[ja]    = upBnd[ja] - lwrBnd[ja] + 1;
      arrGblEnd[ja]   = gblOrigin[ja] + arrSizeL[ja] - 1;
      arrSizePadL[ja] = upBnd[ja] - lwrBnd[ja] + 1 + 2*pad[ja];
      padG[ja]        = 0;
      arrTotSize     *= arrSizeG[ja];
      if (arrSizeL[ja] != arrSizeG[ja] && resetInd == -1)
        resetInd = ja;
    }
  if (nprocs == 1)
    resetInd = nDim - PROC_DIM;
  assert(resetInd > -1);

  arrFactorL[0] = 1;
  arrFactorG[0] = 1;
  for (ja = 1; ja < nDim; ja++)
    {    
      arrFactorL[ja] = arrSizePadL[ja-1] * arrFactorL[ja-1];
      arrFactorG[ja] = arrSizeG[ja-1] * arrFactorG[ja-1];
    }

  gblcnt = 0;

  /* \subsection{Calculate number of blocks, [[nblocks]]}                     */
  /*                                                                          */
  /*                                                                          */
  /* <Calculate number of blocks, [[nblocks]]>=                               */
  nblocks = (arrTotSize + blksize - 1)/blksize;


  for (iblock = 0; iblock < nblocks; iblock++)
    {
      int size = MIN(blksize, arrTotSize - gblcnt);
      if (i_am != nprocs - 1)
        /* \section{Sending, Receiving and Writing Blocks}                          */
        /* Here we send and receive the blocks of memory between neighbor           */
        /* processors in the chain.                                                 */
        /*                                                                          */
        /*                                                                          */
        /*                                                                          */
        /* \subsection{Receive current block from the [[i_am + 1]] processor}       */
        /*                                                                          */
        /*                                                                          */
        /* <Receive current block from the [[i_am + 1]] processor>=                 */
        {
          /* RecvTypeX(1000+i_am,buffer,size); */
	  MPI_Status status;
	  MPI_Recv((void*)buffer,size,MPI_FLOAT,MPI_ANY_SOURCE,
		   1000+i_am,comm,&status);
        }



      /* \section{Copy portions of my array into [[buffer]]}                      */
      /*                                                                          */
      /*                                                                          */
      /* <Copy portions of my array into [[buffer]]>=                             */
      done = 0;
      while (!done)
        {
          /* \subsection{Find Start location in [[buffer]] and in the Parallel Array} */
          /* Here we find the maximum of the global counter for buffer [[gblcnt]]     */
          /*  and [[arrStartG]] (the global 1D index of [[arrIdxG]]).                 */
          /* Later we will check to see if there is any overlap.                      */
          /* Note that we use the rule that processors don't copy out nodes at        */
          /*  ``left'' boundary unless there are at the ``left'' edge.                */
          /*                                                                          */
          /*                                                                          */
          /* <Find Start location in [[buffer]] and in the Parallel Array>=           */
          ip = -1;
          /* Move start of origin stored in [[arrIdxG]] if node is
             shared and note a left boundary */
          for (ja = resetInd; ja < nDim; ja++)
            {
              ip++;
              assert(ip < PROC_DIM);
              if (myDom[ip] > 0 && shared[ja] && arrIdxG[ja] == gblOrigin[ja])
                arrIdxG[ja] = gblOrigin[ja] + 1;
            }

          arrStartG = index1D(arrIdxG, nDim, lwrBnd, arrFactorG, padG);

          if (gblcnt <= arrStartG)
            bufStartG = arrStartG;
          else
            {
              bufStartG = gblcnt;
              convert1DtoIndex(bufStartG,nDim, lwrBnd, padG, arrSizeG, arrIdxG);
            }

          onproc = convertGblIndex2LocalIndex(arrIdxG, nDim, lwrBnd, gblOrigin,
                                              arrSizeL, arrIdxL);
          if (! onproc)
            {
              done = 1;
              break;
            }

          /* \subsection{Find End of row location in [[buffer]], [[break]] if no overlap} */
          /*                                                                          */
          /*                                                                          */
          /* <Find End of row location in [[buffer]], [[break]] if no overlap>=       */
          for (ja = 0; ja <= resetInd; ja++)
            arrIdxEndG[ja] = arrGblEnd[ja];

          for (ja = resetInd + 1; ja < nDim; ja++)
            arrIdxEndG[ja] = arrIdxG[ja];

          arrEndRowG = index1D(arrIdxEndG, nDim, lwrBnd, arrFactorG, padG);
          bufEndG    = MIN(arrEndRowG,gblcnt+size-1);

          /* break if no overlap. */
          if (bufEndG < bufStartG)
            break;

          /* \subsection{Store Row in [[buffer]]}                                     */
          /*                                                                          */
          /*                                                                          */
          /* <Store Row in [[buffer]]>=                                               */
          arrStartL = index1D(arrIdxL, nDim, lwrBnd, arrFactorL, pad);
          bufStartL = bufStartG - gblcnt;
          rowsize   = bufEndG - bufStartG + 1;
          for (ja = 0; ja < rowsize; ja++)
            buffer[ja + bufStartL] = array[ja + arrStartL];

          /* \subsection{bump counters past end of row, if necessary}                 */
          /* Here we need to move [[arrIdxG]] to the beginning of the next row.       */
          /* So we move the ``index'' up one and handle the carries.                  */
          /* If the buffer ended in the middle of a row then we set [[arrIdxG]] to    */
          /*  the next location and set the done flag as there can be no more to      */
          /*  add to the current buffer.                                              */
          /*                                                                          */
          /*                                                                          */
          /* <bump counters past end of row, if necessary>=                           */
          if (bufEndG < arrEndRowG)
            {
              convert1DtoIndex(bufEndG+1, nDim, lwrBnd, pad, arrSizeG,
                               arrIdxG);
              done = 1;
            }
          else
            {
              assert (bufEndG == arrEndRowG);
              for (ja = 0; ja <= resetInd; ja++)
                arrIdxG[ja] = gblOrigin[ja];
              for (ja = resetInd+1;  ; ja++)
                {
                  if (ja == nDim)
                    {
                      done = 1;
                      break;
                    }
                  i = arrIdxG[ja] + 1;
                  if ( i <= arrGblEnd[ja] )
                    {
                      arrIdxG[ja] = i;
                      break;
                    }
                  else
                    arrIdxG[ja] = gblOrigin[ja];
                }
            }

        }
      gblcnt += size;


      if (i_am > 0)
        /* \subsection{Send current block to the [[i_am - 1]] processor}            */
        /*                                                                          */
        /*                                                                          */
        /* <Send current block to the [[i_am - 1]] processor>=                      */
        {
          /* SendTypeX(1000+i_am-1, buffer, size, i_am-1); */
	  MPI_Send((void*)buffer,size,MPI_FLOAT,i_am-1,1000+i_am-1,comm);
        }

      else
        /* \subsection{Write out current block to output stream}                    */
        /*                                                                          */
        /*                                                                          */
        /* <Write out current block to output stream>=                              */
        (*writeBuf)(buffer, sizeof(TYPE), size, fp);


    }
  /* <Deallocate working arrays>=                                             */
  free(arrGblEnd);
  free(arrFactorG);
  free(arrFactorL);
  free(padG);
  free(arrSizePadL);
  free(arrSizeL);
  free(arrIdxL);
  free(arrIdxEndG);
  free(arrIdxG);

  /* \subsection{Deallocate [[buffer]]}                                       */
  /*                                                                          */
  /*                                                                          */
  /* <Deallocate [[buffer]]>=                                                 */
  free(buffer);

}


