#include <stdio.h>
#include <stdlib.h>
#include "debug.h"

static int myProc;
static int opened = 0;
static FILE* debugFile[2]; /* = { stdout, stdout }; */

/******************************************************************************
Title: initDebug
Description: Initialize standard out for debugging.
*/
void initDebug(int myProc_in) {
  debugFile[0] = stdout;
  debugFile[1] = stdout;

  myProc = myProc_in;
}

/******************************************************************************
Title: openDebug
Description: Open debugging files.
*/
void openDebug(int file, int myProc_in) {
  char name[10];

  debugFile[0] = stdout;

  myProc = myProc_in;
  if(file) {
    if(!opened) {
      sprintf(name,"debug%4.4d",myProc);
      debugFile[1] = fopen(name,"w");
      if(debugFile[1] == NULL) {
	fprintf(stderr,"Error opening debug file on processor %d\n",myProc);
	debugFile[1] = stdout;
	opened = 0;
      } else {
	opened = 1;
      }
    }
  } else {
    if(opened) closeDebug();
  }
}

/******************************************************************************
Title: closeDebug
Description: Close debugging files.
*/
void closeDebug() {
  if(opened) fclose(debugFile[1]);
  debugFile[1] = stdout;
  opened = 0;
}

/******************************************************************************
Title: debugPlace
Description: Tag a place executed in the code.
*/
void debugPlace(int file, char* fileName, int fileLine) {
  fprintf(debugFile[file], "Proc %d, File %s, Line %d\n",
	  myProc,fileName,fileLine);
}

/******************************************************************************
Title: debug<TYPE>
Description: Print out the given array (in serial mode) for debugging purposes.
Input: file    Boolean as to whether to write to the debug files or the screen.
       name    The name of the array.
       array   The array.
       len     Its length (unless the TYPE is string).
*/
/* TYPE == double */
void debugDouble(int file, char* name, double* array, int len) {
  int i;
  char processor[13] = "";
  if(file) {
    file = 1;
  } else {
    sprintf(processor,"Proc: %d, ",myProc);
  }

  if(array == NULL) {
    fprintf(debugFile[file],"%s%s is null\n",processor,name);
    return;
  }
  if (len <= 0) {
    fprintf(debugFile[file],"%s%s has length %d\n",processor,name,len);
    return;
  }
  for(i=0; i<len-4; i = i+5)
    fprintf(debugFile[file],
	    "%s%s:  %10g  %10g  %10g  %10g  %10g\n",processor,name,
	    array[i],array[i+1],array[i+2],array[i+3],array[i+4]);

  switch(len-i) {
  case 4: fprintf(debugFile[file],
		  "%s%s:  %10g  %10g  %10g  %10g\n",processor,name,
		  array[i],array[i+1],array[i+2],array[i+3]); break;
  case 3: fprintf(debugFile[file],
		  "%s%s:  %10g  %10g  %10g\n",processor,name,
		  array[i],array[i+1],array[i+2]); break;
  case 2: fprintf(debugFile[file],
		  "%s%s:  %10g  %10g\n",processor,name,
		  array[i],array[i+1]); break;
  case 1: fprintf(debugFile[file],
		  "%s%s:  %10g\n",processor,name, array[i]); break;
  case 0: break;
  default: fprintf(stderr,"ERROR IN DEBUGGING OUTPUT ROUTINE\n");
  }
}

/* TYPE == float */
void debugFloat(int file, char* name, float* array, int len) {
  int i;
  char processor[13] = "";
  if(file) {
    file = 1;
  } else {
    sprintf(processor,"Proc: %d, ",myProc);
  }

  if (array == NULL) {
    fprintf(debugFile[file],"%s%s is null\n",processor,name);
    return;
  }
  if (len <= 0) {
    fprintf(debugFile[file],"%s%s has length %d\n",processor,name,len);
    return;
  }
  for(i=0; i<len-4; i = i+5)
    fprintf(debugFile[file],
	    "%s%s:  %10g  %10g  %10g  %10g  %10g\n",processor,name,
	    array[i],array[i+1],array[i+2],array[i+3],array[i+4]);

  switch(len-i) {
  case 4: fprintf(debugFile[file],
		  "%s%s:  %10g  %10g  %10g  %10g\n",processor,name,
		  array[i],array[i+1],array[i+2],array[i+3]); break;
  case 3: fprintf(debugFile[file],
		  "%s%s:  %10g  %10g  %10g\n",processor,name,
		  array[i],array[i+1],array[i+2]); break;
  case 2: fprintf(debugFile[file],
		  "%s%s:  %10g  %10g\n",processor,name,
		  array[i],array[i+1]); break;
  case 1: fprintf(debugFile[file],
		  "%s%s:  %10g\n",processor,name, array[i]); break;
  case 0: break;
  default: fprintf(stderr,"ERROR IN DEBUGGING OUTPUT ROUTINE\n");
  }
}

/* TYPE == int */
void debugInt(int file, char* name, int* array, int len) {
  int i;
  char processor[13] = "";
  if(file) {
    file = 1;
  } else {
    sprintf(processor,"Proc: %d, ",myProc);
  }

  if (array == NULL) {
    fprintf(debugFile[file],"%s%s is null\n",processor,name);
    return;
  }
  if (len <= 0) {
    fprintf(debugFile[file],"%s%s has length %d\n",processor,name,len);
    return;
  }
  for(i=0; i<len-4; i = i+5)
    fprintf(debugFile[file],
	    "%s%s:  %10d  %10d  %10d  %10d  %10d\n",processor,name,
	    array[i],array[i+1],array[i+2],array[i+3],array[i+4]);

  switch(len-i) {
  case 4: fprintf(debugFile[file],
		  "%s%s:  %10d  %10d  %10d  %10d\n",processor,name,
		  array[i],array[i+1],array[i+2],array[i+3]); break;
  case 3: fprintf(debugFile[file],
		  "%s%s:  %10d  %10d  %10d\n",processor,name,
		  array[i],array[i+1],array[i+2]); break;
  case 2: fprintf(debugFile[file],
		  "%s%s:  %10d  %10d\n",processor,name,
		  array[i],array[i+1]); break;
  case 1: fprintf(debugFile[file],
		  "%s%s:  %10d\n",processor,name, array[i]); break;
  case 0: break;
  default: fprintf(stderr,"ERROR IN DEBUGGING OUTPUT ROUTINE\n");
  }
}

/* TYPE == string */
void debugString(int file, char* name, char* array) {
  char processor[13] = "";
  if(file) {
    file = 1;
  } else {
    sprintf(processor,"Proc: %d, ",myProc);
  }

  fprintf(debugFile[file],"%s%s:  %s\n",processor,name,array);
}

int isNaN(double x) {
  double zero = 0.0;
  return !(zero*x == zero);
}

void checkArrayNaN(double* a, int n, int callNumber) {
  int i, abort = 0;

  for(i=0; i<n; i++) {
    if(isNaN(a[i])) {
      abort = 1;
      fprintf(stderr,"NaN found at %d in call %d\n",i,callNumber);
    }
  }
  if(abort && callNumber >= 0) exit(1);
}
