c=======================================================================
c
      subroutine scalPerm(modelMiscDisp,
     $     perm,facePerm,permType,visc_ratio,molwgt,rho,conc,scalingDir,
     $     nx,ny,nz,padx,pady,padz,ncomp)
      implicit none
      integer modelMiscDisp,permType,nx,ny,nz,padx,pady,padz
      integer ncomp,scalingDir
      real*8  perm(6,nx,ny,nz),facePerm(*)
      real*8  molwgt(ncomp),rho
      real*8  visc_ratio(*)
      real*4  conc(*)

      if(modelMiscDisp.eq.1) then
         call viscosity_quarter_power(conc,visc_ratio,molwgt,
     $        rho,perm,facePerm,permType,scalingDir,
     &        nx,ny,nz,padx,pady,padz,ncomp)
       else
          write(*,*)
     &         "FATAL ERROR: Unsupported miscible displacement model"
          stop
       endif

ccc:
c       do 100 k=1,nz
c       do 100 j=1,ny
c       do 100 i=1,nx
c         if(perm(1,i,j,k).lt.0) then
c           write(*,*) 'Negative K11 entry at ',i,j,k
c           STOP
c         elseif(perm(4,i,j,k).lt.0) then 
c           write(*,*) 'Negative K22 entry at ',i,j,k
c           STOP
c         elseif(perm(6,i,j,k).lt.0) then
c           write(*,*) 'Negative K33 entry at ',i,j,k
c           STOP
c         endif
c 100  continue
ccc.

       return
       end

C======================================================================
      subroutine viscosity_quarter_power(conc,visc_ratio,molwgt,
     $           rho,perm,facePerm,permType,scalingDir,
     $           nx,ny,nz,padx,pady,padz,ncomp)
C  INPUTS:
C  real*4 conc             -> concentrations of components
C                              these are in moles/volume
C  real*8 visc_ratio       -> ratio component(i) viscosity to resident
C  real*8 molwgt           -> holds the molecular weight of component(i)
C                              units (mass/mol)
C  real*8  rho             -> the density
C  real*8  perm            -> the permeabilities to scale (if permType != 1)
C  real*8  facePerm        -> the permeabilities to scale (if permType=1)
C  integer permType        -> scale facePerm if permType = 1
C  integer scalingDir      -> scaling direction (1 to scale, 0 to undo)
C  integer nx,ny,nz        -> dimensions of the grid
C  integer padx,pady,padz  -> padding of the grid

C  OUTPUTS:
C  real*8  perm            -> the scaled permeabilities (if permType != 1)
C  real*8  facePerm        -> the scaled permeabilities (if permType = 1)

C The rule is
C 
C \mu(\xi_1,...\xi_n) = \mu_0(\sum_{i=0}^n r_i^{1/4} \xi_i)^4,
C 
C where \mu is the viscosity, \mu_0 is the resident fluid viscosity of
C species 0, r_i is the ratio of the viscosity of the i species to the
C resident fluid viscosity, and \xi_i is the mass fraction of the ith
C species.  The mass fraction is
C 
C \xi_i = c_i w_i / \rho,
C 
C where c_i is the molar concentration, w_i is the molecular weight, and
C \rho is the flowing phase mass density.  The resident mass fraction
C \xi_0 is given by solving
C 
C \sum_{i=0}^n \xi_i = 1.
C 
C The permeability is assumed scaled by 1/\mu (scalingDir = 0)
C or 1/mu_0 (scalingDir = 1) on input.  We compute \mu/\mu_0 below.

      implicit none
      INTRINSIC sqrt
      integer nx,ny,nz,padx,pady,padz,ncomp
      integer i,j,k,n,permType,scalingDir
      integer mynodex
      real*4  conc(-padx:nx+padx+1,-pady:ny+pady+1,
     $             -padz:nz+padz+1,ncomp)
      real*8  perm(6,nx,ny,nz),
     $        facePerm((nx+1)*ny*nz+nx*(ny+1)*nz+nx*ny*(nz+1))
      real*8 visc_ratio(ncomp)
      real*8 molwgt(ncomp),rho
      real*8 vc,vcPrev,xi0,val

      real*8  a,b,HARMONIC
      integer indX,indY,indZ
      HARMONIC(a,b) = 2.0*(a)*(b)/((a)+(b))
      indX(i,j,k) = i + (nx+1)*(j-1 + ny*(k-1))
      indY(i,j,k) = (nx+1)*ny*nz + i + nx*(j-1 + (ny+1)*(k-1))
      indZ(i,j,k) = (nx+1)*ny*nz + nx*(ny+1)*nz
     $     + i + nx*(j-1 + ny*(k-1))

      if(permType.ne.1) then

      do 100 k=1,nz
      do 100 j=1,ny
      do 100 i=1,nx

         xi0=rho
         do 10 n = 1, ncomp
c           if visc_ratio(i)< 0 then ignore this component
            if(visc_ratio(n).gt.0) xi0 = xi0-molwgt(n)*conc(i,j,k,n)
 10      continue
         if(xi0.lt.0.) xi0 = 0.

         vc = xi0
         do 20 n=1,ncomp
            if(visc_ratio(n).gt.0) then
               vc = vc + conc(i,j,k,n)*molwgt(n)
     $              /sqrt(sqrt(visc_ratio(n)))
            endif
 20      continue
         vc = vc/rho
         vc = 1/(vc*vc*vc*vc)

         if(scalingDir.eq.1) vc = 1/vc

         perm(1,i,j,k) = perm(1,i,j,k)*vc
         perm(2,i,j,k) = perm(2,i,j,k)*vc
         perm(3,i,j,k) = perm(3,i,j,k)*vc
         perm(4,i,j,k) = perm(4,i,j,k)*vc
         perm(5,i,j,k) = perm(5,i,j,k)*vc
         perm(6,i,j,k) = perm(6,i,j,k)*vc
 100  continue

      else

c x-dir
      do k=1,nz
      do j=1,ny
         i = 1

         xi0=rho
         do n = 1, ncomp
c           if visc_ratio(i)< 0 then ignore this component
            if(visc_ratio(n).gt.0) xi0 = xi0-molwgt(n)*conc(i,j,k,n)
         enddo
         if(xi0.lt.0.) xi0 = 0.

         vc = xi0
         do n=1,ncomp
            if(visc_ratio(n).gt.0) then
               vc = vc + conc(i,j,k,n)*molwgt(n)
     $              /sqrt(sqrt(visc_ratio(n)))
            endif
         enddo
         vc = vc/rho
         vc = 1/(vc*vc*vc*vc)
         vcPrev = vc

         if(scalingDir.eq.1) vc = 1/vc
         facePerm(indX(i,j,k)) = facePerm(indX(i,j,k))*vc

         do i=2,nx

            xi0=rho
            do n = 1, ncomp
c              if visc_ratio(i)< 0 then ignore this component
               if(visc_ratio(n).gt.0) xi0 = xi0-molwgt(n)*conc(i,j,k,n)
            enddo
            if(xi0.lt.0.) xi0 = 0.

            vc = xi0
            do n=1,ncomp
               if(visc_ratio(n).gt.0) then
                  vc = vc + conc(i,j,k,n)*molwgt(n)
     $                 /sqrt(sqrt(visc_ratio(n)))
               endif
            enddo
            vc = vc/rho
            vc = 1/(vc*vc*vc*vc)

            val = HARMONIC(vc,vcPrev)
            if(scalingDir.eq.1) val = 1/val
            facePerm(indX(i,j,k)) = facePerm(indX(i,j,k))*val
            vcPrev = vc
         enddo
         i = nx+1
         if(scalingDir.eq.1) vc = 1/vc
         facePerm(indX(i,j,k)) = facePerm(indX(i,j,k))*vc
      enddo
      enddo
     
c y-dir
      do k=1,nz
      do i=1,nx
         j = 1

         xi0=rho
         do n = 1, ncomp
c           if visc_ratio(i)< 0 then ignore this component
            if(visc_ratio(n).gt.0) xi0 = xi0-molwgt(n)*conc(i,j,k,n)
         enddo
         if(xi0.lt.0.) xi0 = 0.

         vc = xi0
         do n=1,ncomp
            if(visc_ratio(n).gt.0) then
               vc = vc + conc(i,j,k,n)*molwgt(n)
     $              /sqrt(sqrt(visc_ratio(n)))
            endif
         enddo
         vc = vc/rho
         vc = 1/(vc*vc*vc*vc)
         vcPrev = vc

         if(scalingDir.eq.1) vc = 1/vc
         facePerm(indY(i,j,k)) = facePerm(indY(i,j,k))*vc

         do j=2,ny

            xi0=rho
            do n = 1, ncomp
c              if visc_ratio(i)< 0 then ignore this component
               if(visc_ratio(n).gt.0) xi0 = xi0-molwgt(n)*conc(i,j,k,n)
            enddo
            if(xi0.lt.0.) xi0 = 0.

            vc = xi0
            do n=1,ncomp
               if(visc_ratio(n).gt.0) then
                  vc = vc + conc(i,j,k,n)*molwgt(n)
     $                 /sqrt(sqrt(visc_ratio(n)))
               endif
            enddo
            vc = vc/rho
            vc = 1/(vc*vc*vc*vc)

            val = HARMONIC(vc,vcPrev)
            if(scalingDir.eq.1) val = 1/val
            facePerm(indY(i,j,k)) = facePerm(indY(i,j,k))*val
            vcPrev = vc
         enddo
         j = ny+1
         if(scalingDir.eq.1) vc = 1/vc
         facePerm(indY(i,j,k)) = facePerm(indY(i,j,k))*vc
      enddo
      enddo
     
c z-dir
      do j=1,ny
      do i=1,nx
         k = 1

         xi0=rho
         do n = 1, ncomp
c           if visc_ratio(i)< 0 then ignore this component
            if(visc_ratio(n).gt.0) xi0 = xi0-molwgt(n)*conc(i,j,k,n)
         enddo
         if(xi0.lt.0.) xi0 = 0.

         vc = xi0
         do n=1,ncomp
            if(visc_ratio(n).gt.0) then
               vc = vc + conc(i,j,k,n)*molwgt(n)
     $              /sqrt(sqrt(visc_ratio(n)))
            endif
         enddo
         vc = vc/rho
         vc = 1/(vc*vc*vc*vc)
         vcPrev = vc

         if(scalingDir.eq.1) vc = 1/vc
         facePerm(indZ(i,j,k)) = facePerm(indZ(i,j,k))*vc

         do k=2,nz

            xi0=rho
            do n = 1, ncomp
c              if visc_ratio(i)< 0 then ignore this component
               if(visc_ratio(n).gt.0) xi0 = xi0-molwgt(n)*conc(i,j,k,n)
            enddo
            if(xi0.lt.0.) xi0 = 0.

            vc = xi0
            do n=1,ncomp
               if(visc_ratio(n).gt.0) then
                  vc = vc + conc(i,j,k,n)*molwgt(n)
     $                 /sqrt(sqrt(visc_ratio(n)))
               endif
            enddo
            vc = vc/rho
            vc = 1/(vc*vc*vc*vc)

            val = HARMONIC(vc,vcPrev)
            if(scalingDir.eq.1) val = 1/val
            facePerm(indZ(i,j,k)) = facePerm(indZ(i,j,k))*val
            vcPrev = vc
         enddo
         k = nz+1
         if(scalingDir.eq.1) vc = 1/vc
         facePerm(indZ(i,j,k)) = facePerm(indZ(i,j,k))*vc
      enddo
      enddo

      endif
     
      return 
      end
