#ifndef _MSG_H
#define _MSG_H

#if nullComm

#include <time.h>

#define comm 0

#define MPI_BYTE   1
#define MPI_CHAR   1
#define MPI_INT    4
#define MPI_FLOAT  4
#define MPI_DOUBLE 8
#define MPI_REAL8  8

#define MPI_MIN 0
#define MPI_MAX 0
#define MPI_SUM 0

#define MPI_ANY_TAG -1
#define MPI_ANY_SOURCE -1

#define MPI_REQUEST_NULL 0
typedef int MPI_Request;
struct MPI_Status_struct {
  int MPI_TAG;
  int MPI_SOURCE;
  int MPI_ERROR;
};
typedef struct MPI_Status_struct MPI_Status;
typedef int MPI_Comm;

#define MPI_Allgather(buf1,len1,type1,buf2,len2,type2,_comm)
#define MPI_Allgatherv(buf1,len1,type1,buf2,len2,displacement,type2,_comm)
#define MPI_Allreduce(buf1,buf2,len,type,op,_comm)
#define MPI_Barrier(_comm)
#define MPI_Bcast(buf,len,type,source,_comm)
#define MPI_Bsend(buf,len,type,to,tag,_comm)
#define MPI_Comm_free(_comm)
#define MPI_Comm_size(_comm,nProcs) (*nProcs)=1
#define MPI_Comm_split(_comm,_color,_key,_newcomm)
#define MPI_Comm_rank(_comm,myProc) (*myProc)=0
#define MPI_Finalize()
#define MPI_Get_count(status,type,len) (*len)=0
#define MPI_Init(argc,argv)
#define MPI_Iprobe(source,tag,_comm,flag,status) \
        do{ (*flag) = 0; (*status).MPI_TAG=tag; \
            (*status).MPI_SOURCE=source; }while(0)
#define MPI_Irecv(buf,len,type,from,tag,_comm,request)
#define MPI_Isend(buf,len,type,to,tag,_comm,request)
#define MPI_Request_free(req)
#define MPI_Recv(buf,len,type,from,tag,_comm,status)
#define MPI_Send(buf,len,type,to,tag,_comm)
#define MPI_Sendrecv(buf1,n1,type1,to,tag1,buf2,n2,type2,from,tag2,_comm,stat)
#define MPI_Waitany(count,requests,index,status)
#define MPI_Wtime() clock()/CLOCKS_PER_SEC

#define MSG_Gimin(buf,len,_comm)
#define MSG_Gimin1(value,_comm)
#define MSG_Gimin2(buf1,buf2,len,_comm) \
     do{ int i; for(i=0; i<len; i++) buf2[i] = buf1[i]; }while(0)
#define MSG_Gimax(buf,len,_comm)
#define MSG_Gimax1(value,_comm)
#define MSG_Gimax2(buf1,buf2,len,_comm) \
     do{ int i; for(i=0; i<len; i++) buf2[i] = buf1[i]; }while(0)
#define MSG_Gdmin(buf,len,_comm)
#define MSG_Gdmin1(value,_comm)
#define MSG_Gdmin2(buf1,buf2,len,_comm) \
     do{ int i; for(i=0; i<len; i++) buf2[i] = buf1[i]; }while(0)
#define MSG_Gdmax(buf,len,_comm)
#define MSG_Gdmax1(value,_comm)
#define MSG_Gdmax2(buf1,buf2,len,_comm) \
     do{ int i; for(i=0; i<len; i++) buf2[i] = buf1[i]; }while(0)
#define MSG_Gisum(buf,len,_comm)
#define MSG_Gisum1(value,_comm)
#define MSG_Gisum2(buf1,buf2,len,_comm) \
     do{ int i; for(i=0; i<len; i++) buf2[i] = buf1[i]; }while(0)
#define MSG_Gdsum(buf,len,_comm)
#define MSG_Gdsum1(value,_comm)
#define MSG_Gdsum2(buf1,buf2,len,_comm) \
     do{ int i; for(i=0; i<len; i++) buf2[i] = buf1[i]; }while(0)

#else

#include <mpi.h>

#define comm MPI_COMM_WORLD

#define MSG_Gimin(buf,len,commun) \
  do{ int i; int* tmp = (int*) malloc(len*sizeof(int));\
      MPI_Allreduce((void*)buf,(void*)tmp,len,MPI_INT,MPI_MIN,commun);\
      for(i=0; i<len; i++) buf[i] = tmp[i]; free(tmp); }while(0)
#define MSG_Gimin1(value,commun) \
  do{ int tmp; \
      MPI_Allreduce((void*)value,(void*)&tmp,1,MPI_INT,MPI_MIN,commun);\
      *value = tmp; }while(0)
#define MSG_Gimin2(buf1,buf2,len,commun) \
  do{ int i; \
      MPI_Allreduce((void*)buf1,(void*)buf2,len,MPI_INT,MPI_MIN,commun);\
      for(i=0; i<len; i++) buf1[i] = buf2[i]; }while(0)
#define MSG_Gimax(buf,len,commun) \
  do{ int i; int* tmp = (int*) malloc(len*sizeof(int));\
      MPI_Allreduce((void*)buf,(void*)tmp,len,MPI_INT,MPI_MAX,commun);\
      for(i=0; i<len; i++) buf[i] = tmp[i]; free(tmp); }while(0)
#define MSG_Gimax1(value,commun) \
  do{ int tmp; \
      MPI_Allreduce((void*)value,(void*)&tmp,1,MPI_INT,MPI_MAX,commun);\
      *value = tmp; }while(0)
#define MSG_Gimax2(buf1,buf2,len,commun) \
  do{ int i; \
      MPI_Allreduce((void*)buf1,(void*)buf2,len,MPI_INT,MPI_MAX,commun);\
      for(i=0; i<len; i++) buf1[i] = buf2[i]; }while(0)
#define MSG_Gdmin(buf,len,commun) \
  do{int i; double* tmp = (double*) malloc(len*sizeof(double));\
     MPI_Allreduce((void*)buf,(void*)tmp,len,MPI_DOUBLE,MPI_MIN,commun);\
     for(i=0; i<len; i++) buf[i] = tmp[i]; free(tmp); }while(0)
#define MSG_Gdmin1(value,commun) \
  do{double tmp; \
     MPI_Allreduce((void*)value,(void*)&tmp,1,MPI_DOUBLE,MPI_MIN,commun);\
     *value = tmp; }while(0)
#define MSG_Gdmin2(buf1,buf2,len,commun) \
  do{int i; \
     MPI_Allreduce((void*)buf1,(void*)buf2,len,MPI_DOUBLE,MPI_MIN,commun);\
     for(i=0; i<len; i++) buf1[i] = buf2[i]; }while(0)
#define MSG_Gdmax(buf,len,commun) \
  do{int i; double* tmp = (double*) malloc(len*sizeof(double));\
     MPI_Allreduce((void*)buf,(void*)tmp,len,MPI_DOUBLE,MPI_MAX,commun);\
     for(i=0; i<len; i++) buf[i] = tmp[i]; free(tmp); }while(0)
#define MSG_Gdmax1(value,commun) \
  do{double tmp; \
     MPI_Allreduce((void*)value,(void*)&tmp,1,MPI_DOUBLE,MPI_MAX,commun);\
     *value = tmp; }while(0)
#define MSG_Gdmax2(buf1,buf2,len,commun) \
  do{int i; \
     MPI_Allreduce((void*)buf1,(void*)buf2,len,MPI_DOUBLE,MPI_MAX,commun);\
     for(i=0; i<len; i++) buf1[i] = buf2[i]; }while(0)
#define MSG_Gisum(buf,len,commun) \
  do{int i; int* tmp = (int*) malloc(len*sizeof(int));\
     MPI_Allreduce((void*)buf,(void*)tmp,len,MPI_INT,MPI_SUM,commun);\
     for(i=0; i<len; i++) buf[i] = tmp[i]; free(tmp); }while(0)
#define MSG_Gisum1(value,commun) \
  do{int tmp; \
     MPI_Allreduce((void*)value,(void*)&tmp,1,MPI_INT,MPI_SUM,commun);\
     *value = tmp; }while(0)
#define MSG_Gisum2(buf1,buf2,len,commun) \
  do{int i; \
     MPI_Allreduce((void*)buf1,(void*)buf2,len,MPI_INT,MPI_SUM,commun);\
     for(i=0; i<len; i++) buf1[i] = buf2[i]; }while(0)
#define MSG_Gdsum(buf,len,commun) \
  do{int i; double* tmp = (double*) malloc(len*sizeof(double));\
     MPI_Allreduce((void*)buf,(void*)tmp,len,MPI_DOUBLE,MPI_SUM,commun);\
     for(i=0; i<len; i++) buf[i] = tmp[i]; free(tmp); }while(0)
#define MSG_Gdsum1(value,commun) \
  do{double tmp; \
     MPI_Allreduce((void*)value,(void*)&tmp,1,MPI_DOUBLE,MPI_SUM,commun);\
     *value = tmp; }while(0)
#define MSG_Gdsum2(buf1,buf2,len,commun) \
  do{int i; \
     MPI_Allreduce((void*)buf1,(void*)buf2,len,MPI_DOUBLE,MPI_SUM,commun);\
     for(i=0; i<len; i++) buf1[i] = buf2[i]; }while(0)

#endif

#endif
