#include "mymem.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static long startLive = 0xd0d0f1ea;
static long endLive = 0xae1f0d0d;
static long startDead = 0xDEADBEEF;
static long endDead = 0xFEEBDAED;

static void setLive(ptr, size)
     double* ptr;
     long size;
{
  char* p = (char*)ptr - 2*sizeof(long);
  *(long*)p = size;
  *(long*)(p+sizeof(long)) = startLive;
  memcpy(p+2*sizeof(long)+size, &endLive, sizeof(long));
}

static void setDead(ptr)
     double* ptr;
{
  char* p = (char*)ptr - 2*sizeof(long);
  long oldSize = *(long*)p;
  *(long*)(p+sizeof(long)) = startDead;
  memcpy(p+2*sizeof(long)+oldSize, &endDead, sizeof(long));
}

long mymemtest(ptr, file, line)
     double* ptr;
     char* file;
     long line;
{
  char* p = (char*)ptr - 2*sizeof(long);
  long oldSize = *(long*)p;
  long startCheck, endCheck;
  long ok = 1;
  if (oldSize <= 0)
    {
      fprintf(stderr,
	      "Negative memory size?!?, (had %ld bytes at addr %p) in line %ld of file %s\n",
	      oldSize, ptr, line, file);
      ok = 0;
    }
  startCheck = *(long*)(p+sizeof(long));
  memcpy(&endCheck, p+2*sizeof(long)+oldSize, sizeof(long));

  if (startCheck == startDead && endCheck == endDead)
    {
      fprintf
	(stderr,
	 "Memory already freed!, (had %ld bytes at addr %p) in line %ld of file %s\n",
	 oldSize, ptr, line, file);
      ok = 0;
    }
  else
    {
      if (startCheck != startLive)
	{
	  fprintf
	    (stderr,
	     "Low memory corrupted!, (had %ld bytes at addr %p) in line %ld of file %s\n",
	     oldSize, ptr, line, file);
	  ok = 0;
	}
  
      if (endCheck != endLive)
	{
	  fprintf
	    (stderr,
	     "Hi memory corrupted!, (had %ld bytes at addr %p) in line %ld of file %s\n",
	     oldSize, ptr, line, file);
	  ok = 0;
	}
    }
  return ok;
}

static long padded(size)
     long size;
{
  return size + 3*sizeof(long);
}

static char* userArea(p)
     char* p;
{
  return p+2*sizeof(long);
}

static long bufSize(ptr)
     double* ptr;
{
  char* p = (char*)ptr - 2*sizeof(long);
  return *(long*)p;
}

static char* rawPtr(ptr)
     double* ptr;
{
  return (char*)ptr - 2*sizeof(long);
}

double* mymalloc(size, file, line)
     long size;
     char* file;
     long line;
{
  double* ptr;
  char* p = (char*) malloc(padded(size));
  if (p == 0)
    fprintf(stderr,
	    "malloc: Out of memory!, requested %ld bytes in line %ld of file %s\n",
	    size, line, file);

  ptr = (double*)userArea(p);
  setLive(ptr, size);
  return ptr;
}

double* myrealloc(ptr, size, file, line)
     double* ptr;
     long size;
     char* file;
     long line;
{
  char* p = rawPtr(ptr);
  mymemtest(ptr, file, line);
  setDead(ptr);

  p = (char*) realloc(p, padded(size));
  if (p == 0)
    fprintf(stderr,
	    "Out of memory!, want %ld bytes (had %ld at addr %p) in line %ld of file %s\n",
	    size, bufSize(ptr), ptr, line, file);

  ptr = (double*)userArea(p);
  setLive(ptr, size);
  return ptr;
}

void myfree(ptr, file, line)
     double* ptr;
     char* file;
     long line;
{
  mymemtest(ptr, file, line);
  setDead(ptr);
  free(rawPtr(ptr));
}
