/*
Todd Arbogast
The University of Texas at Austin
August 1996
*/
#include <stdlib.h>
#include <stdio.h>
#include "ftn.h"

#include "driverGlobals.h"
#include "typeNumbers.h"

/* DECLARATIONS OF MACROS */
#define ABORT(err) do{error = err; goto exit;}while(0)
#define ERR_MEMORY        1
#define ERR_OPENING_FILE  2
#define ERR_PARALLEL_COMM 3
#define ERR_FATAL_F       4
#define ERR_FATAL_T       5
#define ERR_WORKSPACE_F   6
#define ERR_IN_CODE       7

#define MAX(i,j)  ((i>j) ? i : j)

/* DECLARATIONS OF ROUTINES */
double timeSince();
void checkExtraMsg(char* routine);

void ParcelFlowInit
  (int* _nProcs, int* _outFlags_f, int* _myNbr,
   int* _pcTypeIF_f, int* _maxIterIF_f,
   double* _relTolIF_f, double* _absTolIF_f, int* _diagMapPerm);
void ParcelFlow
  (double* _time_f, double* _permx, double* _permy, double* _phaseViscosity,
   double* _jacInv, float* _map, int* _nx, int* _ny, int* _nz,
   double* _xpComp_0, double* _ypComp_0, double* _zpComp_0,
   int* _hWellIndex, int* _zWellIndexLo, int* _zWellIndexHi,
   int* _wellType, int* _nWells, double* _wellRadius, double* _wellPres,
   double* _perm, double* _permPtr, double* _parcelAlpha, double* _parcelF,
   int* _bcRegion1, int* _bcRegion2, int* _bcRegion3, int* _bcType_f_1,
   int* _nBCRegions,
   int* _bcPresType_in, int* _PRES_TYPE_POTENTIAL, int* _PRES_TYPE_PRESSURE,
   int* _PRES_TYPE_HYDRAULIC_HEAD, int* _PRES_TYPE_HEAD,
   double* _gravity, double* _phaseDensity,
   double* _parcelBC_0, double* _parcelBC_1, double* _parcelBC_2,
   double* _lagrangePresX, double* _lagrangePresY, double* _lagrangePresZ,
   double* _pressure, double* _velX, double* _velY, double* _velZ,
   int* _modelMiscDisp, double*  _viscosityRatio, double* _molecularWeight,
   float*  _conc, int* _padx, int* _pady, int* _padz, int* _nSpecies,
   double* _dwk, int* _dimDwk, int* _iwk, int* _dimIwk,
   int* _myNbr, int* _nDom, int* _error);
int pssMappingSetup();
int pssTransportSetup();
int transport(double* dt , double* time_chem);
void pssTransportCloseup();
void flow_to_transport();

int outputPres();
int outputVel();
int outputWork();
int outputConc();
int outputGrid();
int outputConcHist();

void MapVel(double* _velX, double* _velY, double* _velZ,float* _map,
	    double* _xpComp_0, double* _ypComp_0, double* zpComp_0,
	    int* _nx, int* _ny, int* _nz);
void SendVelPorSatData(double* velx0, double* vely0, double* velz0,
		       double* porsat, double* t);

int singleFlowStep(double* _dt_f, double* _time_flow, double* _time_out);

/* DECLARATIONS OF STATIC VARIABLES */
static int stepOut_pres;
static double tOut_pres;
static double tOut_pres_old;
static int stepOut_vel;
static double tOut_vel;
static double tOut_vel_old;
static int stepOut_conc;
static double tOut_conc;

static int monitor;
static double* permPtr;

static double* permx;
static double* permy;
static double* parcelBC[3];
static double* parcelAlpha;
static double* parcelF;
static int iWorkspace_f;

/******************************************************************************
Title: harmAvgPerm
Description: Harmonic average face permeabilities
*/
#define HARMONIC(a,b) 2*(a)*(b)/((a)+(b))
void harmAvgPerm() {
  int i,j,k;

  for(k=0; k<nz; k++)
  for(j=0; j<ny; j++) {
    facePermX_(0,j,k) = perm_(PERM_XX,0,j,k);
    for(i=1; i<nx; i++) {
      facePermX_(i,j,k)
        = HARMONIC(perm_(PERM_XX,i-1,j,k),perm_(PERM_XX,i,j,k));
    }
    facePermX_(nx,j,k) = perm_(PERM_XX,nx-1,j,k);
  }

  for(k=0; k<nz; k++)
  for(i=0; i<nx; i++) {
    facePermY_(i,0,k) = perm_(PERM_YY,i,0,k);
    for(j=1; j<ny; j++) {
      facePermY_(i,j,k)
        = HARMONIC(perm_(PERM_YY,i,j-1,k),perm_(PERM_YY,i,j,k));
    }
    facePermY_(i,ny,k) = perm_(PERM_YY,i,ny-1,k);
  }

  for(j=0; j<ny; j++)
  for(i=0; i<nx; i++) {
    facePermZ_(i,j,0) = perm_(PERM_ZZ,i,j,0);
    for(k=1; k<nz; k++) {
      facePermZ_(i,j,k)
        = HARMONIC(perm_(PERM_ZZ,i,j,k-1),perm_(PERM_ZZ,i,j,k));
    }
    facePermZ_(i,j,nz) = perm_(PERM_ZZ,i,j,nz-1);
  }
}
#undef HARMONIC

/******************************************************************************
Title: pssDriverInit
Description: Initialize the pss driver.
Return: Error code.
*/
int pssDriverInit(double* time_flow, double* time_out) {
  int error = 0;
  int i,j,k,m;
  int writeFlow_pres=0, writeFlow_vel=0, writeTrans_conc=0;

  /* Restart or start run */

/*  if(restartRun) {
    if(monitor) printf("Processing Restart Information\n");
    error = readRestart(); if(error) goto exit;
  } else */ {
    time_f = tInitial; time_t = tInitial;
    step_f = 0; step_t = 0;
  }

  /* Initialize Output */

  if(computeFlow) {
    double max = dtMax_f[0];
    for(i=1; i<ndtMax_f; i++) { if(max < dtMax_f[i]) max = dtMax_f[i]; }

    if(dStepOut_pres == 0) dStepOut_pres = 1;
    if(dStepOut_pres < 0) dStepOut_pres = nStepsMax_f + 1;
    if(dtOut_pres == 0) dStepOut_pres = 1;
    if(dtOut_pres <= 0) dtOut_pres = tFinal + max + 1;
    writeFlow_pres = (dStepOut_pres <= nStepsMax_f)
      || (dtOut_pres <= tFinal + max);
    stepOut_pres = (initialOut || (writeFlow_pres && steadyFlow)) ?
      step_f : step_f + dStepOut_pres;
    tOut_pres = time_f + dtOut_pres;
    tOut_pres_old = tOut_pres;

    if(dStepOut_vel == 0) dStepOut_vel = 1;
    if(dStepOut_vel < 0) dStepOut_vel = nStepsMax_f + 1;
    if(dtOut_vel == 0) dStepOut_vel = 1;
    if(dtOut_vel <= 0) dtOut_vel = tFinal + max + 1;
    writeFlow_vel = (dStepOut_vel <= nStepsMax_f)
      || (dtOut_vel <= tFinal + max);
    stepOut_vel = (initialOut || (writeFlow_pres && steadyFlow)) ?
      step_f : step_f + dStepOut_vel;
    tOut_vel = time_f + dtOut_vel;
    tOut_vel_old = tOut_vel;
  }

  if(callTransport) {
    double max = dtMax_t[0];
    for(i=1; i<ndtMax_t; i++) { if(max < dtMax_t[i]) max = dtMax_t[i]; }

    if(dStepOut_conc == 0) dStepOut_conc = 1;
    if(dStepOut_conc < 0) dStepOut_conc = nStepsMax_t + 1;
    if(dtOut_conc == 0) dStepOut_conc = 1;
    if(dtOut_conc <= 0) dtOut_conc = tFinal + max + 1;
    writeTrans_conc = (dStepOut_conc <= nStepsMax_t)
      || (dtOut_conc <= tFinal + max);
    stepOut_conc = step_t + dStepOut_conc;
    tOut_conc = time_t + dtOut_conc;
  }

  if(writeFlow_pres || writeFlow_vel || writeTrans_conc || initialOut) {
    if(monitor) printf("Output Grid\n");
    error = outputGrid(); if(error) goto exit;
  }

  /* Initialize mapping */

  if(monitor) printf("Setting up Mapping\n");
  error = pssMappingSetup(); if(error) ABORT(ERR_PARALLEL_COMM);

  /* Fortran workspace */

  dimDwk = 0;
  dimDwk = MAX(dimDwk,dWorkspace_f);
  dimDwk = MAX(dimDwk,dWorkspace_t);
  dimDwk = MAX(dimDwk,dWorkspace_c);
  if(dimDwk) {
    dwk = (double*) malloc(dimDwk*sizeof(double));
    if(dwk == NULL) ABORT(ERR_MEMORY);
  }

  iWorkspace_f = (computeFlow) ? nBCRegions + 8 : 0;
  dimIwk = 0;
  dimIwk = MAX(dimIwk,iWorkspace_f);
  dimIwk = MAX(dimIwk,iWorkspace_c);
  if(dimIwk) {
    iwk = (int*) malloc(dimIwk*sizeof(int));
    if(iwk == NULL) ABORT(ERR_MEMORY);
  }

  /* Initialize flow */

  if(computeFlow) {
    if(monitor) printf("Setting up Flow\n");

    permx = (double*) malloc(nx*ny*nz*sizeof(double));
    permy = (double*) malloc(nx*ny*nz*sizeof(double));
    lagrangePresX
      = (double*) malloc(2*(ny*nz + nx*nz + nx*ny)*sizeof(double));
    parcelBC[0]
      = (double*) malloc(2*(ny*nz + nx*nz + nx*ny)*sizeof(double));
    parcelAlpha = (double*) malloc(nx*ny*nz*sizeof(double));
    parcelF = (double*) malloc(nx*ny*nz*sizeof(double));
    if(permx == NULL || permy == NULL
       || lagrangePresX == NULL || parcelBC[0] == NULL
       || parcelAlpha == NULL || parcelF == NULL) ABORT(ERR_MEMORY);
    lagrangePresY = lagrangePresX + 2*ny*nz;
    lagrangePresZ = lagrangePresY + 2*nx*nz;
    parcelBC[1] = parcelBC[0] + 2*ny*nz;
    parcelBC[2] = parcelBC[1] + 2*nx*nz;

    m = 0;
    for(k=0; k<nz; k++)
    for(j=0; j<ny; j++)
    for(i=0; i<nx; i++) {
      permx[m] = perm_(PERM_XX,i,j,k);
      permy[m] = perm_(PERM_YY,i,j,k);
      m++;
    }
    for(i=0; i<2*(ny*nz + nx*nz + nx*ny); i++) lagrangePresX[i]=0;

    if(diagMapPerm) {
      if(permType != PERM_FACE) {
	if(facePerm) ABORT(ERR_IN_CODE);
	facePerm = (double*) malloc( ( (nx+1)*ny*nz + nx*(ny+1)*nz
				       + nx*ny*(nz+1) )*sizeof(double) );
	if(facePerm == NULL) ABORT(ERR_MEMORY);
	harmAvgPerm();
      }
      permPtr = facePerm;
    } else {
      permPtr = perm;
    }

    ParcelFlowInit(&nProcs,outFlags_f,myNbr,
		   &pcTypeIF_f,&maxIterIF_f,&relTolIF_f,&absTolIF_f,
		   &diagMapPerm);

    timeSince();
    {
      double dt_f=0;
      step_f--;
      error = singleFlowStep(&dt_f, time_flow, time_out);
    }
    if(error) goto exit;

    if(steadyFlow) {
      free(permx);
      free(permy);
      free(lagrangePresX);
      free(parcelBC[0]);
      free(parcelAlpha);
      free(parcelF);
      if(callTransport) {
	dimDwk = 0;
	dimDwk = MAX(dimDwk,dWorkspace_t);
	dimDwk = MAX(dimDwk,dWorkspace_c);
	if(dimDwk) {
	  dwk = (double*) realloc((void*)dwk,dimDwk*sizeof(double));
	  if(dwk == NULL) ABORT(ERR_MEMORY);
	} else if(dwk) free(dwk);
	dimIwk = 0;
	dimIwk = MAX(dimIwk,iWorkspace_c);
	if(dimIwk) {
	  iwk = (int*) realloc((void*)iwk,dimIwk*sizeof(int));
	  if(iwk == NULL) ABORT(ERR_MEMORY);
	} else if(iwk) free(iwk);
      }
      time_f = tFinal;
    }
    if(debug >= DEBUG_PARALLEL) checkExtraMsg("FLOW");
  } else {
    /* map velocities */
    if(!identityMap) {
      MapVel(velX,velY,velZ, map, &xpComp_(0),&ypComp_(0),&zpComp_(0),
	     &nx,&ny,&nz);
    }
  }
  if(steadyFlow) time_f = tFinal;

  /* Initialize transport and flow-to-transport */

  if(callTransport) {
    if(monitor) printf("Setting up Transport\n");
    error = pssTransportSetup(); if(error) ABORT(ERR_FATAL_T);

    if(step_t == 0 && initialOut) {
      if(monitor) printf("Initial conc output\n");
      if (computeChemistry) {
         error = outputWork(); if(error) goto exit; 
      }
      error = outputConc(); if(error) goto exit;
      *time_out += timeSince();
      if(debug >= DEBUG_PARALLEL) checkExtraMsg("OUTPUT CONC");

      if(!computeFlow) {
	if(monitor) printf("Initial velocities output\n");
	error = outputVel(); if(error) goto exit;
	*time_out += timeSince();
	if(debug >= DEBUG_PARALLEL) checkExtraMsg("OUTPUT VEL");
      }
    }
    if(step_t == 0 && nHistories_t > 0) {
      if(monitor) printf("Initial conc history output\n");
      error = outputConcHist(); if(error) goto exit;
      *time_out += timeSince();
    }
  }

  /* Misc initialization */

  if(debug >= DEBUG_SETUP) dumpDriverGlobals(1,2);

 exit:
  return error;
}

/******************************************************************************
Title: singleFlowStep
Description: Take a single flow step.  Write output if necessary.
Return: Error code.
*/
int singleFlowStep(double* dt_f, double* time_flow, double* time_out) {
  int error = 0;
  int i, j, k;

  step_f++;
  if(monitor) printf("\nFLOW STEP %d FROM TIME %g\n",step_f,time_f);
  time_f += *dt_f;
  {
    int pres0 = PRES_TYPE_POTENTIAL, pres1 = PRES_TYPE_PRESSURE,
      pres2 = PRES_TYPE_HYDRAULIC_HEAD, pres3 = PRES_TYPE_HEAD;
    ParcelFlow(&time_f,permx,permy,&phaseViscosity,
	       jacInv,map,&nx,&ny,&nz,&xpComp_(0),&ypComp_(0),&zpComp_(0),
	       hWellIndex,zWellIndexLo,zWellIndexHi,
               wellType,&nWells,wellRadius,wellPres,
	       perm,permPtr,parcelAlpha,parcelF,
	       bcRegion1,bcRegion2,bcRegion3,&bcType_f_(1),&nBCRegions,
	       bcPresType_in, &pres0, &pres1, &pres2, &pres3,
	       &gravity, &phaseDensity,
	       parcelBC[0],parcelBC[1],parcelBC[2],
	       lagrangePresX,lagrangePresY,lagrangePresZ,
	       pressure,velX,velY,velZ,
	       &modelMiscDisp,viscosityRatio,molecularWeight,
	       conc,&padx,&pady,&padz,&nSpecies,
	       dwk,&dimDwk,iwk,&dimIwk,myNbr,nDom,&error);
  }
  if(error) goto exit;

  if(nProcs > 1) {
    /* Parcel overwrites these values, so reset to -1 */

    if(myNbr_(-1,0,0) > -1) {
      for(j=0; j<ny; j++)
      for(k=0; k<nz; k++)
        bcRegion1_(j,k,0) = -1;
    }
    if(myNbr_(1,0,0) > -1) {
      for(j=0; j<ny; j++)
      for(k=0; k<nz; k++)
        bcRegion1_(j,k,1) = -1;
    }
    if(myNbr_(0,-1,0) > -1) {
      for(i=0; i<nx; i++)
      for(k=0; k<nz; k++)
        bcRegion2_(i,k,0) = -1;
    }
    if(myNbr_(0,1,0) > -1) {
      for(i=0; i<nx; i++)
      for(k=0; k<nz; k++)
        bcRegion2_(i,k,1) = -1;
    }
    if(myNbr_(0,0,-1) > -1) {
      for(i=0; i<nx; i++)
      for(j=0; j<ny; j++)
        bcRegion3_(i,j,0) = -1;
    }
    if(myNbr_(0,0,1) > -1) {
      for(i=0; i<nx; i++)
      for(j=0; j<ny; j++)
        bcRegion3_(i,j,1) = -1;
    }
  }

  *time_flow += timeSince();
  if(debug >= DEBUG_PARALLEL) checkExtraMsg("FLOW");

  /* OUTPUT */
  if(step_f == stepOut_pres || time_f >= tOut_pres) {
    if(monitor) printf("Output pressure\n");
    error = outputPres(); if(error) goto exit;
    *time_out += timeSince();
    if(debug >= DEBUG_PARALLEL) checkExtraMsg("OUTPUT PRES");
    if(step_f == stepOut_pres) stepOut_pres += dStepOut_pres;
    if(time_f >= tOut_pres) {
      tOut_pres_old = tOut_pres;
      while(time_f >= tOut_pres) { tOut_pres += dtOut_pres; }
    }
  }
  if(step_f == stepOut_vel || time_f >= tOut_vel) {
    if(monitor) printf("Output velocities\n");
    error = outputVel(); if(error) goto exit;
    *time_out += timeSince();
    if(debug >= DEBUG_PARALLEL) checkExtraMsg("OUTPUT VEL");
    if(step_f == stepOut_vel) stepOut_vel += dStepOut_vel;
    if(time_f >= tOut_vel) {
      tOut_vel_old = tOut_vel;
      while(time_f >= tOut_vel) { tOut_vel += dtOut_vel; }
    }
  }

 exit:
  if(error < 0 && !myProc) {
    fprintf(stderr,"Fatal Error in Parcel: code %d\n",error);
    if(error == -15) {
      error = ERR_WORKSPACE_F;
    } else {
      error = ERR_FATAL_F;
    }
  }
  return error;
}

/******************************************************************************
Title: flow_to_transport
Description: Send flow data to transport functions
*/
void flow_to_transport() {
  SendVelPorSatData(velX,velY,velZ,effPorosity,&time_f);
  if(debug >= DEBUG_PARALLEL) checkExtraMsg("FLOW/TRANS INTERFACE");
}

/******************************************************************************
Title: singleTransportStep
Description: Take a single transport step.  Write output if necessary.
Return: Error code.
*/
int singleTransportStep(double* dt_t, double* time_trans, double* time_out, double* time_chem) {
  int error = 0;
/*  double ratio; */

  step_t++;
  if(monitor) printf("\nREACTIVE TRANSPORT STEP %d FROM TIME %g\n",
		     step_t,time_t);
  error = transport(dt_t , time_chem); /* ??? */
  *time_trans += timeSince();
/*   ratio = *time_chem / *time_trans ;
     if(verbosity) printf("%d: ratio of chem/trans time = %3.2g \n",
                          myProc , ratio);    */ 
  if(error) { step_t--; ABORT(ERR_FATAL_T); }
  if(debug >= DEBUG_PARALLEL) checkExtraMsg("TRANSPORT");

  /* OUTPUT */
  if(step_t == stepOut_conc || time_t >= tOut_conc) {
    if(monitor) printf("Output conc and work\n");
    if (computeChemistry) {
       error = outputWork(); if(error) goto exit; 
    }
    error = outputConc(); if(error) goto exit;
    *time_out += timeSince();
    if(debug >= DEBUG_PARALLEL) checkExtraMsg("OUTPUT CONC");
    if(step_t == stepOut_conc) stepOut_conc += dStepOut_conc;
    while(time_t >= tOut_conc) { tOut_conc += dtOut_conc; }
  }
  if(nHistories_t > 0) {
    error = outputConcHist(); if(error) goto exit;
    *time_out += timeSince();
  }

 exit:
  return error;
}

/******************************************************************************
Title: pssDriver
Description: Control (drive) the simulation.  Also provide time stepping
             control and output.
Return: Error code.
*/
int pssDriver(double* time_flow, double* time_trans, double* time_out, double* time_chem) {
  int error = 0, status = 0;
  int idt_f = 0, idt_t = 0;
  double dt_f, dt_t;

  monitor = verbosity > 0 && !myProc;
  error = pssDriverInit(time_flow,time_out); if(error) goto exit;
  timeSince();

  /* FLOW ONLY */
  if(computeFlow && !callTransport) {
    while(step_f < nStepsMax_f && time_f < tFinal) {
      while(idt_f < ndtMax_f-1 && time_f >= tdtMax_f[idt_f+1]) { idt_f++; }
      dt_f = dtMax_f[idt_f];
      error = singleFlowStep(&dt_f, time_flow, time_out);
      if(error) goto exit;
    }
    goto done;
  }

  /* TRANSPORT ONLY */
  if(steadyFlow && callTransport) {	
    while(step_t < nStepsMax_t && time_t < tFinal) {
      while(idt_t < ndtMax_t-1 && time_t >= tdtMax_t[idt_t+1]) { idt_t++; }
      dt_t = dtMax_t[idt_t];
      error = singleTransportStep(&dt_t, time_trans, time_out, time_chem);
      if(error) goto exit;
    }
    goto done;
  }

  /* FLOW AND TRANSPORT */
  if(computeFlow && callTransport) {
    while(idt_t < ndtMax_t-1 && time_t >= tdtMax_t[idt_t+1]) { idt_t++; }
    dt_t = dtMax_t[idt_t];
    if(time_t + dt_t > tFinal) dt_t = tFinal - time_t;

    while(time_f < tFinal || time_t < tFinal) {
      int microCounter = 0;

      /* Flow */
      while(time_f < time_t + dt_t && time_f < tFinal) {
	if(step_f >= nStepsMax_f) goto done;

	while(idt_f < ndtMax_f-1 && time_f >= tdtMax_f[idt_f+1]) { idt_f++; }
	dt_f = dtMax_f[idt_f];
	if(time_f + dt_f > tFinal) dt_f = tFinal - time_f;
	error = singleFlowStep(&dt_f, time_flow, time_out);
	if(error) goto exit;
      }

      /* Flow to Transport Interface */
      flow_to_transport();

      /* Transport */

      microCounter = 0;
      while(time_t + .5*dt_t < time_f && time_t < tFinal) {
	if(step_t >= nStepsMax_t) goto done;

        /* back up to recompute flow if necessary */
	if(nStepsMax_t_per_f > 0 && microCounter >= nStepsMax_t_per_f) {
	  if(step_f >= nStepsMax_f) goto done;
	  time_f -= dt_f;
	  tOut_pres = tOut_pres_old;
	  tOut_vel = tOut_vel_old;
	  if(monitor) printf("\nREPEATING FLOW STEP");
	  error = singleFlowStep(&dt_f, time_flow, time_out);
	  if(error) goto exit;
	  time_f -= dt_f;
	  flow_to_transport();
	  time_f += dt_f;
	  flow_to_transport();
	  microCounter = 0;
	}

	error = singleTransportStep(&dt_t, time_trans, time_out, time_chem);
	if(error) goto exit;
	microCounter++;
	while(idt_t < ndtMax_t-1 && time_t >= tdtMax_t[idt_t+1]) { idt_t++; }
	dt_t = dtMax_t[idt_t];
	if(time_t + dt_t > tFinal) dt_t = tFinal - time_t;
      }

    }
  }

 done:
#if WRITE_VEL
{
  int i,j,k;
  FILE* fl;
  char nm[] = "vx0000.dat";
  sprintf(nm,"vx%4.4d",myProc);
  fl = fopen(nm,"w");
  fprintf(fl,"%d %d %d   %d %d %d\n",nx_g,ny_g,nz_g,nx,ny,nz);
  fprintf(fl,"%d %d %d   %d %d %d\n",nDom[0],nDom[1],nDom[2],
	  myDom[0],myDom[1],myDom[2]);
  for(k=0; k<nz; k++)
  for(j=0; j<ny; j++)
  for(i=0; i<=nx; i++) {
    fprintf(fl,"%g\n", velX_(i,j,k));
  }
  fclose(fl);
  sprintf(nm,"vy%4.4d",myProc);
  fl = fopen(nm,"w");
  fprintf(fl,"%d %d %d\n",nx,ny,nz);
  for(k=0; k<nz; k++)
  for(j=0; j<=ny; j++)
  for(i=0; i<nx; i++) {
    fprintf(fl,"%g\n", velY_(i,j,k));
  }
  fclose(fl);
  sprintf(nm,"vz%4.4d",myProc);
  fl = fopen(nm,"w");
  fprintf(fl,"%d %d %d\n",nx,ny,nz);
  for(k=0; k<=nz; k++)
  for(j=0; j<ny; j++)
  for(i=0; i<nx; i++) {
    fprintf(fl,"%g\n", velZ_(i,j,k));
  }
  fclose(fl);
}
#endif

  if(callTransport) pssTransportCloseup();

  if(!myProc) {
    printf("\nSIMULATION FINISHED\n");
    if(computeFlow)   printf("  Flow:      Steps %d, Time %g\n",step_f,time_f);
    if(callTransport) printf("  Transport: Steps %d, Time %g\n",step_t,time_t);
  }

 exit:
  switch(error) {
  case 0:
    break;
  case ERR_MEMORY:
    fprintf(stderr,"Error allocating memory on processor %d\n",myProc);
    break;
  case ERR_OPENING_FILE:
    fprintf(stderr,"Error opening output file\n");
    break;
  case ERR_PARALLEL_COMM:
    fprintf(stderr,
	    "Error in parallel communications on processor %d\n",myProc);
    break;
  case ERR_FATAL_F:
    fprintf(stderr,
	    "Fatal error in flow on processor %d\n",myProc);
    break;
  case ERR_FATAL_T:
    fprintf(stderr,
	    "Fatal error in transport on processor %d\n",myProc);
    break;
  case ERR_WORKSPACE_F:
    fprintf(stderr,
	    "Insufficient Fortran double workspace for flow on processor %d\n",
	    myProc);
    break;
  case ERR_IN_CODE:
    fprintf(stderr,
	    "Error in code seen on processor %d\n",myProc);
    break;
  default:
    fprintf(stderr,
	    "Fatal error on processor %d\n",myProc);
  }
  return error;
}
