/*
Todd Arbogast
The University of Texas at Austin
August 1996
*/
#include <stdio.h>
#include "read.h"

#include "driverGlobals.h"

#define DEFINE_WRITETYPEINFO
#include "typeNumbers.h"

/* DECLARATIONS OF SECTION READING ROUTINES */
int readGeneral();
int readSolutionParams();
int readTime();
int readGrid();
int readMaterials(int *sorpTypePorosity, int* nSorpTypes_in);
int readPhase();
int readChemGen();
int readSpecies(int sorpTypePorosity, int nSorpTypes_in);
int readMiscDisp();
int readRND();
int readRXN();
int pre_readICS();
int readICS();
int readBCS();
int readWells();
int readLeaks();
int readOutput();

void internalPerm();
int echoChemData(char* _echoChemFileName);
int packChemData();

/******************************************************************************
Title: pssInitialize
Description: Read the input file, allocate memory, and preprocess data.
*/
#define UNPACK() do{ unPackDouble(); unPackInt(); unPackFloat(); }while(0)
#define FIRSTCALL(fcn) do{ prs_beginSkip(); \
  error = prs_readSection(1); if(error) goto exit; prs_endSkip(); \
  error = (fcn); if(error) goto exit; UNPACK(); }while(0)
#define CALL(fcn) do{ \
  error = prs_readSection(0); if(error) goto exit; \
  error = (fcn); if(error) goto exit; UNPACK(); }while(0)
#define PRE_CALL(fcn) do{ error = (fcn); if(error) goto exit; }while(0)

int pssInitialize(char* inFileName0, char* echoFileName,
		  char* echoChemFileName, int echo, int echoChem) {
  int error = 0;
  int sorpTypePorosity,nSorpTypes_in;

  error = openReader(inFileName0, echoFileName, SIZE_FILENAME-1, echo);
  if(error) goto exit;

  FIRSTCALL(readGeneral());

  if( !computeFlow && !callTransport ) goto exit;

  CALL(readSolutionParams());
  CALL(readTime());
  CALL(readGrid());
  CALL(readMaterials(&sorpTypePorosity,&nSorpTypes_in));
  CALL(readPhase());
  if(computeChemistry) CALL(readChemGen());
  if(callTransport) CALL(readSpecies(sorpTypePorosity,nSorpTypes_in));
  if(computeFlow && callTransport) CALL(readMiscDisp());
  if(computeRND) CALL(readRND());
  if(computeRXN) CALL(readRXN());
  PRE_CALL(pre_readICS());
  if(callTransport) CALL(readICS());
  if(computeFlow || computeTransport) {
    if(!periodicBC[0] || !periodicBC[1] || !periodicBC[2]) CALL(readBCS());
    CALL(readWells());
  }
  if(callTransport) CALL(readLeaks());
/*   debug  */
  CALL(readOutput());
  if(!myProc) {
    error = prs_readSubSubSection(0);
    if(!error) error = ERR_SUBSUBSECTION_READ;
    if(error == ERR_SUBSUBSECTION_NOT_READ || error == ERR_EOF) error = 0;
    if(error) goto exit;
  }

  if(computeFlow) internalPerm();
  if(computeChemistry) {
    if(echoChem) {
      error = echoChemData(echoChemFileName);
      if(error) goto exit;
    }
    error = packChemData(); if(error) goto exit;
  }

 exit:
  if(error != ERR_OPENING_FILE) {
    closeReader(error);
  } else {
    prs_processReaderError(error);
  }
  return error;
}
