/*
Todd Arbogast
The University of Texas at Austin
August 1996 and February 1997
*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "ftn.h"
#include "msgs.h"
#include "typeNumbers.h"

#include "driverGlobals.h"

/* DECLARATIONS OF ERROR CODES */
#define ERR_MEMORY 1
#define ERR_GRID 2
#define ERR_IN_CODE 3
#define ERR_PADDING 4
#define ERR_PARALLEL_COMM 5

#define ABORT(err) do{error = err; goto exit;}while(0)
#define MAX(i,j) ( ( (i)>(j) ) ? (i) : (j) )

void SetupMapCoefs(float* _dMap, double* _jacInv,
   double* _effPorosity, float* _effIdentity,
   double* _perm, double* _porosity, int* _nSorpTypes, double* _sorp,
   double* _xpComp_neg1, double* _ypComp_neg1, double* _zpComp_neg1,
   float* _map, int* _identityMap, int* _nx, int* _ny, int* _nz, int* _myNbr);

int padGrid(int msg) {
  int m;
  int padx1 = padx+1, pady1 = pady+1, padz1 = padz+1;
  int maxPad = MAX(padx1,MAX(pady1,padz1));
  MPI_Status status;

  double* wk = (double*) malloc(maxPad*sizeof(double));
  if(wk == NULL) return ERR_MEMORY;

  /* COMMUNICATE INTERIOR OF PROCESSOR TOPOLOGY */

  /* X-direction */
  if(myDom[0] < nDom[0]-1) {
    for(m = 0; m < padx1; m++) wk[m] = xpComp_(nx-padx1+m);
    MPI_Send((void*)wk,padx1,MPI_DOUBLE,myNbr_(1,0,0),msg,comm);
  }
  if(myDom[0] > 0) {
    MPI_Recv((void*)wk,padx1,MPI_DOUBLE,myNbr_(-1,0,0),msg,comm,&status);
    for(m = 0; m < padx1; m++) xpComp_(m-padx1) = wk[m];
  }
  msg++;
  if(myDom[0] > 0) {
    for(m = 0; m < padx1; m++) wk[m] = xpComp_(m+1);
    MPI_Send((void*)wk,padx1,MPI_DOUBLE,myNbr_(-1,0,0),msg,comm);
  }
  if(myDom[0] < nDom[0]-1) {
    MPI_Recv((void*)wk,padx1,MPI_DOUBLE,myNbr_(1,0,0),msg,comm,&status);
    for(m = 0; m < padx1; m++) xpComp_(nx+1+m) = wk[m];
  }

  /* Y-direction */
  msg++;
  if(myDom[1] < nDom[1]-1) {
    for(m = 0; m < pady1; m++) wk[m] = ypComp_(ny-pady1+m);
    MPI_Send((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,1,0),msg,comm);
  }
  if(myDom[1] > 0) {
    MPI_Recv((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,-1,0),msg,comm,&status);
    for(m = 0; m < pady1; m++) ypComp_(m-pady1) = wk[m];
  }
  msg++;
  if(myDom[1] > 0) {
    for(m = 0; m < pady1; m++) wk[m] = ypComp_(m+1);
    MPI_Send((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,-1,0),msg,comm);
  }
  if(myDom[1] < nDom[1]-1) {
    MPI_Recv((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,1,0),msg,comm,&status);
    for(m = 0; m < pady1; m++) ypComp_(ny+1+m) = wk[m];
  }

  /* Z-direction */
  msg++;
  if(myDom[2] < nDom[2]-1) {
    for(m = 0; m < padz1; m++) wk[m] = zpComp_(nz-padz1+m);
    MPI_Send((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,1),msg,comm);
  }
  if(myDom[2] > 0) {
    MPI_Recv((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,-1),msg,comm,&status);
    for(m = 0; m < padz1; m++) zpComp_(m-padz1) = wk[m];
  }
  msg++;
  if(myDom[2] > 0) {
    for(m = 0; m < padz1; m++) wk[m] = zpComp_(m+1);
    MPI_Send((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,-1),msg,comm);
  }
  if(myDom[2] < nDom[2]-1) {
    MPI_Recv((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,1),msg,comm,&status);
    for(m = 0; m < padz1; m++) zpComp_(nz+1+m) = wk[m];
  }

  /* COMMUNICATE OR ARTIFICIALLY PAD EXTERIOR OF PROCESSOR TOPOLOGY */

  /* X-direction */
  msg++;
  if(myDom[0] == nDom[0]-1 && periodicBC[0]) {
    for(m = 0; m < padx1; m++) wk[m] = xpComp_(nx) - xpComp_(nx-padx1+m);
    MPI_Send((void*)wk,padx1,MPI_DOUBLE,myNbr_(1,0,0),msg,comm);
  }
  if(myDom[0] == 0) {
    if(periodicBC[0]) {
      MPI_Recv((void*)wk,padx1,MPI_DOUBLE,myNbr_(-1,0,0),msg,comm,&status);
    } else {
      for(m = 0; m < padx1; m++) wk[m] = xpComp_(padx1-m) - xpComp_(0);/*(padx1-m)*(xpComp_(nx)-xpComp_(0));*/
    }
    for(m = 0; m < padx1; m++) xpComp_(m-padx1) = xpComp_(0) - wk[m];
  }
  msg++;
  if(myDom[0] == 0 && periodicBC[0]) {
    for(m = 0; m < padx1; m++) wk[m] = xpComp_(m+1) - xpComp_(0);
    MPI_Send((void*)wk,padx1,MPI_DOUBLE,myNbr_(-1,0,0),msg,comm);
  }
  if(myDom[0] == nDom[0]-1) {
    if(periodicBC[0]) {
      MPI_Recv((void*)wk,padx1,MPI_DOUBLE,myNbr_(1,0,0),msg,comm,&status);
    } else {
      for(m = 0; m < padx1; m++) wk[m] = xpComp_(nx) - xpComp_(nx-1-m);/*(m+1)*(xpComp_(nx)-xpComp_(0));*/
    }
    for(m = 0; m < padx1; m++) xpComp_(nx+1+m) = xpComp_(nx) + wk[m];
  }

 /* Y-direction */
  msg++;
  if(myDom[1] == nDom[1]-1 && periodicBC[1]) {
    for(m = 0; m < pady1; m++) wk[m] = ypComp_(ny) - ypComp_(ny-pady1+m);
    MPI_Send((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,1,0),msg,comm);
  }
  if(myDom[1] == 0) {
    if(periodicBC[1]) {
      MPI_Recv((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,-1,0),msg,comm,&status);
    } else {
      for(m = 0; m < pady1; m++) wk[m] = ypComp_(pady1-m) - ypComp_(0);/*(pady1-m)*(ypComp_(ny)-ypComp_(0));*/
    }
    for(m = 0; m < pady1; m++) ypComp_(m-pady1) = ypComp_(0) - wk[m];
  }
  msg++;
  if(myDom[1] == 0 && periodicBC[1]) {
    for(m = 0; m < pady1; m++) wk[m] = ypComp_(m+1) - ypComp_(0);
    MPI_Send((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,-1,0),msg,comm);
  }
  if(myDom[1] == nDom[1]-1) {
    if(periodicBC[1]) {
      MPI_Recv((void*)wk,pady1,MPI_DOUBLE,myNbr_(0,1,0),msg,comm,&status);
    } else {
      for(m = 0; m < pady1; m++) wk[m] = ypComp_(ny) - ypComp_(ny-1-m);/*(m+1)*(ypComp_(ny)-ypComp_(0));*/
    }
    for(m = 0; m < pady1; m++) ypComp_(ny+1+m) = ypComp_(ny) + wk[m];
  }

  /* Z-direction */
  msg++;
  if(myDom[2] == nDom[2]-1 && periodicBC[2]) {
    for(m = 0; m < padz1; m++) wk[m] = zpComp_(nz) - zpComp_(nz-padz1+m);
    MPI_Send((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,1),msg,comm);
  }
  if(myDom[2] == 0) {
    if(periodicBC[2]) {
      MPI_Recv((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,-1),msg,comm,&status);
    } else {
      for(m = 0; m < padz1; m++) wk[m] = zpComp_(padz1-m) - zpComp_(0);/*(padz1-m)*(zpComp_(nz)-zpComp_(0));*/
    }
    for(m = 0; m < padz1; m++) zpComp_(m-padz1) = zpComp_(0) - wk[m];
  }
  msg++;
  if(myDom[2] == 0 && periodicBC[2]) {
    for(m = 0; m < padz1; m++) wk[m] = zpComp_(m+1) - zpComp_(0);
    MPI_Send((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,-1),msg,comm);
  }
  if(myDom[2] == nDom[2]-1) {
    if(periodicBC[2]) {
      MPI_Recv((void*)wk,padz1,MPI_DOUBLE,myNbr_(0,0,1),msg,comm,&status);
    } else {
      for(m = 0; m < padz1; m++) wk[m] = zpComp_(nz) - zpComp_(nz-1-m);/*(m+1)*(zpComp_(nz)-zpComp_(0));*/
    }
    for(m = 0; m < padz1; m++) zpComp_(nz+1+m) = zpComp_(nz) + wk[m];
  }

  free(wk);
  return 0;
}

int padMap(int msg) {
  int i,j,k,m,maxFace,error=0;
  float* wk;
  MPI_Status status;

  int three = 3, one = 1;
  int nx1 = nx+1, ny1 = ny+1, nz1 = nz+1;

  /* Communicate padded points */

  RPadStencil(map,&three,&nx1,&ny1,&nz1,
	      &one,&one,&one,&one,&one,&one,myNbr,&msg,&error);
  if(error>0) return ERR_IN_CODE;
  if(error<0) return ERR_PARALLEL_COMM;

  /* Correct periodically distributed points */

  maxFace = 0;
  if(periodicBC[0]) maxFace = MAX(maxFace,(ny+1)*(nz+1));
  if(periodicBC[1]) maxFace = MAX(maxFace,(ny+1)*(nz+1));
  if(periodicBC[2]) maxFace = MAX(maxFace,(nx+1)*(ny+1));

  if(maxFace == 0) return 0;

  wk = (float*) malloc(maxFace*sizeof(float));
    
  /* X-direction */
  msg++;
  if(periodicBC[0]) {
    if(myDom[0] == nDom[0]-1) {
      m = 0;
      for(k=0; k<=nz; k++)
      for(j=0; j<=ny; j++) {
	wk[m++] = map_(0,nx,j,k) - map_(0,nx-1,j,k);
      }
      MPI_Send((void*)wk,(ny+1)*(nz+1),MPI_FLOAT,myNbr_(1,0,0),msg,comm);
    }
    if(myDom[0] == 0) {
      MPI_Recv((void*)wk,(ny+1)*(nz+1),MPI_FLOAT,
	       myNbr_(-1,0,0),msg,comm,&status);
      m = 0;
      for(k=0; k<=nz; k++)
      for(j=0; j<=ny; j++) {
	map_(0,-1,j,k) = map_(0,0,j,k) - wk[m++];
      }
    }
    msg++;
    if(myDom[0] == 0) {
      m = 0;
      for(k=0; k<=nz; k++)
      for(j=0; j<=ny; j++) {
	wk[m++] = map_(0,1,j,k) - map_(0,0,j,k);
      }
      MPI_Send((void*)wk,(ny+1)*(nz+1),MPI_FLOAT,myNbr_(-1,0,0),msg,comm);
    }
    if(myDom[0] == nDom[0]-1) {
      MPI_Recv((void*)wk,(ny+1)*(nz+1),MPI_FLOAT,
	       myNbr_(1,0,0),msg,comm,&status);
      m = 0;
      for(k=0; k<=nz; k++)
      for(j=0; j<=ny; j++) {
	map_(0,nx+1,j,k) = map_(0,nx,j,k) + wk[m++];
      }
    }
  } else msg++;

  /* Y-direction */
  msg++;
  if(periodicBC[1]) {
    if(myDom[1] == nDom[1]-1) {
      m = 0;
      for(k=0; k<=nz; k++)
      for(i=0; i<=nx; i++) {
	wk[m++] = map_(1,i,ny,k) - map_(1,i,ny-1,k);
      }
      MPI_Send((void*)wk,(nx+1)*(nz+1),MPI_FLOAT,myNbr_(0,1,0),msg,comm);
    }
    if(myDom[1] == 0) {
      MPI_Recv((void*)wk,(nx+1)*(nz+1),MPI_FLOAT,
	       myNbr_(0,-1,0),msg,comm,&status);
      m = 0;
      for(k=0; k<=nz; k++)
      for(i=0; i<=nx; i++) {
	map_(1,i,-1,k) = map_(1,i,0,k) - wk[m++];
      }
    }
    msg++;
    if(myDom[1] == 0) {
      m = 0;
      for(k=0; k<=nz; k++)
      for(i=0; i<=nx; i++) {
	wk[m++] = map_(1,i,1,k) - map_(1,i,0,k);
      }
      MPI_Send((void*)wk,(nx+1)*(nz+1),MPI_FLOAT,myNbr_(0,-1,0),msg,comm);
    }
    if(myDom[1] == nDom[1]-1) {
      MPI_Recv((void*)wk,(nx+1)*(nz+1),MPI_FLOAT,
	       myNbr_(0,1,0),msg,comm,&status);
      m = 0;
      for(k=0; k<=nz; k++)
      for(i=0; i<=nx; i++) {
	map_(1,i,ny+1,k) = map_(1,i,ny,k) + wk[m++];
      }
    }
  } else msg++;

  /* Z-direction */
  msg++;
  if(periodicBC[2]) {
    if(myDom[2] == nDom[2]-1) {
      m = 0;
      for(j=0; j<=ny; j++)
      for(i=0; i<=nx; i++) {
	wk[m++] = map_(2,i,j,nz) - map_(2,i,j,nz-1);
      }
      MPI_Send((void*)wk,(nx+1)*(ny+1),MPI_FLOAT,myNbr_(0,0,1),msg,comm);
    }
    if(myDom[2] == 0) {
      MPI_Recv((void*)wk,(nx+1)*(ny+1),MPI_FLOAT,
	       myNbr_(0,0,-1),msg,comm,&status);
      m = 0;
      for(j=0; j<=ny; j++)
      for(i=0; i<=nx; i++) {
	map_(2,i,j,-1) = map_(2,i,j,0) - wk[m++];
      }
    }
    msg++;
    if(myDom[2] == 0) {
      m = 0;
      for(j=0; j<=ny; j++)
      for(i=0; i<=nx; i++) {
	wk[m++] = map_(2,i,j,1) - map_(2,i,j,0);
      }
      MPI_Send((void*)wk,(nx+1)*(ny+1),MPI_FLOAT,myNbr_(0,0,-1),msg,comm);
    }
    if(myDom[2] == nDom[2]-1) {
      MPI_Recv((void*)wk,(nx+1)*(ny+1),MPI_FLOAT,
	       myNbr_(0,0,1),msg,comm,&status);
      m = 0;
      for(j=0; j<=ny; j++)
      for(i=0; i<=nx; i++) {
	map_(2,i,j,nz+1) = map_(2,i,j,nz) + wk[m++];
      }
    }
  } else msg++;

  free(wk);
  return 0;
}

/******************************************************************************
Title: dsum2d
Description: Parallel double sum of a vector in a 2-D cross-section
             of processors.
Inputs: perp   The direction perpendicular to the 2-D cross-section.
        data   The data items on the processor.
        n      Length of the data array.
        msg    Message type number.
Output: data   The parallel vector sum of all data items in the 2-D
               cross-section of processors.
Return: Error code.
*/
int dsum2d(int perp, double* data, int n, int msg) {
  MPI_Comm slice;

  MPI_Comm_split(comm,myDom[perp],0,&slice);
  MSG_Gdsum(data,n,slice);
  MPI_Comm_free(&slice);

  return 0;
}

/******************************************************************************
Title: defineComputationalGrid
Description: Define a computational grid.
Return: Error code.
*/
int defineComputationalGrid() {
  int error = 0;
  int i,j,k;
  double dist;
  int nxOwned = nx-1, nyOwned = ny-1, nzOwned = nz-1;
  MPI_Status status;

  int nxGbl = (periodicBC[0]) ? nx_g : nx_g+1;
  int nyGbl = (periodicBC[1]) ? ny_g : ny_g+1;
  int nzGbl = (periodicBC[2]) ? nz_g : nz_g+1;

#define square(x) ((x)*(x))
#define distance(i2,j2,k2, i1,j1,k1) \
  sqrt( square(xp_(i2,j2,k2) - xp_(i1,j1,k1)) \
       + square(yp_(i2,j2,k2) - yp_(i1,j1,k1)) \
       + square(zp_(i2,j2,k2) - zp_(i1,j1,k1)) )

  if(myNbr_(1,0,0) < 0) nxOwned++;
  if(myNbr_(0,1,0) < 0) nyOwned++;
  if(myNbr_(0,0,1) < 0) nzOwned++;

  for(i=1; i<nx+1; i++) {
    xpComp_(i) = 0;
    for(k=0; k<=nzOwned; k++)
    for(j=0; j<=nyOwned; j++) {
      dist = distance(i,j,k, i-1,j,k);
      if(dist == 0) return ERR_GRID;
      xpComp_(i) += dist;
    }
  }
  if(nDom[1]*nDom[2] > 1) {
    error = dsum2d(0,&xpComp_(1),nx,MSG_XPCOMP1); if(error) return error;
  }
  for(i=1; i<nx+1; i++) xpComp_(i) /= (nyGbl*nzGbl);
  if(myDom[0] == 0) {
    xpComp_(0) = 0;
  } else {
    /* recvRealX(MSG_XPCOMP2,&xpComp_(0),1); */
    MPI_Recv((void*)&xpComp_(0),1,MPI_DOUBLE,
	     MPI_ANY_SOURCE,MSG_XPCOMP2,comm,&status);
  }
  for(i=1; i<nx+1; i++) xpComp_(i) += xpComp_(i-1);
  if(myDom[0] != nDom[0]-1) {
    /* sendRealX(MSG_XPCOMP2,&xpComp_(nx),1,myNbr_(1,0,0)); */
    MPI_Send((void*)&xpComp_(nx),1,MPI_DOUBLE,myNbr_(1,0,0),MSG_XPCOMP2,comm);
  }

  for(j=1; j<ny+1; j++) {
    ypComp_(j) = 0;
    for(k=0; k<=nzOwned; k++)
    for(i=0; i<=nxOwned; i++) {
      dist = distance(i,j,k, i,j-1,k);
      if(dist == 0) return ERR_GRID;
      ypComp_(j) += dist;
    }
  }
  if(nDom[0]*nDom[2] > 1) {
    error = dsum2d(1,&ypComp_(1),ny,MSG_YPCOMP1); if(error) return error;
  }
  for(j=1; j<ny+1; j++) ypComp_(j) /= (nxGbl*nzGbl);
  if(myDom[1] == 0) {
    ypComp_(0) = 0;
  } else {
    /* recvRealX(MSG_YPCOMP2,&ypComp_(0),1); */
    MPI_Recv((void*)&ypComp_(0),1,MPI_DOUBLE,
	     MPI_ANY_SOURCE,MSG_YPCOMP2,comm,&status);
  }
  for(j=1; j<ny+1; j++) ypComp_(j) += ypComp_(j-1);
  if(myDom[1] != nDom[1]-1) {
    /* sendRealX(MSG_YPCOMP2,&ypComp_(ny),1,myNbr_(0,1,0)); */
    MPI_Send((void*)&ypComp_(ny),1,MPI_DOUBLE,myNbr_(0,1,0),MSG_YPCOMP2,comm);
  }

  for(k=1; k<nz+1; k++) {
    zpComp_(k) = 0;
    for(j=0; j<=nyOwned; j++)
    for(i=0; i<=nxOwned; i++) {
      dist = distance(i,j,k, i,j,k-1);
      if(dist == 0) return ERR_GRID;
      zpComp_(k) += dist;
    }
  }
  if(nDom[0]*nDom[1] > 1) {
    error = dsum2d(2,&zpComp_(1),nz,MSG_ZPCOMP1); if(error) return error;
  }
  for(k=1; k<nz+1; k++) zpComp_(k) /= (nxGbl*nyGbl);
  if(myDom[2] == 0) {
    zpComp_(0) = 0;
  } else {
    /* recvRealX(MSG_ZPCOMP2,&zpComp_(0),1); */
    MPI_Recv((void*)&zpComp_(0),1,MPI_DOUBLE,
	     MPI_ANY_SOURCE,MSG_ZPCOMP2,comm,&status);
  }
  for(k=1; k<nz+1; k++) zpComp_(k) += zpComp_(k-1);
  if(myDom[2] != nDom[2]-1) {
    /* sendRealX(MSG_ZPCOMP2,&zpComp_(nz),1,myNbr_(0,0,1)); */
    MPI_Send((void*)&zpComp_(nz),1,MPI_DOUBLE,myNbr_(0,0,1),MSG_ZPCOMP2,comm);
  }

  return 0;
}

/******************************************************************************
Title: pssMappingSetup
Description: Driver initialization related to the mapping.
Return: Error code.
*/
int pssMappingSetup() {
  int error = 0;
  int i,j,k;

  /* padding needed for transport */

  if( (nx < padx+1 && (myNbr_(-1,0,0) != -1 || myNbr_(1,0,0) != -1)) || 
      (ny < pady+1 && (myNbr_(0,-1,0) != -1 || myNbr_(0,1,0) != -1)) || 
      (nz < padz+1 && (myNbr_(0,0,-1) != -1 || myNbr_(0,0,1) != -1)) )
    ABORT(ERR_PADDING);

  xpComp = (double*) malloc((nx+1+2*(padx+1))*sizeof(double));
  ypComp = (double*) malloc((ny+1+2*(pady+1))*sizeof(double));
  zpComp = (double*) malloc((nz+1+2*(padz+1))*sizeof(double));
  if(xpComp == NULL || ypComp == NULL || zpComp == NULL) ABORT(ERR_MEMORY);

  error = defineComputationalGrid(); if(error) goto exit;
  error = padGrid(MSG_PADGRID); if(error) goto exit;

  map = (float*) malloc(3*(nx+3)*(ny+3)*(nz+3)*sizeof(float));
  dMap = (float*) malloc(9*(nx+1)*(ny+1)*(nz+1)*sizeof(float));
  jacInv = (double*) malloc(nx*ny*nz*sizeof(double));
  effIdentity = (float*) malloc(6*(nx+1)*(ny+1)*(nz+1)*sizeof(float));
  effPorosity = (double*) malloc(nx*ny*nz*sizeof(double));
  if(map == NULL || dMap == NULL || jacInv == NULL || effIdentity == NULL
     || effPorosity == NULL) ABORT(ERR_MEMORY);

  for(i=0; i<3*(nx+3)*(ny+3)*(nz+3); i++) map[i] = 0;
  for(i=0; i<9*(nx+1)*(ny+1)*(nz+1); i++) dMap[i] = 0;

  for(k=0; k<=nz; k++)
  for(j=0; j<=ny; j++)
  for(i=0; i<=nx; i++) {
    map_(0,i,j,k) = xp_(i,j,k);
    map_(1,i,j,k) = yp_(i,j,k);
    map_(2,i,j,k) = zp_(i,j,k);
  }
  error = padMap(MSG_PADMAP); if(error) goto exit;

  /* Compute Jacobian matrix and modify coefficients */

  SetupMapCoefs(dMap,jacInv,effPorosity,effIdentity,perm,
		porosity,&nSorpTypes,sorp,
		&xpComp_(-1),&ypComp_(-1),&zpComp_(-1),
		map,&identityMap,&nx,&ny,&nz,myNbr);

  if(debug >= DEBUG_SETUP) {
    debugDouble(1,"xpComp",xpComp,(nx+1+2*(padx+1)));
    debugDouble(1,"ypComp",ypComp,(ny+1+2*(pady+1)));
    debugDouble(1,"zpComp",zpComp,(nz+1+2*(padz+1)));
    debugFloat(1,"map",map,3*(nx+3)*(ny+3)*(nz+3));
    debugFloat(1,"dMap",dMap,9*(nx+1)*(ny+1)*(nz+1));
    debugDouble(1,"jacInv",jacInv,nx*ny*nz);
    debugFloat(1,"effIdentity",effIdentity,6*(nx+1)*(ny+1)*(nz+1));
    debugDouble(1,"effPorosity",effPorosity,nx*ny*nz);
  }

 exit:
  switch(error) {
  case 0: break;
  case ERR_MEMORY:
    fprintf(stderr,"Error allocating memory on processor %d\n",myProc);
    break;
  case ERR_GRID:
    fprintf(stderr,"Improper grid detected on processor %d\n",myProc);
    break;
  case ERR_IN_CODE:
    fprintf(stderr,"Error in code on processor %d\n",myProc);
    break;
  case ERR_PADDING:
    fprintf(stderr,"Padding beyond subdomain size on processor %d\n",myProc);
    break;
  case ERR_PARALLEL_COMM:
    fprintf(stderr,"Parallel communications error on processor %d\n",myProc);
    break;
  default:
    fprintf(stderr,"Error on processor %d\n",myProc);
  }
  return error;
}
