/*
Todd Arbogast
The University of Texas at Austin
August 1996
*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "ftn.h"
#include "msgs.h"

#include "driverGlobals.h"

/* DECLARATIONS OF ERROR CODES */
#define ERR_NO_MESSAGE -1
#define ERR_MEMORY 1
#define ERR_IN_CODE 2

/* DECLARATIONS OF ROUTINES */
void TransportSetup(int* _nDom, int* _myDom, int* _myNbr,
      int* _nx_f_g, int* _ny_f_g, int* _nz_f_g,
      int* _nx_f, int* _ny_f, int* _nz_f,
      int* _padx_f, int* _pady_f, int* _padz_f,
      int* _nx_t_g, int* _ny_t_g, int* _nz_t_g,
      int* _nx_t, int* _ny_t, int* _nz_t,
      int* _padx_t, int* _pady_t, int* _padz_t,
      double* _xpComp_f, double* _ypComp_f, double* _zpComp_f,
      double* _xpComp_t, double* _ypComp_t, double* _zpComp_t,
      int* _identityMap, float* _dMap,
      int* _xOffset_g, int* _yOffset_g, int* _zOffset_g,
      double* _effPorosity, int* _nSorpTypes, double* _sorp,
      int* _nSpecies, int* _steadyFlow, double* _tInitial, double* _tFinal,
      double* _velX, double* _velY, double* _velZ);
int transportInit();
void transportClose();
void FreeFcnmem();

#define ABORT(err) do{error = err; goto exit;}while(0)

/******************************************************************************
Title: pssTransportSetup
Description: Driver initialization related to transport.
Return: Error code.
*/
int pssTransportSetup() {
  int error = 0;

  TransportSetup(nDom, myDom, myNbr,
     &nx_g, &ny_g, &nz_g, &nx, &ny, &nz, &padx, &pady, &padz,
     &nx_g, &ny_g, &nz_g, &nx, &ny, &nz, &padx, &pady, &padz,
     &xpComp_(-padx), &ypComp_(-pady), &zpComp_(-padz), xpComp, ypComp, zpComp,
     &identityMap, dMap, &xOffset_g_(-1), &yOffset_g_(-1), &zOffset_g_(-1),
     effPorosity, &nSorpTypes, sorp,
     &nSpecies, &steadyFlow, &tInitial, &tFinal,
     velX, velY, velZ);

  error = transportInit(); if(error) ABORT(ERR_NO_MESSAGE);

 exit:
  switch(error) {
  case 0:
  case ERR_NO_MESSAGE:
    break;
  case ERR_MEMORY:
    fprintf(stderr,"Memory allocation error on processor %d\n",myProc);
    break;
  default:
    fprintf(stderr,"Error on processor %d\n",myProc);
  }
  return error;
}

/******************************************************************************
Title: pssTransportCloseup
Description: End the transport code.
*/
void pssTransportCloseup() {
  FreeFcnmem();
  transportClose();
}
