#include "reader.h"

/******************************************************************************
Title: MACROS for reading input
Description: General helpful macros and stubs for the proper calls to
             the utility reader routines.
*/
#define ABORT(err) do { error = (err); goto exit; } while(0)
#define STRING_IS(str) !strcmp(string,str)

/* general */
#define openReader prs_openReader
#define closeReader prs_closeReader
#define processReaderError prs_processReaderError

#define beginSkip() prs_beginSkip()
#define endSkip() prs_endSkip()

/* Skip through the next whitespace (WS), key word, ... */
#define skipWS() do{ error = prs_skipWS(0); if(error) goto exit; }while(0)
#define skipWS_EOL() \
  do{ error = prs_skipWS(1); if(error) goto exit; }while(0)
#define skipKeys(str,n,k) \
  do{ error = prs_readKeys(str,n,k,1); if(error) goto exit; }while(0)
#define skipKey(str) \
  do{ error = prs_readKey(str,1); if(error) goto exit; }while(0)

#define readSection(skip) \
  do{ error = prs_readSection(skip); if(error) goto exit; }while(0)
#define readSubSection(skip) \
  do{ error = prs_readSubSection(skip); if(error) goto exit; }while(0)
#define readSubSubSection(skip) \
  do{ error = prs_readSubSubSection(skip); if(error) goto exit; }while(0)

#define skipToSection(str) \
  do{ error = prs_skipToSection(str); if(error) goto exit; }while(0)
#define skipToSubSection(str) \
  do{ error = prs_skipToSubSection(str); if(error) goto exit; }while(0)
#define skipToSubSubSection(str) \
  do{ error = prs_skipToSubSubSection(str); if(error) goto exit; }while(0)
#define skipToSections(str,n,k) \
  do{ error = prs_skipToSections(str,n,k); if(error) goto exit; }while(0)
#define skipToSubSections(str,n,k) \
  do{ error = prs_skipToSubSections(str,n,k); if(error) goto exit; }while(0)
#define skipToSubSubSections(str,n,k) \
  do{ error = prs_skipToSubSubSections(str,n,k); if(error) goto exit; }while(0)

#define beginList() \
  do{ error = prs_beginList(); if(error) goto exit; }while(0)
#define endList() \
  do{ error = prs_endList(); if(error) goto exit; }while(0)
#define skipToEndList() \
  do{ error = prs_skipToEndList(); if(error) goto exit; }while(0)

/* Read the next key word, ... */
#define readKeys(str,n,k) \
  do{ error = prs_readKeys(str,n,k,0); if(error) goto exit; }while(0)
#define readKey(str) \
  do{ error = prs_readKey(str,0); if(error) goto exit; }while(0)

/* Pack the next integer, double, float, ... */
#define packInt(var) \
  do{ error = prs_packInt(&(var)); if(error) goto exit; }while(0)
#define packDouble(var) \
  do{ error = prs_packDouble(&(var)); if(error) goto exit; }while(0)
#define packFloat(var) \
  do{ error = prs_packFloat(&(var)); if(error) goto exit; }while(0)
#define packDoubleUnits(var,dimen) \
  do{ int dimensions[] = {dimen}; \
      error = prs_packDoubleUnits(&(var),dimensions); \
      if(error) goto exit; }while(0)
#define packFloatUnits(var,dimen) \
  do{ int dimensions[] = {dimen}; \
      error = prs_packFloatUnits(&(var),dimensions); \
      if(error) goto exit; }while(0)
#define packDoubleTemp(var) \
  do{ error = prs_packDoubleTemp(&(var)); if(error) goto exit; }while(0)
#define packFloatTemp(var) \
  do{ error = prs_packFloatTemp(&(var)); if(error) goto exit; }while(0)

/* unpack integer, double, float, ... */
#define unPackInt prs_unPackInt
#define unPackDouble prs_unPackDouble
#define unPackFloat prs_unPackFloat

/* Read the next string, line, ... */
#define readString(string,len) \
  do{ error = prs_readString(string,len); if(error) goto exit; }while(0)
#define readLine(string,len) \
  do{ error = prs_readLine(string,len); if(error) goto exit; }while(0)
#define readIntArray(array,len,nRead,nToReadTotal) \
  do{ error = prs_readIntArray(array,len,nRead,nToReadTotal); \
    if(error) goto exit; }while(0)
#define readDoubleArray(array,len,nRead,nToReadTotal) \
  do{ error = prs_readDoubleArray(array,len,nRead,nToReadTotal); \
    if(error) goto exit; }while(0)
#define readFloatArray(array,len,nRead,nToReadTotal) \
  do{ error = prs_readFloatArray(array,len,nRead,nToReadTotal); \
    if(error) goto exit; }while(0)
#define readDoubleArrayUnits(array,len,nRead,nToReadTotal,units) \
  do{ error = prs_readDoubleArrayUnits(array,len,nRead,nToReadTotal,units); \
    if(error) goto exit; }while(0)
#define readFloatArrayUnits(array,len,nRead,nToReadTotal,units) \
  do{ error = prs_readFloatArrayUnits(array,len,nRead,nToReadTotal,units); \
    if(error) goto exit; }while(0)

/******************************************************************************
Declarations for the 3D reader functions
*/
int getUnits(double* units, int* dimen);

int read3DDouble(double* array, double units,
		 int pointx, int pointy, int pointz,
		 int n, int px, int py, int pz);
int read3DInt(int* array,
	      int pointx, int pointy, int pointz,
	      int n, int px, int py, int pz);
int read3DFloat(float* array, double units,
		int pointx, int pointy, int pointz,
		int n, int px, int py, int pz);

int setup3DDouble(double* array, int* dimen,
		  int pointx, int pointy, int pointz,
		  int n, int px, int py, int pz,
		  char* string, int charSize, int* nearly);
int setup3DInt(int* array,
	       int pointx, int pointy, int pointz,
	       int n, int px, int py, int pz,
	       char* string, int charSize, int* nearly);
int setup3DFloat(float* array, int* dimen,
		 int pointx, int pointy, int pointz,
		 int n, int px, int py, int pz,
		 char* string, int charSize, int* nearly);

int read3DBndryDouble(double* arrayBndry, double units,
		      int side, int point1, int point2, int n, int p1, int p2);
int read3DBndryInt(int* arrayBndry,
		   int side, int point1, int point2, int n, int p1, int p2);
int read3DBndryFloat(float* arrayBndry, double units,
		     int side, int point1, int point2, int n, int p1, int p2);

int setup3DBndryDouble(double* arrayBndry, int* dimen,
		       int side, int point1, int point2, int n, int p1, int p2,
		       char* string, int charSize, int* nearly);
int setup3DBndryInt(int* arrayBndry,
		    int side, int point1, int point2, int n, int p1, int p2,
		    char* string, int charSize, int* nearly);
int setup3DBndryFloat(float* arrayBndry, int* dimen,
		      int side, int point1, int point2, int n, int p1, int p2,
		      char* string, int charSize, int* nearly);

#define SIDE_X 0
#define SIDE_Y 1
#define SIDE_Z 2

