#include <stdio.h>
#include "read.h"
#include "typeNumbers.h"
#include "expr.h"

#include "driverGlobals.h"

#define EPS 1e-6

/******************************************************************************
Title: readGeneral
Description: Read and preprocess the general input data.
Return: Error code (see reader.h).
*/
int readGeneral() {
  static char GENERAL_INFO[] = {"GENERAL INFO"};

  int error = 0;

  skipToSection(GENERAL_INFO);

  packInt(computeFlow);

  packInt(computeTransport);
  unPackInt();
  switch(computeTransport) {
  case 0:
    break;
  case ADV_CMM:
  case ADV_HOG:
  case ADV_GODUNOV:
  case ADV_DIFF_ONLY:
    break;
  default:
    ABORT(ERR_UNSUPPORTED_OPTION);
  }

  packInt(computeChemistry);
  unPackInt();
  if(computeChemistry == CHEM_SPECIALIZED) {
    computeChemistry = 0;
    computeRXN = 1;
  } else {
    computeRXN = 0;
  }
  if(computeChemistry != 0 && computeChemistry != CHEM_VERSION) ABORT(-1);

  packInt(computeRND);

  unPackInt();
  steadyFlow = 1;
  diagMapPerm = 1;
  callTransport = computeTransport || computeChemistry
    || computeRND || computeRXN;

  /* Base Units */
  if(!myProc) {
    double baseLength,baseMass,baseTime,baseTemp;
    int base, newBase;

    {
      int dimen[] = {DIMEN_LENGTH};
      skipWS();
      error = prs_readUnits(&baseLength,dimen);
      if(error) { if(error == NO_UNITS) ABORT(ERR_ON_LINE); ABORT(error); }
      prs_echoChar('\n');
    }
    {
      int dimen[] = {DIMEN_MASS};
      skipWS();
      error = prs_readUnits(&baseMass,dimen);
      if(error) { if(error == NO_UNITS) ABORT(ERR_ON_LINE); ABORT(error); }
      prs_echoChar('\n');
    }
    {
      int dimen[] = {DIMEN_TIME};
      skipWS();
      error = prs_readUnits(&baseTime,dimen);
      if(error) { if(error == NO_UNITS) ABORT(ERR_ON_LINE); ABORT(error); }
      prs_echoChar('\n');
    }
    {
      double zeroTemp = 0;

      skipWS();
      error = prs_readTempUnits(&zeroTemp,0);
      if(error) { if(error == NO_UNITS) ABORT(ERR_ON_LINE); ABORT(error); }
      prs_echoChar('\n');

      base = expr_baseTempIndex();
      if(base < 0) ABORT(ERR_IN_CODE);
      newBase = expr_convertedFromTempIndex();

      if(base == newBase ||
	 (base != INDEX_DEG_F && newBase != INDEX_DEG_F) ) {
	baseTemp = 1;
      } else {
	baseTemp = (base == INDEX_DEG_F) ? 5/9 : 1.8;
      }
    }
    error = expr_setBaseUnits(baseLength, baseMass, baseTime,
			      baseTemp, 1, newBase);
    if(error) ABORT(ERR_UNITS);
  }

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case -1:
      fprintf(stderr,"Code compiled for chemistry option %d or %d\n",
	      CHEM_VERSION,CHEM_SPECIALIZED);
      error = ERR_ON_LINE;
      break;
    }
  }
  return error;
}
