#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "read.h"

#include "driverGlobals.h"

/******************************************************************************
Title: pre_readICS
Description: Define storage for some initial conditions data.
Return: Error code (see reader.h).
*/
int pre_readICS() {
  int error = 0;
  int i;

  int flowICs = !computeFlow && computeTransport;

  pressure = (double*) malloc(nx*ny*nz*sizeof(double));
  velX = (double*) malloc((nx+1)*ny*nz*sizeof(double));
  velY = (double*) malloc(nx*(ny+1)*nz*sizeof(double));
  velZ = (double*) malloc(nx*ny*(nz+1)*sizeof(double));
  if(pressure == NULL || velX == NULL
     || velY == NULL || velZ == NULL) ABORT(ERR_MEMORY);
  for(i=0; i<nx*ny*nz; i++) pressure[i] = 0;
  if(!flowICs) {
    for(i=0; i<(nx+1)*ny*nz; i++) velX[i] = 0;
    for(i=0; i<nx*(ny+1)*nz; i++) velY[i] = 0;
    for(i=0; i<nx*ny*(nz+1); i++) velZ[i] = 0;
  }

exit:
  return error;
}

/******************************************************************************
Title: readICS
Description: Read and preprocess the initial conditions input data.
Return: Error code (see reader.h).
*/
int readICS() {
  static char INITIAL_CONDITIONS[] = {"INITIAL CONDITIONS"};
  static char Flow[] = {"Flow"};
  static char Transport[] = {"Transport"};

  int error = 0;
  int i,s;
  int nCells_t = (nx+2*(padx+1))*(ny+2*(pady+1))*(nz+2*(padz+1));
  char string[21];
  int nearly;

  int flowICs = !computeFlow && computeTransport;

  skipToSection(INITIAL_CONDITIONS);

  if(flowICs) {
    int dimen[] = {DIMEN_VELOCITY};
    readSubSection(0);
    skipToSubSection(Flow);
    nearly = 0;
    error = setup3DDouble(velX, dimen, 1,0,0, 1, 0,0,0, string, 21, &nearly);
    if(error) goto exit;
    
    nearly = 0;
    error = setup3DDouble(velY, dimen, 0,1,0, 1, 0,0,0, string, 21, &nearly);
    if(error) goto exit;
    
    nearly = 0;
    error = setup3DDouble(velZ, dimen, 0,0,1, 1, 0,0,0, string, 21, &nearly);
    if(error) goto exit;
  }

  conc = (float*) malloc(nCells_t*nSpecies*sizeof(float));
  if(conc == NULL) ABORT(ERR_MEMORY);
  for(i=0; i<nCells_t*nSpecies; i++) conc[i] = 0;
    
  readSubSection(0);
  skipToSubSection(Transport);
  beginList();
  {
    int dimen[] = {DIMEN_MOLAR_CONC};
    for(s=0; s<nSpecies; s++) {
      nearly = 0;
      error = setup3DFloat(&conc[nCells_t*s], dimen,
			   0,0,0, 1, padx+1,pady+1,padz+1,
			   string, 21, &nearly);
      if(error) goto exit;
    }
  }
  endList();

 exit:
  return error;
}
