#include <stdlib.h>
#include <stdio.h>
#include "read.h"
#include "msgs.h"
#include "ftn.h"
#include "debug.h"
#include <string.h>

#include "driverGlobals.h"
#include "typeNumbers.h"

/******************************************************************************
Title: readOutput
Description: Read and preprocess the output control input data.
Return: Error code (see reader.h).
*/
int readOutput() {
  static char OUTPUT[] = {"OUTPUT"};
  static char Flow[] = {"Flow"};
  static char Transport[] = {"Transport"};
  static char History[] = {"History"};

  int error = 0;
  char string[21];
  int m,i=0,hist=0;

  skipToSection(OUTPUT);

  packInt(runNumber);
  readLine(runDescription,SIZE_LINE);
  packInt(verbosity);
  packInt(debug);
  readString(outDir,SIZE_DIRECTORYNAME);

  packInt(outFormat3D);
  unPackInt();
  if(debug >= DEBUG_MISC) {
    openDebug(1,myProc);
  } else {
    openDebug(0,myProc);
  }
  switch(outFormat3D) {
  case OUT_3D_FORMAT_EYE:
  case OUT_3D_FORMAT_RAW:
  case OUT_3D_FORMAT_TECPLOT:
    break;
  default:
    ABORT(ERR_UNSUPPORTED_OPTION);
  }
  packInt(nSpeciesPerOutfile);
  packInt(initialOut);

  packInt(outFormat1D);
  unPackInt();
  if(nSpeciesPerOutfile < 1) nSpeciesPerOutfile = 1;
  switch(outFormat1D) {
  case OUT_1D_FORMAT_TECPLOT:
    break;
  default:
    ABORT(ERR_UNSUPPORTED_OPTION);
  }

  if(computeFlow) {
    readSubSection(0);
    skipToSubSection(Flow);
    packInt(dStepOut_pres);
    packInt(dStepOut_vel);
    packDoubleUnits(dtOut_pres,DIMEN_TIME);
    packDoubleUnits(dtOut_vel,DIMEN_TIME);

    readString(string,21);
    if(STRING_IS("potential")) {
      presType_out = PRES_TYPE_POTENTIAL;
    } else if(STRING_IS("pressure")) {
      presType_out = PRES_TYPE_PRESSURE;
    } else if(STRING_IS("hydraulicHead")) {
      presType_out = PRES_TYPE_HYDRAULIC_HEAD;
    } else if(STRING_IS("head")) {
      presType_out = PRES_TYPE_HEAD;
    } else ABORT(-1);

    for(i=0; i<2; i++) {
      packInt(outFlags_f[i]);
    }
  }

  if(callTransport) {
    readSubSection(0);
    skipToSubSection(Transport);
    packInt(dStepOut_conc);
    packDoubleUnits(dtOut_conc,DIMEN_TIME);
    packInt(nHistories_t);
    for(i=0; i<9; i++) {
      packInt(outFlags_t[i]);
    }

    unPackInt();
    if(nHistories_t > 0) {
      histType_t = (int *) malloc(nHistories_t*sizeof(int));
      histIndexLo_t = (int *) malloc(3*nHistories_t*sizeof(int));
      histIndexHi_t = (int *) malloc(3*nHistories_t*sizeof(int));
      histVolFactor_t = (double *) malloc(nHistories_t*sizeof(double));
      nHistSpecies_t = (int*) malloc(nHistories_t*sizeof(int));
      histSpecie_t = (int**) malloc(nHistories_t*sizeof(int*));
      dStepOut_hist_t = (int*) malloc(nHistories_t*sizeof(int));
      dtOut_hist_t = (double*) malloc(nHistories_t*sizeof(double));
      if(histType_t == NULL ||
	 histIndexLo_t == NULL || histIndexHi_t == NULL ||
	 histVolFactor_t == NULL || nHistSpecies_t == NULL ||
	 histSpecie_t == NULL || dStepOut_hist_t == NULL ||
	 dtOut_hist_t == NULL) ABORT(ERR_MEMORY);

      for(hist=0; hist<nHistories_t; hist++) {
	readSubSubSection(0);
	skipToSubSubSection(History);
	packInt(histType_t[hist]);
	packInt(histIndexLo_t_(0,hist));
	packInt(histIndexLo_t_(1,hist));
	packInt(histIndexLo_t_(2,hist));
	packInt(histIndexHi_t_(0,hist));
	packInt(histIndexHi_t_(1,hist));
	packInt(histIndexHi_t_(2,hist));
	packInt(nHistSpecies_t[hist]);

	unPackInt();
        switch(histType_t[hist]) {
	case HIST_MOLES_PHASE:
	case HIST_MOLES_TOTAL:
	  histVolFactor_t[hist] = 1;
	  break;
	case HIST_CONC_PHASE:
	case HIST_CONC_TOTAL:
	  histVolFactor_t[hist] = 0;
	  break;
	default:
	  ABORT(-4);
	}

	if(histIndexLo_t_(0,hist)<1 || histIndexLo_t_(0,hist)>nx_g) ABORT(-2);
	if(histIndexLo_t_(1,hist)<1 || histIndexLo_t_(1,hist)>ny_g) ABORT(-2);
	if(histIndexLo_t_(2,hist)<1 || histIndexLo_t_(2,hist)>nz_g) ABORT(-2);
	if(histIndexHi_t_(0,hist)<1 || histIndexHi_t_(0,hist)>nx_g) ABORT(-2);
	if(histIndexHi_t_(1,hist)<1 || histIndexHi_t_(1,hist)>ny_g) ABORT(-2);
	if(histIndexHi_t_(2,hist)<1 || histIndexHi_t_(2,hist)>nz_g) ABORT(-2);

	if(nHistSpecies_t[hist] <= 0) ABORT(ERR_ON_LINE);
	histSpecie_t[hist] = (int*) malloc(nHistSpecies_t[hist]*sizeof(int));
	if(histSpecie_t[hist] == NULL) ABORT(ERR_MEMORY);

	beginList();
	for(i=0; i<nHistSpecies_t[hist]; i++)
	  packInt(histSpecie_t_(i,hist));
	endList();

	unPackInt();
	for(i=0; i<nHistSpecies_t[hist]; i++)
	  if(histSpecie_t_(i,hist) < 1 || histSpecie_t_(i,hist) > nSpecies)
	    ABORT(-3);

	packInt(dStepOut_hist_t[hist]);
	packDoubleUnits(dtOut_hist_t[hist],DIMEN_TIME);

	/* Localize */
	for(m=0; m<3; m++) {
	  if(histIndexLo_t_(m,hist) > histIndexHi_t_(m,hist)) {
	    int ind = histIndexLo_t_(m,hist);
	    histIndexLo_t_(m,hist) = histIndexHi_t_(m,hist);
	    histIndexHi_t_(m,hist) = ind;
	  }
	}
	histIndexLo_t_(0,hist) = histIndexLo_t_(0,hist) - xOffset_g_(0);
	histIndexLo_t_(1,hist) = histIndexLo_t_(1,hist) - yOffset_g_(0);
	histIndexLo_t_(2,hist) = histIndexLo_t_(2,hist) - zOffset_g_(0);
	histIndexHi_t_(0,hist) = histIndexHi_t_(0,hist) - xOffset_g_(0);
	histIndexHi_t_(1,hist) = histIndexHi_t_(1,hist) - yOffset_g_(0);
	histIndexHi_t_(2,hist) = histIndexHi_t_(2,hist) - zOffset_g_(0);
      }
    }
  }

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case -1:
      fprintf(stderr,"Unsupported pressure type\n");
      error = ERR_ON_LINE;
      break;
    case -2:
      fprintf(stderr,
	      "History %d from (%d,%d,%d) to (%d,%d,%d) is outside domain\n",
	      hist+1,histIndexLo_t_(0,hist),histIndexLo_t_(1,hist),
	      histIndexLo_t_(2,hist),histIndexHi_t_(0,hist),
	      histIndexHi_t_(1,hist),histIndexHi_t_(2,hist));
      error = ERR_BEFORE_LINE;
      break;
    case -3:
      fprintf(stderr,"History %d specie %d is out of range\n",
	      hist+1,histSpecie_t_(i,hist));
      error = ERR_BEFORE_LINE;
      break;
    case -4:
      fprintf(stderr,"History %d has unsupported type\n",hist+1);
      error = ERR_BEFORE_LINE;
      break;
    }
  }
  return error;
}
