#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "read.h"

#include "driverGlobals.h"

#define STRING_IS(str) !strcmp(string,str)

/*****************************************************************************
Title: blankPad
Description: pad with `size' blanks, and then append '\0'
Return: none
*/
void blankPad(char* string, int size) {
  int i;

  for (i = 0 ; i < size; i++) string[i] = ' ';
  string[size] = '\0';
}

/******************************************************************************
Title: readPhase
Description: Read and preprocess the fluid properties input data.
Return: Error code (see reader.h).
*/
int readPhase() {
  static char PHASE_PROPERTIES[] = {"PHASE PROPERTIES"};
  static char Phase[] = {"Phase"};

  int error = 0;
  int phase,len;
  char string[21];

  skipToSection(PHASE_PROPERTIES);

  packDoubleUnits(phaseViscosity,DIMEN_VISCOSITY);
  packDoubleUnits(phaseDensity,DIMEN_MASS_CONC);
  packInt(chemPhases);

  if(computeChemistry) {
    unPackInt();
    if(chemPhases <= 0) ABORT(-2);

    phaseNames = (char *) malloc((chemPhases*PhaseNameSZ+1)*sizeof(char));
    if(phaseNames == NULL) return ERR_MEMORY;

    chmPhaseDensity = (double *) malloc(chemPhases*sizeof(double));
    if(chmPhaseDensity == NULL) return ERR_MEMORY;

    for(phase = 0; phase < chemPhases; phase++) {
      readSubSection(0);
      skipToSubSection(Phase);
      readString(&phaseNames_(0,phase),PhaseNameSZ+1);
      len = strlen(&phaseNames_(0,phase));
      blankPad(&phaseNames_(len,phase),PhaseNameSZ-len);
      readString(string,21);
      if(STRING_IS("multi-species")) {
	/* phaseType = 0; */
	packDoubleUnits(chmPhaseDensity[phase],DIMEN_MOLAR_CONC);
      } else if(STRING_IS("single-species")) {
	/* phaseType = 1; */
	chmPhaseDensity[phase] = 1.0;
      } else ABORT(-1);
    }
  }
 
 exit:
  if(myProc == 0) {
    switch(error) {
    case -1:
      fprintf(stderr,"Unknown phase type (multi-species or single-species)\n");
      error = ERR_ON_LINE;
      break;
    case -2:
      fprintf(stderr,"Need at least one phase\n");
      error = ERR_ON_LINE;
      break;
    }
  }
  return error;
}
