#include <stdlib.h>
#include <stdio.h>
#include "read.h"

#include "driverGlobals.h"

/******************************************************************************
Title: readRXN
Description: Read and preprocess the reaction (rxn, e.g., biodegradation) data.
Return: Error code (see reader.h).
*/
int readRXN() {
  static char SPECIALIZED_REACTIONS[] = {"SPECIALIZED REACTIONS"};

  int i,j;
  int error = 0;

  skipToSection(SPECIALIZED_REACTIONS);

  packInt(rxn_nComps);
  unPackInt();
  if(rxn_nComps > 0) {
    rxn_comp = (int*) malloc(rxn_nComps*sizeof(int));
    if(rxn_comp == NULL) ABORT(ERR_MEMORY);
  } else {
    goto exit;
  }
  beginList();
  for(i=0; i<rxn_nComps; i++) {
    packInt(rxn_comp[i]);
  }
  endList();
  unPackInt();
  for(i=0; i<rxn_nComps; i++) {
    rxn_comp[i] -= 1; /* Decrement to change from Fortran to C indexing */
    if(rxn_comp[i] < 0 || rxn_comp[i] >= nSpecies) ABORT(-1);
  }
  for(i=0; i<rxn_nComps; i++) {
    for(j=i+1; j<rxn_nComps; j++) {
      if(rxn_comp[i] == rxn_comp[j]) ABORT(-2);
    }
  }

  packInt(rxn_nMicroSteps);
  packDoubleUnits(rxn_tolerance,DIMEN_NUMBER);

  packInt(rxn_nDParams);
  unPackInt();
  if(rxn_nDParams > 0) {
    rxn_dParam = (double*) malloc(rxn_nDParams*sizeof(double));
    if(rxn_dParam == NULL) ABORT(ERR_MEMORY);
  }
  beginList();
  for(i=0; i<rxn_nDParams; i++) {
    packDoubleUnits(rxn_dParam[i],DIMEN_NUMBER);
  }
  endList();

  packInt(rxn_nIParams);
  unPackInt();
  if(rxn_nIParams > 0) {
    rxn_iParam = (int*) malloc(rxn_nIParams*sizeof(int));
    if(rxn_iParam == NULL) ABORT(ERR_MEMORY);
  }
  beginList();
  for(i=0; i<rxn_nIParams; i++) {
    packInt(rxn_iParam[i]);
  }
  endList();

  unPackInt();
  unPackDouble();

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case -1:
      fprintf(stderr,"Species number out of range\n");
      error = ERR_ON_LINE;
      break;
    case -2:
      fprintf(stderr,"Species numbers duplicated\n");
      error = ERR_ON_LINE;
      break;
    }
  }
  return error;
}
