#include <stdlib.h>
#include <stdio.h>
#include "read.h"

#include "driverGlobals.h"

/******************************************************************************
Title: readTime
Description: Read and preprocess the time input data.
Return: Error code (see reader.h).
*/

int readTime() {
  static char TIME[] = {"TIME"};
  static char Flow[] = {"Flow"};
  static char Transport[] = {"Transport"};

  int error = 0;
  int i;

  skipToSection(TIME);

  packDoubleUnits(tInitial,DIMEN_TIME);
  packDoubleUnits(tFinal,DIMEN_TIME);

  if(computeFlow) {
    readSubSection(0);
    skipToSubSection(Flow);
    packInt(nStepsMax_f);
    packInt(ndtMax_f);
    unPackInt();
    if(ndtMax_f < 1) ABORT(ERR_ON_LINE);
    tdtMax_f = (double*) malloc(ndtMax_f*sizeof(double));
    dtMax_f = (double*) malloc(ndtMax_f*sizeof(double));
    if(tdtMax_f == NULL || dtMax_f == NULL) ABORT(ERR_MEMORY);
    beginList();
    for(i=0; i<ndtMax_f; i++) {
      packDoubleUnits(tdtMax_f[i],DIMEN_TIME);
      packDoubleUnits(dtMax_f[i],DIMEN_TIME);
    }
    endList();
  }

  if(callTransport) {
    readSubSection(0);
    skipToSubSection(Transport);
    packInt(nStepsMax_t);
    packInt(nStepsMax_t_per_f);
    packInt(ndtMax_t);
    unPackInt();
    if(ndtMax_t < 1) ABORT(ERR_ON_LINE);
    tdtMax_t = (double*) malloc(ndtMax_t*sizeof(double));
    dtMax_t = (double*) malloc(ndtMax_t*sizeof(double));
    if(tdtMax_t == NULL || dtMax_t == NULL) ABORT(ERR_MEMORY);
    beginList();
    for(i=0; i<ndtMax_t; i++) {
      packDoubleUnits(tdtMax_t[i],DIMEN_TIME);
      packDoubleUnits(dtMax_t[i],DIMEN_TIME);
    }
    endList();
  }

  unPackDouble();
  if(computeFlow) {
    if(tdtMax_f[0] > tInitial) tdtMax_f[0] = tInitial;
    for(i=1; i<ndtMax_f; i++) { if(tdtMax_f[i] < tdtMax_f[i-1]) ABORT(-1); }
    for(i=0; i<ndtMax_f; i++) { if(dtMax_f[i] <= 0) ABORT(-1); }
  }
  if(callTransport) {
    if(tdtMax_t[0] > tInitial) tdtMax_t[0] = tInitial;
    for(i=1; i<ndtMax_t; i++) { if(tdtMax_t[i] < tdtMax_t[i-1]) ABORT(-1); }
    for(i=0; i<ndtMax_t; i++) { if(dtMax_t[i] <= 0) ABORT(-1); }
  }

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case -1:
      fprintf(stderr,"Inconsistent time data");
      error = ERR_BEFORE_LINE;
      break;
    }
  }
  return error;
}
