#ifndef READER_INCLUDED
#define READER_INCLUDED

#ifndef NULL
# define NULL 0
#endif

#include "dimen.h"

/* DEFINE ERROR AND RETURN CODES */
#define ERR_ON_LINE                     1
#define ERR_BEFORE_LINE                 2
#define ERR_MEMORY                      3
#define ERR_OPENING_FILE                4
#define ERR_NESTING_FILE                5
#define ERR_EOF                         6
#define ERR_UNKNOWN_CMD                 7
#define ERR_UNMATCHED_END_COMMENT_CMD   8
#define ERR_IN_CODE                     9
#define ERR_UNSUPPORTED_OPTION         10
#define ERR_IN_DATA_BLOCK              11
#define ERR_IN_GRID_ARRAY              12
#define ERR_KEY_NOT_READ               13
#define ERR_BEGIN_LIST_NOT_READ        14
#define ERR_END_LIST_NOT_READ          15
#define ERR_SECTION_READ               16
#define ERR_SUBSECTION_READ            17
#define ERR_SUBSUBSECTION_READ         18
#define ERR_SECTION_NOT_READ           19
#define ERR_SUBSECTION_NOT_READ        20
#define ERR_SUBSUBSECTION_NOT_READ     21
#define ERR_SECTION_KEY_NOT_READ       22
#define ERR_SUBSECTION_KEY_NOT_READ    23
#define ERR_SUBSUBSECTION_KEY_NOT_READ 24
#define READ_LITERAL_CMD               25

#define NO_UNITS                   100
#define ERR_UNITS_NEGATIVE         101
#define ERR_NONNESTED_PARENTHESES  102
#define ERR_ARITHMETIC_OP          103
#define ERR_UNITS_ADDITION_FOUND   104
#define ERR_UNITS_WRONG_DIMEN      105
#define ERR_UNITS_TOO_LONG         106
#define ERR_UNITS_IMPROPER_EXPR    107
#define ERR_UNKNOWN_UNIT           108
#define ERR_UNITS                  109

/* DECLARE FUNCTIONS */
int prs_openReader(char* inFileName0, char* echoFileName,
		   int length_fileName, int echo);
void prs_closeReader(int error);
void prs_processReaderError(int error);

int prs_skipWS(int stop_after_EOL);
void prs_skipWord(int stopNonPWS, int stopListSymbols, int echoWord);
int prs_getc();
void prs_ungetc(int c);
void prs_echoString(char* str);
void prs_echoChar(int c);

void prs_beginSkip();
void prs_endSkip();

int prs_readKey(char* key, int skip_echoWord);
int prs_readKeys(char** keys, int nKeys, int* keyRead, int skip_echoWord);

int prs_readSection(int skip);
int prs_readSubSection(int skip);
int prs_readSubSubSection(int skip);

int prs_skipToSection(char* key);
int prs_skipToSubSection(char* key);
int prs_skipToSubSubSection(char* key);
int prs_skipToSections(char** keys, int nKeys, int* keyRead);
int prs_skipToSubSections(char** keys, int nKeys, int* keyRead);
int prs_skipToSubSubSections(char** keys, int nKeys, int* keyRead);

int prs_beginList();
int prs_endList();
int prs_skipToEndList();

int prs_readUnits(double* value, int* dimen);
int prs_readTempUnits(double* temp, int echoConversion);

int prs_packInt(int* var);
int prs_packDouble(double* var);
int prs_packDoubleUnits(double* var, int* dimen);
int prs_packDoubleTemp(double* var);
int prs_packFloat(float* var);
int prs_packFloatUnits(float* var, int* dimen);
int prs_packFloatTemp(float* var);

void prs_unPackInt();
void prs_unPackDouble();
void prs_unPackFloat();

int prs_readString(char* str, int len);
int prs_readLine(char* str, int len);
int prs_readIntArray(int* array, int len, int* nRead, int nToReadTotal);
int prs_readDoubleArray(double* array, int len, int* nRead, int nToReadTotal);
int prs_readDoubleArrayUnits(double* array, int len, int* nRead,
			     int nToReadTotal, double units);
int prs_readFloatArray(float* array, int len, int* nRead, int nToReadTotal);
int prs_readFloatArrayUnits(float* array, int len, int* nRead,
			    int nToReadTotal, double units);

#endif
