#ifndef _TYPENUMBERS_H
#define _TYPENUMBERS_H

/* FLOW */
/* The following must match defines.hf in Flow */
#define PC_IF_NOPRECOND_F 0
#define PC_IF_DIAGONAL_F 1
#define PC_IF_BALANCESTYLE_F 5

#define PERM_SCALAR 1
#define PERM_DIAGONAL 2
#define PERM_SYMMETRIC 3
#define PERM_FACE 4

/* ADVECTION OPTIONS */
#define ADV_DIFF_ONLY -1
#define ADV_CMM 1
#define ADV_HOG 2
#define ADV_GODUNOV 3

/* CHEMISTRY */
#define CHEM_RNSF 1
#define CHEM_SF 2
#define CHEM_UNSF 3
#define CHEM_SPECIALIZED -1

#define CHEM_RXN_EQ 0
#define CHEM_RXN_K_MASS_ACTION 1
#define CHEM_RXN_K_MONOD 2

#define CHEM_INIT_GUESS_AUTO -1
#define CHEM_INIT_GUESS_TRAN 0
#define CHEM_INIT_GUESS_PREV 1

#if CHEM==RNSF
# define CHEM_VERSION CHEM_RNSF
#elif CHEM==SF
# define CHEM_VERSION CHEM_SF
#elif CHEM==UNSF
# define CHEM_VERSION CHEM_UNSF
#else
# define CHEM_VERSION 0
#endif

/* MISCIBLE DISPLACEMENT */
#define UMD_QUARTER_POWER 1

/* BOUNDARY CONDITIONS */
/* The following must match BCTypes.hf in Flow */
#define BC_NEUMANN_F   1
#define BC_DIRICHLET_F 2
#define BC_NOFLOW_F    3
#define BC_ROBIN_F     4

#define BC_NEUMANN_F_INPUT   1
#define BC_DIRICHLET_F_INPUT 2
#define BC_NOFLOW_F_INPUT    0

/* The following must match bctypes.hf in Transport */
#define BC_INTERNAL_T -1
#define BC_NOFLOW_T    0
#define BC_FLOW_T      1
#define BC_DIRICHLET_T 2

/* WELLS */
/******************************************************************************
Well types
  INACTIVE:        same as no well
  INJECTION:       specified positive injection rate
  EXTRACTION:      (or Production) specified negative injection rate
  BOTTOMHOLE_PROD: bottomhole production well
  REINJECTION:     (or Sparging) specified injection rate, two piece well
                   one piece is an EXTRACTION well, the other is an
                   INJECTION well with injection fluid equal to that
                   produced by the rest of the well (with perhaps
                   concentrations topped off)
******************************************************************************/
/* The following must match well.hf */
#define WELL_INACTIVE 0
#define WELL_INJECTION 1
#define WELL_EXTRACTION 2
#define WELL_BOTTOMHOLE_PROD 3
#define WELL_REINJECTION 4

/* OUTPUT */
#define OUT_3D_FORMAT_EYE    -1
#define OUT_3D_FORMAT_RAW     0
#define OUT_3D_FORMAT_TECPLOT 1

#define OUT_1D_FORMAT_TECPLOT 1

#define HIST_CONC_PHASE    1
#define HIST_CONC_TOTAL 2
#define HIST_MOLES_PHASE    3
#define HIST_MOLES_TOTAL 4

/* DEBUGGING LEVELS */
#define DEBUG_PARALLEL 1
#define DEBUG_MISC     2
#define DEBUG_INPUT    3
#define DEBUG_SETUP    4

/* MISCELLANEOUS */
#define PRES_TYPE_POTENTIAL      0
#define PRES_TYPE_PRESSURE       1
#define PRES_TYPE_HYDRAULIC_HEAD 2
#define PRES_TYPE_HEAD           3

#define PERM_TYPE_PERMEABILITY 0
#define PERM_TYPE_CONDUCTIVITY 1

#ifndef DEFINE_WRITETYPEINFO
void writeTypeInfo();
#else
void writeTypeInfo() {
  printf("FLOW\n");
  printf("Interface preconditioner:\n");
  printf("  %2d = None\n",PC_IF_NOPRECOND_F);
  printf("  %2d = Diagonal\n",PC_IF_DIAGONAL_F);
  printf("  %2d = Balancing\n",PC_IF_BALANCESTYLE_F);

  printf("\nTRANSPORT\n");
  printf("  %2d = None\n",0);
  printf("  %2d = Characteristics Mixed Method\n",ADV_CMM);
  printf("  %2d = Higher order Godunov Method\n",ADV_HOG);
  printf("  %2d = Low order Godunov Method\n",ADV_GODUNOV);
  printf("  %2d = Diffusion (and dispersion) only\n",ADV_DIFF_ONLY);

  printf("\nCHEMISTRY\n");
  printf("  Chemistry package option\n");
  printf("    %2d = None\n",0);
  switch(CHEM_VERSION) {
  case CHEM_RNSF:
    printf("    %2d = RNSF\n",CHEM_RNSF);
    break;
  case CHEM_SF:
    printf("    %2d = SF\n",CHEM_SF);
    break;
  case CHEM_UNSF:
    printf("    %2d = UNSF\n",CHEM_UNSF); 
    break;
  default:
    printf("ERROR IN CODE\n");
  }
  printf("    %2d = specialized\n",CHEM_SPECIALIZED);
  printf("    Note that the code can be compiled for general options:\n");
  printf("      %2d = RNSF\n",CHEM_RNSF);
  printf("      %2d = SF\n",CHEM_SF);
  printf("      %2d = UNSF\n",CHEM_UNSF);
  printf("  General chemistry reaction types\n");
  printf("    %2d = Equilibrium\n",CHEM_RXN_EQ);
  printf("    %2d = Mass action kinetic\n",CHEM_RXN_K_MASS_ACTION);
  printf("    %2d = Monod kinetic\n",CHEM_RXN_K_MONOD);
  printf("  Chemistry initial guess for equilibrium\n");
  printf("    %2d = Transported\n",CHEM_INIT_GUESS_TRAN);
  printf("    %2d = Previous\n",CHEM_INIT_GUESS_PREV);
  printf("    %2d = Auto\n",CHEM_INIT_GUESS_AUTO);

  printf("\nMISCIBLE DISPLACEMENT\n");
  printf("  %2d = None\n",0);
  printf("  %2d = Quarter power mixing rule\n",UMD_QUARTER_POWER);

  printf("\nBOUNDARY CONDITIONS\n");
  printf("Flow\n");
  printf("  %2d = Noflow\n",BC_NOFLOW_F_INPUT);
  printf("  %2d = Neumann\n",BC_NEUMANN_F_INPUT);
  printf("  %2d = Dirichlet\n",BC_DIRICHLET_F_INPUT);
  printf("Transport\n");
  printf("  %2d = Noflow/outflow\n",BC_NOFLOW_T);
  printf("  %2d = Inflow\n",BC_FLOW_T);
  printf("  %2d = Dirichlet\n",BC_DIRICHLET_T);

  printf("\nWELLS\n");
  printf("  %2d = Inactive\n",WELL_INACTIVE);
  printf("  %2d = Injection\n",WELL_INJECTION);
  printf("  %2d = Extraction\n",WELL_EXTRACTION);
  printf("  %2d = Bottomhole production\n",WELL_BOTTOMHOLE_PROD);
  printf("  %2d = Reinjection/Sparging\n",WELL_REINJECTION);

  printf("\nHISTORY TYPES\n");
  printf("  %2d = Pore volume average of phase concentration\n",
	 HIST_CONC_PHASE);
  printf("  %2d = Pore volume average of total moles\n",HIST_CONC_TOTAL);
  printf("  %2d = Total phase moles\n",HIST_MOLES_PHASE);
  printf("  %2d = Total moles\n",HIST_MOLES_TOTAL);

  printf("\nMISCELLANEOUS\n");
  printf("Output Formats for 3-D\n");
  printf("  %2d = Eye\n",OUT_3D_FORMAT_EYE);
  printf("  %2d = Raw data\n",OUT_3D_FORMAT_RAW);
  printf("  %2d = Tecplot\n",OUT_3D_FORMAT_TECPLOT);
  printf("Output Formats for 1-D\n");
  printf("  %2d = Tecplot\n",OUT_1D_FORMAT_TECPLOT);
  printf("Debug Level\n");
  printf("  %2d = Debug the parallel set-up\n",DEBUG_PARALLEL);
  printf("  %2d = Debug miscellaneous items\n",DEBUG_MISC);
  printf("  %2d = Debug (dump) the input data variables\n",DEBUG_INPUT);
  printf("  %2d = Debug (dump) the internal global variables\n",DEBUG_SETUP);
}

#endif

#endif
