c=======================================================================
c  Set up the 19-diagonal matrix for the diffusion operator only
c
c ... In:
c     n?       : number of cells in each direction
c     d??      : dispersion tensor at the grid points
c     h?       : mesh widths h?(i) = x?(i)-x?(i-1), i = 1,n?
c                h?(0) = h?(n?+1) = 0. on external boundary
c     byd?     : byd?(i) = 1./(h?(i) + h?(i+1)), i = 0,n?
c     bcRegion?: boundary condition regions
c     bcType   : boundary condition flags
c                see "bctypes.hf"
c                bcType = CBCINTERNAL  : internal boundary
c                bcType = CBCNOFLOW    : noflow (outflow) (z.n = 0)
c                bcType = CBCFLOW      : inflow (Robin) (z.n + c*u.n = cIN*u.n)
c                bcType = CBCDIRICHLET : Dirichlet (c = cD)
c     bcFlux?  : normal component of the velocity on the boundary
c                (these are stored in face?0 to save storage, since
c                face?? is used only for CBCDIRICHLET and bcFlux? is
c                used only for CBCFLOW if the flow is into the domain)
c
c ... Out:
c     a        : cell-centered finite difference matrix
c                unknowns: p, lambda's on external faces,
c                phantom lambda's on external edges (to simplify
c                data structure)
c                matrix structure:
c                  band  0 = diag (i,j,k)
c                  band -1 = connection to (i-1,j,k)
c                  band -2 = connection to (i+1,j-1,k)
c                  band -3 = connection to (i,j-1,k)
c                  band -4 = connection to (i-1,j-1,k)
c                  band -5 = connection to (i,j+1,k-1)
c                  band -6 = connection to (i+1,j,k-1)
c                  band -7 = connection to (i,j,k-1)
c                  band -8 = connection to (i-1,j,k-1)
c                  band -9 = connection to (i,j-1,k-1)
c
c     cdmatrix : diagonal of a
c     face??   : coefficients of a on the faces; stored for computing
c                the action of a on given boundary concentrations and
c                adding to the right-hand side
c
c ... Comments
c 
c     The lowest order RTN spaces with quadrature are used.
c
      subroutine csetupmat (a,n1,n2,n3,d11,d12,d13,d22,d23,d33,
     $     h1,h2,h3,byd1,byd2,byd3,cdmatrix,face10,face11,
     $     face20,face21,face30,face31,bcRegion1,bcRegion2,bcRegion3,
     $     bcType,idebug,debugunit,mynbr)
c
      implicit none
      include 'bctypes.hf'
c
      integer n1,n2,n3,idebug,debugunit,mynbr(-1:1,-1:1,-1:1)
      real*8  a(0:n1+1,0:n2+1,0:n3+1,-9:0),cdmatrix(n1,n2,n3),
     $        d11(0:n1,0:n2,0:n3),d12(0:n1,0:n2,0:n3),
     $        d13(0:n1,0:n2,0:n3),d22(0:n1,0:n2,0:n3),
     $        d23(0:n1,0:n2,0:n3),d33(0:n1,0:n2,0:n3),
     $        h1(0:n1+1),h2(0:n2+1),h3(0:n3+1),
     $        byd1(0:n1),byd2(0:n2),byd3(0:n3),
     $        face10(n2,n3,5),face11(n2,n3,5),
     $        face20(n1,n3,5),face21(n1,n3,5),
     $        face30(n1,n2,5),face31(n1,n2,5)
      integer bcRegion1(n2,n3,0:1),bcRegion2(n1,n3,0:1),
     $        bcRegion3(n1,n2,0:1),bcType(-1:1)
c
      integer i,j,k,index
      real*8  diag,fac
c
c Inline function
c
      real*8 coeff,x,y,dd,z1,z2,z3
      coeff(x,y,dd,z1,z2,z3)=0.125d0*(x)*(y)*(dd)*(z1)*(z2)*(z3)
c
c-----------------------------------------------------------------------
ccc: check positivity of determinant of dispersion
c      do k=0,n3
c      do j=0,n2
c      do i=0,n1
c         fac = d11(i,j,k)*d22(i,j,k)*d33(i,j,k)
c     $        + d12(i,j,k)*d23(i,j,k)*d13(i,j,k)
c     $        + d13(i,j,k)*d23(i,j,k)*d12(i,j,k)
c     $        - d13(i,j,k)*d22(i,j,k)*d13(i,j,k)
c     $        - d12(i,j,k)*d12(i,j,k)*d33(i,j,k)
c     $        - d11(i,j,k)*d23(i,j,k)*d23(i,j,k)
c         if(fac.lt.0.) write(*,*) 'det:',mynbr(0,0,0),i,j,k,fac
c      enddo
c      enddo
c      enddo
ccc.
c Initialize matrix bands
c
      do 10 index=-9,0
      do 10 k=0,n3+1
      do 10 j=0,n2+1
      do 10 i=0,n1+1
         a(i,j,k,index) = 0.0d0
 10   continue
c
      do 20 k=1,n3+1
      do 20 j=1,n2+1
      do 20 i=1,n1
         a(i,j,k,-9) = 
     $        coeff(byd2(j-1),byd3(k-1),d23(i-1,j-1,k-1)+d23(i,j-1,k-1),
     $              h2(j-1),-h1(i),h3(k))
     $      + coeff(byd3(k-1),byd2(j-1),d23(i-1,j-1,k-1)+d23(i,j-1,k-1),
     $              h3(k-1),-h1(i),h2(j))
 20   continue
      do 30 k=1,n3+1
      do 30 j=1,n2
      do 30 i=1,n1+1
         a(i,j,k,-8) = 
     $        coeff(byd1(i-1),byd3(k-1),d13(i-1,j-1,k-1)+d13(i-1,j,k-1),
     $               h1(i-1),-h2(j),h3(k)) 
     $      + coeff(byd3(k-1),byd1(i-1),d13(i-1,j-1,k-1)+d13(i-1,j,k-1),
     $               h3(k-1),-h1(i),h2(j))
 30   continue
      do 40 k=1,n3+1
      do 40 j=1,n2
      do 40 i=0,n1
         a(i,j,k,-6) =
     $        coeff(byd1(i),byd3(k-1),d13(i,j-1,k-1)+d13(i,j,k-1),
     $              h1(i+1),h2(j),h3(k))
     $      - coeff(byd3(k-1),byd1(i),d13(i,j-1,k-1)+d13(i,j,k-1),
     $              h3(k-1),-h1(i),h2(j))
 40   continue
      do 50 k=1,n3+1
      do 50 j=0,n2
      do 50 i=1,n1
         a(i,j,k,-5) =
     $        coeff(byd2(j),byd3(k-1),d23(i-1,j,k-1)+d23(i,j,k-1),
     $              h2(j+1),h1(i),h3(k))
     $      - coeff(byd3(k-1),byd2(j),d23(i-1,j,k-1)+d23(i,j,k-1),
     $              h3(k-1),-h1(i),h2(j))
 50   continue
      do 60 k=1,n3
      do 60 j=1,n2+1
      do 60 i=1,n1+1
         a(i,j,k,-4) =
     $        coeff(byd1(i-1),byd2(j-1),d12(i-1,j-1,k-1)+d12(i-1,j-1,k),
     $              h1(i-1),-h2(j),h3(k))
     $      + coeff(byd2(j-1),byd1(i-1),d12(i-1,j-1,k-1)+d12(i-1,j-1,k),
     $              h2(j-1),-h1(i),h3(k))
 60   continue
      do 70 k=1,n3
      do 70 j=1,n2+1
      do 70 i=0,n1
         a(i,j,k,-2) =
     $        coeff(byd1(i),byd2(j-1),d12(i,j-1,k-1)+d12(i,j-1,k),
     $              h1(i+1),h2(j),h3(k))
     $      - coeff(byd2(j-1),byd1(i),d12(i,j-1,k-1)+d12(i,j-1,k),
     $              h2(j-1),-h1(i),h3(k))
 70   continue
c
      do 500 k=0,n3+1
      do 500 j=0,n2+1
      do 500 i=0,n1+1
         diag = 0.0d0
c
         if ( (k.ne.0).and.(i.ne.0).and.(i.ne.n1+1).
     $        and.(j.ne.0).and.(j.ne.n2+1) ) then
         fac =
     $        coeff(byd3(k-1),byd3(k-1),
     $              d33(i-1,j-1,k-1) + d33(i,j-1,k-1) 
     $              + d33(i-1,j,k-1) + d33(i,j,k-1),
     $              h3(k-1) + h3(k),-h1(i),h2(j))
     $      + coeff(byd1(i),byd3(k-1),d13(i,j-1,k-1)+d13(i,j,k-1),
     $              h1(i),h2(j),h3(k))
     $      + coeff(byd1(i-1),byd3(k-1),d13(i-1,j-1,k-1)+d13(i-1,j,k-1),
     $              h1(i),-h2(j),h3(k))
     $      + coeff(byd2(j),byd3(k-1),d23(i-1,j,k-1)+d23(i,j,k-1),
     $              h2(j),h1(i),h3(k))
     $      + coeff(byd2(j-1),byd3(k-1),d23(i-1,j-1,k-1)+d23(i,j-1,k-1),
     $              h2(j),-h1(i),h3(k))
         diag = diag + fac
c
         a(i,j,k,-7) = fac
     $      + coeff(byd3(k-1),byd1(i),d13(i,j-1,k-1)+d13(i,j,k-1),
     $              h3(k-1),-h1(i),h2(j))
     $      - coeff(byd3(k-1),byd1(i-1),d13(i-1,j-1,k-1)+d13(i-1,j,k-1),
     $              h3(k-1),-h1(i),h2(j))
     $      + coeff(byd3(k-1),byd2(j),d23(i-1,j,k-1)+d23(i,j,k-1),
     $              h3(k-1),-h1(i),h2(j))
     $      - coeff(byd3(k-1),byd2(j-1),d23(i-1,j-1,k-1)+d23(i,j-1,k-1),
     $              h3(k-1),-h1(i),h2(j))
         endif
c
c
         if ( (j.ne.0).and.(i.ne.0).and.(i.ne.n1+1).
     $        and.(k.ne.0).and.(k.ne.n3+1) ) then
         fac =
     $       coeff(byd2(j-1),byd2(j-1),
     $             d22(i-1,j-1,k-1) + d22(i,j-1,k-1) 
     $              + d22(i-1,j-1,k) + d22(i,j-1,k),
     $             h2(j-1) + h2(j),-h1(i),h3(k))
     $     + coeff(byd1(i),byd2(j-1),d12(i,j-1,k-1)+d12(i,j-1,k),
     $             h1(i),h2(j),h3(k))
     $     + coeff(byd1(i-1),byd2(j-1),d12(i-1,j-1,k-1)+d12(i-1,j-1,k),
     $             h1(i),-h2(j),h3(k))
     $     + coeff(byd3(k),byd2(j-1),d23(i-1,j-1,k)+d23(i,j-1,k),
     $             h3(k),h1(i),h2(j))
     $     + coeff(byd3(k-1),byd2(j-1),d23(i-1,j-1,k-1)+d23(i,j-1,k-1),
     $             h3(k),-h1(i),h2(j))
         diag = diag + fac
c
         a(i,j,k,-3) = fac 
     $      + coeff(byd2(j-1),byd1(i),d12(i,j-1,k-1)+d12(i,j-1,k),
     $              h2(j-1),-h1(i),h3(k))
     $      - coeff(byd2(j-1),byd1(i-1),d12(i-1,j-1,k-1)+d12(i-1,j-1,k),
     $              h2(j-1),-h1(i),h3(k))
     $      + coeff(byd2(j-1),byd3(k),d23(i-1,j-1,k)+d23(i,j-1,k),
     $              h2(j-1),-h1(i),h3(k))
     $      - coeff(byd2(j-1),byd3(k-1),d23(i-1,j-1,k-1)+d23(i,j-1,k-1),
     $              h2(j-1),-h1(i),h3(k))
         endif
c
c
         if ( (i.ne.0).and.(j.ne.0).and.(j.ne.n2+1).
     $        and.(k.ne.0).and.(k.ne.n3+1) ) then
         fac =
     $        coeff(byd1(i-1),byd1(i-1),
     $              d11(i-1,j-1,k-1) + d11(i-1,j-1,k) 
     $              + d11(i-1,j,k-1) + d11(i-1,j,k),
     $              h1(i-1) + h1(i),-h2(j),h3(k))
     $      + coeff(byd2(j),byd1(i-1),d12(i-1,j,k-1)+d12(i-1,j,k),
     $              h2(j),h1(i),h3(k))
     $      + coeff(byd2(j-1),byd1(i-1),d12(i-1,j-1,k-1)+d12(i-1,j-1,k),
     $              h2(j),-h1(i),h3(k))
     $      + coeff(byd3(k),byd1(i-1),d13(i-1,j-1,k)+d13(i-1,j,k),
     $              h3(k),h1(i),h2(j))
     $      + coeff(byd3(k-1),byd1(i-1),d13(i-1,j-1,k-1)+d13(i-1,j,k-1),
     $              h3(k),-h1(i),h2(j))
         diag = diag + fac
c
         a(i,j,k,-1) = fac 
     $      - coeff(byd1(i-1),byd2(j-1),d12(i-1,j-1,k-1)+d12(i-1,j-1,k),
     $              h1(i-1),-h2(j),h3(k))
     $      + coeff(byd1(i-1),byd2(j),d12(i-1,j,k-1)+d12(i-1,j,k),
     $              h1(i-1),-h2(j),h3(k))
     $      - coeff(byd1(i-1),byd3(k-1),d13(i-1,j-1,k-1)+d13(i-1,j,k-1),
     $              h1(i-1),-h2(j),h3(k))
     $      + coeff(byd1(i-1),byd3(k),d13(i-1,j-1,k)+d13(i-1,j,k),
     $              h1(i-1),-h2(j),h3(k))
cDEBUG:
c         if (a(i,j,k,-1).ne.0.) then
c            write(*,*)'-1:',i,j,k,a(i,j,k,-1)
c            write(*,*)'fac:',fac,d12(i-1,j-1,k-1),d12(i-1,j-1,k),
c     $           d12(i-1,j,k-1),d12(i-1,j,k),
c     $           d13(i-1,j-1,k-1),d13(i-1,j-1,k),
c     $           d13(i-1,j,k-1),d13(i-1,j,k),
c     $           d11(i-1,j-1,k-1) ,d11(i-1,j-1,k), 
c     $           d11(i-1,j,k-1) ,d11(i-1,j,k)
c         endif
         endif
c
c
         if ( (k.ne.0).and.(j.ne.0).and.(k.ne.n3+1).and.(j.ne.n2+1).
     $        and.(i.ne.n1+1) ) then
         fac =
     $       -coeff(byd1(i),byd1(i),
     $              d11(i,j-1,k-1) + d11(i,j-1,k) 
     $              + d11(i,j,k-1) + d11(i,j,k),
     $              h1(i) + h1(i+1),h2(j),h3(k))
     $      - coeff(byd1(i),byd2(j-1),d12(i,j-1,k-1)+d12(i,j-1,k),
     $              h1(i),-h2(j),h3(k))
     $      + coeff(byd1(i),byd2(j),d12(i,j,k-1)+d12(i,j,k),
     $              h1(i),-h2(j),h3(k))
     $      - coeff(byd1(i),byd3(k-1),d13(i,j-1,k-1)+d13(i,j,k-1),
     $              h1(i),-h2(j),h3(k))
     $      + coeff(byd1(i),byd3(k),d13(i,j-1,k)+d13(i,j,k),
     $              h1(i),-h2(j),h3(k))
         diag = diag + fac
         endif
c
         if ( (k.ne.0).and.(i.ne.0).and.(k.ne.n3+1).and.(i.ne.n1+1).
     $        and.(j.ne.n2+1) ) then
         fac =
     $       -coeff(byd2(j),byd2(j),
     $             d22(i-1,j,k-1) + d22(i,j,k-1) 
     $             + d22(i-1,j,k) + d22(i,j,k),
     $             h2(j) + h2(j+1),h1(i),h3(k))
     $      + coeff(byd2(j),byd1(i),d12(i,j,k-1)+d12(i,j,k),
     $              h2(j),-h1(i),h3(k))
     $      - coeff(byd2(j),byd1(i-1),d12(i-1,j,k-1)+d12(i-1,j,k),
     $              h2(j),-h1(i),h3(k))
     $      + coeff(byd2(j),byd3(k),d23(i-1,j,k)+d23(i,j,k),
     $              h2(j),-h1(i),h3(k))
     $      - coeff(byd2(j),byd3(k-1),d23(i-1,j,k-1)+d23(i,j,k-1),
     $              h2(j),-h1(i),h3(k))
         diag = diag + fac
         endif
c
         if ( (j.ne.0).and.(i.ne.0).and.(j.ne.n2+1).and.(i.ne.n1+1).
     $        and.(k.ne.n3+1) ) then
         fac =
     $        -coeff(byd3(k),byd3(k),
     $              d33(i-1,j-1,k) + d33(i,j-1,k) 
     $              + d33(i-1,j,k) + d33(i,j,k),
     $              h3(k) + h3(k+1),h1(i),h2(j))
     $      + coeff(byd3(k),byd1(i),d13(i,j-1,k)+d13(i,j,k),
     $              h3(k),-h1(i),h2(j))
     $      - coeff(byd3(k),byd1(i-1),d13(i-1,j-1,k)+d13(i-1,j,k),
     $              h3(k),-h1(i),h2(j))
     $      + coeff(byd3(k),byd2(j),d23(i-1,j,k)+d23(i,j,k),
     $              h3(k),-h1(i),h2(j))
     $      - coeff(byd3(k),byd2(j-1),d23(i-1,j-1,k)+d23(i,j-1,k),
     $              h3(k),-h1(i),h2(j))
         diag = diag + fac
         endif
c
         a(i,j,k,0) = -diag
 500  continue
c
ccc: check positivity of diagonal
c      do k=0,n3+1
c      do j=0,n2+1
c      do i=0,n1+1
c         if( (a(i,j,k,0).lt.0.).and.
c     $        ((i.ge.1).or.(mynbr(-1,0,0).eq.-1)).and.
c     $        ((i.le.n1).or.(mynbr(1,0,0).eq.-1)).and.
c     $        ((j.ge.1).or.(mynbr(0,-1,0).eq.-1)).and.
c     $        ((j.le.n2).or.(mynbr(0,1,0).eq.-1)).and.
c     $        ((k.ge.1).or.(mynbr(0,0,-1).eq.-1)).and.
c     $        ((k.le.n3).or.(mynbr(0,0,1).eq.-1)) )
c     $        write(*,*) 'negF:',mynbr(0,0,0),i,j,k,a(i,j,k,0)
c      enddo
c      enddo
c      enddo
ccc.
      do 510 k=1,n3
      do 510 j=1,n2
      do 510 i=1,n1
         cdmatrix(i,j,k) = a(i,j,k,0)
 510  continue
c
c-----------------------------------------------------------------------
c DIRICHLET bc's and symmetrization
c
c bottom face
c
      do j = 1,n2
      do i = 1,n1
         if (bcType(bcRegion3(i,j,0)) .eq. CBCDIRICHLET) then
            a(i,j,0,0) = 1.0d0*h1(i)*h2(j)
c
            face30(i,j,1) = a(i,j-1,1,-5)
            face30(i,j,2) = a(i-1,j,1,-6)
            face30(i,j,3) = a(i,j,1,-7)
            face30(i,j,4) = a(i+1,j,1,-8)
            face30(i,j,5) = a(i,j+1,1,-9)
c
            a(i,j-1,1,-5) = 0.0
            a(i-1,j,1,-6) = 0.0
            a(i,j,1,-7) = 0.0
            a(i+1,j,1,-8) = 0.0
            a(i,j+1,1,-9) = 0.0
         else if (bcType(bcRegion3(i,j,0)).eq.CBCFLOW) then
            if(face30(i,j,1).lt.0) a(i,j,0,0) = 
     $           a(i,j,0,0) - face30(i,j,1)*h1(i)*h2(j)
         endif
      enddo
      enddo
c-----------------------------------------------------------------------
c top face
c
      do j = 1,n2
      do i = 1,n1
         if (bcType(bcRegion3(i,j,1)) .eq. CBCDIRICHLET) then
            a(i,j,n3+1,0) = 1.0d0*h1(i)*h2(j)
c
            face31(i,j,1) = a(i,j,n3+1,-5)
            face31(i,j,2) = a(i,j,n3+1,-6)
            face31(i,j,3) = a(i,j,n3+1,-7)
            face31(i,j,4) = a(i,j,n3+1,-8)
            face31(i,j,5) = a(i,j,n3+1,-9)
c
            a(i,j,n3+1,-5) = 0.0
            a(i,j,n3+1,-6) = 0.0
            a(i,j,n3+1,-7) = 0.0
            a(i,j,n3+1,-8) = 0.0
            a(i,j,n3+1,-9) = 0.0
         else if (bcType(bcRegion3(i,j,1)).eq.CBCFLOW) then
            if(face30(i,j,2).lt.0) a(i,j,n3+1,0) =
     $           a(i,j,n3+1,0) - face30(i,j,2)*h1(i)*h2(j)
         endif
      enddo
      enddo
c-----------------------------------------------------------------------
c front face
c
      do k = 1,n3
      do i = 1,n1
         if (bcType(bcRegion2(i,k,0)).eq.CBCDIRICHLET) then
            a(i,0,k,0) = 1.0d0*h1(i)*h3(k)
c
            face20(i,k,1) = a(i,1,k+1,-9)
            face20(i,k,2) = a(i+1,1,k,-4)
            face20(i,k,3) = a(i,1,k,-3)
            face20(i,k,4) = a(i-1,1,k,-2)
            face20(i,k,5) = a(i,0,k,-5)
c
            a(i,1,k+1,-9) = 0.0
            a(i+1,1,k,-4) = 0.0
            a(i,1,k,-3) = 0.0
            a(i-1,1,k,-2) = 0.0
            a(i,0,k,-5) = 0.0
         else if (bcType(bcRegion2(i,k,0)).eq.CBCFLOW) then
            if(face20(i,k,1).lt.0) a(i,0,k,0) =
     $           a(i,0,k,0) - face20(i,k,1)*h1(i)*h3(k)
         endif
      enddo
      enddo
c-----------------------------------------------------------------------
c back face
c
      do k = 1,n3
      do i = 1,n1
         if (bcType(bcRegion2(i,k,1)).eq.CBCDIRICHLET) then
            a(i,n2+1,k,0) = 1.0d0*h1(i)*h3(k)
c
            face21(i,k,1) = a(i,n2+1,k,-9)
            face21(i,k,2) = a(i,n2+1,k,-4)
            face21(i,k,3) = a(i,n2+1,k,-3)
            face21(i,k,4) = a(i,n2+1,k,-2)
            face21(i,k,5) = a(i,n2,k+1,-5)
c
            a(i,n2+1,k,-9) = 0.0
            a(i,n2+1,k,-4) = 0.0
            a(i,n2+1,k,-3) = 0.0
            a(i,n2+1,k,-2) = 0.0
            a(i,n2,k+1,-5) = 0.0
         else if (bcType(bcRegion2(i,k,1)).eq.CBCFLOW) then
            if(face20(i,k,2).lt.0) a(i,n2+1,k,0) =
     $           a(i,n2+1,k,0) - face20(i,k,2)*h1(i)*h3(k)
         endif
      enddo
      enddo
c-----------------------------------------------------------------------
c left face
c
      do k = 1,n3
      do j = 1,n2
         if (bcType(bcRegion1(j,k,0)).eq.CBCDIRICHLET) then
            a(0,j,k,0) = 1.0d0*h2(j)*h3(k)
c
            face10(j,k,1) = a(1,j,k+1,-8)
            face10(j,k,2) = a(1,j+1,k,-4)
            face10(j,k,3) = a(1,j,k,-1)
            face10(j,k,4) = a(0,j,k,-2)
            face10(j,k,5) = a(0,j,k,-6)
c
            a(1,j,k+1,-8) = 0.0
            a(1,j+1,k,-4) = 0.0
            a(1,j,k,-1) = 0.0
            a(0,j,k,-2) = 0.0
            a(0,j,k,-6) = 0.0
         else if (bcType(bcRegion1(j,k,0)).eq.CBCFLOW) then
            if(face10(j,k,1).lt.0) a(0,j,k,0) =
     $           a(0,j,k,0) - face10(j,k,1)*h2(j)*h3(k)
         endif
      enddo
      enddo
c-----------------------------------------------------------------------
c right face
c
      do k = 1,n3
      do j = 1,n2
         if (bcType(bcRegion1(j,k,1)).eq.CBCDIRICHLET) then
            a(n1+1,j,k,0) = 1.0d0*h2(j)*h3(k)
c
            face11(j,k,1) = a(n1+1,j,k,-8)
            face11(j,k,2) = a(n1+1,j,k,-4)
            face11(j,k,3) = a(n1+1,j,k,-1)
            face11(j,k,4) = a(n1,j+1,k,-2)
            face11(j,k,5) = a(n1,j,k+1,-6)
c
            a(n1+1,j,k,-8) = 0.0
            a(n1+1,j,k,-4) = 0.0
            a(n1+1,j,k,-1) = 0.0
            a(n1,j+1,k,-2) = 0.0
            a(n1,j,k+1,-6) = 0.0
         else if (bcType(bcRegion1(j,k,1)).eq.CBCFLOW) then
            if(face10(j,k,2).lt.0) a(n1+1,j,k,0) =
     $           a(n1+1,j,k,0) - face10(j,k,2)*h2(j)*h3(k)
         endif
      enddo
      enddo
c
c-----------------------------------------------------------------------
c Account for edge unknowns
c
      do i = 0,n1+1
         a(i,0,0,0) = 1.0
         a(i,0,n3+1,0) = 1.0
         a(i,n2+1,0,0) = 1.0
         a(i,n2+1,n3+1,0) = 1.0
      enddo
      do j = 0, n2+1
         a(0,j,0,0) = 1.0
         a(n1+1,j,0,0) = 1.0
         a(0,j,n3+1,0) = 1.0
         a(n1+1,j,n3+1,0) = 1.0
      enddo
      do k = 0, n3+1
         a(0,0,k,0) = 1.0
         a(n1+1,0,k,0) = 1.0
         a(0,n2+1,k,0) = 1.0
         a(n1+1,n2+1,k,0) = 1.0
      enddo 
c
c-----------------------------------------------------------------------
c debugging output
c
      if (idebug.ge.1) then
         write(debugunit,*) '--- CSETUPMAT ---'
         write(debugunit,1010) (index,index=-9,0)
         do 130 k = 0,n3+1
         do 130 j = 0,n2+1
         do 130 i = 0,n1+1
            write(debugunit,1000) i,j,k,(a(i,j,k,index),index=-9,0)
 130     continue
         write(debugunit,*)
 1010    format('matrix:   ',10(1x,i11))
 1000    format('(',i2,',',i2,',',i2,')',10(1x,f11.7))
      endif
c
      return
      end
c
c=======================================================================
c CSETUPRHS.f
c Set up diagonal of matrix, initial guess, and right-hand-side
c
c ... In:
c     n?       : number of cells in each direction
c     h?       : mesh widths h?(i) = x?(i)-x?(i-1)
c     bcRegion?: boundary condition regions
c     bcType   : boundary condition flags
c     bcFlux?  : normal component of the velocity on the boundary
c                (these are stored in face?0 to save storage, since
c                face?? is used only for CBCDIRICHLET and bcFlux? is
c                used only for CBCFLOW)
c     conc     : contains the boundary concentrations in the 
c                padded region
c     pad?     : padded region widths
c     face??   : coefficients of a on the faces; stored for computing
c                the action of a on given boundary concentrations and
c                adding to the right-hand side
c     cdmatrix : diagonal of a
c
c ... Out:
c     a        : cell-centered finite difference matrix
c     crhs     : right-hand side
c
      subroutine csetuprhs (a,n1,n2,n3,h1,h2,h3,
     $     face10,face11,face20,face21,face30,face31,
     $     cdmatrix,crhs,csol,conc,
     $     pad1,pad2,pad3,bcRegion1,bcRegion2,
     $     bcRegion3,bcType,cbar,totCoef,byDeltaT)
c
      implicit none
      include 'bctypes.hf'
c
      integer n1,n2,n3,pad1,pad2,pad3
      real*8  a(0:n1+1,0:n2+1,0:n3+1,-9:0),cdmatrix(n1,n2,n3),
     $        h1(0:n1+1),h2(0:n2+1),h3(0:n3+1),
     $        face10(n2,n3,5),face11(n2,n3,5),
     $        face20(n1,n3,5),face21(n1,n3,5),
     $        face30(n1,n2,5),face31(n1,n2,5),
     $        crhs(0:n1+1,0:n2+1,0:n3+1),csol(0:n1+1,0:n2+1,0:n3+1),
     $        cbar(n1,n2,n3),byDeltaT
      real*4  conc(-pad1:n1+pad1+1,-pad2:n2+pad2+1,-pad3:n3+pad3+1)
      real*4  totCoef(1-pad1:n1+pad1,1-pad2:n2+pad2,1-pad3:n3+pad3)
c
      integer bcRegion1(n2,n3,0:1),bcRegion2(n1,n3,0:1),
     $        bcRegion3(n1,n2,0:1),bcType(-1:1)
c
      integer i,j,k,n
      real*8  volPerDT
c
c-----------------------------------------------------------------------
c

      do k = 0,n3+1
      do j = 0,n2+1
      do i = 0,n1+1
         crhs(i,j,k) = 0.0d0
         csol(i,j,k) = 0.0d0
      enddo
      enddo
      enddo
c
c-----------------------------------------------------------------------
c

      do 10 k = 1, n3
      do 10 j = 1, n2
      do 10 i = 1, n1
         volPerDT = h1(i)*h2(j)*h3(k)*byDeltaT
         a(i,j,k,0) = cdmatrix(i,j,k) + volPerDT*totCoef(i,j,k)
         crhs(i,j,k) = volPerDT*totCoef(i,j,k)*cbar(i,j,k)
         csol(i,j,k) = cbar(i,j,k)
 10   continue
c
c-----------------------------------------------------------------------
c Boundary conditions...
c
c ... bottom face
c
      do j = 1,n2
      do i = 1,n1
      if (bcType(bcRegion3(i,j,0)).ne.CBCINTERNAL) then
         if (bcType(bcRegion3(i,j,0)).eq.CBCNOFLOW) then
            csol(i,j,0) = csol(i,j,1)
         else
            csol(i,j,0) = conc(i,j,0)
c
            if (bcType(bcRegion3(i,j,0)) .eq. CBCDIRICHLET) then
               crhs(i,j,0) = csol(i,j,0)*h1(i)*h2(j)
               crhs(i,j-1,1) = crhs(i,j-1,1) - face30(i,j,1)*csol(i,j,0)
               crhs(i-1,j,1) = crhs(i-1,j,1) - face30(i,j,2)*csol(i,j,0)
               crhs(i,j,  1) = crhs(i,j,  1) - face30(i,j,3)*csol(i,j,0)
               crhs(i+1,j,1) = crhs(i+1,j,1) - face30(i,j,4)*csol(i,j,0)
               crhs(i,j+1,1) = crhs(i,j+1,1) - face30(i,j,5)*csol(i,j,0)
            else if (bcType(bcRegion3(i,j,0)).eq.CBCFLOW) then
               if(face30(i,j,1).lt.0.) crhs(i,j,0) = crhs(i,j,0) 
     $              - face30(i,j,1)*csol(i,j,0)*h1(i)*h2(j)
            else
               stop 'FATAL TRANSPORT ERROR: unsupported bc type'
            endif
         endif
      endif
      enddo
      enddo
c
c ... top face
c
      do j = 1,n2
      do i = 1,n1
      if (bcType(bcRegion3(i,j,1)).ne.CBCINTERNAL) then
         if (bcType(bcRegion3(i,j,1)).eq.CBCNOFLOW) then
            csol(i,j,n3+1) = csol(i,j,n3)
         else
            csol(i,j,n3+1) = conc(i,j,n3+1)
c
            if (bcType(bcRegion3(i,j,1)) .eq. CBCDIRICHLET) then
               crhs(i,j,n3+1) = csol(i,j,n3+1)*h1(i)*h2(j)
               crhs(i,j+1,n3) = crhs(i,j+1,n3) 
     $              - face31(i,j,1)*csol(i,j,n3+1)
               crhs(i+1,j,n3) = crhs(i+1,j,n3) 
     $              - face31(i,j,2)*csol(i,j,n3+1)
               crhs(i,j,n3) = crhs(i,j,n3) 
     $              - face31(i,j,3)*csol(i,j,n3+1)
               crhs(i-1,j,n3) = crhs(i-1,j,n3) 
     $              - face31(i,j,4)*csol(i,j,n3+1)
               crhs(i,j-1,n3) = crhs(i,j-1,n3) 
     $              - face31(i,j,5)*csol(i,j,n3+1)
            else if (bcType(bcRegion3(i,j,1)).eq.CBCFLOW) then
               if(face30(i,j,2).lt.0.) crhs(i,j,n3+1) = crhs(i,j,n3+1)
     $              - face30(i,j,2)*csol(i,j,n3+1)*h1(i)*h2(j)
            else
               stop 'FATAL TRANSPORT ERROR: unsupported bc type'
            endif
         endif
      endif
      enddo
      enddo
c
c ... front face
c
      do k = 1,n3
      do i = 1,n1
      if (bcType(bcRegion2(i,k,0)).ne.CBCINTERNAL) then
         if (bcType(bcRegion2(i,k,0)).eq.CBCNOFLOW) then
            csol(i,0,k) = csol(i,1,k)
         else
            csol(i,0,k) = conc(i,0,k)
c
            if (bcType(bcRegion2(i,k,0)).eq.CBCDIRICHLET) then
               crhs(i,0,k) = csol(i,0,k)*h1(i)*h3(k)
               crhs(i,1,k+1) = crhs(i,1,k+1) - face20(i,k,1)*csol(i,0,k)
               crhs(i+1,1,k) = crhs(i+1,1,k) - face20(i,k,2)*csol(i,0,k)
               crhs(i,1,k) = crhs(i,1,k) - face20(i,k,3)*csol(i,0,k)
               crhs(i-1,1,k) = crhs(i-1,1,k) - face20(i,k,4)*csol(i,0,k)
               crhs(i,1,k-1) = crhs(i,1,k-1) - face20(i,k,5)*csol(i,0,k)
            else if (bcType(bcRegion2(i,k,0)).eq.CBCFLOW) then
               if(face20(i,k,1).lt.0.) crhs(i,0,k) = crhs(i,0,k) 
     $              - face20(i,k,1)*csol(i,0,k)*h1(i)*h3(k)
            else
               stop 'FATAL TRANSPORT ERROR: unsupported bc type'
            endif
         endif
      endif
      enddo
      enddo
c
c ... back face
c
      do k = 1,n3
      do i = 1,n1
      if (bcType(bcRegion2(i,k,1)).ne.CBCINTERNAL) then
         if (bcType(bcRegion2(i,k,1)).eq.CBCNOFLOW) then
            csol(i,n2+1,k) = csol(i,n2,k)
         else
            csol(i,n2+1,k) = conc(i,n2+1,k)
c
            if (bcType(bcRegion2(i,k,1)).eq.CBCDIRICHLET) then
               crhs(i,n2+1,k) = csol(i,n2+1,k)*h1(i)*h3(k)
               crhs(i,n2,k-1) = crhs(i,n2,k-1) 
     $              - face21(i,k,1)*csol(i,n2+1,k)
               crhs(i-1,n2,k) = crhs(i-1,n2,k) 
     $              - face21(i,k,2)*csol(i,n2+1,k)
               crhs(i,n2,k) = crhs(i,n2,k) 
     $              - face21(i,k,3)*csol(i,n2+1,k)
               crhs(i+1,n2,k) = crhs(i+1,n2,k) 
     $              - face21(i,k,4)*csol(i,n2+1,k)
               crhs(i,n2,k+1) = crhs(i,n2,k+1) 
     $              - face21(i,k,5)*csol(i,n2+1,k)
            else if (bcType(bcRegion2(i,k,1)).eq.CBCFLOW) then
               if(face20(i,k,2).lt.0.) crhs(i,n2+1,k) = crhs(i,n2+1,k)
     $              - face20(i,k,2)*csol(i,n2+1,k)*h1(i)*h3(k)
            else
               stop 'FATAL TRANSPORT ERROR: unsupported bc type'
            endif
         endif
      endif
      enddo
      enddo
c
c ... left face
c
      do k = 1,n3
      do j = 1,n2
      if (bcType(bcRegion1(j,k,0)).ne.CBCINTERNAL) then
         if (bcType(bcRegion1(j,k,0)).eq.CBCNOFLOW) then
            csol(0,j,k) = csol(1,j,k)
         else
            csol(0,j,k) = conc(0,j,k)
c
            if (bcType(bcRegion1(j,k,0)).eq.CBCDIRICHLET) then
               crhs(0,j,k) = csol(0,j,k)*h2(j)*h3(k)
               crhs(1,j,k+1) = crhs(1,j,k+1) - face10(j,k,1)*csol(0,j,k)
               crhs(1,j+1,k) = crhs(1,j+1,k) - face10(j,k,2)*csol(0,j,k)
               crhs(1,j,k) = crhs(1,j,k) - face10(j,k,3)*csol(0,j,k)
               crhs(1,j-1,k) = crhs(1,j-1,k) - face10(j,k,4)*csol(0,j,k)
               crhs(1,j,k-1) = crhs(1,j,k-1) - face10(j,k,5)*csol(0,j,k)
            else if (bcType(bcRegion1(j,k,0)).eq.CBCFLOW) then
               if(face10(j,k,1).lt.0.) then
                  crhs(0,j,k) = crhs(0,j,k) 
     $              - face10(j,k,1)*csol(0,j,k)*h2(j)*h3(k)
               endif
            else
               stop 'FATAL TRANSPORT ERROR: unsupported bc type'
            endif
         endif
      endif
      enddo
      enddo
c
c ... right face
c
      do k = 1,n3
      do j = 1,n2
      if (bcType(bcRegion1(j,k,1)).ne.CBCINTERNAL) then
         if (bcType(bcRegion1(j,k,1)).eq.CBCNOFLOW) then
            csol(n1+1,j,k) = csol(n1,j,k)
         else
            csol(n1+1,j,k) = conc(n1+1,j,k)
c
            if (bcType(bcRegion1(j,k,1)).eq.CBCDIRICHLET) then
               crhs(n1+1,j,k) = csol(n1+1,j,k)*h2(j)*h3(k)
               crhs(n1,j,k-1) = crhs(n1,j,k-1) 
     $              - face11(j,k,1)*csol(n1+1,j,k)
               crhs(n1,j-1,k) = crhs(n1,j-1,k) 
     $              - face11(j,k,2)*csol(n1+1,j,k)
               crhs(n1,j,k) = crhs(n1,j,k) 
     $              - face11(j,k,3)*csol(n1+1,j,k)
               crhs(n1,j+1,k) = crhs(n1,j+1,k) 
     $              - face11(j,k,4)*csol(n1+1,j,k)
               crhs(n1,j,k+1) = crhs(n1,j,k+1) 
     $              - face11(j,k,5)*csol(n1+1,j,k)
            else if (bcType(bcRegion1(j,k,1)).eq.CBCFLOW) then
               if(face10(j,k,2).lt.0.) crhs(n1+1,j,k) = crhs(n1+1,j,k)
     $              - face10(j,k,2)*csol(n1+1,j,k)*h2(j)*h3(k)
            else
               stop 'FATAL TRANSPORT ERROR: unsupported bc type'
            endif
         endif
      endif
      enddo
      enddo
c
ccc: check positivity of diagonal
c      do 1510 k=1,n3
c      do 1510 j=1,n2
c      do 1510 i=1,n1
c         if(a(i,j,k,0).lt.0.)
c     $        write(*,*) 'negA:',i,j,k,a(i,j,k,0)
c 1510  continue
ccc.
      return
      end
c=======================================================================
c Set up bcFlux
c
c ... In:
c     x?       : grid
c     n?       : number of cells in each direction
c     bcRegion?: boundary condition regions
c     bcType   : boundary condition flags (see 'bctypes.hf')
c                bcType < 0 is an internal boundary
c                bcType = CBCFLOW
c                   outflow is z.n = 0
c                   inflow is z.n + c*u.n = cIN*u.n (Robin)
c                bcType = CBCDIRICHLET is c = cD
c                bcType = CBCNOFLOW is z.n = 0
c     comp     : a component with the desired dispType
c     t        : time
c
c ... Out:
c     bcFlux?  : normal component of the velocity on the boundary
c
      subroutine setupbcFlux(bcFlux1,bcFlux2,bcFlux3,x1,x2,x3,
     $     n1,n2,n3,bcRegion1,bcRegion2,bcRegion3,bcType,comp,t)
      implicit none
      include 'bctypes.hf'
c
      integer n1,n2,n3,comp
      integer bcRegion1(n2,n3,0:1),bcRegion2(n1,n3,0:1),
     $        bcRegion3(n1,n2,0:1),bcType(-1:1)
      real*8  bcFlux1(n2,n3,0:1),bcFlux2(n1,n3,0:1),bcFlux3(n1,n2,0:1),
     $        x1(1:n1+1),x2(1:n2+1),x3(1:n3+1),t
c
      integer i,j,k
      real*8  vel(3)
c
      do 10 k=1,n3
      do 10 j=1,n2
         if(bcType(bcRegion1(j,k,0)).eq.CBCFLOW) then
            call evalDispVel(comp,x1(1),.5*(x2(j)+x2(j+1)),
     $           .5*(x3(k)+x3(k+1)),t,vel)
            bcFlux1(j,k,0) = -vel(1)
         endif
         if(bcType(bcRegion1(j,k,1)).eq.CBCFLOW) then
            call evalDispVel(comp,x1(n1+1),.5*(x2(j)+x2(j+1)),
     $           .5*(x3(k)+x3(k+1)),t,vel)
            bcFlux1(j,k,1) = vel(1)
         endif
 10   continue
c
      do 20 k=1,n3
      do 20 i=1,n1
         if(bcType(bcRegion2(i,k,0)).eq.CBCFLOW) then
            call evalDispVel(comp,.5*(x1(i)+x1(i+1)),x2(1),
     $           .5*(x3(k)+x3(k+1)),t,vel)
            bcFlux2(i,k,0) = -vel(2)
         endif
         if(bcType(bcRegion2(i,k,1)).eq.CBCFLOW) then
            call evalDispVel(comp,.5*(x1(i)+x1(i+1)),x2(n2+1),
     $           .5*(x3(k)+x3(k+1)),t,vel)
            bcFlux2(i,k,1) = vel(2)
         endif
 20   continue
c
      do 30 j=1,n2
      do 30 i=1,n1
         if(bcType(bcRegion3(i,j,0)).eq.CBCFLOW) then
            call evalDispVel(comp,.5*(x1(i)+x1(i+1)),.5*(x2(j)+x2(j+1)),
     $           x3(1),t,vel)
            bcFlux3(i,j,0) = -vel(3)
         endif
         if(bcType(bcRegion3(i,j,1)).eq.CBCFLOW) then
            call evalDispVel(comp,.5*(x1(i)+x1(i+1)),.5*(x2(j)+x2(j+1)),
     $           x3(n3+1),t,vel)
            bcFlux3(i,j,1) = vel(3)
         endif
 30   continue
c
      return
      end
