c=======================================================================
c Evaluate the diffusion/dispersion tensor
c   at points of the concentration mesh
c
c We assume that the physical domain has been mapped to the
c   computational domain
c
c For each dispType i,
c
c   D_i(v,s) = D_i^{mol}(s) + D_i^{disp}(v)
c
c   D_i^{mol}(s)
c     = diffM * (Jacobian * porosity)
c       * \sum_{j=1}^NPHASES [ (saturation)_j * phaseDist(j,i) ] * I
c
c   D_i^{disp}(v)
c     = \sum_{=1}^NPHASES { phaseDist(j,i) * s_j
c                          * [ (displ - dispt) * P(v_j) + dispt * I ] }
c
c   I = effective identity tensor = dMap^{-1} * (dMap^{-1})^T
c
c   P(v_j)_{m,n} = (v_j)_m * (v_j)_n / s_j**2
c     = projection onto v_j direction
c
c   s_j = |dMap v_j| is the effective speed
c
c We require the work arrays:
c   vel(3,NPHASES),speed(NPHASES)
c   sat(NPHASES)
c     Total size: 5*NPHASES
c
c-----------------------------------------------------------------------
c
      subroutine dispVarD(d11,d12,d13,d22,d23,d33,
     $     xc,yc,zc,tc,diffMAry,dispLAry,dispTAry,
     $     phaseDist,ncomp,comp,effIdentity,
     $     vel,sat,speed,nxc,nyc,nzc,idebug,debugunit)
      implicit none
c
      integer NPHASES
      parameter(NPHASES = 1)
c
      integer nxc,nyc,nzc,idebug,debugunit
      integer ncomp,comp
      real*8  d11(nxc+1,nyc+1,nzc+1),d12(nxc+1,nyc+1,nzc+1),
     $        d13(nxc+1,nyc+1,nzc+1),d22(nxc+1,nyc+1,nzc+1),
     $        d23(nxc+1,nyc+1,nzc+1),d33(nxc+1,nyc+1,nzc+1)
      real*8  xc(nxc+1),yc(nyc+1),zc(nzc+1)
      real*8  tc,
     $        diffMAry(nxc, nyc, nzc),  diffMa, 
     $        dispLAry(nxc, nyc, nzc),  dispLa,
     $        dispTAry(nxc, nyc, nzc),  dispTa
c      disp?Ary for dispersion/diffusion parameter array
c      disp?a for average dispersion/diffusion parameter at the corner point
      real*8  phaseDist(0:NPHASES,ncomp),vel(3,NPHASES),sat(NPHASES),
     $        speed(NPHASES)
      real*4  effIdentity(6,nxc+1,nyc+1,nzc+1)
c
      integer i,j,k,phase,m
      real*8  dm,id(6), sumM, sumL, sumT
      integer n_sum
      logical nonZeroM, nonZeroL, nonZeroT
c
c-----------------------------------------------------------------------
c loop over the grid points
c
      do 10 k = 1, nzc+1
      do 10 j = 1, nyc+1
      do 10 i = 1, nxc+1

c --------------------------------------- start block -- shuyu
c     calculate diffMa, dispLa, dispTa
         sumM = 0.0
         sumL = 0.0
         sumT = 0.0
         nonZeroM = .true.
         nonZeroL = .true.
         nonZeroT = .true.
         n_sum = 0
         if (i.ge.2) then
           if (j.ge.2) then
             if (k.ge.2) then
               nonZeroM = nonZeroM.and.( diffMAry(i-1,j-1,k-1).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i-1,j-1,k-1).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i-1,j-1,k-1).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i-1,j-1,k-1)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i-1,j-1,k-1)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i-1,j-1,k-1)
               n_sum = n_sum + 1
             endif
             if (k.le.nzc) then     
               nonZeroM = nonZeroM.and.( diffMAry(i-1,j-1,k).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i-1,j-1,k).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i-1,j-1,k).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i-1,j-1,k)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i-1,j-1,k)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i-1,j-1,k)
               n_sum = n_sum + 1
             endif
           endif
           if (j.le.nyc) then
             if (k.ge.2) then
               nonZeroM = nonZeroM.and.( diffMAry(i-1,j,k-1).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i-1,j,k-1).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i-1,j,k-1).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i-1,j,k-1)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i-1,j,k-1)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i-1,j,k-1)
               n_sum = n_sum + 1
             endif
             if (k.le.nzc) then     
               nonZeroM = nonZeroM.and.( diffMAry(i-1,j,k).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i-1,j,k).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i-1,j,k).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i-1,j,k)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i-1,j,k)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i-1,j,k)
               n_sum = n_sum + 1
             endif
           endif
         endif
         if (i.le.nxc) then
           if (j.ge.2) then
             if (k.ge.2) then
               nonZeroM = nonZeroM.and.( diffMAry(i,j-1,k-1).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i,j-1,k-1).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i,j-1,k-1).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i,j-1,k-1)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i,j-1,k-1)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i,j-1,k-1)
               n_sum = n_sum + 1
             endif
             if (k.le.nzc) then     
               nonZeroM = nonZeroM.and.( diffMAry(i,j-1,k).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i,j-1,k).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i,j-1,k).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i,j-1,k)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i,j-1,k)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i,j-1,k)
               n_sum = n_sum + 1
             endif
           endif
           if (j.le.nyc) then
             if (k.ge.2) then
               nonZeroM = nonZeroM.and.( diffMAry(i,j,k-1).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i,j,k-1).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i,j,k-1).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i,j,k-1)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i,j,k-1)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i,j,k-1)
               n_sum = n_sum + 1
             endif
             if (k.le.nzc) then     
               nonZeroM = nonZeroM.and.( diffMAry(i,j,k).ne.0.0 )
               nonZeroL = nonZeroL.and.( dispLAry(i,j,k).ne.0.0 )
               nonZeroT = nonZeroT.and.( dispTAry(i,j,k).ne.0.0 )
               if(nonZeroM) sumM = sumM + 1.0d0 / diffMAry(i,j,k)
               if(nonZeroL) sumL = sumL + 1.0d0 / dispLAry(i,j,k)
               if(nonZeroT) sumT = sumT + 1.0d0 / dispTAry(i,j,k)
               n_sum = n_sum + 1
             endif
           endif
         endif
         
         diffMa = 0.0
         dispLa = 0.0
         dispTa = 0.0
         if(nonZeroM) diffMa = n_sum * 1.0d0 / sumM
         if(nonZeroL) dispLa = n_sum * 1.0d0 / sumL
         if(nonZeroT) dispTa = n_sum * 1.0d0 / sumT

c-------------------------------------------------end block -- shuyu

c
         call evalVel(xc(i),yc(j),zc(k),tc,vel)
         call evalPorSat(xc(i),yc(j),zc(k),tc,sat)
         call evalEffSpeed(vel,i,j,k,speed)
c
c molecular diffusion
c
         dm = 0.
         do 20 phase=1,NPHASES
            dm = dm + sat(phase)*phaseDist(phase,comp)
 20      continue

         dm = diffMa*dm
         do m = 1,6
            id(m) = effIdentity(m,i,j,k)
         enddo
         d11(i,j,k) = dm*id(1)
         d12(i,j,k) = dm*id(2)
         d13(i,j,k) = dm*id(3)
         d22(i,j,k) = dm*id(4)
         d23(i,j,k) = dm*id(5)
         d33(i,j,k) = dm*id(6)
c
c dispersion
c
         d12(i,j,k) = 0.d0
         d13(i,j,k) = 0.d0
         d23(i,j,k) = 0.d0
         do 30 phase=1,NPHASES
            if (speed(phase).gt.0.d0) then
               d11(i,j,k) = d11(i,j,k) + phaseDist(phase,comp) * (
     $            (dispLa-dispTa)*vel(1,phase)*vel(1,phase)/speed(phase)
     $              + speed(phase)*id(1)*dispTa )
               d12(i,j,k) = d12(i,j,k) + phaseDist(phase,comp) * (
     $            (dispLa-dispTa)*vel(1,phase)*vel(2,phase)/speed(phase)
     $              + speed(phase)*id(2)*dispTa )
               d13(i,j,k) = d13(i,j,k) + phaseDist(phase,comp) * (
     $            (dispLa-dispTa)*vel(1,phase)*vel(3,phase)/speed(phase)
     $              + speed(phase)*id(3)*dispTa )
               d22(i,j,k) = d22(i,j,k) + phaseDist(phase,comp) * (
     $            (dispLa-dispTa)*vel(2,phase)*vel(2,phase)/speed(phase)
     $              + speed(phase)*id(4)*dispTa )
               d23(i,j,k) = d23(i,j,k) + phaseDist(phase,comp) * (
     $            (dispLa-dispTa)*vel(2,phase)*vel(3,phase)/speed(phase)
     $              + speed(phase)*id(5)*dispTa )
               d33(i,j,k) = d33(i,j,k) + phaseDist(phase,comp) * (
     $            (dispLa-dispTa)*vel(3,phase)*vel(3,phase)/speed(phase)
     $              + speed(phase)*id(6)*dispTa )
            endif
 30      continue
c
 10   continue
c
      if (idebug.ge.1) then
         write(debugunit,*) '--- DISP ---'
         write(debugunit,1010)
         do 200 k = 1, nzc+1
         do 200 j = 1, nyc+1
         do 200 i = 1, nxc+1
            write(debugunit,1000) i,j,k,
     $           d11(i,j,k),d12(i,j,k),d13(i,j,k),
     $           d22(i,j,k),d23(i,j,k),d33(i,j,k)
 200     continue
         write(debugunit,*)
 1010    format('disp:     ',
     $        6x,'d11',11x,'d12',11x,'d13',11x,'d22',11x,'d23',
     $        11x,'d33')
 1000    format('(',i2,',',i2,',',i2,')',6(1x,f13.10))
      endif
c
      return
      end
