#include <stdlib.h>
#include <stdio.h>
#include "../Main/ftn.h"
#include "../Main/bool.h"
#include "../Main/mymem.h"
#include "../Main/mymisc.h"
#include "../Main/mybits.h"

#define NPHASES 1
#define WATERPHASE 1

double EvalOneSorp();
int traceTypeComp(int);

typedef double double3[3];

static int nComps;
static int nTtypes;
static double3 *phaseVel;
static double *porsat;
static double *porsatint;
static double *sorpint;
static int* sorptype;
static double *phasedist;

void setupEvalTraceVel(int ncmp, int ntt, int nSorps,
		       int* srptype, double* phsedist) {
  nComps = ncmp;
  nTtypes = ntt;
  phaseVel = (double3*) MYMALLOC(NPHASES * sizeof(double3));
  porsat = (double*) MYMALLOC(NPHASES * sizeof(double));
  porsatint = (double*) MYMALLOC(NPHASES * sizeof(double));
  sorpint = (double*) MYMALLOC((nSorps+1) * sizeof(double));
  sorptype = srptype;
  phasedist = phsedist;
}

void cleanupEvalTraceVel()
{
  MYFREE(phaseVel);
  MYFREE(porsat);
  MYFREE(porsatint);
  MYFREE(sorpint);
}

#define index(phase,comp) ((phase)+(NPHASES+1)*((comp)-1))

bool EvalTraceCornerVels(tracetype, xPos, yPos, zPos, t, vel)
     integer tracetype;
     real8 xPos, yPos, zPos;
     real8 t;
     double vel[8][3];
{
  typedef double double_8_3[8][3];
  double_8_3* phaseVels = (double_8_3*) MYMALLOC(NPHASES * sizeof(double_8_3));
  int k;
  int phase;
  int tType = *tracetype; /* tType >= 1, ie tracetype is mobile */
  int comp = traceTypeComp(tType)+1;
  double sum = 0.;
  bool ok;
  ok = EvalCornerVels(xPos, yPos, zPos, t, phaseVels);
  if (!ok)
    {
      MYFREE((char*) phaseVels);
      return FALSE;
    }
  EvalPorSat(xPos, yPos, zPos, t, porsat);
  for (k = 0; k < 8; ++k)
    {
      vel[k][0] = vel[k][1] = vel[k][2] = 0.;
      for (phase = 1; phase <= NPHASES; ++phase)
	{
	  double phdist = phasedist[index(phase,comp)];
	  sum += phdist * porsat[phase-1];
	  vel[k][0] += phdist * phaseVels[phase-1][k][0];
	  vel[k][1] += phdist * phaseVels[phase-1][k][1];
	  vel[k][2] += phdist * phaseVels[phase-1][k][2];
	}
      sum += phasedist[index(0,comp)]*EvalOneSorp(xPos,yPos,zPos,&sorptype[comp-1]);
      if (sum == 0.)
	vel[k][0] = vel[k][1] = vel[k][2] = 0.;
      else
	{
	  sum = 1. / sum;
	  vel[k][0] *= sum;
	  vel[k][1] *= sum;
	  vel[k][2] *= sum;
	}
    }
  MYFREE((char*) phaseVels);
  return TRUE;
}

bool Hassink(tracetype, xPos, yPos, zPos, t)
     integer tracetype;
     real8 xPos, yPos, zPos;
     real8 t;
{
  double vel[8][3];
  bool possiblyIsSink = TRUE;
  int j;
  EvalTraceCornerVels(tracetype, xPos, yPos, zPos, t, vel);
  for (j = 0; possiblyIsSink && j < 3; ++j)
    {
      int mask = clrbit(0x7, j);
      int k = 0;
      int kbit = setbit(0,j);
      do
	if (vel[k][j] > 0. || vel[k | kbit][j] < 0.)
	  possiblyIsSink = FALSE;
      while (possiblyIsSink && !!maskedIncrement(mask,k));
    }
  if (!possiblyIsSink) return FALSE;
  {
    double bary[2][3];
    bary[0][0] = bary[0][1] = bary[0][2] = 0.;
    bary[1][0] = bary[1][1] = bary[1][2] = 1.;
    while (bary[1][0] - bary[0][0] > 0.001)
      {
	for (j = 0; j < 3; ++j)
	  {
	    double baryMid = 0.5 * (bary[0][j]+bary[1][j]);
	    double velMid = 0.;
	    int mask = clrbit(0x7, j);
	    int k;
	    bool way;
	    for (k = 0; k < 8; ++k)
	      velMid += vel[k][j];
	    way = velMid >= 0;
	    k = 0;
	    do
	      {
		int hither = k;
		int yon = setbit(k,j);
		int thither = way? yon: hither;
		int jj;
		for (jj = 0; jj < 3; ++jj)
		  vel[thither][jj] = 0.5 * (vel[hither][jj] + vel[yon][jj]);
	      }
	    while (!!maskedIncrement(mask,k));
	    bary[way][j] = baryMid;
	  }
      }
    BarytoNatural(bary[0], xPos, yPos, zPos);
    return TRUE;
  }
}

bool EvalTraceVel(tracetype, xPos, yPos, zPos, t, vel)
     integer tracetype;
     real8 xPos, yPos, zPos;
     real8 t;
     real8 vel;
{
  int phase;
  int tType = *tracetype; /* tType >= 1, ie tracetype is mobile */
  int comp = traceTypeComp(tType)+1;
  double sum = 0.;
  bool ok;
  ok = EvalVel(xPos, yPos, zPos, t, phaseVel);
  if (!ok) return FALSE;
  EvalPorSat(xPos, yPos, zPos, t, porsat);
  vel[0] = vel[1] = vel[2] = 0.;
  for (phase = 1; phase <= NPHASES; ++phase)
    {
      double phdist = phasedist[index(phase,comp)];
      sum += phdist * porsat[phase-1];
      vel[0] += phdist * phaseVel[phase-1][0];
      vel[1] += phdist * phaseVel[phase-1][1];
      vel[2] += phdist * phaseVel[phase-1][2];
    }
  sum += phasedist[index(0,comp)]
    *EvalOneSorp(xPos,yPos,zPos,&sorptype[comp-1]);
  if (sum == 0.)
    vel[0] = vel[1] = vel[2] = 0.;
  else
    {
      sum = 1. / sum;
      vel[0] *= sum;
      vel[1] *= sum;
      vel[2] *= sum;
    }
  return TRUE;
}

void EvalTotCoefInt(xPos, yPos, zPos, t, totcoefint)
     real8 xPos, yPos, zPos;
     real8 t;
     real8 totcoefint;
{
  int phase, ttype;
  EvalPorSatInt(xPos,yPos,zPos,t,porsatint);
  EvalSorpInt(xPos,yPos,zPos,sorpint);
  for (ttype = 1; ttype <= nTtypes; ++ttype)
    {
      int comp = traceTypeComp(ttype)+1;
      double sum = phasedist[index(0,comp)] * sorpint[sorptype[comp-1]];
      for (phase = 1; phase <= NPHASES; ++phase)
	sum += phasedist[index(phase,comp)] * porsatint[phase-1];
      totcoefint[ttype-1] = sum;
    }
}

double EvalRXNVolInt(xPos, yPos, zPos, t)
     real8 xPos, yPos, zPos;
     real8 t;
{
  EvalPorSatInt(xPos,yPos,zPos,t,porsatint);
  return porsatint[WATERPHASE-1];
}

bool EvalDispVel(comp, xPos, yPos, zPos, t, vel)
     integer comp;
     real8 xPos, yPos, zPos;
     real8 t;
     real8 vel;
{
  int phase;
  bool ok;
  ok = EvalVel(xPos, yPos, zPos, t, phaseVel);
  if (!ok) return FALSE;
  vel[0] = vel[1] = vel[2] = 0.;
  for (phase = 1; phase <= NPHASES; ++phase)
    {
      double phdist = phasedist[index(phase,*comp)];
      vel[0] += phdist * phaseVel[phase-1][0];
      vel[1] += phdist * phaseVel[phase-1][1];
      vel[2] += phdist * phaseVel[phase-1][2];
    }
  return TRUE;
}
