/******************************************************************************
 * Initialize transport
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "../Main/msgs.h"
#include "../Main/debug.h"
#include "../Main/driverGlobals.h"
#include "transportGlobals.h"
#include "transport.h"

#define NPHASES 1
#define MIN(x,y) ((x) < (y)) ? (x) : (y)
#define MAX(x,y) ((x) > (y)) ? (x) : (y)

void traceSetup(int[3],int*,int,int,int*,int*,int*,
		double*,double*,double*,int,int);
void setupEvalTraceVel(int,int,int,int*,double*);
void setupTrilinear(double*,double*,double*,double*,double*,double*,int*);

/******************************************************************************
 * True evaluation of totCoef (Assume saturation=1, i.e., N_FLOWING_PHASES=1)
 *****************************************************************************/
#if N_FLOWING_PHASES != 1
This code has an error!!!
#endif

#if 0
void defineTotCoefInt() {
  int i,j,k,specie,ttype;

  for(k=0; k<nz; k++)
  for(j=0; j<ny; j++)
  for(i=0; i<nx; i++) {
    for(specie = 0; specie < nSpecies; specie++) {
      ttype = traceType[specie]-1;
      if(ttype < 0) continue;
      if(sorpType[specie]==0) {
	totCoefInt_(i,j,k,ttype) =
	  ( phaseDist_(0,specie) + phaseDist_(1,specie) )*effPorosity_(i,j,k)
	  *hx_(i)*hy_(j)*hz_(k);
      } else {
	totCoefInt_(i,j,k,ttype) =
	  ( phaseDist_(0,specie)*sorp_(i,j,k,sorpType[specie]-1)
	    + phaseDist_(1,specie)*effPorosity_(i,j,k) )*hx_(i)*hy_(j)*hz_(k);
      }
    }
  }
}
#endif

int defineTotCoef() {
  int i,j,k,specie,ttype,error;

  for(k=0; k<nz; k++)
  for(j=0; j<ny; j++)
  for(i=0; i<nx; i++) {
    for(specie = 0; specie < nSpecies; specie++) {
      ttype = traceType[specie]-1;
      if(ttype < 0) continue;
      if(sorpType[specie]==0) {
	totCoef_(i,j,k,ttype) =
	  ( phaseDist_(0,specie) + phaseDist_(1,specie) )*effPorosity_(i,j,k);
      } else {
	totCoef_(i,j,k,ttype) =
	  phaseDist_(0,specie)*sorp_(i,j,k,sorpType[specie]-1)
	  + phaseDist_(1,specie)*effPorosity_(i,j,k);
      }
    }
  }

  /* Communicate padded information for totCoef (mobile types only) */

  for(ttype = 0; ttype < nTraceTypes; ttype++) {
    int zero = 0, one = 1;
    int msg = MSG_PADTOTCOEF;
    MPI_Barrier(comm);
    RPadStencil(&totCoef_array_(ttype),&one,&nx,&ny,&nz,
                &zero,&zero,&zero,&padx,&pady,&padz,myNbr,&msg,&error);
    if(error) return 2;
    RPadExtFace(&totCoef_array_(ttype),&one,&nx,&ny,&nz,
		&padx,&pady,&padz,myNbr,&zero,&zero);
    if(error) return 2;
  }

  return 0;
}

/******************************************************************************
 * Some initialization
 *****************************************************************************/
int initTrans() {
  int i,j,k,specie,phase;
  int error = 0;

  /* Define grid coefficients */

  for(i=-padx-1; i<nx+padx+1; i++) hx_(i) = xpComp_(i+1) - xpComp_(i);
  for(i=-padx; i<nx+padx; i++) byhx_(i) = 1/hx_(i);
  for(i=-padx; i<=nx+padx; i++) byDistx_(i) = 2/(xpComp_(i+1)-xpComp_(i-1));
  if(myNbr_(-1,0,0) == -1) {
    for(i=-padx-1; i < 0; i++) hx_(i) = 0;
    byDistx_(0) = 2/(xpComp_(1)-xpComp_(0));
  }
  if(myNbr_(1,0,0) == -1) {
    for(i=nx; i<=nx+padx; i++) hx_(i) = 0;
    byDistx_(nx) = 2/(xpComp_(nx)-xpComp_(nx-1));
  }

  for(j=-pady-1; j<ny+pady+1; j++) hy_(j) = ypComp_(j+1) - ypComp_(j);
  for(j=-pady; j<ny+pady; j++) byhy_(j) = 1/hy_(j);
  for(j=-pady; j<=ny+pady; j++) byDisty_(j) = 2/(ypComp_(j+1)-ypComp_(j-1));
  if(myNbr_(-1,0,0) == -1) {
    for(j=-pady-1; j < 0; j++) hy_(j) = 0;
    byDisty_(0) = 2/(ypComp_(1)-ypComp_(0));
  }
  if(myNbr_(1,0,0) == -1) {
    for(j=ny; j<=ny+pady; j++) hy_(j) = 0;
    byDisty_(ny) = 2/(ypComp_(ny)-ypComp_(ny-1));
  }

  for(k=-padz-1; k<nz+padz+1; k++) hz_(k) = zpComp_(k+1) - zpComp_(k);
  for(k=-padz; k<nz+padz; k++) byhz_(k) = 1/hz_(k);
  for(k=-padz; k<=nz+padz; k++) byDistz_(k) = 2/(zpComp_(k+1)-zpComp_(k-1));
  if(myNbr_(-1,0,0) == -1) {
    for(k=-padz-1; k < 0; k++) hz_(k) = 0;
    byDistz_(0) = 2/(zpComp_(1)-zpComp_(0));
  }
  if(myNbr_(1,0,0) == -1) {
    for(k=nz; k<=nz+padz; k++) hz_(k) = 0;
    byDistz_(nz) = 2/(zpComp_(nz)-zpComp_(nz-1));
  }

  /* Define minimum padding distance (within domain) */

  padDist = MAX(MAX(xpComp_(nx+padx) - xpComp_(-padx),
		    ypComp_(ny+pady) - ypComp_(-pady)),
		zpComp_(nz+padz) - zpComp_(-padz));
  if(myNbr_(-1,0,0) != -1) padDist=MIN(padDist, xpComp_(padx)-xpComp_(0));
  if(myNbr_( 1,0,0) != -1) padDist=MIN(padDist, xpComp_(nx)-xpComp_(nx-padx));
  if(myNbr_(0,-1,0) != -1) padDist=MIN(padDist, ypComp_(pady)-ypComp_(0));
  if(myNbr_(0, 1,0) != -1) padDist=MIN(padDist, ypComp_(ny)-ypComp_(ny-pady));
  if(myNbr_(0,0,-1) != -1) padDist=MIN(padDist, zpComp_(padz)-zpComp_(0));
  if(myNbr_(0,0, 1) != -1) padDist=MIN(padDist, zpComp_(nz)-zpComp_(nz-padz));

  /* Define the trace types
   *   Trace:
   *     0 = immobile (no tracing needed)
   *     1,2,...,nTraceTypes = distinct mobile characteristic tracing types */

  k = nTraceTypes;
  nTraceTypes = 0;
  for(specie = 0; specie < nSpecies; specie++) {
    /* check if immobile */
    int immobile = 1;
    for(phase = 1; phase <= NPHASES; phase++) {
      immobile = immobile && (phaseDist_(phase,specie) == 0);
    }
    if(immobile) {
      traceType[specie] = 0;
    } else {
      /* check if new trace type */
      for(i=0; i<specie; i++) {
	/* check mobile and immobile parts */
	int sameSorp = (sorpType[specie] == sorpType[i]);
	for(phase = 0; phase <= NPHASES; phase++) {
	  sameSorp
	    = sameSorp && (phaseDist_(phase,specie) == phaseDist_(phase,i));
	}
	if(sameSorp) {
	  traceType[specie] = traceType[i];
	  goto nextTraceSpecie;
	}
      }
      nTraceTypes++;
      traceType[specie] = nTraceTypes;
    }
  nextTraceSpecie:;
  }
  if(k != nTraceTypes) return ERR_IN_CODE;

  /* Define the disp types
   *   Disp:
   *     0 = no dispersion, i.e., immobile or zeroD flag set
   *     1,2,... = distinct dispersion types */

  k = nDispTypes;
  nDispTypes = 0;
  for(specie = 0; specie < nSpecies; specie++) {
    /* check if immobile or if no dispersion flag is set */
    int immobile = 1;
    for(phase = 1; phase <= NPHASES; phase++) {
      immobile = immobile && (phaseDist_(phase,specie) == 0);
    }
    if(immobile || zeroD) {
      dispType[specie] = 0;
    } else {
      /* check if new disp type */
      for(i=0; i<specie; i++) {
	/* check mobile part only */
	int samePhaseDist = 1;
	for(phase = 1; phase <= NPHASES; phase++) {
	  samePhaseDist = samePhaseDist &&
	    (phaseDist_(phase,specie) == phaseDist_(phase,i));
	}
	if(samePhaseDist) {
	  dispType[specie] = dispType[i];
	  goto nextDispSpecie;
	}
      }
      nDispTypes++;
      dispType[specie] = nDispTypes;
    }
  nextDispSpecie:;
  }
  if(k != nDispTypes) return ERR_IN_CODE;

  /* Set up trace back and evaluation routines */

  traceSetup(volRefine_t,traceType,nTraceTypes,
	     nWells,hWellIndex,zWellIndexLo,zWellIndexHi,
	     wellInjConc,wellProd,wellTotCoefInt,
	     MSG_TRACEBCK,steadyFlow);
  setupEvalTraceVel(nSpecies,nTraceTypes,
		    nSorpTypes,sorpType,phaseDist);
  setupTrilinear(hx,hy,hz,byDistx,byDisty,byDistz,myNbr);

  /* Define initial totCoefInt and totCoef */

  DefineTotCoefInt(totCoefInt,&time_t,&xpComp_(0),&ypComp_(0),&xpComp_(0),
		   &byhx_(0),&byhy_(0),&byhz_(0),
		   &nx,&ny,&nz,&padx,&pady,&padz,&nTraceTypes,dwk);
  error = defineTotCoef(); if(error) return 2;

  return error;
}

/******************************************************************************
 * Set some sizes for memory allocation
 *  nTraceTypes and nDispTypes
 *
 * This code must match the code that actually defines the trace and disp types
 *****************************************************************************/
void derivedSizes() {
  int specie,i,phase;

  /* Determine the number of trace types */

  nTraceTypes = 0;
  for(specie = 0; specie < nSpecies; specie++) {
    /* check if immobile */
    int immobile = 1;
    for(phase = 1; phase <= NPHASES; phase++) {
      immobile = immobile && (phaseDist_(phase,specie) == 0);
    }
    if(!immobile) {
      /* check if new trace type */
      for(i=0; i<specie; i++) {
	/* check mobile and immobile parts */
	int sameSorp = (sorpType[specie] == sorpType[i]);
	for(phase = 0; phase <= NPHASES; phase++) {
	  sameSorp
	    = sameSorp && (phaseDist_(phase,specie) == phaseDist_(phase,i));
	}
	if(sameSorp) goto nextTraceSpecie;
      }
      nTraceTypes++;
    }
  nextTraceSpecie:;
  }

  /* Define the disp types
   *   Disp:
   *     0 = no dispersion, i.e., immobile or zeroD flag set
   *     1,2,... = distinct dispersion types */

  nDispTypes = 0;
  for(specie = 0; specie < nSpecies; specie++) {
    /* check if immobile or if no dispersion flag is set */
    int immobile = 1;
    for(phase = 1; phase <= NPHASES; phase++) {
      immobile = immobile && (phaseDist_(phase,specie) == 0);
    }
    if( !(immobile || zeroD) ) {
      /* check if new disp type */
      for(i=0; i<specie; i++) {
	/* check mobile part only */
	int samePhaseDist = 1;
	for(phase = 1; phase <= NPHASES; phase++) {
	  samePhaseDist = samePhaseDist &&
	    (phaseDist_(phase,specie) == phaseDist_(phase,i));
	}
	if(samePhaseDist) goto nextDispSpecie;
      }
      nDispTypes++;
    }
  nextDispSpecie:;
  }

  return;
}
