#include <stdio.h>
#include <stdlib.h>
#include "../Main/ftn.h"
#include "../Main/bool.h"
#include "../Main/float_types.h"
#include "../Main/cglobals.h"
#include "../Main/mymem.h"
#include "../Main/mymisc.h"
#include "../Main/msg.h"

extern void traceback_setup();
extern void traceBack();

/* extern int mynodeX(); */
/* extern int nnodesX(); */

#define MsgSize 3
enum MsgNames {nCuts=0, nMsgsSent, sender};
enum MsgType  {TB_timestep=0, TB_sync};

static int* cut;
static int* msgsRecvd;
static int* msgsSent;
static int nProcs, me;
static int msgtype_TB_timestep;
static int msgtype_TB_sync;

void traceSetup(int volrefine[3], int* tracetype, int ntracetypes,
		int nWells, int (*locWell)[2],
		int (*topWell)[2], int (*botWell)[2],
		double* wellInj, double* wellProd, double* wellTotCoef,
		int msgroot, int steadFlow)
{
  /* nProcs = nnodesX(); */
  /* me = mynodeX(); */
  MPI_Comm_size(comm,&nProcs);
  MPI_Comm_rank(comm,&me);
  cut = (int*) MYMALLOC(nProcs*sizeof(int));
  msgsRecvd = (int*) MYMALLOC(nProcs*sizeof(int));
  msgsSent = (int*) MYMALLOC(nProcs*sizeof(int));
  
  traceback_setup(tracetype, ntracetypes, volrefine,
		  locWell, topWell, botWell, nWells,
		  wellInj, wellProd, wellTotCoef, steadFlow);
  msgtype_TB_timestep = msgroot + (int)(TB_timestep);
  msgtype_TB_sync = msgroot + (int)(TB_sync);
}

void traceCleanup ()
{
  MYFREE((char*) msgsSent);
  MYFREE((char*) msgsRecvd);
  MYFREE((char*) cut);
}

int needToCutStep(int bad, char *warnmsg, int *tcut,
		  double *deltat, int verbose)
{
  int maxGlobalCut = *tcut;
  int flag;
  MPI_Status status;
  /* MPI_Request request; */

  /* while (iprobeX(msgtype_TB_timestep)) */
  while(1)
    {
      MPI_Iprobe(MPI_ANY_SOURCE,msgtype_TB_timestep,comm,&flag,&status);
      if(!flag) break;
      {
	int msg[MsgSize];
	/* recvIntX(msgtype_TB_timestep, &msg[0], MsgSize); */
	MPI_Recv((void*)msg,MsgSize,MPI_INT,MPI_ANY_SOURCE,msgtype_TB_timestep,
		 comm,&status);
	if (msgsSent[msg[sender]] < msg[nMsgsSent])
	  msgsSent[msg[sender]] = msg[nMsgsSent];
	++msgsRecvd[msg[sender]];
	if (maxGlobalCut < msg[nCuts])
	  maxGlobalCut = msg[nCuts];
      }
    } 

  if (bad && maxGlobalCut == *tcut)
    {
      int msg[MsgSize];
      int i;
      msg[nCuts] = *tcut+1;
      msg[nMsgsSent] = ++msgsSent[me];
      msg[sender] = me;
      /* sendIntX(msgtype_TB_timestep, &msg[0], MsgSize, -1); */
      for(i=0; i<nProcs; i++) {
	if(i!=me)
	  /* MPI_Bsend((void*)msg,MsgSize,MPI_INT,i,msgtype_TB_timestep,
		    comm,request); */
	  MPI_Send((void*)msg,MsgSize,MPI_INT,i,msgtype_TB_timestep,comm);
      }
      ++msgsRecvd[me];
      ++maxGlobalCut;
      if (verbose)
	{
	  fprintf(stderr, warnmsg);
	  fprintf(stderr, "\nCutting time-step Deltat : %f\n", 0.5* *deltat);
	}
    }
  bad = (*tcut < maxGlobalCut) ? 1 : 0;
    
  /*
  if (maxGlobalCut > maxtcut)
    {
      if (n_domSum == 0)
      fprintf(stderr,
	      "\nFATAL TRANSPORT ERROR: Cutting time-step too many times\n");
      exit(-1);
    }
  */

  while (*tcut < maxGlobalCut)
    {
      *deltat *= 0.5;
      ++(*tcut);
    }
  return bad;
}

void setup_collaboration()
{
  int i;
  for (i = 0; i < nProcs; ++i)
    {
      cut[i] = -1;
      msgsRecvd[i] = 0;
      msgsSent[i] = 0;
    }
}

bool moreCuttingSomewhere(int* tcut, double* deltat)
{
  int i,flag;
  int msg[MsgSize];
  MPI_Status status;
  /* MPI_Request request; */

  msg[nCuts] = *tcut;
  msg[nMsgsSent] = ++msgsSent[me];
  msg[sender] = me;
  /* sendIntX(msgtype_TB_sync, &msg[0], MsgSize, -1); */
  for(i=0; i<nProcs; i++) {
    if(i!=me)
      /* MPI_Bsend((void*)msg,MsgSize,MPI_INT,i,msgtype_TB_sync,comm,request); */
      MPI_Send((void*)msg,MsgSize,MPI_INT,i,msgtype_TB_sync,comm);
  }
  ++msgsRecvd[me];
  cut[me] = *tcut;

  i = 0;
  while (i < nProcs)
    {
      if (cut[i] == *tcut)
	++i;
      else
	{
	  int msg[MsgSize];
	  int msgtype;
	  /* int msgtype = (iprobeX(msgtype_TB_timestep))?
	     msgtype_TB_timestep: msgtype_TB_sync; */
	  MPI_Iprobe(MPI_ANY_SOURCE,msgtype_TB_timestep,comm,&flag,&status);
	  msgtype = (flag)? msgtype_TB_timestep: msgtype_TB_sync;

	  /* recvIntX(msgtype, &msg[0], MsgSize); */
	  MPI_Recv((void*)msg,MsgSize,MPI_INT,MPI_ANY_SOURCE,msgtype,
		   comm,&status);
	  if (msgsSent[msg[sender]] < msg[nMsgsSent])
	    msgsSent[msg[sender]] = msg[nMsgsSent];
	  ++msgsRecvd[msg[sender]];
	  if (msg[nCuts] > cut[msg[sender]])
	    cut[msg[sender]] = msg[nCuts];

	  if (msg[nCuts] > cut[me])
	    {
	      while (*tcut < msg[nCuts])
		{
		  *deltat *= 0.5;
		  ++(*tcut);
		}
	      return TRUE;
	    }
	}
    }
  /* consume possible additional messages */
  i = 0;
  while (i < nProcs)
    if (msgsRecvd[i] == msgsSent[i])
      ++i;
    else
      {
	int msg[MsgSize];
	int type[2];
	int msgIndex = 0;
	type[0] = msgtype_TB_timestep;
	type[1] = msgtype_TB_sync;
	/* while (!iprobeX(type[msgIndex])) */
	while(1) {
	  MPI_Iprobe(MPI_ANY_SOURCE,type[msgIndex],comm,&flag,&status);
          if(flag) break;
	  msgIndex ^= 1;
	}
	/* recvIntX(type[msgIndex], &msg[0], MsgSize); */
	MPI_Recv((void*)msg,MsgSize,MPI_INT,MPI_ANY_SOURCE,type[msgIndex],
		 comm,&status);
	++msgsRecvd[msg[sender]];
      }
  return FALSE;
}
