#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "../Main/ftn.h"
#include "../Main/cglobals.h"
#include "../Main/bool.h"
#include "../Main/float_types.h"
#include "../Main/mymem.h"
#include "../Main/mybits.h"
#include "../Main/msg.h"

bool moreCuttingSomewhere(int* tcut, double* deltat);
bool needToCutStep(int bad, char *warnmsg, int *tcut,
		   double *deltat, int verbose);

typedef double vector3[3];
#define between(a,b,c) ((a)<=(b) && (b)<=(c))
#define strictlyBetween(a,b,c) ((a)<(b) && (b)<(c))

extern bool tracevel ();
extern void setup_collaboration();
extern void inoutletvel();
extern void EvalTotCoefInt();
extern void Trilinear();

void cubeint();
void traceview();
double cubevol();

static int nTtypes;
static tpoint_t *tpoints;
static bool firstParity;
static int  volrefine[3];
static int rf_l_concSize[3];
static bool steadyFlow;
static int* firstComp;
static int* compList;
static double* compSum;
static int* compId;
static conc_t* sum_cbar;
static double* res;
static int nWells;
static int (*locWell)[2];
static int (*topWell)[2];
static int (*botWell)[2];
static double* wellInj;
static double* wellProd;
static double* wellTotbar;
#define INJECT 0
#define PRODUCT 1

#define TETRA_INT

#ifdef LOBATTO
/* source: Z. Kopal, Numerical Analysis, Wiley, 1955 */
/* Note: the Legendre roots are remapped to the interval [0,1] */
#ifndef TETRA_INT
static double gaussWght[10][10] =
{
  {1.},
  {1./2., 1./2.},
  {1./6., 4./6., 1./6.},
  {1./12., 5./12., 5./12., 1./12.},
  {1./20., 49./180., 64./180., 49./180., 1./20.},
  {1./30., .18923748, .27742919, .27742919, .18923748, 1./30.},
  {1./42., .13841302, .21587269, .24380952, .21587269, .13841302, 1./42.},
  {1./56., .10535211, .17056135, .206122940, .206122940, .17056135, .10535211, 1./56.},
  {1./72., .0827476808, .1372693563, .1732142555, .1857596372, .17321425555, .1372693563, .0827476808, 1./72.},
  {1./90., .0666529954, .1124446710, .1460213418, .1637698806, .1637698806, .1460213418, .1124446710, .0666529954, 1./90.}
};
#endif
static double gaussPt[10][10] =
{
  {.5},
  {0., 1.},
  {0., .5, 1.},
  {0., .276393200, .723606800, 1.},
  {0., .172673165, .5, .827326835, 1.},
  {0., .117472340, .357384240, .642615760, .88252766, 1.},
  {0., .084888050, .265575605, .5, .734424395, .91511195, 1.},
  {0., .064129925, .204149910, .39535039, .60464961, .79585009, .935870075, 1.},
  {0., .0501210023, .16140686025, .3184412681, .5, .6815587319, .83859313975, .9498789977, 1.},
  {0., .0402330459, .13061306745, .2610375251, .41736052115, .58263947885, .7389624749, .86938693255, .9597669541, 1.}
};
#else /* trapezoid */
#ifndef TETRA_INT
static double gaussWght[10][10] =
{
  {1.},
  {1./2., 1./2.},
  {1./4., 1./2., 1./4.},
  {1./6., 1./3., 1./3., 1./6.},
  {1./8., 1./4., 1./4., 1./4., 1./8.},
  {1./10.,1./5., 1./5., 1./5., 1./5., 1./10.},
  {1./12.,1./6., 1./6., 1./6., 1./6., 1./6., 1./12.},
  {1./14.,1./7., 1./7., 1./7., 1./7., 1./7., 1./7., 1./14.},
  {1./16.,1./8., 1./8., 1./8., 1./8., 1./8., 1./8., 1./8., 1./16.},
  {1./18.,1./9., 1./9., 1./9., 1./9., 1./9., 1./9., 1./9., 1./9., 1./18.}
};
#endif
static double gaussPt[10][10] =
{
  {.5},
  {0., 1.},
  {0., 1./2., 1.},
  {0., 1./3., 2./3., 1.},
  {0., 1./4., 2./4., 3./4., 1.},
  {0., 1./5., 2./5., 3./5., 4./5., 1.},
  {0., 1./6., 2./6., 3./6., 4./6., 5./6., 1.},
  {0., 1./7., 2./7., 3./7., 4./7., 5./7., 6./7., 1.},
  {0., 1./8., 2./8., 3./8., 4./8., 5./8., 6./8., 7./8., 1.},
  {0., 1./9., 2./9., 3./9., 4./9., 5./9., 6./9., 7./9., 8./9., 1.}
};
#endif

#define wiIndex(comp,well) ((comp)+(nComps)*((well)-1))

void traceback_setup(int *ttypeId, int nt, int vrefine[3],
                     int (*locW)[2], int (*topW)[2], int (*botW)[2], int nw,
                     double* winj, double* wprod, double* wtotprod,
                     bool steadFlow)
{
  int j;
  int maxNTrComps;
  int s, ci;
  nTtypes = nt;
  nWells = nw;
  locWell = locW;
  wellInj = winj;
  wellProd = wprod;
  wellTotbar = wtotprod;
  topWell = topW;
  botWell = botW;

  for (j = 0; j < 3; ++j)
    {
      volrefine[j] = vrefine[j];
      rf_l_concSize[j] = l_concSize[j] * volrefine[j] + 1;
    }
  firstComp = (int*) MYMALLOC((2+nTtypes)*sizeof(int));
  res = (double*) MYMALLOC(nTtypes*sizeof(double));
  for (s = 0; s <= nTtypes; ++s)
    firstComp[s] = 0;
  for (ci = 0; ci < nComps; ++ci)
    ++firstComp[ttypeId[ci]];
  maxNTrComps = firstComp[0];
  for (s = 1; s <= nTtypes; ++s)
    {
      if (maxNTrComps < firstComp[s])
        maxNTrComps = firstComp[s];
      firstComp[s] += firstComp[s-1];
    }
  firstComp[nTtypes+1] = firstComp[nTtypes];
  compList = (int*) MYMALLOC(firstComp[nTtypes]*sizeof(int));
  compSum = (double*) MYMALLOC(firstComp[nTtypes]*sizeof(*compSum));
  for (ci = nComps; ci-- > 0; )
    compList[--firstComp[ttypeId[ci]]] = ci;
  compId = (int*) MYMALLOC((1+maxNTrComps) * sizeof(int));
  sum_cbar = (conc_t*) MYMALLOC((1+maxNTrComps) * sizeof( *sum_cbar));

  firstParity = (n_dom[0]*volrefine[0] + n_dom[1]*volrefine[1] + n_dom[2]*volrefine[2]) % 2;

  steadyFlow = steadFlow;

/*  printf("maxNTrComps %g, compList=%g %g %g %g %g %g\n", maxNTrComps,compList[0],
         compList[1],compList[2],compList[3],compList[4],compList[5]);
*/
}

int traceTypeComp(int ttype)
{
  return compList[firstComp[ttype]];
}

void tpoint_setup()
{
  tpoints = (tpoint_t*) MYMALLOC
    (3 * nTtypes *
     rf_l_concSize[2] *
     rf_l_concSize[1] *
     rf_l_concSize[0] * sizeof(*tpoints));
}

void tpoint_cleanup()
{
  MYFREE((char*)tpoints);
}

void traceback_cleanup()  
{
  if (steadyFlow)
    tpoint_cleanup();
  MYFREE((char*)res);
  MYFREE((char*)firstComp);
  MYFREE((char*)compList);
  MYFREE((char*)compSum);
  MYFREE((char*)compId);
  MYFREE((char*)sum_cbar);
}

bool isWell(int ck[3])
{
  int w;
  for (w = 0; w < nWells; ++w)
    if (ck[0]-concPad[0]+1 == locWell[w][0] &&
        ck[1]-concPad[1]+1 == locWell[w][1] &&
        (between(topWell[w][INJECT], ck[2]-concPad[2]+1, botWell[w][INJECT]) ||
         between(topWell[w][PRODUCT], ck[2]-concPad[2]+1, botWell[w][PRODUCT])))
      return TRUE;
  return FALSE;
}

bool myWell(int w)
{
  if (between(1, locWell[w][0], l_concSize[0]) &&
      between(1, locWell[w][1], l_concSize[1]))
    return TRUE;
  return FALSE;
}

bool isReinjection(int w)
{
  if ((topWell[w][INJECT] > botWell[w][INJECT]) ||
      (topWell[w][PRODUCT] > botWell[w][PRODUCT]))
    return FALSE;
  return TRUE;
}

bool isSource(vector3 x, vector3 y)
{
  int w;
  for (w = 0; w < nWells; ++w)
    {
      int i = locWell[w][0]+concPad[0]-1;
      int j = locWell[w][1]+concPad[1]-1;
      int kTopI = topWell[w][INJECT]+concPad[2]-1;
      int kBotI1 = botWell[w][INJECT]+concPad[2];
      if (strictlyBetween(pl_concPlane[0][i],
                          x[0], pl_concPlane[0][i+1]) &&
          strictlyBetween(pl_concPlane[1][j],
                          x[1], pl_concPlane[1][j+1]) &&
          between(pl_concPlane[2][kTopI],
                  x[2], pl_concPlane[2][kBotI1]))
        {
/* usual wells
          y[0] = 0.5 * (pl_concPlane[0][i+1] + pl_concPlane[0][i]);
          y[1] = 0.5 * (pl_concPlane[1][j+1] + pl_concPlane[1][j]);
*/
          if (locWell[w][0] == 1 && nbr[0][0][-1] == -1)
            y[0] = pl_concPlane[0][i];
          else if (locWell[w][0] == l_concSize[0] && nbr[0][0][1] == -1)
            y[0] = pl_concPlane[0][i+1];
          else
            y[0] = 0.5 * (pl_concPlane[0][i+1] + pl_concPlane[0][i]);

          if (locWell[w][1] == 1 && nbr[0][-1][0] == -1)
            y[1] = pl_concPlane[1][j];
          else if (locWell[w][1] == l_concSize[1] && nbr[0][1][0] == -1)
            y[1] = pl_concPlane[1][j+1];
          else
            y[1] = 0.5 * (pl_concPlane[1][j+1] + pl_concPlane[1][j]);

          y[2] = x[2];
          return TRUE;
        }
    }
  return FALSE;
}

static bool calcCellComp(int ck[3], csol_t* cellCbar, csol_t* cellTotBar,
			 csol_t* totint, conc_t* conc, conc_t* totcoef,
			 double voltol, conc_t* vol,
			 int verbose, int printtrace,
			 int* tcut, double* deltat)
{
  int tr;
  double orig_vol = ((pl_concPlane[0][ck[0]+1] - pl_concPlane[0][ck[0]]) *
                     (pl_concPlane[1][ck[1]+1] - pl_concPlane[1][ck[1]]) *
                     (pl_concPlane[2][ck[2]+1] - pl_concPlane[2][ck[2]]));
  int offset[3], limit[3];
  {
    int j;
    for (j = 0; j < 3; ++j)
      {
        offset[j] = volrefine[j] * (ck[j] - concPad[j]);
        limit [j] = offset[j] + volrefine[j];
      }
  }
  /*
   *  For each immobile component (tr == 0), integrate the easy
   *  way by looking directly at the average value of conc.
   */
  tr = 0;
  {
    int nTrComps = firstComp[tr+1] - firstComp[tr];
    int ci;
    for (ci = 0; ci < nTrComps; ++ci)
      {
        int c = compList[firstComp[tr] + ci];
        cellCbar[c] += (orig_vol *
                        conc[((1+ck[0])+(2+pl_concSize[0])*
                              ((1+ck[1])+(2+pl_concSize[1])*
                               ((1+ck[2])+(2+pl_concSize[2])*
                                (c))))]);
      }
  }

  for (tr = 1; tr <= nTtypes; ++tr)
    {
      bool iAmTwisted;
      int wellWithCorner = -1;
      bool nearWell;
      float tot_int;
      int ci, subk[3];
      int nTrComps = firstComp[tr+1] - firstComp[tr];

      for (subk[2] = offset[2]; subk[2] <= limit[2]; ++subk[2])
        for (subk[1] = offset[1]; subk[1] <= limit[1]; ++subk[1])
          for (subk[0] = offset[0]; subk[0] <= limit[0]; ++subk[0])
            {
              /*
               *  check all wells to see if, perchance, this point of the twisted
               *  box lies within it
               */
              tpoint_t* tPt = &tpoints[3*(tr-1 + nTtypes *
                                          (subk[0] + rf_l_concSize[0] *
                                           (subk[1] + rf_l_concSize[1] *
                                            (subk[2]))))];
              int w;
              for (w = 0; w < nWells; ++w)
                {
                  int i = locWell[w][0]+concPad[0]-1;
                  int j = locWell[w][1]+concPad[1]-1;
                  int kTopI = topWell[w][INJECT]+concPad[2]-1;
                  int kBotI1 = botWell[w][INJECT]+concPad[2];
                  
                  if (strictlyBetween(pl_concPlane[0][i],
                                      tPt[0], pl_concPlane[0][i+1]) &&
                      strictlyBetween(pl_concPlane[1][j],
                                      tPt[1], pl_concPlane[1][j+1]) &&
                      between(pl_concPlane[2][kTopI],
                              tPt[2], pl_concPlane[2][kBotI1]))
                    {
                      wellWithCorner = w;
                      /* fprintf(stderr,
                         "(%d,%d,%d) wellfind %d at ck (%2d %2d %2d).\n",
                         n_dom[0], n_dom[1], n_dom[2], w, ck[0], ck[1], ck[2]); */
                      break;
                    }
                }
            }
      
      for (ci = 0; ci < nTrComps; ++ci)
        {
          compId[ci] = compList[firstComp[tr] + ci];
          sum_cbar[ci] = 0.;
        }
      sum_cbar[nTrComps] = 0.;
      compId[nTrComps] = tr;

#ifdef TETRA_INT
      for (subk[2] = offset[2]; subk[2] < limit[2]; ++subk[2])
        for (subk[1] = offset[1]; subk[1] < limit[1]; ++subk[1])
          for (subk[0] = offset[0]; subk[0] < limit[0]; ++subk[0])
            {
              int index;
              int sign = ((subk[0]+subk[1]+subk[2])%2 != firstParity)? -1: 1;
              double box[8][3];
              for (index = 0; index < 8; ++index)
                {
                  tpoint_t* tPt
                    = &tpoints[3*(tr-1 + nTtypes *
                                  (subk[0] + bit(index,0) + rf_l_concSize[0] *
                                   (subk[1] + bit(index,1) + rf_l_concSize[1] *
                                    (subk[2] + bit(index,2)))))];
                  int j;
                  for (j = 0; j < 3; ++j)
                    box[index][j] = tPt[j];
                }
              cubeint(conc, totcoef, sign, sum_cbar, nTrComps+1, box,
                      pl_concPlane, pl_concSize, compId, concPad);
            }
#else
      for (subk[2] = offset[2]; subk[2] <= limit[2]; ++subk[2])
        for (subk[1] = offset[1]; subk[1] <= limit[1]; ++subk[1])
          for (subk[0] = offset[0]; subk[0] <= limit[0]; ++subk[0])
            {
              double box[8], wght[3];
              int cck[3], j;
              tpoint_t* tPt = &tpoints[3*(tr-1 + nTtypes *
                                          (subk[0] + rf_l_concSize[0] *
                                           (subk[1] + rf_l_concSize[1] *
                                            (subk[2]))))];
              for (j = 0; j < 3; ++j)
                {
                  for (cck[j] = 0; cck[j] < pl_concSize[j]; ++cck[j])
                    if (tPt[j] < pl_concPlane[j][cck[j]])
                      break;
                  wght[j] =           (tPt[j] - pl_concPlane[j][cck[j]]) /
                    (pl_concPlane[j][cck[j]+1] - pl_concPlane[j][cck[j]]);
                }
              
              for (ci = 0; ci < nTrComps; ++ci)
                {
                  int c = compId[ci];
                  Trilinear(conc, box, &cck[0], &cck[1], &cck[2], &c);
                  for (j = 3; j-- > 0;)
                    {
                      int i;
                      int k = 1<<j;
                      for (i = 0; i < k; ++i)
                        box[i] += wght[j] * (box[i+k] - box[i]);
                    }
                  sum_cbar[ci] +=
                    (
                     gaussWght[volrefine[2]][subk[2]-offset[2]] *
                     gaussWght[volrefine[1]][subk[1]-offset[1]] *
                     gaussWght[volrefine[0]][subk[0]-offset[0]] *
                     box[0]);
                }
              Trilinear(totcoef, box, &cck[0], &cck[1], &cck[2], &tr);
              for (j = 3; j-- > 0;)
                {
                  int i;
                  int k = 1<<j;
                  for (i = 0; i < k; ++i)
                    box[i] += wght[j] * (box[i+k] - box[i]);
                }
              sum_cbar[nTrComps] +=
                (
                 gaussWght[volrefine[2]][subk[2]-offset[2]] *
                 gaussWght[volrefine[1]][subk[1]-offset[1]] *
                 gaussWght[volrefine[0]][subk[0]-offset[0]] *
                 box[0]);
            }
#endif
      tot_int = totint[ck[0]-concPad[0] + l_concSize[0] *
                       (ck[1]-concPad[1] + l_concSize[1] *
                        (ck[2]-concPad[2] + l_concSize[2] *
                         (tr-1)))];
      if (wellWithCorner >= 0)
        for (ci = 0; ci < nTrComps; ++ci)
          {
            int c = compList[firstComp[tr] + ci];
            sum_cbar[ci] += wellInj[wellWithCorner*nComps+c] *
              (tot_int - sum_cbar[nTrComps]);
          }
      cellTotBar[tr-1] += sum_cbar[nTrComps];
      nearWell = (wellWithCorner >= 0) || isWell(ck);
      iAmTwisted = !nearWell &&
        fabs(tot_int - sum_cbar[nTrComps]) >
          voltol * (tot_int>sum_cbar[nTrComps]? tot_int: sum_cbar[nTrComps]);
      if (iAmTwisted)
        fprintf(stderr,
                "twisted: ttype=%d totbar=%g totint=%g voltol=%g node=%d cell=(%d %d %d)\n",
                tr, sum_cbar[nTrComps], tot_int, voltol, nbr[0][0][0],
                ck[0]-concPad[0]+1, ck[1]-concPad[1]+1, ck[2]-concPad[2]+1);

      if (needToCutStep(iAmTwisted, "volume imbalance",
			tcut, deltat, verbose))
	  return FALSE;

      if (printtrace > 0)
        {
          vol[tr-1 + nTtypes *
              (ck[0] - concPad[0] + l_concSize[0]*
               (ck[1] - concPad[1] + l_concSize[1]*
                (ck[2] - concPad[2])))] =
                  (tot_int == 0. || nearWell)?
                    1. : (sum_cbar[nTrComps] / tot_int);
        }
      for (ci = 0; ci < nTrComps; ++ci)
        {
          int c = compList[firstComp[tr] + ci];
          if (sum_cbar[ci] < 0.)
            {
              if (sum_cbar[ci] < -1e-4)
                if (verbose > 1)
                  fprintf(stderr, "cbar %3d %3d %3d %f\n",
                          ck[0], ck[1], ck[2], sum_cbar[ci]);
              sum_cbar[ci] = 0.;
            }
          /* no maximum for now:
             else if (sum_cbar[ci] > 1.)
             {
             if (sum_cbar[ci] - 1. > 1e-4)
             if (verbose > 1)
             fprintf(stderr, "cbar %3d %3d %3d %f\n",
             ck[0], ck[1], ck[2], sum_cbar[ci]);
             sum_cbar[ci] = 1.;
             }
             */
          /*
           *  volume fix that causes mass balance error is to multiply
           *  by tot_int/cellTotBar[tr-1]
           */
	  /*   if (!nearWell)
            if (cellTotBar[tr-1] == 0.)
              sum_cbar[ci] = 0.;
            else
              sum_cbar[ci] *= tot_int / cellTotBar[tr-1];
	      */
          cellCbar[c] += sum_cbar[ci];
        }
    }
  return TRUE;
}
  

int traceBack(csol_t* cbar, csol_t* totint, conc_t* conc, conc_t* totcoef,
	      double voltol, int verbose, int printtrace, int traceunit,
	      int maxtcut, int tstep, double t0, double* deltat, int newDT)
{
  conc_t* vol = 0;
  double t1;
  int tcut = 0;
  int ck[3], offset[3], limit[3];
  int j, base;
  double frac;
 
  setup_collaboration();
  for (j = 0; j < 3; ++j)
    {
      offset[j] = concPad[j]*volrefine[j];
      limit[j] = (pl_concSize[j] - concPad[j])*volrefine[j];
    }
  if (printtrace > 0)
    {
      int i;
      vol = (conc_t*) MYMALLOC
        (nTtypes*l_concSize[2]*l_concSize[1]*l_concSize[0]*sizeof(*vol));
      for (i = 0; i < nTtypes*l_concSize[2]*l_concSize[1]*l_concSize[0]; ++i)
        vol[i] = 1.;
    }
  if (tstep == 0 || newDT || !steadyFlow)
    {
      tpoint_setup();
    restart:
      if (n_domSum == 0)
        printf("  Traceback points\n");
      t1 = t0 + *deltat;
      for (ck[2] = offset[2]; ck[2] <= limit[2]; ++ck[2])
        {
          double startTpt[3];
          base = ck[2]/volrefine[2];
          frac = gaussPt[volrefine[2]][ck[2]%volrefine[2]];
          startTpt[2] = (pl_concPlane[2][base] +
                         frac * (pl_concPlane[2][base+1] - pl_concPlane[2][base]));
          for (ck[1] = offset[1]; ck[1] <= limit[1]; ++ck[1])
            {
              base = ck[1]/volrefine[1];
              frac = gaussPt[volrefine[1]][ck[1]%volrefine[1]];
              startTpt[1] = (pl_concPlane[1][base] +
                             frac * (pl_concPlane[1][base+1] - pl_concPlane[1][base]));
              for (ck[0] = offset[0]; ck[0] <= limit[0]; ++ck[0])
                {
                  base = ck[0]/volrefine[0];
                  frac = gaussPt[volrefine[0]][ck[0]%volrefine[0]];
                  startTpt[0] = (pl_concPlane[0][base] +
                                 frac * (pl_concPlane[0][base+1] - pl_concPlane[0][base]));

                  {
                    bool tooFar =
                      tracevel(&tpoints[3 * nTtypes *
                                        (ck[0]-offset[0] + rf_l_concSize[0] *
                                         (ck[1]-offset[1] + rf_l_concSize[1] *
                                          (ck[2]-offset[2])))],
                               startTpt, t1, t0, nTtypes);
                    if (needToCutStep(tooFar,"traced too far",
                                      &tcut, deltat, verbose))
		      {
			if(tcut > maxtcut) return 1;
			goto restart;
		      }
                  }
                }
            }
        }
    } else {
      t1 = t0 + *deltat;
    }

  for (ck[2] = concPad[2]; ck[2] < l_concSize[2] + concPad[2]; ++ck[2])
    for (ck[1] = concPad[1]; ck[1] < l_concSize[1] + concPad[1]; ++ck[1])
      for (ck[0] = concPad[0]; ck[0] < l_concSize[0] + concPad[0]; ++ck[0])
        {
          int tr;
	  EvalTotCoefInt(&pl_concPlane[0][ck[0]],
			 &pl_concPlane[1][ck[1]],
			 &pl_concPlane[2][ck[2]],
			 &t1, res);
	  for (tr = 1; tr <= nTtypes; ++tr)
	    totint[ck[0]-concPad[0] + l_concSize[0] *
		  (ck[1]-concPad[1] + l_concSize[1] *
		   (ck[2]-concPad[2] + l_concSize[2] *
		    (tr-1)))] = res[tr-1];
        }
        
  /*
   *  compute concentrations for injection wells from the production wells
   */
  if (n_domSum == 0)
    printf("  Production wells\n");
  {
    int w;
    for (w = 0; w < nWells; ++w)
      {
        int tr, c;
        double xsectionArea;
        bool reinject = isReinjection(w);
        if(!myWell(w))
          continue;
        ck[0] = locWell[w][0]+concPad[0]-1;
        ck[1] = locWell[w][1]+concPad[1]-1;
        xsectionArea = ((pl_concPlane[0][ck[0]+1] - pl_concPlane[0][ck[0]]) *
                        (pl_concPlane[1][ck[1]+1] - pl_concPlane[1][ck[1]]));
        for (c = 0; c < nComps; ++c)
          wellProd[w*nComps+c] = 0.;
        for (tr = 0; tr < nTtypes; ++tr)
          wellTotbar[w*nTtypes+tr] = 0.;

        for (ck[2] = topWell[w][PRODUCT]+concPad[2]-1;
	     ck[2] <= botWell[w][PRODUCT]+concPad[2]-1; ++ck[2])
          {
            if (!calcCellComp(ck, &wellProd[w*nComps], &wellTotbar[w*nTtypes],
			      totint, conc, totcoef, voltol, vol,
			      verbose, printtrace, &tcut, deltat)) {
	      if(tcut > maxtcut) return 1;
              goto restart;
	    }
          }

        for (ck[2] = topWell[w][PRODUCT]+concPad[2]-1;
	     ck[2] <= botWell[w][PRODUCT]+concPad[2]-1; ++ck[2])
          {
            double by_orig_vol = 1. /
              (xsectionArea*(pl_concPlane[2][ck[2]+1]-pl_concPlane[2][ck[2]]));
            for (tr = 1; tr <= nTtypes; ++tr)
              {
                int ci;
                int nTrComps = firstComp[tr+1] - firstComp[tr];
                double totvol = totint[ck[0]-concPad[0] + l_concSize[0] *
                                       (ck[1]-concPad[1] + l_concSize[1] *
                                        (ck[2]-concPad[2] + l_concSize[2] *
                                         (tr-1)))];
                double tv, bytotbar = wellTotbar[w*nTtypes+tr-1];
                if (bytotbar != 0.) bytotbar = 1./ bytotbar;
                tv = totvol * by_orig_vol * bytotbar;
                for (ci = 0; ci < nTrComps; ++ci)
                  {
                    int c = compList[firstComp[tr] + ci];
                    conc[((1+ck[0]) + (2+pl_concSize[0]) *
                          ((1+ck[1]) + (2+pl_concSize[1]) *
                           ((1+ck[2]) + (2+pl_concSize[2]) *
                            (c))))]
                              = wellProd[w*nComps+c] * tv;
                  }
              }
          }

        for (ck[2] = topWell[w][INJECT]+concPad[2]-1; ck[2] <= botWell[w][INJECT]+concPad[2]-1; ++ck[2])
          {
            double by_orig_vol = 1. /
              (xsectionArea * (pl_concPlane[2][ck[2]+1] - pl_concPlane[2][ck[2]]));
            for (tr = 1; tr <= nTtypes; ++tr)
              {
                int ci;
                int nTrComps = firstComp[tr+1] - firstComp[tr];
                double tiByV = totint[ck[0]-concPad[0] + l_concSize[0] *
                                      (ck[1]-concPad[1] + l_concSize[1] *
                                       (ck[2]-concPad[2] + l_concSize[2] *
                                        (tr-1)))] * by_orig_vol;
                double totbar = wellTotbar[w*nTtypes+tr-1];
                double bytotbar = totbar;
                if (bytotbar != 0.) bytotbar = 1./ bytotbar;
                for (ci = 0; ci < nTrComps; ++ci)
                  {
                    int c = compList[firstComp[tr] + ci];
                    conc[((1+ck[0]) + (2+pl_concSize[0]) *
                          ((1+ck[1]) + (2+pl_concSize[1]) *
                           ((1+ck[2]) + (2+pl_concSize[2]) *
                            (c))))] = tiByV *
                              ((!reinject || (wellProd[w*nComps+c] < wellInj[c] * totbar)) ?
                               wellInj[c] : wellProd[w*nComps+c] * bytotbar);
                  }
              }
          }
      }
  }
  /*
   *  communicate conc data in padded region to neighbors
   */
  {
    int one = 1;
    int zero = 0;
    int MSG = 72727; /* this needs to be fixed later */
    int monitorunit = 6;
    int ierr;
    int pad0 = concPad[0] + 1;
    int pad1 = concPad[1] + 1;
    int pad2 = concPad[2] + 1;
    double * rwk;
    int size  = (pl_concSize[0] + 2)*(pl_concSize[1] + 2)*pad2;
    int sizeA = (pl_concSize[0] + 2)*(pl_concSize[2] + 2)*pad1;
    int sizeB = (pl_concSize[1] + 2)*(pl_concSize[2] + 2)*pad0;
    int ratio = sizeof(double)/sizeof(conc_t);
    int c;

    size = ( size >= sizeA ? size : sizeA );
    size = ( size >= sizeB ? size : sizeB );
    size += 2;
    size = (size + ratio - 1)/ratio;
    rwk = (double*) MYMALLOC(size*sizeof(double));

    for (c = 0; c < nComps; ++c)
      {
        int i = (2+pl_concSize[0])*(2+pl_concSize[1])*(2+pl_concSize[2])*c;

        /* gsyncX(); */
	MPI_Barrier(comm);
        RPadStencil(&conc[i],&one,
		    &l_concSize[0],&l_concSize[1],&l_concSize[2],
		    &zero,&zero,&zero,&pad0,&pad1,&pad2,
	            &nbr[-1][-1][-1],&MSG,&ierr);
        if (ierr)
          fprintf(stderr,"TRANSPORT ERROR: padding concentrations %d\n",
                  nbr[0][0][0]);
      }
    MYFREE((char*) rwk);
  }

  if (n_domSum == 0)
    printf("  Integrations\n");

  /*
   *  compute twisted volumes for ttype_components and
   *  compute concentrations for components
   */
  {
    csol_t* cellCbar = (csol_t*) MYMALLOC(nComps*sizeof(csol_t));
    csol_t* cellTotbar = (csol_t*) MYMALLOC(nTtypes*sizeof(csol_t));
    for (ck[2] = concPad[2]; ck[2] < l_concSize[2] + concPad[2]; ++ck[2])
      for (ck[1] = concPad[1]; ck[1] < l_concSize[1] + concPad[1]; ++ck[1])
        for (ck[0] = concPad[0]; ck[0] < l_concSize[0] + concPad[0]; ++ck[0])
          {
            int c, tr;
            for (c = 0; c < nComps; ++c)
              cellCbar[c] = 0.;
            for (tr = 0; tr < nTtypes; ++tr)
              cellTotbar[tr] = 0.;
            if (isWell(ck))
              {
                double orig_vol = ((pl_concPlane[0][ck[0]+1] - pl_concPlane[0][ck[0]]) *
                                   (pl_concPlane[1][ck[1]+1] - pl_concPlane[1][ck[1]]) *
                                   (pl_concPlane[2][ck[2]+1] - pl_concPlane[2][ck[2]]));
                for (c = 0; c < nComps; ++c)
                  {
                    cbar[ck[0]-concPad[0] + l_concSize[0] *
                         (ck[1]-concPad[1] + l_concSize[1] *
                          (ck[2]-concPad[2] + l_concSize[2] *
                           (c)))] = orig_vol *
                             conc[((1+ck[0])+(2+pl_concSize[0])*
                                   ((1+ck[1])+(2+pl_concSize[1])*
                                    ((1+ck[2])+(2+pl_concSize[2])*
                                     (c))))];
                  }
              }
            else
              {
                if(!calcCellComp(ck, cellCbar, cellTotbar, totint, conc,
				 totcoef, voltol, vol,
				 verbose, printtrace, &tcut, deltat))
		  {
		    if(tcut > maxtcut) return 1;
		    goto restart;
		  }
                for (c = 0; c < nComps; ++c)
                  cbar[ck[0]-concPad[0] + l_concSize[0] *
                       (ck[1]-concPad[1] + l_concSize[1] *
                        (ck[2]-concPad[2] + l_concSize[2] *
                         (c)))] = cellCbar[c];
              }
          }
    MYFREE((char*) cellCbar);
    MYFREE((char*) cellTotbar);
  }
  if (moreCuttingSomewhere(&tcut, deltat))
    {
      if(tcut > maxtcut) return 1;
      goto restart;
    }
  if ( printtrace != 0 && (tstep == 0 || newDT || !steadyFlow) )
    {
      if(printtrace > 0)
        {
          int nxc = l_concSize[0];
          int nyc = l_concSize[1];
          int nzc = l_concSize[2];
          TraceDraw(tpoints, vol, &printtrace,
                     &nxc, &nyc, &nzc, &nTtypes,
                     &volrefine[0],&volrefine[1],&volrefine[2], &traceunit);
	  /* MYFREE((char*)vol); */
        }
      else
        traceview(tpoints, tstep, nTtypes, volrefine);
    }

  if (!steadyFlow)
    tpoint_cleanup();
  return 0;
}

static int foundConcCell[3];

static bool findConcCell(double sought[3])
{
  int j;
  for (j = 0; j < 3; ++j)
    {
      double* plane = pl_concPlane[j];
      double goal = sought[j];
      int lBound = 0;
      int uBound = pl_concSize[j];
      int k0 = foundConcCell[j]/2;
      int k1 = k0 + 1;
      int step = 1;
      while (goal < plane[k0])
        {
          k0 -= step;
          if (k0 < lBound)
            {
              k0 = lBound;
              break;
            }
          step *= 2;
        }
      step = 1;
      while (plane[k1] <= goal)
        {
          k1 += step;
          if (uBound < k1)
            {
              k1 = uBound;
              break;
            }
          step *= 2;
        }
      while (k0 + 1 != k1)
        {
          int mid = (k0+k1)/2;
          if (goal < plane[mid])
            k1 = mid;
          else
            k0 = mid;
        }
      foundConcCell[j] = 2*k0;
      if (goal >= 0.5*(plane[k0]+plane[k1]))
        ++foundConcCell[j];
      if (foundConcCell[j] == lBound || foundConcCell[j] == 2*uBound-1)
        return FALSE;
    }
  return TRUE;
}

static bool concGridInterp(double xPos, double yPos, double zPos,
			   int nDatasets, conc_t* dataset, double* result)
{
  int i, j, k;
  typedef double datacube[8];
  datacube* dat = (datacube*) MYMALLOC(nDatasets*sizeof(datacube));
  int bb[3];
  double baryPos[3];
  double sought[3];
  sought[0] = xPos;
  sought[1] = yPos;
  sought[2] = zPos;
  if (!findConcCell(sought)) return FALSE;
  for (j = 3; j-- > 0;)
    {
      int bigCell = foundConcCell[j] / 2;
      double* plane = pl_concPlane[j];
      bb[j] = (foundConcCell[j]-1)/2;
      baryPos[j] = ((sought[j]        - plane[bigCell]) /
                    (plane[bigCell+1] - plane[bigCell]));
      if (baryPos[j] >= 0.5)
        baryPos[j] = 2*baryPos[j] - 1.;
    }
  
  /* Find data at 8 corners of containing box */
  for (i = 0; i < nDatasets; ++i)
    for (k = 0; k < 8; ++k)
      {
        int index = (1+bb[0]+bit(k,0) + (2+pl_concSize[0]) *
                     (1+bb[1]+bit(k,1) + (2+pl_concSize[1]) *
                      (1+bb[2]+bit(k,2) + (2+pl_concSize[2]) *
                       (i))));
        dat[i][k] = dataset[index];
      }
  
  /* Adjust data for bounding sub-box */
  for (j = 3; j-- > 0;)
    {
      double* plane = pl_concPlane[j];
      double w = ((plane[bb[j]+1] - plane[bb[j]]) /
                  (plane[bb[j]+2] - plane[bb[j]]));
      int kbit = !!(foundConcCell[j]%2)? setbit(0,j): 0;
      int mask = clrbit(0x7, j);
      int k = 0;
      do
        {
          int dest = k | kbit;
          int k2 = setbit(k,j);
          for (i = 0; i < nDatasets; ++i)
            dat[i][dest] = dat[i][k] + w*(dat[i][k2] - dat[i][k]);
        }
      while (!!maskedIncrement(mask,k));
    }

  /* Fold corner sorptions into sorption at point */
  for (j = 3; j-- > 0 ;)
    {
      int gap = 1 << j;
      for (k = 0; k < gap; ++k)
        for (i = 0; i < nDatasets; ++i)
          dat[i][k] += baryPos[j] * (dat[i][k+gap]-dat[i][k]);
    }
  for (i = 0; i < nDatasets; ++i)
    result[i] = dat[i][0];
  MYFREE((char*)dat);
  return TRUE;
}

bool EvalConc(conc_t* conc,
	      double *xPos, double *yPos, double *zPos, double *result)
{
  return concGridInterp(*xPos, *yPos, *zPos, nComps, conc, result);
}
