#include <values.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
/* #include <assert.h> */
#include "../Main/ftn.h"
#include "../Main/float_types.h"
#include "../Main/bool.h"
#include "../Main/mymem.h"
#include "../Main/mymisc.h"

bool EvalTraceVel();
bool SetCell();
double EvalCellFace();
double EvalCellSize();
bool isSource();

#define EPS 0.01

#define getCellSize(x) (EvalCellSize(&x[0], &x[1], &x[2]))
#define constraint(x,i) (EvalCellFace(&x[0],&x[1],&x[2],&i))
#define calc_vel(dx,x,t,tr)						      \
  BEGIN									      \
    int __tr__ = tr+1;							      \
    EvalTraceVel(&__tr__,&x[0],&x[1],&x[2],&t,dx);       		      \
  END
static int zzyzx;
#define nCellFaces(x,t) (SetCell(&x[0], &x[1], &x[2], &t, &zzyzx), zzyzx)
#define step(dst,src1,mpy2,src2)					      \
BEGIN									      \
  int __j__;								      \
  for (__j__ = 0; __j__ < 3; ++__j__)					      \
    dst[__j__] = src1[__j__] + mpy2*src2[__j__];			      \
END

#define assign(dst,src)							      \
BEGIN									      \
  int __j__;								      \
  for (__j__ = 0; __j__ < 3; ++__j__)					      \
    dst[__j__] = src[__j__];						      \
END

#if 1
#define combine3(dst,src1,src2,src3)					      \
BEGIN									      \
  int __j__;								      \
  for (__j__ = 0; __j__ < 3; ++__j__)					      \
    dst[__j__] = (src1[__j__] + 4*src2[__j__] + src3[__j__])/6.;	      \
END
#else
#define combine3(dst,src1,src2,src3)					      \
BEGIN									      \
  int __j__;								      \
  for (__j__ = 0; __j__ < 3; ++__j__)					      \
    dst[__j__] = (src1[__j__] + 2*src2[__j__] + src3[__j__])/4.;	      \
END
#endif

#define lincomb4(dst,src1,wght1,src2,wght2,src3,wght3,src4,wght4)	      \
BEGIN									      \
  int __j__;								      \
  for (__j__ = 0; __j__ < 3; ++__j__)					      \
    dst[__j__] = wght1*src1[__j__] + wght2*src2[__j__] +		      \
                 wght3*src3[__j__] + wght4*src4[__j__];			      \
END

#define lincomb3(dst,src1,wght1,src2,wght2,src3,wght3)			      \
BEGIN									      \
  int __j__;								      \
  for (__j__ = 0; __j__ < 3; ++__j__)					      \
    dst[__j__] = wght1*src1[__j__] + wght2*src2[__j__] +		      \
                 wght3*src3[__j__];					      \
END

#define lincomb2(dst,src1,wght1,src2,wght2)				      \
BEGIN									      \
  int __j__;								      \
  for (__j__ = 0; __j__ < 3; ++__j__)					      \
    dst[__j__] = wght1*src1[__j__] + wght2*src2[__j__];			      \
END

typedef double vector3[3];

static double norm(v0, v1)
     vector3 v0, v1;
{
  int j;
  double sumSq = 0.;
  vector3 diff;
  double diffMax = 0.;
  for (j = 0; j < 3; ++j)
    {
      diff[j] = fabs(v0[j]-v1[j]);
      if (diffMax < diff[j])
	diffMax = diff[j];
    }
  if (diffMax == 0.) return 0.;
  for (j = 0; j < 3; ++j)
    {
      diff[j] /= diffMax;
      sumSq += diff[j]*diff[j];
    }
  return diffMax * sqrt(sumSq);
}

bool
tracevel (endTpt, startTpt, t0, t1, nTtypes)
     tpoint_t (*endTpt)[3];
     vector3 startTpt;
     double t0, t1;
     int nTtypes;
{
  double timeSize = fabs(t1 - t0);
  int tr;
  for (tr = 0; tr < nTtypes; ++tr)
    /* Integrate from startTPt with velocities until
       t moves to t1.  Record final x value.  */
    {
      double t = t0;
      vector3 x, dxdt;
/* */
      assign(x, startTpt);
      /* MPI_Comm_rank(comm,&mynod);
         printf("0. mynod %d, tType %d, starting (%f %f %f)\n",mynod,
	 tr, startTpt[0], startTpt[1], startTpt[2]); */
      
      do { /* for each path segment */
	/* Integrate from x until the point traces
	   across a cell face or consumes all time. */
	int i,j,k;
	int tryDiff;
	double h_est[8];
	vector3* pos_est[8];
	int pos_steps[8];
	double dt;
	int nFaces = nCellFaces(x,t);
	double cellSize = getCellSize(x);

	if (nFaces == 0) return TRUE;

/*
	if (isSource(x,dxdt))
	  {
	    t = t1;
	    / *
	      fprintf(stderr,
	      "Trace point in well: type %d, start (%g %g %g), end (%g %g %g)\n",
	      tr+1, startTpt[0], startTpt[1], startTpt[2], x[0], x[1], x[2]);
	      * /
	    assign(x, dxdt);
	    break;
	  }
*/

	calc_vel(dxdt, x, t, tr);
	/* find the largest stepLen dt such that x+dt*dxdt
	   lies in this cell */
	{
	  vector3 x1;
	  double old_dt = dt = t1 - t;
	  step(x1, x, old_dt, dxdt);
	  /* MPI_Comm_rank(comm,&mynod);
	     printf("1. mynod %d, dt = %g, t1 = %g, t = %g, x = (%g, %g, %g), x1 = (%g, %g, %g)\n",
	     mynod,dt,t1,t, x[0],x[1],x[2],x1[0],x1[1],x1[2]); */
	  for (i = 0; i < nFaces; ++i)
	    {
	      double dist0 = constraint(x, i);
	      double dist1 = constraint(x1, i);
	      double distChange = dist1 - dist0;
	      if (distChange < 0.)
		{
		  double stepLen = -old_dt * dist0 / distChange;
		  if (fabs(dt) > fabs(stepLen))
		    {
		      dt = stepLen;
		      /* MPI_Comm_rank(comm,&mynod);
			 printf("2. mynod %d, steplen1 = %g, dist0 = %g, dist1 = %g\n",
			 mynod,dt,dist0,dist1); */
		    }
		}
	    }
	}

	/* Newton estimates that we can trace forward for time dt until
	   the first of the boundary constraints is satisfied.
	   We are now going to take that big a step (in little pieces),
	   though we'll adjust as we find ourself missing our constraint target. */
	
	tryDiff = 2;
	for (k = 0; ; ++k)
	  {
	    vector3 root_est[8];
	    int nViolations;
	    int nSteps;
	    double next_dt = dt;
/* */
	    if (k == 8)
	      /* This is the eighth iteration.  Rather than go to higher
		 degree extrapolation, we rewrite history to indicate that
		 we've been dividing the big step into 2z, 4z, 6z substeps
		 all along, and not into z, 2z, 3z as we've really been doing.
		 Then we resume regular programming with k == 4. */
	      {
		for (i = 0; i < 4; ++i)
		  {
		    MYFREE((char*)pos_est[2*i]);
		    pos_est[i] = pos_est[2*i+1];
		    pos_steps[i] = pos_steps[2*i+1];
		    h_est[i] = h_est[2*i+1];
		  }
		tryDiff *= 2;
		k = 4;
	      }

	    /* initialization for mod-midpt method for kth iteration */
	    nSteps = tryDiff*(k+1);
	    h_est[k] = dt/nSteps;
	    /* MPI_Comm_rank(comm,&mynod);
	       printf("3. mynod %d, dt = %g, nSteps = %d, k = %d, x = (%g %g %g)\n",
	       mynod,dt,nSteps,k,x[0],x[1],x[2]); */
	    pos_est[k] = (vector3*) MYMALLOC((nSteps+3)*sizeof(vector3));
	    pos_steps[k] = 1;
	    assign(pos_est[k][0], x);
	    step(pos_est[k][1], pos_est[k][0], h_est[k], dxdt);

	    /* apply the mod-midpt method until we can estimate x */
	    /* in a neighborhood of t+dt for all sequences */
	    /* up to the kth.  Usually, only the last iter of this */
	    /* loop does significant work. */
	    for (i = 0; i <= k; ++i)
	      {
		double h = h_est[i];
		double r = dt/h;
		int nSteps = (int)r;
		/* MPI_Comm_rank(comm,&mynod);
		   printf("4. mynod %d, h = %g, dt = %g, r = %g\n",
		   mynod,h,dt,r); */

		/* assert (r>0.); */

		if (pos_steps[i] < nSteps + 2)
		  {
		    double *prev, *curr, *next;
		    pos_est[i] = (vector3*)
		      MYREALLOC((char*)pos_est[i],(nSteps+3)*sizeof(vector3));
		    curr = &pos_est[i][pos_steps[i]-1][0];
		    next = &pos_est[i][pos_steps[i]][0];
		    for (j = pos_steps[i] + 1; j <= nSteps+2; ++j)
		      {
			vector3 vel;
			double tick = t + (j-1)*h;
			prev = curr; curr = next; next = &pos_est[i][j][0];
			calc_vel(vel, curr, tick, tr);
			step(next, prev, 2*h, vel);
			/* what if this mod-midpt try reveals
			   a border crossing that we missed before? */
			if (fabs((j-1)*h) < fabs(next_dt))
			  {
			    int f;
			    for (f = 0; f < nFaces; ++f)
			      {
				float currVal = constraint(curr,f);
				if (currVal < 0.)
				  {
				    float prevVal = constraint(prev,f);
				    if (prevVal >= 0.)
				      next_dt = (j-currVal/(currVal-prevVal))*h;
				  }
			      }
			  }
		      }
		    pos_steps[i] = nSteps+2;
		  }
		/* use cubic spline to est function value at t+dt */
		{
		  double a1 = r - nSteps;
		  double a12 = a1*a1;
		  double a0 = 1.-a1;
		  double w_p2 = a12*a1/6.;
		  double w_p1 = (1.+3.*a1*(1.+a1*a0))/6.;
		  double w_p0 = (4.-3.*a12*(2.-a1))/6.;
		  double w_m1 = a0*a0*a0/6.;
		  if (nSteps > 0)
		    lincomb4(root_est[i],
			     pos_est[i][nSteps+2], w_p2,
			     pos_est[i][nSteps+1], w_p1,
			     pos_est[i][nSteps  ], w_p0,
			     pos_est[i][nSteps-1], w_m1);
		  else
		    {
		      w_p1 -= w_m1;
		      w_p0 += 2 * w_m1;
		      lincomb3(root_est[i],
			       pos_est[i][nSteps+2], w_p2,
			       pos_est[i][nSteps+1], w_p1,
			       pos_est[i][nSteps  ], w_p0);
		    }
		}
	      }

	    if (fabs(next_dt) < fabs(dt))
	      {
		dt = next_dt;
		/* MPI_Comm_rank(comm,&mynod);
		   printf("5. mynod %d, next_dt = %g\n",mynod,dt); */
	      }
	    else
	      {
		/* extrapolate x(t+guess, h) to h=0 using last several values */
		for (i = 0; i < k; ++i)
		  {
		    for (j = 0; i+j < k; ++j)
		      {
			double delta = h_est[i+j+1] - h_est[j];
			double wj =  h_est[i+j+1]/delta;
			double wj1 = 1.-wj;
			if (delta==0.)
			  {
			    fprintf(stderr, "Uh-oh\n");
			  }

			/* assert(delta!=0.); */

			lincomb2(root_est[j],
				 root_est[j],   wj,
				 root_est[j+1], wj1);
		      }
		  }
		/* we think that the curve passes thru root_est[0]; */
		/* Adjust dt by a Newton step from that point so that */
		/* we take the biggest step possible without crossing */
		{
		  vector3 xx;
		  double old_dt = dt;
		  int g = k;
		  double normVal;
		  while (g > 0 &&
			 (normVal = norm(root_est[g], root_est[g-1]),
			  normVal*normVal) < EPS)
		    --g;
		  {
		    double tick = t + dt;
		    vector3 vel;
		    calc_vel(vel, root_est[g], tick, tr);
		    step(xx, root_est[g], -dt, vel);
		  }
	      
		  nViolations = 0;
		  dt = t1 - t;
		  /* MPI_Comm_rank(comm,&mynod);
		     printf("6. mynod %d, dt2 = %g, t1 = %g, t = %g\n",
		     mynod,dt,t1,t); */
		  for (i = 0; i < nFaces; ++i)
		    {
		      double dist0 = constraint(xx, i);
		      double dist1 = constraint(root_est[g], i);
		      double distChange = dist1 - dist0;
		      if (dist1 <= 0.) ++nViolations;
		      if (dist0 >= 0. && distChange < 0.)
			{
			  double stepLen = -old_dt * dist0 / distChange;
			  if (fabs(dt) > fabs(stepLen))
			    {
			      dt = stepLen;
			      /* MPI_Comm_rank(comm,&mynod);
				 printf("7. mynod %d, steplen2 = %g\n",
				 mynod,dt); */
			    }
			}
		    }
		}
		
		if (k > 0 &&
		    (dt == t1 - t || nViolations == 1) &&
		    norm(root_est[1], root_est[0]) < cellSize*EPS)
		  {
		    isSource(x, root_est[0]);
		    assign(x, root_est[0]);
		    /* MPI_Comm_rank(comm,&mynod);
		       if printf("8. mynod %d, Success!\n", mynod); */
		    break;
		  }
	      }
	    if (tryDiff == 64) /* An arbitrary power of two */
	      {
		if (isSource(x, root_est[0]))
		  t = t1;
		/*
		  fprintf(stderr,
		  "Trace point may be inaccurate: type %d, start (%g %g %g), end (%g %g %g)\n",
		  tr+1, startTpt[0], startTpt[1], startTpt[2], x[0], x[1], x[2]);
		  */
		assign(x, root_est[0]);
		break;
	      }
	  }
	for (i = 0; i <= k; ++i)
	  MYFREE((char*)pos_est[i]);
	t += dt;
      } while (fabs(t - t1) > timeSize*EPS);
      assign(endTpt[tr], x);
    }
  return FALSE;
}
