/* #define TEXT_OUTPUT or BINARY_OUTPUT */

#include <stdlib.h>
#include <unistd.h>

#define TEXT_OUTPUT

#ifdef BINARY_OUTPUT
#include <rpc/types.h>
#include <rpc/xdr.h>
#else
#include <sys/types.h>
#endif
#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "../Main/float_types.h"
#include "../Main/cglobals.h"
#include "../Main/mymem.h"
#include "../Main/msg.h"

/* for fseek */
#define BEGINNING	0
#define CURRENT		1
#define END		2

#define HEADER_WIDTH (TEXTFIELD_WIDTH * FIELDS_FOR_HEADER)

#ifdef TEXT_OUTPUT
#define TEXTFIELD_WIDTH 13
#define FIELDS_FOR_HEADER 3

static void print_open(buf,sz)
{}

static void print_close()
{}

static void print_float(buf,offs,val)
     char* buf;
     int offs;
     double val;
{
  sprintf(&buf[TEXTFIELD_WIDTH*offs], "%*.*g\n",
	  TEXTFIELD_WIDTH-1, TEXTFIELD_WIDTH-9, val);
}

static void print_int(buf,val)
     char* buf;
     int val;
{
  sprintf((buf), "%*d\n", TEXTFIELD_WIDTH-1, (val));
}

static void print_code(buf,c)
     char* buf;
     char c;
{
  sprintf(buf, "%-*c\n", HEADER_WIDTH-1, c);
}

static void print_name(buf,name)
     char* buf;
     char* name;
{
  sprintf(buf, "f %*s\n", HEADER_WIDTH - 3, name);
}

#else /* BINARY_OUTPUT */
#define TEXTFIELD_WIDTH 4
#define FIELDS_FOR_HEADER 10

static XDR xdrs;
static void print_open(buf,sz)
{
  xdrmem_create(&xdrs, buf, sz, XDR_ENCODE);
}

static void print_close()
{
  xdr_destroy(&xdrs);
}

static void print_float(buf,offs,val)
     char* buf;
     int offs;
     double val;
{
  float xdr_arg = val;
  xdr_float(&xdrs,&xdr_arg);
}

static void print_int(buf,val)
     char* buf;
     int val;
{
  int xdr_arg = val;
  xdr_int(&xdrs,&xdr_arg);
}

static void print_code(buf,c)
     char* buf;
     char c;
{
  int i;
  buf[0] = 'b';
  buf[1] = c;
  for (i = 2; i < HEADER_WIDTH-1; ++i)
    buf[i] = ' ';
  buf[HEADER_WIDTH-1] = '\n';
}

static void print_name(buf,name)
     char* buf;
     char* name;
{
  int i;
  strcpy(buf, "bf ");
  for (i = strlen(buf); i < HEADER_WIDTH-1; ++i)
    buf[i] = ' ';
  strcpy(&buf[HEADER_WIDTH-1-strlen(name)], name);
  buf[HEADER_WIDTH-1] = '\n';
}


#endif /* TEXT_OUTPUT */

void traceview(tpoints, tstep, nTtypes, volrefine)
tpoint_t* tpoints;
int tstep;
int nTtypes;
int volrefine[3];
{
  int tr, j;
  double* rf_l_concPlane[3];
  int rf_g_concSize[3], rf_g_concStart[3], rf_l_concSize[3];
  for (j = 0; j < 3; ++j)
    {
      int m;
      rf_l_concSize[j] = l_concSize[j]*volrefine[j]+1;
      rf_l_concPlane[j] = (double*) MYMALLOC(rf_l_concSize[j]*sizeof(double));
      rf_g_concSize[j] = g_concSize[j]*volrefine[j]+1;
      rf_g_concStart[j] = g_concStart[j]*volrefine[j];

      for (m = 0; m < l_concSize[j]; ++m)
	{
	  int p;
	  float lo = l_concPlane[j][m];
	  float dd = l_concPlane[j][m+1] - lo;
	  rf_l_concPlane[j][m*volrefine[j]] = lo;
	  for (p = 1; p < volrefine[j]; ++p)
	    rf_l_concPlane[j][m*volrefine[j]+p] = lo + p*dd/volrefine[j];
	}
      rf_l_concPlane[j][l_concSize[j]*volrefine[j]] = l_concPlane[j][l_concSize[j]];
    }

  for (tr = 0; tr < nTtypes; ++tr)
    {
      int bufSize;
      char* buf;
      int headSize = FIELDS_FOR_HEADER;

      int m[3];
      int ofile;
      {
	char filename[80];
	sprintf(filename, "trace.%1d.%04d", tr+1, tstep+1);
	if (n_domSum == 0) unlink(filename);
	/* gsyncX(); */
	MPI_Barrier(comm);
	ofile = creat(filename,0644);
/*	ofile = open(filename,
		     O_WRONLY | O_CREAT,
		     S_IRUSR|S_IWUSR | S_IRGRP|S_IWGRP | S_IROTH);
*/
      }
      
      if (n_domSum == 0)
	{
	  char buf[1+HEADER_WIDTH];
	  print_open(buf,HEADER_WIDTH);
	  print_code(buf,'r');
	  print_close();
	  lseek(ofile, 0, BEGINNING);
	  write(ofile, buf, HEADER_WIDTH);
	  for (j = 0; j < 3; ++j)
	    {
	      print_open(buf,TEXTFIELD_WIDTH);
	      print_int(buf,rf_g_concSize[j]);
	      print_close();
	      write(ofile, buf, TEXTFIELD_WIDTH);
	      lseek(ofile, TEXTFIELD_WIDTH * rf_g_concSize[j], CURRENT);
	    }
	}

      for (j = 0; j < 3; ++j)
	{
	  if (n_domSum == n_dom[j])
	    {
	      int g_index = rf_g_concStart[j] + headSize + 1;
	      int m;
	      bufSize = TEXTFIELD_WIDTH*rf_l_concSize[j];
	      buf = (char*) MYMALLOC(1+bufSize);
	      print_open(buf,bufSize);
	      for (m = 0; m < rf_l_concSize[j]; ++m)
		print_float(buf,m,rf_l_concPlane[j][m]);
	      print_close();
	      lseek(ofile, TEXTFIELD_WIDTH * g_index, BEGINNING);
	      write(ofile, buf, bufSize);
	      MYFREE(buf);
	    }
	  headSize += 1 + rf_g_concSize[j];
	}

      bufSize = 3*TEXTFIELD_WIDTH*rf_l_concSize[0];
      buf = (char*) MYMALLOC(1+bufSize);
      for (m[2] = 0; m[2] < rf_l_concSize[2]; ++m[2])
	for (m[1] = 0; m[1] < rf_l_concSize[1]; ++m[1])
	  {
	    int g_index =
	      (       rf_g_concStart[0] + rf_g_concSize[0]*
	       (m[1] + rf_g_concStart[1] + rf_g_concSize[1]*
		(m[2] + rf_g_concStart[2])));
	    print_open(buf,bufSize);
	    for (m[0] = 0; m[0] < rf_l_concSize[0]; ++m[0])
	      {
		tpoint_t* endTpt =
		  &tpoints[0 + 3*
			   (tr + nTtypes*
			    (m[0] + rf_l_concSize[0]*
			     (m[1] + rf_l_concSize[1]*
			      (m[2]))))];
		for (j = 0; j < 3; ++j)
		  print_float(buf,j+3*m[0], endTpt[j] - rf_l_concPlane[j][m[j]]);
	      }
	    print_close();
	    lseek(ofile, TEXTFIELD_WIDTH * (headSize + 3 * g_index), BEGINNING);
	    write(ofile, buf, bufSize);
	  }
      MYFREE(buf);
      close(ofile);
    }
  for (j = 0; j < 3; ++j)
    MYFREE(rf_l_concPlane[j]);
}
