#ifndef _TRANSPORT_H
#define _TRANSPORT_H

#include "../Main/typeNumbers.h"
#include "../Main/ftn.h"

/* ERROR CODES */
#define ERR_MEMORY 1
#define ERR_UNSUPPORTED_ADVECTION 2
#define ERR_TOO_MANY_DT_CUTS 3
#define ERR_PCG_CONVERGENCE 4
#define ERR_CHEM_EQUILIBRIUM 5
#define ERR_CHEM_MAX_ITER 6
#define ERR_CHEM_MAX_BACKTRACKS 7
#define ERR_CHEM_ILL_CONDITIONED 8
#define ERR_CHEM_INFEASIBLE 9
#define ERR_CHEM_INTEGRATION 10
#define ERR_CHEM 11
#define ERR_IN_CODE 12

/* OUTPUT CODES */
#define OF_ROUTINES 0            /* monitor calling key routines */
#define OF_DELTAT 1              /* monitor deltat cuts */
#define OF_PCG 2                 /* monitor pcg iterations */
#define OF_TRACEBACK_POINTS 3    /* output traceback points (1=draw, 2=eye) */
#define OF_SOLUTION 4            /* debug solution */
#define OF_INITIALIZATION 5      /* debug initialization (calling and set-up)*/
#define OF_TRACEBACK_INTEGRALS 6 /* debug trace-back integrals */
#define OF_REACTIONS 7           /* debug reactions */
#define OF_LINEARSYSTEM 8        /* debug linear system */

/* MESSAGE NUMBERS */
#define MSG_PADCONC1    1
#define MSG_PADCONC2   10
#define MSG_PADCOEF2   20
#define MSG_RXN        30
#define MSG_HOG1       40
#define MSG_HOG2       50
#define MSG_TRACEBCK   60
#define MSG_PADTOTCOEF 70

/* FUNCTION DECLARATIONS */
void OpenTransDebug(int* _myProc, int* _debugUnit);
void CloseTransDebug(int* _debugUnit);
void DerivedSizes(int* _nDispTypes, int* _nTraceTypes, double* _phaseDist,
      int* _nSpecies, int* _sorpType, int* _zeroD);
void InitTrans(double* _time_t, int* _traceType, int* _nTraceTypes,
      int* _dispType, int* _nDispTypes, int* _sorpType, int* _nSorpTypes,
      double* _totCoefInt, float* _totCoef, double* _phaseDist, int* _nSpecies,
      double* _xpComp, double* _ypComp, double* _zpComp,
      int* _nx, int* _ny, int* _nz,
      int* _volRefine_t, int* _steadyFlow, int* _zeroD,
      double* _hx, double* _hy, double* _hz,
      double* _byhx, double* _byhy, double* _byhz,
      double* _byDistx, double* _byDisty, double* _byDistz, double* _padDist,
      int* _nWells, int* _hWellIndex, int* _zWellIndexLo, int* _zWellIndexHi,
      double* _wellInjConc, double* _wellProd, double* _wellTotCoefInt,
      double* _dwk, int* _dimDwk, int* _padx, int* _pady, int* _padz,
      int* _myNbr, int* _error);

void CflTimeStep(double* _cflFactor_t, double* _dt, double* _padDist,
      double* _time_t, int* _nTraceTypes,
      double* _xpComp_0, double* _ypComp_0, double* _zpComp_0,
      int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz);
void RPadStencil(float* _array, int* _stencil, int* _nx, int* _ny, int* _nz,
      int* _ioff, int* _joff, int* _koff, int* _padx1, int* _pady1,int* _padz1,
      int* _myNbr, int* _msg, int* _error);
void PadinFlow(float* _conc, double* _time_t, 
      int* _bcType_t, int* _bcRegion1, int* _bcRegion2, int* _bcRegion3,
      int* _nSpecies, int* _nx, int* _ny, int* _nz,
      int* _padx, int* _pady, int* _padz, int* _myNbr);
void FillWellConc(float* _conc, double* _time_t, int* _nWells,
      int* _wellType, int* _hWellIndex, int* _zWellIndexLo, int* _zWellIndexHi,
      int* _nSpecies, int* _nx, int* _ny, int* _nz,
      int* _padx, int* _pady, int* _padz, int* _myNbr);
double cWellValue(double _time_t, int _specie, int _well);
double pWellValue(double _time_t, int _phase, int _well);
void SetCBar(double* _cbar, float* _conc, int* _nSpecies,
      int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz);
void DefineTotConc(float* _conc, float* _totCoef,
      int* _nx, int* _ny, int* _nz, int* _nSpecies,
      int* _traceType, int* _padx, int* _pady, int* _padz, int* _myNbr);
int traceBack(double* _cbar, double* _totCoefInt,
	      float* _conc, float* _totCoef, double _volTol_t,
	      int _outFlags_t_DELTAT, int _outFlags_t_TRACEBACK_POINTS,
	      int _traceUnit, int _ntCutMax_t,
	      int _step_t_minus_1, double _time_t, double* _dt, int _newDT);
void DefinePrimaryCBar(double* _cbar, float* _totCoef,
		       double* _byhx_0, double* _byhy_0, double* _byhz_0,
		       int* _nx, int* _ny, int* _nz, int* _padxc, int* _padyc,
		       int* _padzc, int* _nSpecies, int* _traceType);
void Hog(double* _cbar, float* _conc, double* _hx, double* _hy, double* _hz,
	 double* _dt, double* _time_t, double* _time_f, int* cflTimeOnly,
	 int* _nWells, int* _wellType, int* _hWellIndex, int* _zWellIndexLo,
	 int* _zWellIndexHi, double* _wellInjConc,
	 int* _padx, int* _pady, int* _padz, int* _nx, int* _ny, int* _nz,
	 int* _nSpecies, double* _phaseDist,
	 int* _myNbr, double* _xpComp, double* _ypComp, double* _zpComp,
	 int* _traceType,int* _sorpType, int* _computeTransport,
         double* _slx_hog, double* _sly_hog, double* _slz_hog,
	 double* _velx_hog, double* _vely_hog, double* _velz_hog,
	 double* _phio_hog, double* _phi_hog,
	 double* _velX, double* _velY, double* _velZ,
	 int* msg, int* _flagWell, int* _monitorRoutines);
void Decay(double* _cbar, double* _dt, int* _iParmRND, double* _rParmRND,
      int* _nx, int* _ny, int* _nz, double* _dwk);
void Reaction(double* _cbar, float* _totCoef, double* _dt, double* _time_t,
      int* _nMicroStepsRXN, double* _rParmRXN, int* _iParmRXN, int* _traceType,
      double* _xpComp_0, double* _ypComp_0, double* _zpComp_0,
      int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz, 
      double* _dwk, double* _dwk_nSpecies, double* _dwk_2_nSpecies,
      int* _outFlags_t_REACTIONS, int* _debugUnit);
void ReactionHeader(int* _step_t, double* _time_t,
      double* _dt, int* _debugUnit);
void DefineTotCoefSimple(float* _totCoef, double* _totCoefInt,
      double* _byhx_0, double* _byhy_0, double* _byhz_0,
      int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz,
      int* _nTraceTypes, int* _myNbr, int* _error);
void DefineTotCoefInt(double* _totCoefInt, double* _time_t,
      double* _xpComp_0, double* _ypComp_0, double* _zpComp_0,
      double* _byhx_0, double* _byhy_0, double* _byhz_0,
      int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz,
      int* _nTraceTypes, double* _dwk);
void RPadExtFace(float* _totCoef_array_type,
      int* _stencil, int* _nx, int* _ny, int* _nz,
      int* _padx, int* _pady, int* _padz, int* _myNbr, int* _off1, int* _off2);
void AdvanceCBar(double* _cbar_array_specie, float* _conc_array_specie,
     int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz);
void Disp(double* _d11, double* _d12, double* _d13,
      double* _d22, double* _d23, double* _d33,
      double* _xpComp_0, double* _ypComp_0, double* _zpComp_0, double* _time_t,
      double* _molDiff, double* _longDisp, double* _transDisp,
      double* _phaseDist, int* _nSpecies, int* _specie1, float* _effIdentity,
      double* _dwk, double* _dwk_3_N_FLOWING_PHASES,
      double* _dwk_4_N_FLOWING_PHASES,
      int* _nx, int* _ny, int* _nz,
      int* _outFlags_t_LINEARSYSTEM, int* _debugUnit);
void SetupBCFlux(double* _face10_array_index, double* _face20_array_index,
      double* _face30_array_index,
      double* _xpComp_0, double* _ypComp_0, double* _zpComp_0,
      int* _nx, int* _ny, int* _nz,
      int* _bcRegion1, int* _bcRegion2, int* _bcRegion3, int* _bcType_t,
      int* _specie1, double* _time_t);
void CSetupMat(double* _cmatrix_array_index, int* _nx, int* _ny, int* _nz,
      double* _d11, double* _d12, double* _d13,
      double* _d22, double* _d23, double* _d33,
      double* _hx_0, double* _hy_0, double* _hz_0,
      double* _byDistx_0, double* _byDisty_0, double* _byDistz_0,
      double* _cdmatrix_array_index,
      double* _face10_array_index, double* _face11_array_index,
      double* _face20_array_index, double* _face21_array_index,
      double* _face30_array_index, double* _face31_array_index,
      int* _bcRegion1, int* _bcRegion2, int* _bcRegion3, int* _bcType_t,
      int* _outFlags_t_LINEARSYSTEM, int* _debugUnit, int* _myNbr);
void CSetupRHS(double* _cmatrix_array_index, int* _nx, int* _ny, int* _nz,
      double* _hx_0, double* _hy_0, double* _hz_0,
      double* _face10_array_index, double* _face11_array_index,
      double* _face20_array_index, double* _face21_array_index,
      double* _face30_array_index, double* _face31_array_index,
      double* _cdmatrix_array_index, double* _crhs, double* _csol,
      float* _conc_array_specie, int* _padx, int* _pady, int* _padz,
      int* _bcRegion1, int* _bcRegion2, int* _bcRegion3, int* _bcType_t,
      double* _cbar_array_specie,
      float* _totCoef_array_traceType_specie, double* _byDt);
void Pcg(double* _cmatrix_array_index, double* _csol, double* _crhs,
      int* _nx, int* _ny, int* _nz,
      double* _relTol_t, int* _maxIter_t, double* _dwk, double* _dwk_dimGrid2,
      double* _dwk_2_dimGrid2, double* _dwk_3_dimGrid2, int* _myNbr,
      int* _outFlags_t_PCG, int* _monitorUnit, int* _error);
void Smooth(double* _csol, int* _nx, int* _ny, int* _nz);
void SetConc(float* _conc_array_specie, double* _csol,
      int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz);
void updateLeaks(float* _conc, float* _totCoef, double* _tc, double* _dt,
		 int* _leakIndexLo, int* _leakIndexHi,
		 int* _traceType, int* _nLeaks, int* _nSpecies,
		 double* _jacInv, int* _nxc, int* _nyc, int* _nzc,
		 int* _padxc, int* _padyc, int* _padzc, int* _myNbr);

void cleanupEvalTraceVel();
void traceCleanup();
void traceback_cleanup();

int reactionStep(double _dt, int loadlim, int EqOnly,
		  int* _cumNtTaken, int* _cumIter, int* _peakIter,
		  int* _exceptionLoc);
void Conversion(double* _cbar, float* _conc,
      double* _xpComp_0, double* _ypComp_0, double* _zpComp_0,
      int* _nx, int* _ny, int* _nz, int* _padx, int* _pady, int* _padz,
      int* _nSpecies, double* _time_t, int* _directionFlag);

#endif
