#ifndef _TRANSPORTGLOBALS_H
#define _TRANSPORTGLOBALS_H

/* TRANSPORT GLOBAL VARIABLES */

struct transportGlobals {

  /* MISC */
  int monitorUnit;
  int traceUnit;
  int debugUnit;
#define monitorUnit   TRN.monitorUnit
#define traceUnit   TRN.traceUnit
#define debugUnit   TRN.debugUnit

  double padDist;
#define padDist   TRN.padDist

  int nDispTypes;
  /*  int* traceType; */
  int* dispType;
#define nDispTypes   TRN.nDispTypes
  /*  #define traceType   TRN.traceType */
#define dispType   TRN.dispType

  double* totCoefInt;
#define totCoefInt   TRN.totCoefInt
#define totCoefInt_(i,j,k,trace)   totCoefInt[i + nx*(j + ny*(k + nz*(trace)))]
#define totCoefInt_array_(trace)   totCoefInt[nx*ny*nz*(trace)]

#ifdef hz
#undef hz
#endif
  /* Grid */
  double* hx;
  double* hy;
  double* hz;

#define hx   TRN.hx
#define hy   TRN.hy
#define hz   TRN.hz
#define hx_(i)   hx[i+padx+1] /* (-padx-1:nx+padx) */
#define hy_(j)   hy[j+pady+1] /* (-pady-1:ny+pady) */
#define hz_(k)   hz[k+padz+1] /* (-padz-1:nz+padz) */

  double* byhx;
  double* byhy;
  double* byhz;
#define byhx   TRN.byhx
#define byhy   TRN.byhy
#define byhz   TRN.byhz
#define byhx_(i)   byhx[i+padx] /* (-padx:nx+padx-1) */
#define byhy_(j)   byhy[j+pady] /* (-pady:ny+pady-1) */
#define byhz_(k)   byhz[k+padz] /* (-padz:nz+padz-1) */
  double* byDistx;
  double* byDisty;
  double* byDistz;
#define byDistx   TRN.byDistx
#define byDisty   TRN.byDisty
#define byDistz   TRN.byDistz
#define byDistx_(i)   byDistx[i+padx] /* (-padx:nx+padx) */
#define byDisty_(j)   byDisty[j+pady] /* (-pady:ny+pady) */
#define byDistz_(k)   byDistz[k+padz] /* (-padz:nz+padz) */

  /* Diffusion */
  int zeroD;
#define zeroD   TRN.zeroD
  double* cmatrix;
  double* cdmatrix;
#define cmatrix   TRN.cmatrix
#define cdmatrix   TRN.cdmatrix
#define cmatrix_(i,j,k,band,n) \
  cmatrix[i+1 + (nx+2)*(j+1 + (ny+2)*(k+1 + (nz+2)*(band + 10*(n))))]
#define cdmatrix_(i,j,k,n)   cdmatrix[i + nx*(j + ny*(k + nz*(n)))]
#define cmatrix_array_(n) cmatrix[(nx+2)*(ny+2)*(nz+2)*10*(n)]
#define cdmatrix_array_(n)   cdmatrix[nx*ny*nz*(n)]
  double* face10;
  double* face11;
  double* face20;
  double* face21;
  double* face30;
  double* face31;
#define face10   TRN.face10
#define face11   TRN.face11
#define face20   TRN.face20
#define face21   TRN.face21
#define face30   TRN.face30
#define face31   TRN.face31
#define face10_(j,k,m,n)   face10[j + ny*(k + nz*(m + 5*(n)))]
#define face11_(j,k,m,n)   face11[j + ny*(k + nz*(m + 5*(n)))]
#define face20_(i,k,m,n)   face20[i + nx*(k + nz*(m + 5*(n)))]
#define face21_(i,k,m,n)   face21[i + nx*(k + nz*(m + 5*(n)))]
#define face30_(i,j,m,n)   face30[i + nx*(j + ny*(m + 5*(n)))]
#define face31_(i,j,m,n)   face31[i + nx*(j + ny*(m + 5*(n)))]
#define face10_array_(n)   face10[ny*nz*5*(n)]
#define face11_array_(n)   face11[ny*nz*5*(n)]
#define face20_array_(n)   face20[nx*nz*5*(n)]
#define face21_array_(n)   face21[nx*nz*5*(n)]
#define face30_array_(n)   face30[nx*ny*5*(n)]
#define face31_array_(n)   face31[nx*ny*5*(n)]
  double* d11;
  double* d12;
  double* d13;
  double* d22;
  double* d23;
  double* d33;
#define d11   TRN.d11
#define d12   TRN.d12
#define d13   TRN.d13
#define d22   TRN.d22
#define d23   TRN.d23
#define d33   TRN.d33

  /* CMM */
  double* csol;
  double* crhs;
  double* cbar;
#define csol   TRN.csol
#define crhs   TRN.crhs
#define cbar   TRN.cbar
#define cbar_(i,j,k,specie)   cbar[(i) + nx*((j) + ny*((k) + nz*(specie)))]
#define cbar_array_(specie)   cbar[nx*ny*nz*(specie)]

  /* HOG */
  double* slx_hog;
  double* sly_hog;
  double* slz_hog;
#define slx_hog   TRN.slx_hog
#define sly_hog   TRN.sly_hog
#define slz_hog   TRN.slz_hog
  double* velx_hog;
  double* vely_hog;
  double* velz_hog;
#define velx_hog   TRN.velx_hog
#define vely_hog   TRN.vely_hog
#define velz_hog   TRN.velz_hog
  double* phio_hog;
  double* phi_hog;
#define phio_hog   TRN.phio_hog
#define phi_hog   TRN.phi_hog

  /* WELLS */
  
  double* wellProd; /* (nSpecies,nWells) */
  double* wellTotCoefInt; /* (nSpecies,nWells) */
  double* wellInjConc; /* (nSpecies,nWells) */
#define wellProd   TRN.wellProd
#define wellTotCoefInt   TRN.wellTotCoefInt
#define wellInjConc   TRN.wellInjConc
#define wellInjConc_(specie,well)   wellInjConc[specie + nSpecies*(well)]
  int* flagWell;
#define flagWell   TRN.flagWell  /* (0:nx+1,0:ny+1,0:nz+1) */
#define flagWell_(i,j,k)   flagWell[i+1 + (nx+2)*(j+1 + (ny+2)*(k+1))]

  /* CHEMISTRY */
  float* locIter;   /* Number of chemistry equilibrium iterations used */
#define locIter TRN.locIter
#define locIter_(i,j,k) locIter[i + nx*(j + ny*(k))]   /*(nx,ny,nz)*/

  int* recv_list;  /* list of procs and # of cells from each */
#define recv_list TRN.recv_list
#define recv_list_(i,p) recv_list[ (i) + 2*(p)]
  int* send_list;  /* list of procs to send work to and # cells */
#define send_list TRN.send_list
#define send_list_(i,p) send_list[ (i) +2*(p)]
  int* cell_list;  /* list of cells to process on this processor */
#define cell_list TRN.cell_list
#define cell_list_(i,j,k) cell_list[(i) + nx*((j) + ny*(k))] 
  double* moles;
#define moles TRN.moles
  double* InitGuess;
#define InitGuess TRN.InitGuess
  double* rates;
#define rates TRN.rates
  double* refrate;
#define refrate TRN.refrate
  double* avgrate;
#define avgrate TRN.avgrate
};

#ifdef TRN_ALLOCATE_MEMORY
struct transportGlobals TRN;
#else
extern struct transportGlobals TRN;
#endif

#endif
