c ...
c ...   Title:  Blas-like routines that work across all nodes 
c ...
c ...   Description:  
c ...   
c ...   Created:        Thu May 28 11:13:51 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 11:13:51 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
C ...  $RCSfile: myblasG.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/myblasG.f,v $
C  
c
c ... Version $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c These routines are a set of BLAS-like routines that work across 
c all the nodes on the data types used frequenty in: 
c $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
c
c ... Routine: dnorm2BdryG
c
c ... File:  myblasG.f
c
c ... Main function:
c     return square of L2-norm of boundary type data across all nodes
c
c ... Include Files: 
c     msgtypes.h mpif.h
c     
c ... Calls: 
c     dnorm2Bdry, gsum0, bcast0
c
c ... In: 
c     n?  : dimensions 
c     h?  : mesh spacings (weights)
c     f?  : Boundary type data
c     msgtype : base message type
c
c ... Out:
c     returns square of L2 norm of f? across all nodes
c
      real*8 function dnorm2BdryG(n1,n2,n3,h1,h2,h3,f1,f2,f3,msgtype)
      implicit none
c
      include '../Include/msgtypes.hf'
      include 'mpif.h'
c
      integer n1,n2,n3,error,msgtype
      real*8 h1(n1),h2(n2),h3(n3)
      real*8 f1(n2,n3,2),f2(n1,n3,2),f3(n1,n2,2)
c
      real*8 dnorm2Bdry,sum
c
      sum = dnorm2Bdry(n1,n2,n3,h1,h2,h3,f1,f2,f3)
c
cmpi      call gdsumx(msgtype,sum,1)
      call MSG_f_Gdsum1(sum,MPI_COMM_WORLD)
c
c ... Summed each edge twice
c
      dnorm2BdryG = sum/2.0d0
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
c
c ... Routine: ddotBdryG
c
c ... File:  myblasG.f
c
c ... Main function:
c     return L2 dot product of boundary type data across all nodes
c
c ... Include Files: 
c     msgtypes.h mpif.h
c     
c ... Calls: 
c     ddotBdry, gsum0, bcast0
c
c ... In: 
c     n?  : dimensions 
c     h?  : mesh spacings (weights)
c     f?  : Boundary type data
c     d?  : Boundary type data
c     msgtype : base message type
c
c ... Out:
c     returns L2 inner-product of f? and d? across all nodes
c
      real*8 function ddotBdryG(n1,n2,n3,h1,h2,h3,f1,f2,f3,d1,d2,d3,
     $     msgtype)
      implicit none
c
      include '../Include/msgtypes.hf'
      include 'mpif.h'
c
      integer n1,n2,n3,error,msgtype
      real*8 h1(n1),h2(n2),h3(n3)
      real*8 f1(n2,n3,2),f2(n1,n3,2),f3(n1,n2,2)
      real*8 d1(n2,n3,2),d2(n1,n3,2),d3(n1,n2,2)
c
      real*8 ddotBdry,sum
c
      sum = ddotBdry(n1,n2,n3,h1,h2,h3,f1,f2,f3,d1,d2,d3)
c
cmpi      call gdsumx(msgtype,sum,1)
      call MSG_f_Gdsum1(sum,MPI_COMM_WORLD)
c
c ... Summed each edge twice
c
      ddotBdryG = sum/2.0d0
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
c
c ... Routine: dsumBdryG
c
c ... File:  myblasG.f
c
c ... Main function:
c     return integral average of boundary type data across all nodes
c
c ... Include Files: 
c     msgtypes.h mpif.h
c     
c ... Calls: 
c     dsumBdry, gsum0, bcast0
c
c ... In: 
c     n?  : dimensions 
c     h?  : mesh spacings (weights)
c     f?  : Boundary type data
c     msgtype : base message type
c
c ... Out:
c     returns integral average of f? across all nodes
c
      real*8 function dsumBdryG(n1,n2,n3,h1,h2,h3,f1,f2,f3,msgtype)
      implicit none
c
      include '../Include/msgtypes.hf'
      include 'mpif.h'
c
      integer n1,n2,n3,error,msgtype
      real*8 h1(n1),h2(n2),h3(n3)
      real*8 f1(n2,n3,2),f2(n1,n3,2),f3(n1,n2,2)
c
      real*8 dsumBdry,sum
c
      sum = dsumBdry(n1,n2,n3,h1,h2,h3,f1,f2,f3)
c
cmpi      call gdsumx(msgtype,sum,1)
      call MSG_f_Gdsum1(sum,MPI_COMM_WORLD)
c
c ... Summed each edge twice
c
      dsumBdryG = sum/2.0d0
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
c
c ... Routine: dsumBdryGNoArea
c
c ... File:  myblasG.f
c
c ... Main function:
c     return integral average of boundary type data across all nodes
c
c ... Include Files: 
c     msgtypes.h mpif.h
c     
c ... Calls: 
c     dsumBdry, gsum0, bcast0
c
c ... In: 
c     n?  : dimensions 
c     h?  : mesh spacings (weights)
c     f?  : Boundary type data
c     msgtype : base message type
c
c ... Out:
c     returns integral average of f? across all nodes
c
      real*8 function dsumBdryGNoArea(n1,n2,n3,f1,f2,f3,msgtype)
      implicit none
c
      include '../Include/msgtypes.hf'
      include 'mpif.h'
c
      integer n1,n2,n3,error,msgtype
      real*8 f1(n2,n3,2),f2(n1,n3,2),f3(n1,n2,2)
c
      real*8 dsumBdryNoArea,sum
c
      sum = dsumBdryNoArea(n1,n2,n3,f1,f2,f3)
c
cmpi      call gdsumx(msgtype,sum,1)
      call MSG_f_Gdsum1(sum,MPI_COMM_WORLD)
c
c ... Summed each edge twice
c
      dsumBdryGNoArea = sum/2.0d0
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: myblasG.f,v 1.3 1997-08-20 11:46:41-05 arbogast Exp $
c
c ... Routine: dsumPresG
c
c ... File:  myblasG.f
c
c ... Main function:
c     return integral average of pressure type data across all nodes
c
c ... Include Files: 
c     msgtypes.h mpif.h
c     
c ... Calls: 
c     dsumPres, gsum0, bcast0
c
c ... In: 
c     n?  : dimensions 
c     hx?  : mesh spacings (weights)
c     pres  : pressure type data
c     msgtype : base message type
c
c ... Out:
c     returns integral average of pres across all nodes
c
      real*8 function dsumPresG(n1,n2,n3,hx1,hx2,hx3,pres,msgtype)
      implicit none
c
      include '../Include/msgtypes.hf'
      include 'mpif.h'
c
      integer n1,n2,n3,error,msgtype
      real*8 hx1(n1),hx2(n2),hx3(n3),pres(n1,n2,n3)
c
      real*8 dsumPres,sum
c
      sum = dsumPres(n1,n2,n3,hx1,hx2,hx3,pres)
c
cmpi      call gdsumx(msgtype,sum,1)
      call MSG_f_Gdsum1(sum,MPI_COMM_WORLD)
c
c ... Summed each edge twice
c
      dsumPresG = sum
c
      return
      end

     
