c ...
c ...   Title:  Interface preconditioner routines
c ...
c ...   Description:  
c ...   
c ...   Created:        Thu May 28 11:17:30 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 11:17:30 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: pc.f,v 1.3 1997-08-20 11:46:43-05 arbogast Exp $
C ...  $RCSfile: pc.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/pc.f,v $
C  
c
c ... Version $Id: pc.f,v 1.3 1997-08-20 11:46:43-05 arbogast Exp $
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c  Routines that deal with the interface preconditioner
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: pc.f,v 1.3 1997-08-20 11:46:43-05 arbogast Exp $
c
c ... Routine: initPC
c
c ... File:  pc.f
c
c ... Main function:
c     Form the preconditioner out of permeability values
c
c ... Include Files: 
c     msgtypes.h
c     
c ... Calls: dinitbdry, copyiface, swapface, davgbdry
c
c ... In:
c     mynod :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     size:      The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     D     :     Cell Tensor
c     w?    :     Bdry type work space
c     wrk,wrk1 :  Work space
c     iwrk  :     size of wrk and wrk1
c
c ... Out:
c     p?:  preconditioner
c
c ... Interal Variables of note:
c
c ... Comments:
c     iws.lt.1 sets the preconditioner as the identity
c
      subroutine initPC(mynod,mynbr,size,n1,n2,n3,D,
     $     hx1,hx2,hx3,p1,p2,p3,w1,w2,w3,iws,wrk,wrk1,avgtype,aflag)
      implicit none
c
      include '../Include/msgtypes.hf'
      include '../Include/BCtypes.hf'
      include '../Include/defines.hf'
c
      integer mynod,mynbr(3,2),size(3,2,2),n1,n2,n3,avgtype
      real*8 D(*),hx1(n1),hx2(n2),hx3(n3)
      real*8 p1(n2,n3,2),p2(n1,n3,2),p3(n1,n2,2)
      real*8 w1(n2,n3,2),w2(n1,n3,2),w3(n1,n2,2)
      integer iws,aflag
      real*8 wrk(iws),wrk1(iws),a
c
      integer i1,i2,i3,iet1,iet2,iet3
c
c ... Internal Common
c
      integer TensorType
      common /parcelcom/ TensorType
c
c ... Load Preconditioner with one-sided n*K*n
c
      if (avgtype.eq.IDENTITY) then
         a = 1
         do i3=1,n3
            do i2=1,n2
               if (aflag.eq.VOLUMETRIC) a = hx2(i2)*hx3(i3)
               p1(i2,i3,1) = a
               p1(i2,i3,2) = a
            enddo
         enddo
c     
         do  i3=1,n3
            do i1=1,n1
               if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx3(i3)
               p2(i1,i3,1) = a
               p2(i1,i3,2) = a
            enddo
         enddo
c     
         do i2=1,n2
            do i1=1,n1
               if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx2(i2)
               p3(i1,i2,1) = a
               p3(i1,i2,2) = a
            enddo
         enddo
      else
         if (TensorType.eq.UserFuncType)  then
            call initPCwithFunction(n1,n2,n3,hx1,hx2,hx3,p1,p2,p3,aflag,
     $           D)
         else if (TensorType.eq.ArrayType) then
            call initPCwithArray(n1,n2,n3,hx1,hx2,hx3,p1,p2,p3,aflag,D)
         else if (TensorType.eq.EdgeTransType)  then
            iet1 = 1
            iet2 = iet1 + (n1+1)*n2*n3
            iet3 = iet2 + n1*(n2+1)*n3
            call initPCwithEdgeTrans(n1,n2,n3,hx1,hx2,hx3,p1,p2,p3,
     $           aflag,D(iet1),D(iet2),D(iet3))
         else
            call ParcelCrash('initPC:Unknown Tensor Type rep.',-30)
         endif

c
c ... Swap k values across interfaces
c
         call copyIface(mynbr,n1,n2,n3,p1,p2,p3,p1,p2,p3)
c     
         call swapface(mynod,mynbr,size,n1,n2,n3,p1,p2,p3,
     $        w1,w2,w3,iws,wrk,wrk1,PCTYPE,aflag)
c     
c     ... Average the left-side and right-side preconditioners
c     
         call davgBdry(avgtype,n1,n2,n3,p1,p2,p3,w1,w2,w3,p1,p2,p3)
c     
c     ... Make sure outer faces are set to one (don't want divide by zero)
c     
         if (mynbr(1,1).eq.-1) then
            call dinitFace(n2,n3,p1(1,1,1),1.0d0)
         endif
c     
         if (mynbr(1,2).eq.-1) then
            call dinitFace(n2,n3,p1(1,1,2),1.0d0)
         endif
c     
         if (mynbr(2,1).eq.-1) then
            call dinitFace(n1,n3,p2(1,1,1),1.0d0)
         endif
c     
         if (mynbr(2,2).eq.-1) then
            call dinitFace(n1,n3,p2(1,1,2),1.0d0)
         endif
c     
         if (mynbr(3,1).eq.-1) then
            call dinitFace(n1,n2,p3(1,1,1),1.0d0)
         endif
c     
         if (mynbr(3,2).eq.-1) then
            call dinitFace(n1,n2,p3(1,1,2),1.0d0)
         endif
c     
      endif
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: pc.f,v 1.3 1997-08-20 11:46:43-05 arbogast Exp $
c
c ... Routine: applyPC
c
c ... File:  pc.f
c
c ... Main function:
c     Apply the preconditioner to boundary values
c
c ... Include Files: none
c     
c ... Calls: none
c
c ... In:
c     n?    :     number of grid points in each direction
c     pc?   :     preconditioner
c     f?    :     Boundary type values
c
c ... Out:
c     pcf?:  preconditioner applied to f?
c
c ... Interal Variables of note:
c
c ... Comments:
c
      subroutine applyPC(n1,n2,n3,pc1,pc2,pc3,f1,f2,f3,pcf1,pcf2,pcf3,
     $     mynbr)
      implicit none
c
      integer n1,n2,n3,mynbr(3,2)
      real*8 pc1(n2,n3,2),pc2(n1,n3,2),pc3(n1,n2,2)
      real*8 f1(n2,n3,2),f2(n1,n3,2),f3(n1,n2,2)
      real*8 pcf1(n2,n3,2),pcf2(n1,n3,2),pcf3(n1,n2,2)
c
      integer i1,i2,i3
c
      do 10 i3=1,n3
         do 11 i2=1,n2
            pcf1(i2,i3,1) = f1(i2,i3,1)/pc1(i2,i3,1)
            pcf1(i2,i3,2) = f1(i2,i3,2)/pc1(i2,i3,2)
 11      continue
 10   continue
c
      do 20 i3=1,n3
         do 21 i1=1,n1
           pcf2(i1,i3,1) = f2(i1,i3,1)/pc2(i1,i3,1)
           pcf2(i1,i3,2) = f2(i1,i3,2)/pc2(i1,i3,2)
 21     continue
 20   continue
c      
      do 30 i2=1,n2
         do 31 i1=1,n1
            pcf3(i1,i2,1) = f3(i1,i2,1)/pc3(i1,i2,1)
            pcf3(i1,i2,2) = f3(i1,i2,2)/pc3(i1,i2,2)
 31      continue
 30   continue
c
      return
      end
c

      subroutine initPCwithFunction(n1,n2,n3,hx1,hx2,hx3,p1,p2,p3,aflag,
     $     userDfcn)
      implicit none
c
      include '../Include/BCtypes.hf'
c
      integer n1,n2,n3,aflag
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 p1(n2,n3,2),p2(n1,n3,2),p3(n1,n2,2)
      real*8 userDfcn
      external userDfcn
c
      integer i1,i2,i3
      real*8 a
c
      a = 1
      do 10 i3=1,n3
         do 11 i2=1,n2
            if (aflag.eq.VOLUMETRIC) a = hx2(i2)*hx3(i3)
            p1(i2,i3,1) = userDfcn(1,1,1,i2,i3)*a
            p1(i2,i3,2) = userDfcn(1,1,n1,i2,i3)*a
 11      continue
 10   continue
c     
      do 20 i3=1,n3
         do 21 i1=1,n1
            if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx3(i3)
            p2(i1,i3,1) = userDfcn(2,2,i1,1,i3)*a
            p2(i1,i3,2) = userDfcn(2,2,i1,n2,i3)*a
 21      continue
 20   continue
c     
      do 30 i2=1,n2
         do 31 i1=1,n1
            if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx2(i2)
            p3(i1,i2,1) = userDfcn(3,3,i1,i2,1)*a
            p3(i1,i2,2) = userDfcn(3,3,i1,i2,n3)*a
 31      continue
 30   continue      
      end


      subroutine initPCwithEdgeTrans(n1,n2,n3,hx1,hx2,hx3,p1,p2,p3,
     $     aflag,k1,k2,k3)
      implicit none
c
      include '../Include/BCtypes.hf'
c
      integer n1,n2,n3,aflag
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 p1(n2,n3,2),p2(n1,n3,2),p3(n1,n2,2)
      real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
c
      integer i1,i2,i3
      real*8 a
c
      a = 1
      do 10 i3=1,n3
         do 11 i2=1,n2
            if (aflag.eq.VOLUMETRIC) a = hx2(i2)*hx3(i3)
            p1(i2,i3,1) = k1(0,i2,i3)*a
            p1(i2,i3,2) = k1(n1,i2,i3)*a
 11      continue
 10   continue
c     
      do 20 i3=1,n3
         do 21 i1=1,n1
            if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx3(i3)
            p2(i1,i3,1) = k2(i1,0,i3)*a
            p2(i1,i3,2) = k2(i1,n2,i3)*a
 21      continue
 20   continue
c     
      do 30 i2=1,n2
         do 31 i1=1,n1
            if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx2(i2)
            p3(i1,i2,1) = k3(i1,i2,0)*a
            p3(i1,i2,2) = k3(i1,i2,n3)*a
 31      continue
 30   continue      
      end
c
c
      subroutine initPCwithArray(n1,n2,n3,hx1,hx2,hx3,p1,p2,p3,aflag,D)
      implicit none
c
      include '../Include/BCtypes.hf'
c
      integer n1, n2, n3, aflag
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 p1(n2,n3,2),p2(n1,n3,2),p3(n1,n2,2)
      real*8 D(6,n1,n2,n3)
c
      integer i, j, k
      real*8 a
c
      a = 1
      do 10 k=1,n3
         do 11 j=1,n2
            if (aflag.eq.VOLUMETRIC) a = hx2(j)*hx3(k)
            p1(j,k,1) = D(1,1,j,k)*a
            p1(j,k,2) = D(1,n1,j,k)*a
 11      continue
 10   continue
c     
      do 20 k=1,n3
         do 21 i=1,n1
            if (aflag.eq.VOLUMETRIC) a = hx1(i)*hx3(k)
            p2(i,k,1) = D(4,i,1,k)*a
            p2(i,k,2) = D(4,i,n2,k)*a
 21      continue
 20   continue
c     
      do 30 j=1,n2
         do 31 i=1,n1
            if (aflag.eq.VOLUMETRIC) a = hx1(i)*hx2(j)
            p3(i,j,1) = D(6,i,j,1)*a
            p3(i,j,2) = D(6,i,j,n3)*a
 31      continue
 30   continue      
      end
