cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: tensUtil.f,v 1.3 1997-08-20 11:46:47-05 arbogast Exp $
c
c ... Routine: pres2velTens
c
c ... File:  tensUtil.f
c
c ... Main function:
c     Calulate boundary flux from solution and boundary conditions
c
c ... Include Files: none
c
c ... Calls: none
c
c ... In:
c     n1,n2,n3 :   Number of Grid divisions
c     bcType   :   Type of boundary conitions: 1 - Neumann,2 - Dirichlet
c     k        :   Permeability
c     hx?      :   Grid spacings
c     u        :   pressures in 3D cell centered
c     bc1,bc2, :   Boundary conditions
c     bc3      :   Surface integral of A grad u.n for Neumann, just u for Dir.
c
c ... Out:
c     bf1,bf2,bf3: Boundary Fluxes  - A grad u.n
c     vel?     :   Velocity in ? direction
c
c ... Comments:
c     Recall: if bcType indicates Neumann bc holds the component 
c     of -A grad u
c
      subroutine pres2velTens(n1,n2,n3,u,bc1,bc2,bc3,hx1,hx2,hx3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,x1,x2,x3,
     $     k11,k12,k13,k22,k23,k33,
     $     bf1,bf2,bf3,vel1,vel2,vel3,aflag)
      implicit none
c
      include '../Include/BCtypes.hf'
c
      integer n1,n2,n3,aflag
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 u(0:n1+1,0:n2+1,0:n3+1)
      real*8 bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      real*8 x1(0:n1),x2(0:n2),x3(0:n3)
      real*8 hx1(0:n1+1),hx2(0:n2+1),hx3(0:n3+1)
      real*8 k11(n1,n2,n3),k12(n1,n2,n3),k13(n1,n2,n3)
      real*8 k22(n1,n2,n3),k23(n1,n2,n3),k33(n1,n2,n3)
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3)
c
      integer i,j,k
      real*8 tld1,tld2,tld3,tld4,tld5,tld6,tld7,tld8,tld9
ccc      logical isNaN
c
c ... compute velocities in the interior of the domain
c
      call makehs(n1,n2,n3,x1,x2,x3,hx1,hx2,hx3)
c
      do k = 1, n3
         do j = 1, n2
            do i = 1, n1-1
               vel1(i,j,k) = 0.25d0*(1./(hx1(i)+hx1(i+1)))

               tld1 = 2.*(u(i,j,k) - u(i+1,j,k))/(hx1(i)+hx1(i+1))  
               tld2 = 2.*(u(i,j-1,k) - u(i,j,k))/(hx2(j-1)+hx2(j))
               tld3 = 2.*(u(i,j,k) - u(i,j+1,k))/(hx2(j)+hx2(j+1))
               tld4 = 2.*(u(i+1,j-1,k) - u(i+1,j,k))/(hx2(j-1)+hx2(j))
               tld5 = 2.*(u(i+1,j,k) - u(i+1,j+1,k))/(hx2(j)+hx2(j+1))
               tld6 = 2.*(u(i,j,k-1) - u(i,j,k))/(hx3(k-1)+hx3(k))
               tld7 = 2.*(u(i,j,k) - u(i,j,k+1))/(hx3(k)+hx3(k+1))
               tld8 = 2.*(u(i+1,j,k-1) - u(i+1,j,k))/(hx3(k-1)+hx3(k))
               tld9 = 2.*(u(i+1,j,k) - u(i+1,j,k+1))/(hx3(k)+hx3(k+1))
c
               vel1(i,j,k) = vel1(i,j,k) * (
     $             4.*(k11(i,j,k)*hx1(i)+k11(i+1,j,k)*hx1(i+1))*tld1
     $           + 2.*k12(i,j,k)*hx1(i)*(tld2+tld3)
     $           + 2.*k12(i+1,j,k)*hx1(i+1)*(tld4+tld5)
     $           + 2.*k13(i,j,k)*hx1(i)*(tld6+tld7)
     $           + 2.*k13(i+1,j,k)*hx1(i+1)*(tld8+tld9)  )
ccc               if(isNaN(vel1(i,j,k))) then
ccc                  write(*,*) "x",i,j,k,vel1(i,j,k)
ccc               endif
c
            enddo
         enddo
      enddo
c
      do k = 1, n3
         do j = 1, n2-1
            do i = 1, n1
               vel2(i,j,k) = 0.25d0*(1./(hx2(j)+hx2(j+1)))

               tld1 = 2.*(u(i,j,k) - u(i,j+1,k))/(hx2(j)+hx2(j+1))  
               tld2 = 2.*(u(i-1,j,k) - u(i,j,k))/(hx1(i-1)+hx1(i))
               tld3 = 2.*(u(i,j,k) - u(i+1,j,k))/(hx1(i)+hx1(i+1))
               tld4 = 2.*(u(i-1,j+1,k) - u(i,j+1,k))/(hx1(i-1)+hx1(i)) 
               tld5 = 2.*(u(i,j+1,k) - u(i+1,j+1,k))/(hx1(i)+hx1(i+1))
               tld6 = 2.*(u(i,j,k-1) - u(i,j,k))/(hx3(k-1)+hx3(k))
               tld7 = 2.*(u(i,j,k) - u(i,j,k+1))/(hx3(k)+hx3(k+1))
               tld8 = 2.*(u(i,j+1,k-1) - u(i,j+1,k))/(hx3(k-1)+hx3(k))
               tld9 = 2.*(u(i,j+1,k) - u(i,j+1,k+1))/(hx3(k)+hx3(k+1))
c
               vel2(i,j,k) = vel2(i,j,k) * ( 
     $             4.*(k22(i,j,k)*hx2(j)+k22(i,j+1,k)*hx2(j+1))*tld1
     $           + 2.*k12(i,j,k)*hx2(j)*(tld2+tld3)
     $           + 2.*k12(i,j+1,k)*hx2(j+1)*(tld4+tld5)
     $           + 2.*k23(i,j,k)*hx2(j)*(tld6+tld7)
     $           + 2.*k23(i,j+1,k)*hx2(j+1)*(tld8+tld9)  )
c
            enddo 
         enddo
      enddo
c
      do k = 1, n3-1
         do j = 1, n2
            do i = 1, n1
               vel3(i,j,k) = 0.25d0*(1./(hx3(k)+hx3(k+1))) 

               tld1 = 2.*(u(i,j,k) - u(i,j,k+1))/(hx3(k)+hx3(k+1))  
               tld2 = 2.*(u(i-1,j,k) - u(i,j,k))/(hx1(i-1)+hx1(i))
               tld3 = 2.*(u(i,j,k) - u(i+1,j,k))/(hx1(i)+hx1(i+1))
               tld4 = 2.*(u(i-1,j,k+1) - u(i,j,k+1))/(hx1(i-1)+hx1(i)) 
               tld5 = 2.*(u(i,j,k+1) - u(i+1,j,k+1))/(hx1(i)+hx1(i+1))
               tld6 = 2.*(u(i,j-1,k) - u(i,j,k))/(hx2(j-1)+hx2(j))
               tld7 = 2.*(u(i,j,k) - u(i,j+1,k))/(hx2(j)+hx2(j+1))
               tld8 = 2.*(u(i,j-1,k+1) - u(i,j,k+1))/(hx2(j-1)+hx2(j))
               tld9 = 2.*(u(i,j,k+1) - u(i,j+1,k+1))/(hx2(j)+hx2(j+1))
c
               vel3(i,j,k) = vel3(i,j,k) * ( 
     $             4.*(k33(i,j,k)*hx3(k)+k33(i,j,k+1)*hx3(k+1))*tld1
     $           + 2.*k13(i,j,k)*hx3(k)*(tld2+tld3)
     $           + 2.*k13(i,j,k+1)*hx3(k+1)*(tld4+tld5)
     $           + 2.*k23(i,j,k)*hx3(k)*(tld6+tld7)
     $           + 2.*k23(i,j,k+1)*hx3(k+1)*(tld8+tld9)  )
c
            enddo 
         enddo
      enddo
c
c
c ... compute flux along edges
c
      call bdryfluxTens(n1,n2,n3,u,bc1,bc2,bc3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     k11,k12,k13,k22,k23,k33,hx1,hx2,hx3,
     $     bf1,bf2,bf3,aflag)
c
c ... Place boundary velocities in velocity arrays
c
      if (aflag.eq.VOLUMETRIC) then
         do  j = 1,n2
            do  k = 1,n3
               vel1(0,j,k) = -bf1(j,k,1)/(hx2(j)*hx3(k))
               vel1(n1,j,k) = bf1(j,k,2)/(hx2(j)*hx3(k))
            enddo
         enddo
      else
         do  j = 1,n2
            do  k = 1,n3
               vel1(0,j,k) = -bf1(j,k,1)
               vel1(n1,j,k) = bf1(j,k,2)
            enddo
         enddo
      endif
c
      if (aflag.eq.VOLUMETRIC) then
         do  i = 1,n1
            do  k = 1,n3
               vel2(i,0,k) = -bf2(i,k,1)/(hx1(i)*hx3(k))
               vel2(i,n2,k) = bf2(i,k,2)/(hx1(i)*hx3(k))
            enddo
         enddo
      else
         do  i = 1,n1
            do  k = 1,n3
               vel2(i,0,k) = -bf2(i,k,1)
               vel2(i,n2,k) = bf2(i,k,2)
            enddo
         enddo
      endif
c    
      if (aflag.eq.VOLUMETRIC) then
         do  i = 1,n1
            do  j = 1,n2
               vel3(i,j,0) = -bf3(i,j,1)/(hx1(i)*hx2(j))
               vel3(i,j,n3) = bf3(i,j,2)/(hx1(i)*hx2(j))
            enddo
         enddo
      else
         do  i = 1,n1
            do  j = 1,n2
               vel3(i,j,0) = -bf3(i,j,1)
               vel3(i,j,n3) = bf3(i,j,2)
            enddo
         enddo
      endif
c
      return
      end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: tensUtil.f,v 1.3 1997-08-20 11:46:47-05 arbogast Exp $
c
c ... Routine: bdryfluxtens
c
c ... File:  util.f
c
c ... Main function:
c     Calulate boundary flux from solution and boundary conditions
c
c ... Include Files: defines.h
c
c ... Calls: none
c
c ... In:
c     n1,n2,n3 :   Number of Grid divisions
c     bcType   :   Type of boundary conitions: 1 - Neumann,2 - Dirichlet
c     k        :   Permeability
c     hx?      :   Grid spacings
c     u        :   pressures in 3D cell centered
c     bc1,bc2, :   Boundary conditions
c     bc3      :   - A grad u.n for Neumann, just u for Dir.
c                  Changed Tue May 26 16:25:19 1992
c     aflag    :   if aflag = VOLUMETRIC then include area factor
c
c ... Out:
c     bf1,bf2,bf3: Boundary Fluxes  - A grad u.n
c
c
c ... Intenal Variables of note:
c
c ... Comments:
c     Recall: if bcType indicates Neumann bc holds the component 
c     of -A grad u
c
      subroutine bdryfluxTens(n1,n2,n3,u,bc1,bc2,bc3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,k11,k12,k13,
     $     k22,k23,k33,hx1,hx2,hx3,
     $     bf1,bf2,bf3,aflag)
      implicit none
      include '../Include/BCtypes.hf'
c
      integer n1,n2,n3,aflag
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 u(0:n1+1,0:n2+1,0:n3+1)
      real*8 bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      real*8 hx1(0:n1+1),hx2(0:n2+1),hx3(0:n3+1)
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 k11(n1,n2,n3),k12(n1,n2,n3),k13(n1,n2,n3)
      real*8 k22(n1,n2,n3),k23(n1,n2,n3),k33(n1,n2,n3)
c
      real*8 a,tld1,tld2,tld3,tld4,tld5,tld6,tld7,tld8,tld9
c
      integer i,j,k,region
c
      real*8 RobinR
      common/Robin/RobinR
c
c left face
c
      do k = 1, n3
         do j = 1, n2
            region = bcRegion1(j,k,1)
            if (bcType(region).eq.BCDIRICHLET) then
               tld1 = 2.*(u(0,j,k) - u(1,j,k))/hx1(1)
               tld4 = 2.*(u(1,j-1,k) - u(1,j,k))/(hx2(j-1)+hx2(j))
               tld5 = 2.*(u(1,j,k) - u(1,j+1,k))/(hx2(j)+hx2(j+1))
               tld8 = 2.*(u(1,j,k-1) - u(1,j,k))/(hx3(k-1)+hx3(k))
               tld9 = 2.*(u(1,j,k) - u(1,j,k+1))/(hx3(k)+hx3(k+1))
c
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx2(j)*hx3(k)
               bf1(j,k,1) = - a*0.25d0*(1./hx1(1))*(
     $             4.*k11(1,j,k)*hx1(1)*tld1
     $           + 2.*k12(1,j,k)*hx1(1)*(tld4+tld5)
     $           + 2.*k13(1,j,k)*hx1(1)*(tld8+tld9)  )
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
               bf1(j,k,1) = a*bc1(j,k,1)
            endif
         enddo
      enddo
c
c right face
c
      do k = 1, n3
         do j = 1, n2
            region = bcRegion1(j,k,2)
            if (bcType(region).eq.BCDIRICHLET) then
               tld1 = 2.*(u(n1,j,k) - u(n1+1,j,k))/hx1(n1)
               tld2 = 2.*(u(n1,j-1,k) - u(n1,j,k))/(hx2(j-1)+hx2(j))
               tld3 = 2.*(u(n1,j,k) - u(n1,j+1,k))/(hx2(j)+hx2(j+1))
               tld6 = 2.*(u(n1,j,k-1) - u(n1,j,k))/(hx3(k-1)+hx3(k))
               tld7 = 2.*(u(n1,j,k) - u(n1,j,k+1))/(hx3(k)+hx3(k+1))
c
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx2(j)*hx3(k)
               bf1(j,k,2) = a*0.25d0*(1./hx1(n1))*(
     $             4.*k11(n1,j,k)*hx1(n1)*tld1
     $           + 2.*k12(n1,j,k)*hx1(n1)*(tld2+tld3)
     $           + 2.*k13(n1,j,k)*hx1(n1)*(tld6+tld7)  )
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
               bf1(j,k,2) = a*bc1(j,k,2)
            endif
         enddo
      enddo
c
c front face
c
      do k = 1, n3
         do i = 1, n1
            region = bcRegion2(i,k,1)
            if (bcType(region).eq.BCDIRICHLET) then
               tld1 = 2.*(u(i,0,k) - u(i,1,k))/hx2(1)
               tld4 = 2.*(u(i-1,1,k) - u(i,1,k))/(hx1(i-1)+hx1(i))
               tld5 = 2.*(u(i,1,k) - u(i+1,1,k))/(hx1(i)+hx1(i+1))
               tld8 = 2.*(u(i,1,k-1) - u(i,1,k))/(hx3(k-1)+hx3(k))
               tld9 = 2.*(u(i,1,k) - u(i,1,k+1))/(hx3(k)+hx3(k+1))
c
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx1(i)*hx3(k)
               bf2(i,k,1) = - a*0.25d0*(1./hx2(1))*(
     $             4.*k22(i,1,k)*hx2(1)*tld1
     $           + 2.*k12(i,1,k)*hx2(1)*(tld4+tld5)
     $           + 2.*k23(i,1,k)*hx2(1)*(tld8+tld9)  )
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
               bf2(i,k,1) = a*bc2(i,k,1)
            endif
         enddo
      enddo
c
c back face
c
      do k = 1, n3
         do i = 1, n1
            region = bcRegion2(i,k,2)
            if (bcType(region).eq.BCDIRICHLET) then
               tld1 = 2.*(u(i,n2,k) - u(i,n2+1,k))/hx2(n2)
               tld2 = 2.*(u(i-1,n2,k) - u(i,n2,k))/(hx1(i-1)+hx1(i))
               tld3 = 2.*(u(i,n2,k) - u(i+1,n2,k))/(hx1(i)+hx1(i+1))
               tld6 = 2.*(u(i,n2,k-1) - u(i,n2,k))/(hx3(k-1)+hx3(k))
               tld7 = 2.*(u(i,n2,k) - u(i,n2,k+1))/(hx3(k)+hx3(k+1))
c
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx1(i)*hx3(k)
               bf2(i,k,2) = a*0.25d0*(1./hx2(n2))*(
     $             4.*k22(i,n2,k)*hx2(n2)*tld1
     $           + 2.*k12(i,n2,k)*hx2(n2)*(tld2+tld3)
     $           + 2.*k23(i,n2,k)*hx2(n2)*(tld6+tld7)  )
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
               bf2(i,k,2) = a*bc2(i,k,2)
            endif
         enddo
      enddo
c
c bottom face
c
      do j = 1, n2
         do i = 1, n1
            region = bcRegion3(i,j,1)
            if (bcType(region).eq.BCDIRICHLET) then
               tld1 = 2.*(u(i,j,0) - u(i,j,1))/hx3(1)
               tld4 = 2.*(u(i-1,j,1) - u(i,j,1))/(hx1(i-1)+hx1(i))
               tld5 = 2.*(u(i,j,1) - u(i+1,j,1))/(hx1(i)+hx1(i+1))
               tld8 = 2.*(u(i,j-1,1) - u(i,j,1))/(hx2(j-1)+hx2(j))
               tld9 = 2.*(u(i,j,1) - u(i,j+1,1))/(hx2(j)+hx2(j+1))
c
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx1(i)*hx2(j)
               bf3(i,j,1) = -a*0.25d0*(1./hx3(1))*(
     $             4.*k33(i,j,1)*hx3(1)*tld1
     $           + 2.*k13(i,j,1)*hx3(1)*(tld4+tld5)
     $           + 2.*k23(i,j,1)*hx3(1)*(tld8+tld9)  )
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
               bf3(i,j,1) = a*bc3(i,j,1)
            endif
         enddo
      enddo
c
c top face
c
      do j = 1, n2
         do i = 1, n1
            region = bcRegion3(i,j,2)
            if (bcType(region).eq.BCDIRICHLET) then
               tld1 = 2.*(u(i,j,n3) - u(i,j,n3+1))/hx3(n3)
               tld2 = 2.*(u(i-1,j,n3) - u(i,j,n3))/(hx1(i-1)+hx1(i))
               tld3 = 2.*(u(i,j,n3) - u(i+1,j,n3))/(hx1(i)+hx1(i+1))
               tld6 = 2.*(u(i,j-1,n3) - u(i,j,n3))/(hx2(j-1)+hx2(j))
               tld7 = 2.*(u(i,j,n3) - u(i,j+1,n3))/(hx2(j)+hx2(j+1))
c
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx1(i)*hx2(j)
               bf3(i,j,2) = a*0.25d0*(1./hx3(n3))*(
     $             4.*k33(i,j,n3)*hx3(n3)*tld1
     $           + 2.*k13(i,j,n3)*hx3(n3)*(tld2+tld3)
     $           + 2.*k23(i,j,n3)*hx3(n3)*(tld6+tld7) )
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
               bf3(i,j,2) = a*bc3(i,j,2)
            endif
         enddo
      enddo
c
      return
      end
