c ...   
c ...   Title:  PARCEL: Utilities for handling the finite elements
c ...   
c ...   Description:  Miscellaneous utilities for extracting information
c ...                 about the finite element representation.
c ...   
c ...   Created:        Mon May 25 13:53:27 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Tue May 26 16:38:36 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
C ...  $RCSfile: util.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/util.f,v $
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
c
c ... Routine: makemesh
c
c ... File:  util.f
c
c ... Main function:
c     Make related meshes from point centered grid
c
c ... Include Files: none
c     
c ... Calls: none
c
c ... In: 
c     n?   : dimensions
c     x?   : point centered grid
c
c ... Out:
c     hx?  : grid spacing x(i+1) - x(i)
c     x?c  : cell centered grid (x(i+1) + x(i)) / 2
c
c ... Interal Variables of note:
c
c ... Comments:
c
      subroutine makemesh(n1,n2,n3,x1,x2,x3,x1c,x2c,x3c,hx1,hx2,hx3)
      implicit none
      integer    n1,n2,n3,i3,i2,i1
      real*8     x1(0:n1),x2(0:n2),x3(0:n3),
     $           x1c(n1),x2c(n2),x3c(n3),
     $           hx1(n1),hx2(n2),hx3(n3)
c
c ... Form Cell-Centered and Mesh Width Grid
c
      do 10 i1=1,n1
        x1c(i1) = 0.5*(x1(i1)+x1(i1-1))
        hx1(i1) = x1(i1)-x1(i1-1)
   10 continue
      do 20 i2=1,n2
        x2c(i2) = 0.5*(x2(i2)+x2(i2-1))
        hx2(i2) = x2(i2)-x2(i2-1)
   20 continue
      do 30 i3=1,n3
        x3c(i3) = 0.5*(x3(i3)+x3(i3-1))
        hx3(i3) = x3(i3)-x3(i3-1)
   30 continue
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
c
c ... Routine: bdryflux123
c
c ... File:  util.f
c
c ... Main function:
c     Calulate boundary flux from solution and boundary conditions
c
c ... Include Files: defines.h
c     
c ... Calls: none
c
c ... In: 
c     n1,n2,n3 :   Number of Grid divisions
c     bcType   :   Type of boundary conitions: 1 - Neumann,2 - Dirichlet
c     k        :   Permeability 
c     hx?      :   Grid spacings
c     u        :   pressures in 3D cell centered
c     bc1,bc2, :   Boundary conditions
c     bc3      :   - A grad u.n for Neumann, just u for Dir. 
c                  Changed Tue May 26 16:25:19 1992
c     aflag    :   if aflag = VOLUMETRIC then include area factor 
c
c ... Out:
c     bf1,bf2,bf3: Boundary Fluxes  - A grad u.n
c
c ... Interal Variables of note:
c
c ... Comments:
c     Recall: if bcType indicates Neumann bc holds the component 
c     of -A grad u
c
      subroutine bdryflux123(n1,n2,n3,u,bc1,bc2,bc3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,k1,k2,k3,hx1,hx2,hx3,
     $     bf1,bf2,bf3,aflag)
      implicit none
      include '../Include/BCtypes.hf'
c
      integer n1,n2,n3,aflag
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 u(n1,n2,n3),bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
c
      real*8 a
c
      integer i1,i2,i3,region
c
      real*8 RobinR
      common/Robin/RobinR
c
      do i3 = 1, n3
         do i2 = 1, n2
            region = bcRegion1(i2,i3,1)
            if (bcType(region).eq.BCDIRICHLET) then
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx2(i2)*hx3(i3)
               bf1(i2,i3,1) = - k1(0,i2,i3)*a*
     $              (u(1,i2,i3)-bc1(i2,i3,1))/(hx1(1)/2.0d0)*(-1.0d0)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
clcc               if (aflag.eq.VOLUMETRIC) a = hx2(i2)*hx3(i3)
c               bf1(i2,i3,1) = -a*bc1(i2,i3,1)/(hx2(i2)*hx3(i3))
               bf1(i2,i3,1) = a*bc1(i2,i3,1)
c$$$            else if (bcType(region).eq.BCROBIN) then
c$$$               bf1(i2,i3,1) = -(bc1(i2,i3,1)-RobinR*u(1,i2,i3))/
c$$$     $              (0.5*RobinR*hx1(1)/k1(0,i2,i3) + 1.0)
            endif
         enddo
      enddo
c
      do i3 = 1, n3
         do i2 = 1, n2
            region = bcRegion1(i2,i3,2)
            if (bcType(region).eq.BCDIRICHLET) then
               a = 1.0d0
               if (aflag.eq.VOLUMETRIC) a = hx2(i2)*hx3(i3)
               bf1(i2,i3,2) = - k1(n1,i2,i3)*a*
     $              (bc1(i2,i3,2)-u(n1,i2,i3))/(hx1(n1)/2.0d0)*(1.0d0)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               a = 1.0d0
clcc               if (aflag.eq.VOLUMETRIC) a = hx2(i2)*hx3(i3)
clcc               bf1(i2,i3,2) = a*bc1(i2,i3,2)/(hx2(i2)*hx3(i3))
               bf1(i2,i3,2) = a*bc1(i2,i3,2)
c$$$            else if (bcType(region).eq.BCROBIN) then
c$$$               bf1(i2,i3,2) = (-bc1(i2,i3,2)+RobinR*u(n1,i2,i3))/
c$$$     $              (0.5*RobinR*hx1(n1)/k1(n1,i2,i3) - 1.0)
            endif
         enddo
      enddo
c
      do i3 = 1, n3
         do i1 = 1, n1
             region = bcRegion2(i1,i3,1)
             if (bcType(region).eq.BCDIRICHLET) then
                a = 1.0d0
                if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx3(i3)
                bf2(i1,i3,1) = - k2(i1,0,i3)*a*
     $               (u(i1,1,i3)-bc2(i1,i3,1))/(hx2(1)/2.0d0)*(-1.0d0)
             else if ((bcType(region).eq.BCNEUMANN)
     $               .or.(bcType(region).eq.BCNOFLOW)) then
                a = 1.0d0
clcc               if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx3(i3)
clcc               bf2(i1,i3,1) = - a*bc2(i1,i3,1)/(hx1(i1)*hx3(i3))
                bf2(i1,i3,1) =  a*bc2(i1,i3,1)
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$                bf2(i1,i3,1) = - (bc2(i1,i3,1)-RobinR*u(i1,1,i3))/
c$$$     $               (0.5*RobinR*hx2(1)/k2(i1,0,i3) + 1)
             endif
          enddo
       enddo
c
       do i3 = 1, n3
          do i1 = 1, n1
             region = bcRegion2(i1,i3,2)
             if (bcType(region).eq.BCDIRICHLET) then
                a = 1.0d0
                if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx3(i3)
                bf2(i1,i3,2) = - k2(i1,n2,i3)*a*
     $               (bc2(i1,i3,2)-u(i1,n2,i3))/(hx2(n2)/2.0d0)*(1.0d0)
             else if ((bcType(region).eq.BCNEUMANN)
     $               .or.(bcType(region).eq.BCNOFLOW)) then
                a = 1.0d0
clcc               if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx3(i3)
clcc               bf2(i1,i3,2) = a*bc2(i1,i3,2)/(hx1(i1)*hx3(i3))
                bf2(i1,i3,2) = a*bc2(i1,i3,2)
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$                bf2(i1,i3,2) = (-bc2(i1,i3,2)+RobinR*u(i1,n2,i3))/
c$$$     $               (0.5*RobinR*hx2(n2)/k2(i1,n2,i3) - 1)
             endif
          enddo
       enddo
c
       do i2 = 1, n2
          do i1 = 1, n1
             region = bcRegion3(i1,i2,1)
             if (bcType(region).eq.BCDIRICHLET) then
                a = 1.0d0
                if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx2(i2)
                bf3(i1,i2,1) = - k3(i1,i2,0)*a*
     $               (u(i1,i2,1)-bc3(i1,i2,1))/(hx3(1)/2.0d0)*(-1.0d0)
             else if ((bcType(region).eq.BCNEUMANN)
     $               .or.(bcType(region).eq.BCNOFLOW)) then
                a = 1.0d0
clcc               if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx2(i2)
clcc               bf3(i1,i2,1) = - a*bc3(i1,i2,1)/(hx1(i1)*hx2(i2))
                bf3(i1,i2,1) = a*bc3(i1,i2,1)
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$                bf3(i1,i2,1) = - (-RobinR*u(i1,i2,1)+bc3(i1,i2,1))/
c$$$     $               (0.5*RobinR*hx3(1)/k3(i1,i2,0) + 1)
             endif
          enddo
       enddo
c
       do i2 = 1, n2
          do i1 = 1, n1
             region = bcRegion3(i1,i2,2)
             if (bcType(region).eq.BCDIRICHLET) then
                a = 1.0d0
                if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx2(i2)
                bf3(i1,i2,2) = - k3(i1,i2,n3)*a*
     $               (bc3(i1,i2,2)-u(i1,i2,n3))/(hx3(n3)/2.0d0)*(1.0d0)
             else if ((bcType(region).eq.BCNEUMANN)
     $               .or.(bcType(region).eq.BCNOFLOW)) then
                a = 1.0d0
clcc               if (aflag.eq.VOLUMETRIC) a = hx1(i1)*hx2(i2)
clcc               bf3(i1,i2,2) = a*bc3(i1,i2,2)/(hx1(i1)*hx2(i2))
                bf3(i1,i2,2) = a*bc3(i1,i2,2)
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$                bf3(i1,i2,2) = (RobinR*u(i1,i2,n3)-bc3(i1,i2,2))/
c$$$     $               (0.5*RobinR*hx3(n3)/k3(i1,i2,n3) - 1)
             endif
          enddo
       enddo
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
c
c ... Routine: bdrypres123
c
c ... File:  util.f
c
c ... Main function:
c     Calulate boundary pressure from solution and boundary conditions
c
c ... Include Files: none
c     
c ... Calls: none
c
c ... In: 
c     n1,n2,n3 :   Number of Grid divisions
c     bcType   :   Type of boundary conitions: 1 - Neumann,2 - Dirichlet
c     k        :   Permeability 
c     hx?      :   Grid spacings
c     u        :   pressures in 3D cell centered
c     bc1,bc2, :   Boundary conditions
c     bc3      :   - A grad u.n for Neumann, just u for Dir.
c                  Changed from surface integral Tue May 26 16:29:24 1992
c
c ... Out:
c     bf1,bf2,bf3: Boundary Pressures: u_bdry
c
c ... Interal Variables of note:
c
c ... Comments:
c     Recall: if bcType indicates Neumann bc holds the component 
c     of -A grad u
c
      subroutine bdrypres123(n1,n2,n3,u,bc1,bc2,bc3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,k1,k2,k3,hx1,hx2,hx3,
     $     bf1,bf2,bf3,aflag)
      implicit none
      include '../Include/BCtypes.hf'
c
      integer n1,n2,n3,aflag
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 u(n1,n2,n3),bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
c
      integer i1,i2,i3,region
c
      real*8 RobinR
      common/Robin/RobinR
c
      do i3 = 1, n3
         do i2 = 1, n2
            region = bcRegion1(i2,i3,1)
            if (bcType(region).eq.BCDIRICHLET) then
               bf1(i2,i3,1) = bc1(i2,i3,1)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               if (aflag.eq.VOLUMETRIC) then
                  bf1(i2,i3,1) = u(1,i2,i3)+
     $                 (-1)*0.5*hx1(1)*bc1(i2,i3,1) /
     $                 (k1(0,i2,i3)*hx2(i2)*hx3(i3))
               else
                  bf1(i2,i3,1) = u(1,i2,i3)+
     $                 (-1)*0.5*hx1(1)*bc1(i2,i3,1) /

     $                 (k1(0,i2,i3))
               endif
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$               bf1(i2,i3,1) = u(1,i2,i3) + 
c$$$     $              0.5*hx1(1)*(bc1(i2,i3,1)-RobinR*u(1,i2,i3))/
c$$$     $              (0.5*RobinR*hx1(1) + k1(0,i2,i3))
            endif
         enddo
      enddo
c
      do i3 = 1, n3
         do i2 = 1, n2
            region = bcRegion1(i2,i3,2)
            if (bcType(region).eq.BCDIRICHLET) then
               bf1(i2,i3,2) = bc1(i2,i3,2)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               if (aflag.eq.VOLUMETRIC) then
                  bf1(i2,i3,2) = u(n1,i2,i3) - 
     $                 0.5*hx1(n1)*bc1(i2,i3,2) /
     $                 (k1(n1,i2,i3)*hx2(i2)*hx3(i3))
               else
                  bf1(i2,i3,2) = u(n1,i2,i3) - 
     $                 0.5*hx1(n1)*bc1(i2,i3,2) /
     $                 (k1(n1,i2,i3))
               endif
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$               bf1(i2,i3,2) = u(n1,i2,i3) - 
c$$$     $              0.5*hx1(n1)*(-bc1(i2,i3,2)+RobinR*u(n1,i2,i3))/
c$$$     $              (0.5*RobinR*hx1(n1) - k1(n1,i2,i3))
            endif
         enddo
      enddo
c
      do i3 = 1, n3
         do i1 = 1, n1
            region = bcRegion2(i1,i3,1)
            if (bcType(region).eq.BCDIRICHLET) then
               bf2(i1,i3,1) =  bc2(i1,i3,1)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               if (aflag.eq.VOLUMETRIC) then
                  bf2(i1,i3,1) = u(i1,1,i3) + 
     $                 (-1)*0.5*hx2(1)*bc2(i1,i3,1)/
     $                 (k2(i1,0,i3)*hx1(i1)*hx3(i3))
               else
                  bf2(i1,i3,1) = u(i1,1,i3) + 
     $                 (-1)*0.5*hx2(1)*bc2(i1,i3,1)/
     $                 (k2(i1,0,i3))
               endif
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$               bf2(i1,i3,1) = u(i1,1,i3) + 0.5*hx2(1)*
c$$$     $              (bc2(i1,i3,1)-RobinR*u(i1,1,i3))/
c$$$     $              (0.5*RobinR*hx2(1) + k2(i1,0,i3))
            endif
         enddo
      enddo
c
      do i3 = 1, n3
         do i1 = 1, n1
            region = bcRegion2(i1,i3,2)
            if (bcType(region).eq.BCDIRICHLET) then
               bf2(i1,i3,2) = bc2(i1,i3,2)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               if (aflag.eq.VOLUMETRIC) then
                  bf2(i1,i3,2) = u(i1,n2,i3) - 
     $                 0.5*hx2(n2)*bc2(i1,i3,2)/
     $                 (k2(i1,n2,i3)*hx1(i1)*hx3(i3))
               else
                  bf2(i1,i3,2) = u(i1,n2,i3) - 
     $                 0.5*hx2(n2)*bc2(i1,i3,2)/
     $                 (k2(i1,n2,i3))
               endif
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$               bf2(i1,i3,2) = u(i1,n2,i3) - 0.5*hx2(n2)*
c$$$     $              (-bc2(i1,i3,2)+RobinR*u(i1,n2,i3))/
c$$$     $              (0.5*RobinR*hx2(n2) - k2(i1,n2,i3) )
            endif
         enddo
      enddo
c
      do i2 = 1, n2
         do i1 = 1, n1
            region = bcRegion3(i1,i2,1)
            if (bcType(region).eq.BCDIRICHLET) then
               bf3(i1,i2,1) = bc3(i1,i2,1)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               if (aflag.eq.VOLUMETRIC) then
                  bf3(i1,i2,1) = u(i1,i2,1) + 
     $                 (-1)*0.5*hx3(1)*bc3(i1,i2,1)/
     $                 (k3(i1,i2,0)*hx1(i1)*hx2(i2))
               else
                  bf3(i1,i2,1) = u(i1,i2,1) + 
     $                 (-1)*0.5*hx3(1)*bc3(i1,i2,1)/
     $                 (k3(i1,i2,0))
               endif
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$               bf3(i1,i2,1) = u(i1,i2,1) +0.5*hx3(1)*
c$$$     $              (-RobinR*u(i1,i2,1)+bc3(i1,i2,1))/
c$$$     $              (0.5*RobinR*hx3(1)+k3(i1,i2,0))
            endif
         enddo
      enddo
c
      do i2 = 1, n2
         do i1 = 1, n1
            region = bcRegion3(i1,i2,2)
            if (bcType(region).eq.BCDIRICHLET) then
               bf3(i1,i2,2) =  bc3(i1,i2,2)
            else if ((bcType(region).eq.BCNEUMANN)
     $              .or.(bcType(region).eq.BCNOFLOW)) then
               if (aflag.eq.VOLUMETRIC) then
                  bf3(i1,i2,2) = u(i1,i2,n3) - 0.5*hx3(n3)*
     $                 bc3(i1,i2,2)/(k3(i1,i2,n3)*hx1(i1)*hx2(i2))
               else
                  bf3(i1,i2,2) = u(i1,i2,n3) - 0.5*hx3(n3)*
     $                 bc3(i1,i2,2)/(k3(i1,i2,n3))
               endif
c$$$             else if (bcType(region).eq.BCROBIN) then
c$$$               bf3(i1,i2,2) = u(i1,i2,n3) - 0.5*hx3(n3)*
c$$$     $              (RobinR*u(i1,i2,n3)-bc3(i1,i2,2))/
c$$$     $              (0.5*RobinR*hx3(n3)-k3(i1,i2,n3))
            endif
         enddo
      enddo
c
      return
      end
c
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
c
c ... Routine: pres2vel123
c
c ... File:  util.f
c
c ... Main function:
c     Calulate boundary flux from solution and boundary conditions
c
c ... Include Files: none
c     
c ... Calls: none
c
c ... In: 
c     n1,n2,n3 :   Number of Grid divisions
c     bcType   :   Type of boundary conitions: 1 - Neumann,2 - Dirichlet
c     k        :   Permeability 
c     hx?      :   Grid spacings
c     u        :   pressures in 3D cell centered
c     bc1,bc2, :   Boundary conditions
c     bc3      :   Surface integral of A grad u.n for Neumann, just u for Dir.
c
c ... Out:
c     bf1,bf2,bf3: Boundary Fluxes  - A grad u.n
c     vel?     :   Velocity in ? direction
c
c ... Interal Variables of note:
c     h(*,*) = harmonic average
c
c ... Comments:
c     Recall: if bcType indicates Neumann bc holds the component 
c     of -A grad u
c
      subroutine pres2vel123(n1,n2,n3,u,bc1,bc2,bc3,
     $     bcType,nBCRegions,bcRegion1,bcRegion2,bcRegion3,
     $     k1,k2,k3,hx1,hx2,hx3,
     $     bf1,bf2,bf3,vel1,vel2,vel3,aflag)
      implicit none
c
      integer n1,n2,n3,aflag
      integer nBCRegions,bcType(nBCRegions),
     $        bcRegion1(n2,n3,2),bcRegion2(n1,n3,2),bcRegion3(n1,n2,2)
      real*8 u(n1,n2,n3),bc1(n2,n3,2),bc2(n1,n3,2),bc3(n1,n2,2)
      real*8 hx1(n1),hx2(n2),hx3(n3)
      real*8 k1(0:n1,n2,n3),k2(n1,0:n2,n3),k3(n1,n2,0:n3)
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3)
c
      integer i1,i2,i3
c
      include '../Include/BCtypes.hf'
c
c ... compute velocities in the interior of the domain
c
      do 13 i3 = 1,n3
         do 12 i2 = 1,n2
            do 11 i1 = 1,n1-1
clcc               vel1(i1,i2,i3) = -h(k(i1,i2,i3,1),k(i1+1,i2,i3,1))*2.0*
               vel1(i1,i2,i3) = -k1(i1,i2,i3)*2.0*
     $              (u(i1+1,i2,i3)-u(i1,i2,i3))/(hx1(i1+1)+hx1(i1))
 11         continue
 12      continue
 13   continue
      do 23 i3 = 1,n3
         do 22 i2 = 1,n2-1
            do 21 i1 = 1,n1
clcc               vel2(i1,i2,i3) = -h(k(i1,i2,i3,2),k(i1,i2+1,i3,2))*2.0*
               vel2(i1,i2,i3) = -k2(i1,i2,i3)*2.0*
     $              (u(i1,i2+1,i3)-u(i1,i2,i3))/(hx2(i2+1)+hx2(i2))
 21         continue
 22      continue
 23   continue
      do 33 i3 = 1,n3-1
         do 32 i2 = 1,n2
            do 31 i1 = 1,n1
clcc               vel3(i1,i2,i3) = -h(k(i1,i2,i3,3),k(i1,i2,i3+1,3))*2.0*
               vel3(i1,i2,i3) = -k3(i1,i2,i3)*2.0*
     $              (u(i1,i2,i3+1)-u(i1,i2,i3))/(hx3(i3+1)+hx3(i3))
 31         continue
 32      continue
 33   continue
c
c ... compute flux along edges
c
      call bdryflux123(n1,n2,n3,u,bc1,bc2,bc3,bcType,nBCRegions,
     $     bcRegion1,bcRegion2,bcRegion3,k1,k2,k3,hx1,hx2,hx3,
     $     bf1,bf2,bf3,aflag)
c
c ... Place boundary velocities in velocity arrays
c
      if (aflag.eq.VOLUMETRIC) then
         do  i2 = 1,n2
            do  i3 = 1,n3
               vel1(0,i2,i3) = -bf1(i2,i3,1)/(hx2(i2)*hx3(i3))
               vel1(n1,i2,i3) = bf1(i2,i3,2)/(hx2(i2)*hx3(i3))
            enddo
         enddo
      else
         do  i2 = 1,n2
            do  i3 = 1,n3
               vel1(0,i2,i3) = -bf1(i2,i3,1)
               vel1(n1,i2,i3) = bf1(i2,i3,2)
            enddo
         enddo
      endif
c
      if (aflag.eq.VOLUMETRIC) then
         do  i1 = 1,n1
            do  i3 = 1,n3
               vel2(i1,0,i3) = -bf2(i1,i3,1)/(hx1(i1)*hx3(i3))
               vel2(i1,n2,i3) = bf2(i1,i3,2)/(hx1(i1)*hx3(i3))
            enddo
         enddo
      else
         do  i1 = 1,n1
            do  i3 = 1,n3
               vel2(i1,0,i3) = -bf2(i1,i3,1)
               vel2(i1,n2,i3) = bf2(i1,i3,2)
            enddo
         enddo
      endif
c     
      if (aflag.eq.VOLUMETRIC) then
         do  i1 = 1,n1
            do  i2 = 1,n2
               vel3(i1,i2,0) = -bf3(i1,i2,1)/(hx1(i1)*hx2(i2))
               vel3(i1,i2,n3) = bf3(i1,i2,2)/(hx1(i1)*hx2(i2))
            enddo
         enddo
      else
         do  i1 = 1,n1
            do  i2 = 1,n2
               vel3(i1,i2,0) = -bf3(i1,i2,1)
               vel3(i1,i2,n3) = bf3(i1,i2,2)
            enddo
         enddo
      endif
c
      return
      end
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
c
c ... Routine: vel2velc
c
c ... File:  util.f
c
c ... Main function:
c     Calulate boundary flux from solution and boundary conditions
c
c ... Include Files: none
c     
c ... Calls: none
c
c ... In: 
c     n1,n2,n3 :   Number of Grid divisions
c     vel?     :   Linear velocities in ? direction
c
c ... Out:
c     velc     :   Cell centered velocities
c
c ... Interal Variables of note:
c
c ... Comments:
c
      subroutine vel2velc(n1,n2,n3,vel1,vel2,vel3,velc)
      implicit none
c
      integer n1,n2,n3
      real*8  vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3)
      real*8  velc(n1,n2,n3,3)
c
      integer i1,i2,i3
c
c ... compute velocities at the cell centers
c
      do 13 i3 = 1,n3
         do 12 i2 = 1,n2
            do 11 i1 = 1,n1
               velc(i1,i2,i3,1) = 0.5*(vel1(i1-1,i2,i3)+vel1(i1,i2,i3))
               velc(i1,i2,i3,2) = 0.5*(vel2(i1,i2-1,i3)+vel2(i1,i2,i3))
               velc(i1,i2,i3,3) = 0.5*(vel3(i1,i2,i3-1)+vel3(i1,i2,i3))
 11         continue
 12      continue
 13   continue
c
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
c
c ... Routine: writePres
c
c ... File:  util.f
c
c ... Main function:
c     Print out a pressure matrix
c
c ... Include Files: none
c     
c ... Calls: none
c
c ... In: 
c     n? :    dimensions of pressure
c     pres :  pressure
c
c ... Out: none
c     
c ... Interal Variables of note: none
c
c ... Comments:
c
      subroutine writePres(outUnit,pres,n1,n2,n3)
      implicit none
      integer n1,n2,n3,outUnit
      real*8 pres(n1,n2,n3)
c
      integer i1,i2,i3
c
      do 10 i3 = 1,n3
         do 20 i2 = 1,n2
            write(outUnit,'(1000e11.3)') 
     $           (pres(i1,i2,i3),i1=1,n1)
 20      continue
 10   continue
      return
      end
c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: util.f,v 1.4 1997-08-20 11:46:48-05 arbogast Exp $
c
c ... Routine: writeVelMag
c
c ... File:  util.f
c
c ... Main function:
c     Print out a Velocity Magnitudes
c
c ... Include Files: none
c     
c ... Calls: none
c
c ... In: 
c     n? :    dimensions of pressure
c     pres :  pressure
c
c ... Out: none
c     
c ... Interal Variables of note: none
c
c ... Comments:
c
      subroutine writeVelMag(outUnit,velc,n1,n2,n3)
      implicit none
      INTRINSIC sqrt
      integer n1,n2,n3,outUnit
      real*8 velc(n1,n2,n3,3)
c
      integer i1,i2,i3
c
      do 10 i3 = 1,n3
         do 20 i2 = 1,n2
cCRAY_OPT   write(outUnit,'(1000e11.3)') 
cCRAY_OPT     $           (sqrt(velc(i1,i2,i3,1)**2+velc(i1,i2,i3,2)**2+
cCRAY_OPT     $           velc(i1,i2,i3,3)**2),i1=1,n1)
 20      continue
 10   continue
      return
      end

      subroutine WriteVel(outunit,n1,n2,x1,x2,u1,u2)
      implicit none
      integer n1,n2,outunit
      real*8 x1(0:n1),x2(0:n2),u1(n1,n2),u2(n1,n2)
c
      integer i,j
c
      write(outunit,*) 'R'
      do 10 j = 1,n2,2
         do 20 i = 1,n2,2
            write(outunit,1000) (x1(i)+x1(i-1))/2,(x2(j)+x2(j-1))/2,
     $           u1(i,j),u2(i,j)
 20      continue
 10   continue
 1000 format(4e12.4)

      return
      end



               
