c ...
c ...   Title:  Utilities for Parcel() data structures that use the cube
c ...
c     ...  Description: Routines to report progress, average boundary
c ...                data, swap boundary data, ping neighbors, remove
c ...                means
c ...   
c ...   Created:        Thu May 28 11:20:59 1992
c ...   Author:         Lawrence Cowsar
c ...           <cowsar@masc28.rice.edu>
c ...   
c ...   LastEditDate:   "Thu May 28 11:20:59 1992"
C     
C ...  $Author: arbogast $
C ...  $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
C ...  $RCSfile: utilCube.f,v $
C ...  $Source: /usr/public/SOURCE/pss/Flow/Parcel/utilCube.f,v $
C  
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
c
c ... Routine: ReportLevel
c
c ... File:  utilCube.f
c
c ... Main function:
c     output status of computation at the end of a V-cycle level
c
c ... Include Files: msgtypes.h
c     
c ... Calls: recv0, send0
c
c ... In: 
c     mynod:    processor node number 
c     CUBESIZE: number of active processors in the cube
c     unit:     Output unit
c     nfcntT:   Average number of nested factorization iteration on this node
c     iVcycle:  Current Vcycle number
c     Level :   Current level
c     NormIn :  Norm at beginning of this level
c     NormOut:  Norm at end of this level
c
c ... Out:
c     outputs to unit, a status of computation
c
c ... Interal Variables of note:
c
c ... Comments:
c     gathers nfcntT across all nodes
c

      subroutine ReportLevel(mynod,CUBESIZE,unit,nfcntT,iVcycle,Level,
     $     NormIn,NormOut,msg)
      implicit none
c    
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      integer mynod,CUBESIZE,unit,nfcntT,iVcycle,Level
      real*8 NormIn,NormOut
      character*(*) msg
c
      integer icnts(0:MAXCUBE),imsgbuf(2),iprtmax,i,error
c
c ... Gather nfcntT from all nodes into icnts on node 0.
c
      call MPI_GATHER(nfcntT, 1, MPI_INTEGER, icnts, 1, MPI_INTEGER, 0,
     $     MPI_COMM_WORLD, error)

      if (MYNOD.eq.0) then
         iprtmax = min(CUBESIZE-1,10)
         write(unit,1601) msg,iVcycle,Level,NormIn,
     $        NormOut,(icnts(i),i=0,iprtmax)
 1601    format(1x,a,1x,i3,i3,2d12.3,2x,16i3)
      endif
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
c
c ... Routine: avgBdry
c
c ... File:  utilCube.f
c
c ... Main function:
c     Average boundary data across interfaces that potentially have
c     different disretezations on the two nodes
c
c ... Include Files: msgtypes.h
c     
c ... Calls: swapface, daxpbyBdry
c
c ... In: 
c     MYNOD :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     size:      The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     g?    :     Boundary type data
c     iws   :     size of work space
c     wrk,wrk1 :  work space
c     msgtype :   base message type (sent to swapface)
c
c ... Out:
c     w?  :  The avgerage value of Bdry data averaged across the interface
c
c ... Interal Variables of note:
c
c ... Comments:
c     this routine and swapface are key to local grid refinement
c
      subroutine avgBdry(mynod,mynbr,size,n1,n2,n3,g1,g2,g3,w1,w2,w3,
     $        iws,wrk,wrk1,msgtype,BdryVarType)
      implicit none
c
      integer mynod,mynbr(3,2),size(3,2,2),n1,n2,n3,msgtype,BdryVarType
      real*8 g1(n2,n3,2),g2(n1,n3,2),g3(n1,n2,2)
      real*8 w1(n2,n3,2),w2(n1,n3,2),w3(n1,n2,2)
      integer iws
      real*8 wrk(iws),wrk1(iws)
c
c ... Send Boundary data to ajoining nodes
c
      call swapface(mynod,mynbr,size,n1,n2,n3,g1,g2,g3,
     $     w1,w2,w3,iws,wrk,wrk1,msgtype,BdryVarType)
c      
c ... Sum the left-side and right-side contributions
c
      call daxpbyBdry(n1,n2,n3,0.5d0,w1,w2,w3,0.5d0,g1,g2,g3,
     $     w1,w2,w3)
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
c
c ... Routine: addBdry
c
c ... File:  utilCube.f
c
c ... Main function:
c     Add boundary data across interfaces that potentially have
c     different disretezations on the two nodes
c
c ... Include Files: msgtypes.h
c     
c ... Calls: swapface, daxpbyBdry
c
c ... In: 
c     MYNOD :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     size:      The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     g?    :     Boundary type data
c     iws   :     size of work space
c     wrk,wrk1 :  work space
c     msgtype :   base message type (sent to swapface)
c
c ... Out:
c     w?  :  The sum of Bdry data averaged across the interface
c
c ... Interal Variables of note:
c
c ... Comments:
c     this routine and swapface are key to local grid refinement
c
      subroutine addBdry(mynod,mynbr,size,n1,n2,n3,g1,g2,g3,w1,w2,w3,
     $        rwk,msgtype,BdryVarType)
      implicit none
c
      integer mynod,mynbr(3,2),size(3,2,2),n1,n2,n3,msgtype,BdryVarType
      real*8 g1(n2,n3,2),g2(n1,n3,2),g3(n1,n2,2)
      real*8 w1(n2,n3,2),w2(n1,n3,2),w3(n1,n2,2),rwk(*)
      integer iws,iptr
c
      integer memalr
      external memalr
c
      iws = 16*max(max(n1*n2,n1*n3),n2*n3)
      iptr = memalr(2*iws,rwk)
c
c ... Send Boundary data to ajoining nodes
c
      call swapface(mynod,mynbr,size,n1,n2,n3,g1,g2,g3,
     $     w1,w2,w3,iws,rwk(iptr),rwk(iptr+iws),msgtype,BdryVarType)
      call memfreer(iptr,rwk)
c      
c ... Sum the left-side and right-side contributions
c
      call daxpbyBdry(n1,n2,n3,1.0d0,w1,w2,w3,1.0d0,g1,g2,g3,
     $     w1,w2,w3)
c
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
c
c ... Routine: swapface
c
c ... File:  utilCube.f
c
c ... Main function:
c     swaps boundary type data across neighbor nodes potentially have
c     different disretezations on the interface
c
c ... Include Files: msgtypes.h
c     
c ... Calls: Parcelcrash,f2f,send0
c
c ... In: 
c     MYNOD :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     size:      The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     bf?    :     Boundary type data
c     iws   :     size of work space
c     wrk,wrk1 :  work space
c     msgtype :   base message type (sent to swapface)
c
c ... Out:
c     df?  :  The neighbors boundary data
c
c ... Interal Variables of note:
c
c ... Comments:
c     this routine is the key to local grid refinement
c     For each edge:
c       abort if there is not enough space (Parcelcrash)
c       convert data into type receiving node expects (f2f)
c       send data (send0)
c
      subroutine swapface(mynod,mynbr,size,n1,n2,n3,bf1,bf2,bf3,
     $     df1,df2,df3,iws,wrk,wrk1,msgtype,BdryVarType)
c
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      integer mynod,mynbr(6),size(6,2),n1,n2,n3,msgtype
      integer BdryVarType
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      integer iws
      real*8 wrk(iws),wrk1(iws)
c
      integer len,status(MPI_STATUS_SIZE),error
c
c Exchange data across each face of the cube, or init to 0.0
c
c WEST/EAST
c
      if (mynbr(WEST).ne.-1) then
         if (size(WEST,1)*size(WEST,2).gt.iws) then
            call Parcelcrash('Not enough work space to project face',
     $           -16)
         endif
         if (mynod.lt.mynbr(EAST)) then
            call f2f(bf1(1,1,1),n2,n3,wrk,size(WEST,1),size(WEST,2),
     $           wrk1,iws,BdryVarType)
            len = size(WEST,1)*size(WEST,2)
            call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(WEST),
     $           msgtype-1+WEST,df1(1,1,1),len,MPI_REAL8,mynbr(WEST),
     $           msgtype-1+EAST,MPI_COMM_WORLD,status, error)
         endif
      else 
         call dinitFace(n2,n3,df1(1,1,1),0.0d0)
      endif

      if (mynbr(EAST).ne.-1) then
         if (size(EAST,1)*size(EAST,2).gt.iws) then
            call Parcelcrash('Not enough work space to project face',
     $           -16)
         endif
         call f2f(bf1(1,1,2),n2,n3,wrk,size(EAST,1),size(EAST,2),wrk1,
     $        iws,BdryVarType)
         len = size(EAST,1)*size(EAST,2)
         call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(EAST),
     $        msgtype-1+EAST,df1(1,1,2),len,MPI_REAL8,mynbr(EAST),
     $        msgtype-1+WEST,MPI_COMM_WORLD,status, error)
      else 
        call dinitFace(n2,n3,df1(1,1,2),0.0d0)
      endif

      if (mynbr(WEST).ne.-1) then
         if (mynod.ge.mynbr(EAST)) then
            call f2f(bf1(1,1,1),n2,n3,wrk,size(WEST,1),size(WEST,2),
     $           wrk1,iws,BdryVarType)
            len = size(WEST,1)*size(WEST,2)
            call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(WEST),
     $           msgtype-1+WEST,df1(1,1,1),len,MPI_REAL8,mynbr(WEST),
     $           msgtype-1+EAST,MPI_COMM_WORLD,status, error)
         endif
      endif
c
c SOUTH/NORTH
c
      if (mynbr(SOUTH).ne.-1) then
         if (size(SOUTH,1)*size(SOUTH,2).gt.iws) then
            call Parcelcrash('Not enough work space to project face',
     $           -16)
         endif
         if (mynod.lt.mynbr(NORTH)) then
            call f2f(bf2(1,1,1),n1,n3,wrk,size(SOUTH,1),size(SOUTH,2),
     $           wrk1,iws,BdryVarType)
            len = size(SOUTH,1)*size(SOUTH,2)
            call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(SOUTH),
     $           msgtype-1+SOUTH,df2(1,1,1),len,MPI_REAL8,mynbr(SOUTH),
     $           msgtype-1+NORTH,MPI_COMM_WORLD,status, error)
         endif
      else
         call dinitFace(n1,n3,df2(1,1,1),0.0d0)
      endif

      if (mynbr(NORTH).ne.-1) then
         if (size(NORTH,1)*size(NORTH,2).gt.iws) then
            call Parcelcrash('Not enough work space to project face',
     $           -16)
         endif
         call f2f(bf2(1,1,2),n1,n3,wrk,size(NORTH,1),size(NORTH,2),wrk1,
     $        iws,BdryVarType)
         len = size(NORTH,1)*size(NORTH,2)
         call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(NORTH),
     $        msgtype-1+NORTH,df2(1,1,2),len,MPI_REAL8,mynbr(NORTH),
     $        msgtype-1+SOUTH,MPI_COMM_WORLD,status, error)
      else
         call dinitFace(n1,n3,df2(1,1,2),0.0d0)
      endif

      if (mynbr(SOUTH).ne.-1) then
         if (mynod.ge.mynbr(NORTH)) then
            call f2f(bf2(1,1,1),n1,n3,wrk,size(SOUTH,1),size(SOUTH,2),
     $           wrk1,iws,BdryVarType)
            len = size(SOUTH,1)*size(SOUTH,2)
            call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(SOUTH),
     $           msgtype-1+SOUTH,df2(1,1,1),len,MPI_REAL8,mynbr(SOUTH),
     $           msgtype-1+NORTH,MPI_COMM_WORLD,status, error)
         endif
      endif
c
c DOWN/UP
c
      if (mynbr(DOWN).ne.-1) then
         if (size(DOWN,1)*size(DOWN,2).gt.iws) then
            call Parcelcrash('Not enough work space to project face',
     $           -16)
         endif
         if (mynod.lt.mynbr(UP)) then
            call f2f(bf3(1,1,1),n1,n2,wrk,size(DOWN,1),size(DOWN,2),
     $           wrk1,iws,BdryVarType)
            len = size(DOWN,1)*size(DOWN,2)
            call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(DOWN),
     $           msgtype-1+DOWN,df3(1,1,1),len,MPI_REAL8,mynbr(DOWN),
     $           msgtype-1+UP,MPI_COMM_WORLD,status, error)
         endif
      else
         call dinitFace(n1,n2,df3(1,1,1),0.0d0)
      endif

      if (mynbr(UP).ne.-1) then
         if (size(UP,1)*size(UP,2).gt.iws) then
            call Parcelcrash('Not enough work space to project face',
     $           -16)
         endif
         call f2f(bf3(1,1,2),n1,n2,wrk,size(UP,1),size(UP,2),wrk1,iws,
     $        BdryVarType)
         len = size(UP,1)*size(UP,2)
         call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(UP),
     $        msgtype-1+UP,df3(1,1,2),len,MPI_REAL8,mynbr(UP),
     $        msgtype-1+DOWN,MPI_COMM_WORLD,status, error)
      else
         call dinitFace(n1,n2,df3(1,1,2),0.0d0)
      endif

      if (mynbr(DOWN).ne.-1) then
         if (mynod.ge.mynbr(UP)) then
            call f2f(bf3(1,1,1),n1,n2,wrk,size(DOWN,1),size(DOWN,2),
     $           wrk1,iws,BdryVarType)
            len = size(DOWN,1)*size(DOWN,2)
            call MPI_SENDRECV(wrk,len,MPI_REAL8,mynbr(DOWN),
     $           msgtype-1+DOWN,df3(1,1,1),len,MPI_REAL8,mynbr(DOWN),
     $           msgtype-1+UP,MPI_COMM_WORLD,status, error)
         endif
      endif

      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
c
c ... Routine: f2f
c
c ... File:  utilCube.f
c
c ... Main function:
c     transfer face into representation used by neighbor node
c
c ... Include Files: none
c     
c ... Calls: Parcelcrash
c
c ... In: 
c     a     :   Boundary face
c     n?    :   Dimensions of a
c     m?    :   Dimensions of b
c     iws   :     size of work space
c     wrk   :  work space
c
c ... Out:
c     b    : transfered a that is m1 x m2
c
c    ... Interal Variables of note:
c
c ... Comments:
c     transfer to same size node is the identity
c     transfer to a coarser direction is by averaging cells that make 
c              up the coarse cell
c     transfer to a finer direction is by assigning to the fine cell the
c              value of the coarse cell which contains it
c
      subroutine f2f(a,n1,n2,b,m1,m2,wrk,iwks,VarType)
      implicit none
c
      integer n1,n2,m1,m2,iwks,VarType
      real*8 a(n1,n2),b(m1,m2),wrk(m1,n2)
c
      include '../Include/BCtypes.hf'
c
      integer i1,i2,k1,k2,inc
      real*8 wt
c
c ... If same size, transfer operator is the identity
c 
      if ((n1.eq.m1).and.(n2.eq.m2)) then
c
         do 10 i2 = 1,m2
            do 11 i1 = 1,m1
               b(i1,i2) = a(i1,i2)
 11         continue
 10      continue
c     
      else
c     
         if (((mod(n1,m1).ne.0).and.(mod(m1,n1).ne.0)).or.
     $        ((mod(n2,m2).ne.0).and.(mod(m2,n2).ne.0))) then
            write(*,*) n1,n2, ' --> ',m1,m2
            call Parcelcrash('Grids not nested!!! Trans op. failed',
     $           -17)
         endif
         if (m1*n2.gt.iwks) then
            write(*,*) 'Not enough workspace in f2f: ',m1*n2,
     $           ' needed.',iwks,' provided.'
            call Parcelcrash('Not enough work space for trans. op.',
     $           -16)
         endif
c     
c ... Transfer in first direction
c     
         do 20 i2 = 1,n2 
            if (mod(n1,m1).eq.0) then
c     
c ... Reduction to coarser grid via averaging
c
               inc = n1/m1
               if (VarType.eq.VOLUMETRIC) then
                  wt = 1.0d0
               else if (VarType.eq.SCALAR) then
                  wt = 1.0d0/dble(inc)
               endif
               do 22 i1 = 1,m1
                  wrk(i1,i2) = 0.0d0
                  do 21 k1 = (i1-1)*inc+1,i1*inc
                     wrk(i1,i2) = wrk(i1,i2) + wt*a(k1,i2)
 21               continue
 22            continue
c     
            else
c
c ... Expansion via evaluation
c
               inc = m1/n1
               if (VarType.eq.VOLUMETRIC) then
                  wt = 1.0d0/dble(inc)
               else if (VarType.eq.SCALAR) then
                  wt = 1.0d0
               endif
               do 25 i1 = 1,m1
                  k1 = (i1+inc-1)/inc
                  wrk(i1,i2) = a(k1,i2)*wt
 25            continue
c     
            endif
 20      continue
c
c ... Transfer in second direction
c
         do 30 i1 = 1,m1
            if (mod(n2,m2).eq.0) then
c     
c ... Reduction to coarser grid via averaging
c
               inc = n2/m2
               if (VarType.eq.VOLUMETRIC) then
                  wt = 1.0d0
               else if (VarType.eq.SCALAR) then
                  wt = 1.0d0/dble(inc)
               endif
               do 32 i2 = 1,m2
                  b(i1,i2) = 0.0d0
                  do 31 k2 = (i2-1)*inc+1,i2*inc
                     b(i1,i2) = b(i1,i2) + wt*wrk(i1,k2)
 31               continue
 32            continue
c
            else
c
c ... Expansion via evaluation
c
               inc = m2/n2
               if (VarType.eq.VOLUMETRIC) then
                  wt = 1.0d0/dble(inc)
               else if (VarType.eq.SCALAR) then
                  wt = 1.0d0
               endif
               do 35 i2 = 1,m2
                  k2 = (i2+inc-1)/inc
                  b(i1,i2) = wrk(i1,k2)*wt
 35            continue
c
            endif
 30      continue
c
      endif
      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
c
c ... Routine: GetNbrSize
c
c ... File:  utilCube.f
c
c ... Main function:
c     determine neighbor's discretization of shared face
c
c ... Include Files: msgtypes.h
c     
c ... Calls: send0, recv0
c
c ... In: 
c     mynod :    processor id
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     n?    :    number of grid points in each direction for my faces
c
c ... Out:
c     mynbrSize: The dimension of neighbor faces 
c                   (used for local grid refinement)
c
c ... Interal Variables of note:
c
c ... Comments:
c
      subroutine GetNbrSize(mynod,mynbr,mynbrSize,n1,n2,n3)
c
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      integer mynod,mynbr(6),mynbrSize(6,2),n1,n2,n3
c
      integer imsgbuf(2),ircvbuf(2)
      integer status(MPI_STATUS_SIZE),error
c
c ... For each shared face, send mynod's discretization
c     and receive the neighbor's discretization
c
c NORTH/SOUTH
c
      if (mynbr(NORTH).ne.-1) then
         if (mynod.lt.mynbr(NORTH)) then
            imsgbuf(1) = n1
            imsgbuf(2) = n3
            call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(NORTH),
     $           ISIZETYPE-1+NORTH,ircvbuf,2,MPI_INTEGER,
     $           mynbr(NORTH),ISIZETYPE-1+SOUTH,
     $           MPI_COMM_WORLD, status, error)
            mynbrSize(NORTH,1) = ircvbuf(1)
            mynbrSize(NORTH,2) = ircvbuf(2)
         endif
      else
         mynbrSize(NORTH,1) = -1
         mynbrSize(NORTH,2) = -1
      endif

      if (mynbr(SOUTH).ne.-1) then
         imsgbuf(1) = n1
         imsgbuf(2) = n3
         call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(SOUTH),
     $        ISIZETYPE-1+SOUTH,ircvbuf,2,MPI_INTEGER,
     $        mynbr(SOUTH),ISIZETYPE-1+NORTH,
     $        MPI_COMM_WORLD, status, error) 
         mynbrSize(SOUTH,1) = ircvbuf(1)
         mynbrSize(SOUTH,2) = ircvbuf(2)
      else
         mynbrSize(SOUTH,1) = -1
         mynbrSize(SOUTH,2) = -1
      endif

      if (mynbr(NORTH).ne.-1) then
         if (mynod.ge.mynbr(NORTH)) then
            imsgbuf(1) = n1
            imsgbuf(2) = n3
            call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(NORTH),
     $           ISIZETYPE-1+NORTH,ircvbuf,2,MPI_INTEGER,
     $           mynbr(NORTH),ISIZETYPE-1+SOUTH,
     $           MPI_COMM_WORLD, status, error)
            mynbrSize(NORTH,1) = ircvbuf(1)
            mynbrSize(NORTH,2) = ircvbuf(2)
         endif
      endif
c
c EAST/WEST
c
      if (mynbr(EAST).ne.-1) then
         if (mynod.lt.mynbr(EAST)) then
            imsgbuf(1) = n2
            imsgbuf(2) = n3
            call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(EAST),
     $           ISIZETYPE-1+EAST,ircvbuf,2,MPI_INTEGER,
     $           mynbr(EAST),ISIZETYPE-1+WEST,
     $           MPI_COMM_WORLD, status, error) 
            mynbrSize(EAST,1) = ircvbuf(1)
            mynbrSize(EAST,2) = ircvbuf(2)
         endif
      else
         mynbrSize(EAST,1) = -1
         mynbrSize(EAST,2) = -1
      endif

      if (mynbr(WEST).ne.-1) then
         imsgbuf(1) = n2
         imsgbuf(2) = n3
         call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(WEST),
     $        ISIZETYPE-1+WEST,ircvbuf,2,MPI_INTEGER,
     $        mynbr(WEST),ISIZETYPE-1+EAST,
     $        MPI_COMM_WORLD, status, error) 
         mynbrSize(WEST,1) = ircvbuf(1)
         mynbrSize(WEST,2) = ircvbuf(2)
      else
         mynbrSize(WEST,1) = -1
         mynbrSize(WEST,2) = -1
      endif

      if (mynbr(EAST).ne.-1) then
         if (mynod.ge.mynbr(EAST)) then
            imsgbuf(1) = n2
            imsgbuf(2) = n3
            call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(EAST),
     $           ISIZETYPE-1+EAST,ircvbuf,2,MPI_INTEGER,
     $           mynbr(EAST),ISIZETYPE-1+WEST,
     $           MPI_COMM_WORLD, status, error) 
            mynbrSize(EAST,1) = ircvbuf(1)
            mynbrSize(EAST,2) = ircvbuf(2)
         endif
      endif
c
c UP/DOWN
c
      if (mynbr(UP).ne.-1) then
         if (mynod.lt.mynbr(UP)) then
            imsgbuf(1) = n1
            imsgbuf(2) = n2
            call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(UP),
     $           ISIZETYPE-1+UP,ircvbuf,2,MPI_INTEGER,
     $           mynbr(UP),ISIZETYPE-1+DOWN,
     $           MPI_COMM_WORLD, status, error) 
            mynbrSize(UP,1) = ircvbuf(1)
            mynbrSize(UP,2) = ircvbuf(2)
         endif
      else
         mynbrSize(UP,1) = -1
         mynbrSize(UP,2) = -1
      endif

      if (mynbr(DOWN).ne.-1) then
         imsgbuf(1) = n1
         imsgbuf(2) = n2
         call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(DOWN),
     $        ISIZETYPE-1+DOWN,ircvbuf,2,MPI_INTEGER,
     $        mynbr(DOWN),ISIZETYPE-1+UP,
     $        MPI_COMM_WORLD, status, error) 
         mynbrSize(DOWN,1) = ircvbuf(1)
         mynbrSize(DOWN,2) = ircvbuf(2)
      else
         mynbrSize(DOWN,1) = -1
         mynbrSize(DOWN,2) = -1
      endif

      if (mynbr(UP).ne.-1) then
         if (mynod.ge.mynbr(UP)) then
            imsgbuf(1) = n1
            imsgbuf(2) = n2
            call MPI_SENDRECV(imsgbuf,2,MPI_INTEGER,mynbr(UP),
     $           ISIZETYPE-1+UP,ircvbuf,2,MPI_INTEGER,
     $           mynbr(UP),ISIZETYPE-1+DOWN,
     $           MPI_COMM_WORLD, status, error) 
            mynbrSize(UP,1) = ircvbuf(1)
            mynbrSize(UP,2) = ircvbuf(2)
         endif
      endif

      return
      end
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c ... Version $Id: utilCube.f,v 1.4 1997-08-20 11:46:49-05 arbogast Exp $
c
c ... Routine: swapflux123
c
c ... File:  cf.f
c
c ... Main function:
c     Compute flux jump across subdomains
c
c ... Include Files: 
c     msgtypes.h
c     
c ... Calls: 
c     bdryflux,copyiface,swapface,daxpbybdry
c
c ... In:
c     mynbr :    processor ids of neighbor subdomains, -1 if bdry edge
c     size:      The dimension of neighbor faces 
c                   (used for local grid refinement)
c     n?    :     number of grid points in each direction
c     pres  :     Pressure solution
c     l?    :     Dirichlet values along interfaces
c     bcType:     BC's for 1 - Neumann, 2 - Dirichlet, 3 - Noflow (NOT
c                 CURRENTLY PASSED)
c     k?    :     Edge Permeability tensor
c     hx?   :     mesh widths hx?(i) = x?(i)-x?(i-1)
c     wrk,wrk1:   work space
c     iwrk  :     lenght of work space
c
c ... Out:
c     bf?   :  Boundary fluxes for this subdomain
c     df?   :  Jump in Boundary fluxes across sumdomain
c
c ... Interal Variables of note: none
c
c ... Comments: none
c
      subroutine swapflux123(mynod,mynbr,size,n1,n2,n3,
     $     bf1,bf2,bf3,df1,df2,df3,iws,wrk,wrk1,BdryVarType)
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      real*8 frac
      parameter(frac = 1.0d0)
c
      integer mynod,mynbr(3,2),size(3,2,2),n1,n2,n3,BdryVarType
      real*8 bf1(n2,n3,2),bf2(n1,n3,2),bf3(n1,n2,2)
      real*8 df1(n2,n3,2),df2(n1,n3,2),df3(n1,n2,2)
      integer iws
      real*8 wrk(iws),wrk1(iws)
c
c ... The commented out code is used to produce matching of the form 
c ... frac*flux + (1-frac)*pressure
c$$$      real*8 bp1(2*idim2),bp2(2*idim2),bp3(2*idim2)
c$$$      call bdrypres123(n1,n2,n3,pres,l1,l2,l3,bcType,nBCRegions,
c$$$     $     bcRegion1,bcRegion2,bcRegion3,k1,k2,k3,hx1,hx2,hx3,
c$$$     $     bp1,bp2,bp3)
c$$$c
c$$$c ... Change sign of some pressures to reflect their "normal-sign"
c$$$c      
c$$$      call dscal(n2*n3,-1.0d0,bp1,1)
c$$$      call dscal(n1*n3,-1.0d0,bp2,1)
c$$$      call dscal(n1*n2,-1.0d0,bp3,1)
c$$$c
c$$$c ... Form the matching criteria:  frac*flux + (1-frac)*pres
c$$$c
c$$$      call daxpbyBdry(n1,n2,n3,frac,bf1,bf2,bf3,1.0d0-frac,bp1,bp2,bp3,
c$$$     $     bf1,bf2,bf3)
c
c ... Extract those faces that are interfaces
c
      call copyIface(mynbr,n1,n2,n3,bf1,bf2,bf3,bf1,bf2,bf3)
c
c ... Send Fluxes to ajoining nodes
c
      call swapface(mynod,mynbr,size,n1,n2,n3,bf1,bf2,bf3,
     $     df1,df2,df3,iws,wrk,wrk1,FLUXTYPE,BdryVarType)
c      
c ... Sum the left-side and right-side fluxes to compute jump
c
      call daxpbyBdry(n1,n2,n3,-1.0d0,bf1,bf2,bf3,-1.0d0,df1,df2,df3,
     $     df1,df2,df3)
c
      return
      end


      subroutine rmBdry(mynbr,n1,n2,n3,x1,x2,x3,hx1,hx2,hx3,
     $     f1,f2,f3)
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      integer mynbr(3,2),n1,n2,n3
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  f1(n2,n3,2),f2(n1,n3,2),f3(n1,n2,2)
c
      integer i1,i2,i3,i,j,error
      real*8 avg,d1,d2,d3,Area(3),Surface
      real*8 dsumBdryG
c
c ... Integrate f over surface faces
c ... Assumes f == 0 on outer boundarys
c
      avg = dsumBdryG(n1,n2,n3,hx1,hx2,hx3,f1,f2,f3,BDRYSUMTYPE)
c
c ... Calculate measure on all faces
c
      d1 = x1(n1) - x1(0)
      d2 = x2(n2) - x2(0)
      d3 = x3(n3) - x3(0)
      Area(1) = d2*d3
      Area(2) = d1*d3
      Area(3) = d1*d2
c
      Surface = 0.0d0
      do 40 i=1,2
         do 41 j = 1,3
            if (mynbr(j,i).ne.-1) then
               Surface = Surface + Area(j)
            endif
 41      continue
 40   continue
c
      call MSG_f_Gdsum1(Surface,MPI_COMM_WORLD)
      Surface = Surface/2.0d0
c
      avg = avg/Surface
c
c ... Remove Mean of f?
c
      do 10 i3 = 1,n3
         do 11 i2 = 1,n2
            if (mynbr(1,1).ne.-1) then
               f1(i2,i3,1) = f1(i2,i3,1) - avg
            endif
            if (mynbr(1,2).ne.-1) then
               f1(i2,i3,2) = f1(i2,i3,2) - avg
            endif
 11      continue
 10   continue
c
c
      do 20 i3 = 1,n3
         do 21 i1 = 1,n1
            if (mynbr(2,1).ne.-1) then
               f2(i1,i3,1) = f2(i1,i3,1) - avg
            endif
            if (mynbr(2,2).ne.-1) then
               f2(i1,i3,2) = f2(i1,i3,2) - avg
            endif
 21      continue
 20   continue
c
      do 30 i2 = 1,n2
         do 31 i1 = 1,n1
            if (mynbr(3,1).ne.-1) then
               f3(i1,i2,1) = f3(i1,i2,1) - avg
            endif
            if (mynbr(3,2).ne.-1) then
               f3(i1,i2,2) = f3(i1,i2,2) - avg
            endif
 31      continue
 30   continue
c               
c ... Check that mean was removed
c
cc      avg = dsumBdryG(n1,n2,n3,hx1,hx2,hx3,f1,f2,f3,BDRYSUMTYPE)
c
cc      avg = avg/Surface
c
      return
      end

      subroutine rmBdryFluxes(mynbr,n1,n2,n3,x1,x2,x3,hx1,hx2,hx3,
     $     f1,f2,f3,ftype)
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
      include '../Include/BCtypes.hf'
c
      integer mynbr(3,2),n1,n2,n3,ftype
      real*8  x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8  f1(n2,n3,2),f2(n1,n3,2),f3(n1,n2,2)
c
      integer i1,i2,i3,i,j,error
      real*8 avg,d1,d2,d3,FaceArea(3),Surface
      real*8 dsumBdryG,dsumBdryGNoArea
c
c ... Integrate f over surface faces
c ... Assumes f == 0 on outer boundarys
c
      if (ftype.eq.VOLUMETRIC) then
         avg = dsumBdryGNoArea(n1,n2,n3,f1,f2,f3,BDRYSUMTYPE)
      else
         avg = dsumBdryG(n1,n2,n3,hx1,hx2,hx3,f1,f2,f3,BDRYSUMTYPE)
      endif
c
c ... Calculate measure on all faces
c
      if (ftype.eq.VOLUMETRIC) then
         FaceArea(1) = n2*n3
         FaceArea(2) = n1*n3
         FaceArea(3) = n1*n2
      else
         d1 = x1(n1) - x1(0)
         d2 = x2(n2) - x2(0)
         d3 = x3(n3) - x3(0)
         FaceArea(1) = d2*d3
         FaceArea(2) = d1*d3
         FaceArea(3) = d1*d2
      endif
c     
      Surface = 0.0d0
      do 40 i=1,2
         do 41 j = 1,3
            if (mynbr(j,i).ne.-1) then
               Surface = Surface + FaceArea(j)
            endif
 41      continue
 40   continue
c     
      call MSG_f_Gdsum1(Surface,MPI_COMM_WORLD)
      Surface = Surface/2.0d0
c
      avg = avg/Surface
c
c ... Remove Mean of f?
c
      do 10 i3 = 1,n3
         do 11 i2 = 1,n2
            if (mynbr(1,1).ne.-1) then
               f1(i2,i3,1) = f1(i2,i3,1) - avg
            endif
            if (mynbr(1,2).ne.-1) then
               f1(i2,i3,2) = f1(i2,i3,2) - avg
            endif
 11      continue
 10   continue
c
      do 20 i3 = 1,n3
         do 21 i1 = 1,n1
            if (mynbr(2,1).ne.-1) then
               f2(i1,i3,1) = f2(i1,i3,1) - avg
            endif
            if (mynbr(2,2).ne.-1) then
               f2(i1,i3,2) = f2(i1,i3,2) - avg
            endif
 21      continue
 20   continue
c
      do 30 i2 = 1,n2
         do 31 i1 = 1,n1
            if (mynbr(3,1).ne.-1) then
               f3(i1,i2,1) = f3(i1,i2,1) - avg
            endif
            if (mynbr(3,2).ne.-1) then
               f3(i1,i2,2) = f3(i1,i2,2) - avg
            endif
 31      continue
 30   continue
c               
c ... Check that mean was removed
c
cc      avg = dsumBdryG(n1,n2,n3,hx1,hx2,hx3,f1,f2,f3,BDRYSUMTYPE)
c
cc      avg = avg/Surface
c
      return
      end
c
c ... Remove Mean from Pressures across all nodes
c
      subroutine rmPresG(n1,n2,n3,x1,x2,x3,hx1,hx2,hx3,pres)
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      integer n1,n2,n3
      real*8 x1(0:n1),x2(0:n2),x3(0:n3),hx1(n1),hx2(n2),hx3(n3)
      real*8 pres(n1,n2,n3)
c
      integer i1,i2,i3,error
      real*8 avg,d1,d2,d3,Volume
      real*8 dsumPresG
c
      avg = dsumPresG(n1,n2,n3,hx1,hx2,hx3,pres,PRESSUMTYPE)
c
      d1 = x1(n1) - x1(0)
      d2 = x2(n2) - x2(0)
      d3 = x3(n3) - x3(0)
      Volume = d1*d2*d3
c
      call MSG_f_Gdsum1(Volume,MPI_COMM_WORLD)
      avg = avg/Volume
c
      do 10 i3 = 1,n3
         do 15 i2 = 1,n2
            do 20 i1 = 1,n1
               pres(i1,i2,i3) = pres(i1,i2,i3) - avg
 20         continue
 15      continue
 10   continue
c
c
c ... Test for mean removed
c
c      avg = dsumPresG(n1,n2,n3,hx1,hx2,hx3,pres,PRESSUMTYPE)
c      avg = avg/Volume
c
      return
      end



      subroutine concatvec(nproc,vec,msgtype)
      implicit none
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
c
      integer nproc,msgtype,error
      real*8 vec(0:nproc-1)
c
cmpi      call gdsumx(msgtype,vec(0),nproc)
      call MSG_f_Gdsum(vec(0),nproc,MPI_COMM_WORLD)
c
      end


      subroutine projVel(n1,n2,n3,vel1,vel2,vel3,
     $                   mynod,mynbr,mynbrSize,rwk)
      implicit none 
      integer n1,n2,n3,mynod,mynbr(3,2),mynbrSize(*)
      real*8 vel1(0:n1,n2,n3),vel2(n1,0:n2,n3),vel3(n1,n2,0:n3),rwk(*)
c
      include 'mpif.h'
      include '../Include/msgtypes.hf'
      include '../Include/BCtypes.hf'
c
      integer memalr
c
      integer g1ptr,g2ptr,g3ptr,w1ptr,w2ptr,w3ptr,iws,iptr
c
      g1ptr = memalr(2*n2*n3,rwk)
      g2ptr = memalr(2*n1*n3,rwk)
      g3ptr = memalr(2*n1*n2,rwk)
      w1ptr = memalr(2*n2*n3,rwk)
      w2ptr = memalr(2*n1*n3,rwk)
      w3ptr = memalr(2*n1*n2,rwk)
      call getIntBdryVel(mynbr,n1,n2,n3,vel1,vel2,vel3,
     $     rwk(g1ptr),rwk(g2ptr),rwk(g3ptr))

      iws = 16*max(max(n1*n2,n1*n3),n2*n3)
      iptr = memalr(2*iws,rwk)

      call avgBdry(mynod,mynbr,mynbrSize,n1,n2,n3,
     $     rwk(g1ptr),rwk(g2ptr),rwk(g3ptr),
     $     rwk(w1ptr),rwk(w2ptr),rwk(w3ptr),
     $     iws,rwk(iptr),rwk(iptr+iws),AVGBDRYTYPE,SCALAR)

      call putIntBdryVel(mynbr,n1,n2,n3,vel1,vel2,vel3,
     $     rwk(w1ptr),rwk(w2ptr),rwk(w3ptr))
c
      call memfreer(g1ptr,rwk)
c
      end

      subroutine getIntBdryVel(mynbr,n1,n2,n3,v1,v2,v3,g1,g2,g3)
      implicit none
      integer mynbr(3,2),n1,n2,n3
      real*8 v1(0:n1,n2,n3),v2(n1,0:n2,n3),v3(n1,n2,0:n3)
      real*8 g1(n2,n3,2),g2(n1,n3,2),g3(n1,n2,2)
c
      integer i1,i2,i3
c
      if (mynbr(1,1).ne.-1) then
         do i3 = 1,n3
            do i2 = 1,n2
               g1(i2,i3,1) = v1(0,i2,i3)
            enddo
         enddo
      else
         do i3 = 1,n3
            do i2 = 1,n2
               g1(i2,i3,1) = 0
            enddo
         enddo
      endif
      if (mynbr(1,2).ne.-1) then
         do i3 = 1,n3
            do i2 = 1,n2
               g1(i2,i3,2) = v1(n1,i2,i3)
            enddo
         enddo
      else
         do i3 = 1,n3
            do i2 = 1,n2
               g1(i2,i3,2) = 0
            enddo
         enddo
      endif
      if (mynbr(2,1).ne.-1) then
         do i3 = 1,n3
            do i1 = 1,n1
               g2(i1,i3,1) = v2(i1,0,i3)
            enddo
         enddo
      else
         do i3 = 1,n3
            do i1 = 1,n1
               g2(i1,i3,1) = 0
            enddo
         enddo
      endif
      if (mynbr(2,2).ne.-1) then
         do i3 = 1,n3
            do i1 = 1,n1
               g2(i1,i3,2) = v2(i1,n2,i3)
            enddo
         enddo
      else
         do i3 = 1,n3
            do i1 = 1,n1
               g2(i1,i3,2) = 0
            enddo
         enddo
      endif
      if (mynbr(3,1).ne.-1) then
         do i2 = 1,n2
            do i1 = 1,n1
               g3(i1,i2,1) = v3(i1,i2,0)
            enddo
         enddo
      else
         do i2 = 1,n2
            do i1 = 1,n1
               g3(i1,i2,1) = 0
            enddo
         enddo
      endif
      if (mynbr(3,2).ne.-1) then
         do i2 = 1,n2
            do i1 = 1,n1
               g3(i1,i2,2) = v3(i1,i2,n3)
            enddo
         enddo
      else
         do i2 = 1,n2
            do i1 = 1,n1
               g3(i1,i2,2) = 0
            enddo
         enddo
      endif
      end

      subroutine putIntBdryVel(mynbr,n1,n2,n3,v1,v2,v3,g1,g2,g3)
      implicit none
      integer mynbr(3,2),n1,n2,n3
      real*8 v1(0:n1,n2,n3),v2(n1,0:n2,n3),v3(n1,n2,0:n3)
      real*8 g1(n2,n3,2),g2(n1,n3,2),g3(n1,n2,2)
c
      integer i1,i2,i3
c
      if (mynbr(1,1).ne.-1) then
         do i3 = 1,n3
            do i2 = 1,n2
               v1(0,i2,i3) = g1(i2,i3,1)
            enddo
         enddo
      endif
      if (mynbr(1,2).ne.-1) then
         do i3 = 1,n3
            do i2 = 1,n2
               v1(n1,i2,i3) = g1(i2,i3,2)
            enddo
         enddo
      endif
      if (mynbr(2,1).ne.-1) then
         do i3 = 1,n3
            do i1 = 1,n1
               v2(i1,0,i3) = g2(i1,i3,1)
            enddo
         enddo
      endif
      if (mynbr(2,2).ne.-1) then
         do i3 = 1,n3
            do i1 = 1,n1
               v2(i1,n2,i3) = g2(i1,i3,2)
            enddo
         enddo
      endif
      if (mynbr(3,1).ne.-1) then
         do i2 = 1,n2
            do i1 = 1,n1
               v3(i1,i2,0) = g3(i1,i2,1)
            enddo
         enddo
      endif
      if (mynbr(3,2).ne.-1) then
         do i2 = 1,n2
            do i1 = 1,n1
               v3(i1,i2,n3) = g3(i1,i2,2)
            enddo
         enddo
      endif
      end

